<?php

namespace Wino;

interface AdminPageTranslator extends
StartImportSessionCommandTranslator,
FlushImportSessionCommandTranslator,
DeleteDataCommandTranslator,
ConfigurationLocaleTranslator,
ConfigurationApiUrlTranslator,
ConfigurationPriceListTranslator,
ConfigurationCategoriesImportEnabledTranslator,
ConfigurationRootCategoryTranslator,
ConfigurationExplicitCategoriesHierarchyEnabledTranslator,
SaveConfigurationCommandTranslator,
ConfigurationStockManagementOverrideEnabledTranslator,
ConfigurationSkuBasedImportEnabledTranslator,
ConfigurationProductDescriptionFormatTranslator,
ConfigurationProductsNamesUpdateEnabledTranslator,
ConfigurationProductsDescriptionsUpdateEnabledTranslator,
ConfigurationPricesUpdateEnabledTranslator,
ConfigurationSynchronizationEnabledTranslator,
ConfigurationMonitoringEnabledTranslator {
  /** @return string */
  public function debugFieldsGroupLabel();

  /** @return string */
  public function updateFieldsGroupLabel();

  /** @return string */
  public function adminPageTitle();

  /** @return string */
  public function importRequiredToConfirmAPriceListUpdate();
}

class AdminPage {
  public const MENU = 'Wino';
  public const MENU_SLUG = 'wino-admin-page';

  /** @return string */
  public static function title(AdminPageTranslator $translator) {
    return $translator->adminPageTitle();
  }

  /** @return string */
  public static function getUrl() {
    $slug = self::MENU_SLUG;
    return admin_url("options-general.php?page=$slug");
  }

  /** @return void */
  public static function initialize() {
    add_action('admin_menu', function () {
      add_options_page(
        self::title(LocaleTranslator::fromConfiguration()),
        self::MENU,
        Auth::ADMIN_CAPABILITY,
        self::MENU_SLUG,
        function () {
          self::render();
        }
      );
    });
  }

  /** @return Render */
  public static function pluginZipDownloadButton() {
    return Render::fromCallable(function () {
      $url = ReleasePage::getPluginZipDownloadUrl();
      if ($url === null) return "";
      return renderDownloadButton("Download plugin zip", $url);
    });
  }

  /** @return Render */
  public static function pluginZipDownloadLink() {
    $url = ReleasePage::url();
    return Render::concat(
      spanRender(
        [],
        Render::fromString("The plugin zip can be found here: "),
        htmlLinkRender($url, $url, "_blank")
      ),
      formSpacingRender()
    );
  }

  /** @return Render */
  public static function debugFormSection(AdminPageTranslator $translator) {
    if (!isDebugModeEnabled()) return Render::none();
    return formTableRender(
      formFieldsGroupRenderWithLabel(
        $translator->debugFieldsGroupLabel(),
        self::pluginZipDownloadLink(),
        divRender(
          ['style' => "display: flex; gap: 8pt;"],
          self::pluginZipDownloadButton(),
          DeleteDataCommand::ajaxButtonWithConfirmation($translator),
          FlushImportSessionCommand::ajaxButton($translator)
        ),
        formSpacingRender(),
        ConfigurationApiUrl::textField($translator),
      )
    );
  }

  /** @return string */
  public const PROGRESS_LABEL_PAGE_TITLE_SUFFIX_SPAN_ID = "admin-page-progress-label";

  /** @return string */
  public static function progressLabelTitleSuffixSpan(
    ?ImportSession $import,
    ImportSessionTranslator $translator
  ) {
    $id = self::PROGRESS_LABEL_PAGE_TITLE_SUFFIX_SPAN_ID;
    $progressLabel = $import === null ? null : $import->progressLabel($translator);
    $content = $progressLabel === null ? "" : " - $progressLabel";

    return "<span id=\"$id\" style=\"font-size:0.832em;opacity:0.832;\">$content</span>";
  }

  /** @return void */
  public static function render() {
    Auth::verifyAdminCapability();
    Auth::verifyCurrentUserCapability(Auth::WOOCOMMERCE_ADMIN_CAPABILITY);

    $log = settingsErrorLogFromTransient(self::MENU_SLUG, transientLog());
    $apiClient = WinoApiClient::fromConfiguration();
    $translator = LocaleTranslator::fromConfiguration();

    $import = ImportSession::persisted(DataScope::none());
    $emptyPriceList = ConfigurationPriceList::value() === "";
    $monitoringUrl = MonitoringPage::indexUrl();

    echo renderAdminPageWrapStart(self::progressLabelTitleSuffixSpan($import, $translator));

    Render::echo(SaveConfigurationCommand::form(
      $import,
      RunImportSessionCommand::ajaxAction(),
      CheckImportSessionProgressCommand::ajaxAction($import),
      formTableRender(
        ConfigurationApiKey::textField($translator),
        ConfigurationShopId::selectWithSingleOptionAutosave($log, $apiClient, $translator),
        ConfigurationStockManagementOverrideEnabled::checkbox($translator),
        ConfigurationSkuBasedImportEnabled::checkbox($translator),
        ConfigurationCategoriesImportEnabled::checkboxWithConditionalBlock(
          $translator,
          divRender(
            ['style' => "display:flex;flex-direction:column;gap:8pt;"],
            ConfigurationExplicitCategoriesHierarchyEnabled::checkbox($translator),
            ConfigurationRootCategory::select($translator)
          ),
        ),
        ConfigurationProductDescriptionFormat::textArea($translator),
        formFieldsGroupRenderWithLabel(
          $translator->updateFieldsGroupLabel(),
          ConfigurationProductsNamesUpdateEnabled::checkbox($translator),
          ConfigurationProductsDescriptionsUpdateEnabled::checkbox($translator),
          ConfigurationPricesUpdateEnabled::checkbox($translator)
        ),
        ConfigurationSynchronizationEnabled::checkbox($translator),
      ),
      divRender(
        ['style' => "display:flex;"],
        StartImportSessionCommand::ajaxButton($import, $translator),
        ConfigurationPriceList::select($log, $apiClient, $translator),
      ),
      $emptyPriceList ? Render::none() : divRender(
        ['style' => "margin:8pt"],
        emRender(
          [],
          Render::fromString('* '),
          Render::fromString($translator->importRequiredToConfirmAPriceListUpdate())
        )
      ),
      divRender(
        ['style' => "display:flex;gap:12pt;margin-top:18pt;padding-top:16pt;border-top:1px solid #8c8f94"],
        ConfigurationMonitoringEnabled::checkbox($translator),
        htmlLinkRender($monitoringUrl, $monitoringUrl, "_blank")
      ),
      self::debugFormSection($translator)
    ));

    echo renderAdminPageWrapEnd();

    includeAdminStyles();
    includeAjaxScript();
    settings_errors(self::MENU_SLUG);
  }
}
