<?php

namespace Wino;

class MonitoringPage {
  public const PAGE_ID = "monitoring";

  public const LOG_FILE_DATE_QUERY_VAR_NAME = "wino-log-file-date";

  /** @return CustomRoute */
  private static function enable(CustomRoute $route) {
    $context = MonitoringContext::fromConfiguration();
    return $route->enable($context->monitoringEnabled);
  }

  /** @return CustomRoute */
  public static function route() {
    return self::enable(CustomRoute::from(self::PAGE_ID));
  }

  /** @return CustomRoute */
  public static function fileRoute() {
    return self::enable(
      CustomRoute::from(self::PAGE_ID, "/$", self::LOG_FILE_DATE_QUERY_VAR_NAME)
    );
  }

  /** @return void */
  public static function initialize() {
    registerCustomRoute(self::route());
    registerCustomRoute(self::fileRoute());
  }

  /** @return string */
  private static function indexPath() {
    return canonicalCustomRoutePath(self::route());
  }

  /** @return string */
  public static function indexUrl() {
    return customRouteUrl(self::route());
  }

  /** @return string */
  public static function fileUrl(string $logFileDate) {
    return customRouteUrl(self::fileRoute(), $logFileDate);
  }

  /**
   * @param string[] $files
   * @return string
   */
  private static function renderIndex(array $files) {
    if (count($files) === 0) return "No log files available.";
    $links = [];
    foreach ($files as $file) {
      $date = monitoringLogFileDateFromPath($file);
      if ($date === null) continue;
      $links[] = renderHtmlLink(self::fileUrl($date), $file);
    }
    return join("<br/>", $links);
  }

  /**
   * @param string[] $files
   * @return string
   */
  private static function renderBreadcrumb(
    array $files,
    string $fileName
  ) {
    $imax = count($files) - 1;
    $i = array_search($fileName, $files);
    $previous = false;
    $previousUrl = '';
    $next = false;
    $nextUrl = '';
    if (is_int($i)) {
      if ($i > 0) {
        $previousDate = monitoringLogFileDateFromPath($files[$i - 1]);
        if ($previousDate !== null) {
          $previous = true;
          $previousUrl = self::fileUrl($previousDate);
        }
      }
      if ($i < $imax) {
        $nextDate = monitoringLogFileDateFromPath($files[$i + 1]);
        if ($nextDate !== null) {
          $next = true;
          $nextUrl = self::fileUrl($nextDate);
        }
      }
    }

    return str(
      $previous ? renderHtmlLink($previousUrl, "< previous") : '',
      $previous ? ' - ' : '',
      renderHtmlLink(self::indexUrl(), self::indexPath()),
      "/$fileName",
      $next ? ' - ' : '',
      $next ? renderHtmlLink($nextUrl, "next >") : '',
    );
  }

  /** @return string */
  private static function renderBody() {
    try {
      Auth::verifyAdminCapability();
    } catch (\Throwable $err) {
      $message = $err->getMessage();
      $trace = isDebugModeEnabled() ? $err->getTraceAsString() : "";
      return "$message $trace";
    }

    $files = monitoringDirFiles();
    $fileDate = getQueryVar(self::LOG_FILE_DATE_QUERY_VAR_NAME);
    $dateParam = is_string($fileDate);
    $fileName = $dateParam ? monitoringLogFileBasename($fileDate) : '';
    $actualFile = in_array($fileName, $files);
    if (!$actualFile) return str(
      $dateParam ? "No log files found for $fileDate.<br/>" : "",
      self::renderIndex($files)
    );

    $fileContent = file_get_contents(path_join(monitoringDirPath(), $fileName));
    if (!$fileContent) return "Can't read the log file $fileName";
    $breadcrumb = self::renderBreadcrumb($files, $fileName);
    return str(
      "$breadcrumb<hr />",
      str_replace("\n", "<br/>", $fileContent),
      "<hr />$breadcrumb",
    );
  }

  /** @return string */
  public static function render() {
    return renderHtmlPage(renderNoIndexHtmlPageHead(), self::renderBody());
  }
}
