<?php

namespace Wino;

interface ConfigurationPriceListTranslator {
  /** @return string */
  public function configurationPriceListLabel();

  /** @return string */
  public function missingRequiredPriceListConfiguration();

  /** @return string */
  public function apiKeyAndShopIdAreRequiredToSelectPriceList();

  /** @return string */
  public function noPriceListSelectedOptionLabel();
}

class ConfigurationPriceList {
  public const OPTION = 'wino-price-list';
  public const DEFAULT = '';

  /** @return string */
  public static function label(ConfigurationPriceListTranslator $translator) {
    return $translator->configurationPriceListLabel();
  }

  /** @return string */
  private static function missingSelectOptionsNotice(ConfigurationPriceListTranslator $translator) {
    return $translator->apiKeyAndShopIdAreRequiredToSelectPriceList();
  }

  /** @return string */
  public static function value() {
    return textConfigurationValue(self::OPTION, self::DEFAULT);
  }

  /** @return void */
  public static function updateValue(string $value) {
    ConfigurationOption::fromTextValue(self::OPTION, $value, true)->savePostedText();
  }

  private const ID_OPTION_PREFIX = 'wino-price-list-id-';
  /** @return string */
  private static function priceIdOptionSlug(string $priceName) {
    return self::ID_OPTION_PREFIX . $priceName;
  }

  /** @return null|string */
  private static function getPriceIdByName(string $priceName) {
    $optionSlug = self::priceIdOptionSlug($priceName);
    $option = get_option($optionSlug, '');
    if (!is_string($option)) return null;
    $trim = trim($option);
    return $trim === "" ? null : $trim;
  }

  /** @return void */
  private static function saveId(string $name, string $id) {
    $optionSlug = self::priceIdOptionSlug($name);
    $trim = trim($id);
    if ($trim === "") delete_option($optionSlug);
    else update_option($optionSlug, $trim);
  }

  /** @return void */
  private static function savePriceId(WinoApiPrice $price) {
    self::saveId($price->name, $price->id);
  }

  /** @return null|string */
  public static function id() {
    return self::getPriceIdByName(self::value());
  }

  /** @return void */
  public static function save() {
    saveTextConfigurationValue(self::OPTION);
  }

  /** @return SelectOption[] */
  public static function selectOptions(Log $log, WinoApiClient $winoApi) {
    $options = [];
    if (!$winoApi->isConfigured()) return $options;
    $shop = ConfigurationShopId::shopIdentity();
    if ($shop === null) return $options;

    $prices = [];
    try {
      $prices = WinoApiPrice::queryAll($winoApi, $shop);
    } catch (\Throwable $thrown) {
      debugLog($thrown);
      $log->error($thrown);
    }

    foreach ($prices as $price) {
      $name = $price->name;
      $options[] = SelectOption::from($name);
      self::savePriceId($price);
    }
    return $options;
  }

  /** @return ConfigurationFieldMode */
  public static function mode() {
    return ConfigurationFieldMode::adminSave()->disableAjax();
  }

  /** @return SelectOption */
  private static function nothingSelectedOption(ConfigurationPriceListTranslator $translator) {
    return new SelectOption(self::DEFAULT, $translator->noPriceListSelectedOptionLabel());
  }

  /** @return string */
  public static function renderInput(
    Log $log,
    WinoApiClient $winoApi,
    ConfigurationPriceListTranslator $translator
  ) {
    return renderSelectInput(
      self::OPTION,
      self::value(),
      self::mode(),
      self::nothingSelectedOption($translator),
      self::missingSelectOptionsNotice($translator),
      ...self::selectOptions($log, $winoApi)
    );
  }

  /** @return Render */
  public static function select(
    Log $log,
    WinoApiClient $winoApi,
    ConfigurationPriceListTranslator $translator,
    bool $showLabel = false
  ) {
    return configurationSelectRender(
      $showLabel ? self::label($translator) : "",
      self::OPTION,
      self::value(),
      self::mode(),
      self::nothingSelectedOption($translator),
      self::missingSelectOptionsNotice($translator),
      ...self::selectOptions($log, $winoApi)
    );
  }

  /** @return void */
  public static function verify(ConfigurationPriceListTranslator $translator) {
    $trimed = trim(self::value());
    if (strlen($trimed) > 0) return;
    throw msgException($translator->missingRequiredPriceListConfiguration());
  }
}
