<?php

namespace Wino;

interface ConfigurationRootCategoryTranslator {
  /** @return string */
  public function rootCategoryLabel();

  /** @return string */
  public function noRootCategoryOptionLabel();

  /** @return string */
  public function defaultWinoRootCategoryOptionLabel();

  /** @return string */
  public function invalidRootCategoryErrorMessage(\WP_Term $root);
}

/** @return \Error */
function invalidRootCategoryError(
  ConfigurationRootCategoryTranslator $translator,
  \WP_Term $root
) {
  return error($translator->invalidRootCategoryErrorMessage($root));
}

class CustomRootCategory {
  private ?int $termId;
  private ?\WP_Term $term;
  private bool $notFound;
  private bool $invalid;

  private function __construct(
    ?int $termId,
    ?\WP_Term $term,
    bool $notFound,
    bool $invalid
  ) {
    $this->termId = $termId;
    $this->term = $term;
    $this->notFound = $notFound;
    $this->invalid = $invalid;
  }

  /** @return int|null */
  public function getTermId() {
    return $this->termId;
  }

  /** @return \WP_Term|null */
  public function getTerm() {
    return $this->term;
  }

  /** @return string|null */
  public function getTermName() {
    $term = $this->getTerm();
    return $term === null ? null : $term->name;
  }

  /** @return bool */
  public function found() {
    return !$this->notFound;
  }

  /** @return bool */
  public function valid() {
    return !$this->invalid;
  }

  /** @return self|null */
  public static function fromTermId(?int $termId) {
    $invalid = true;
    $valid = !$invalid;
    $notFound = true;
    $found = !$notFound;

    $id = self::termIdValue($termId);
    if ($id === null) return null;
    if ($id === undefinedTermId()) return new self($termId, null, $found, $valid);
    $term = findWordpressTermById($id);
    if ($term === null) return new self($termId, $term, $notFound, $valid);
    if (DataScope::termIsFromAnyScope($term)) return new self($termId, $term, $found, $invalid);
    return new self($termId, $term, $found, $valid);
  }

  /** @return int|null */
  public static function termIdValue(?int $termId) {
    return $termId === null || $termId < undefinedTermId() ? null : $termId;
  }

  /** @return WinoCategory|null */
  public function asWinoCategory(DataScope $scope) {
    $term = $this->getTerm();
    return $term === null ? null : new WinoCategory($scope, $term);
  }

  /** @return void */
  public static function verify(
    ConfigurationRootCategoryTranslator $translator,
    ?int $termId
  ) {
    $customRoot = CustomRootCategory::fromTermId($termId);
    if ($customRoot !== null) $customRoot->assertsValidity($translator);
  }

  /** @return void */
  public function assertsValidity(ConfigurationRootCategoryTranslator $translator) {
    if ($this->valid()) return;
    $term = $this->getTerm();
    if ($term !== null) throw invalidRootCategoryError($translator, $term);
  }
}

class ConfigurationRootCategory {
  public const OPTION = 'wino-root-category';
  public const DEFAULT = -1;

  /** @return string */
  public static function label(ConfigurationRootCategoryTranslator $translator) {
    return $translator->rootCategoryLabel();
  }

  /** @return void */
  public static function save() {
    saveIntegerConfigurationValue(self::OPTION);
  }

  /** @return int */
  public static function value() {
    return integerConfigurationValue(self::OPTION, self::DEFAULT);
  }

  /** @return int */
  private static function compareSelectOptionsTerms(
    WordpressTermsTree $a,
    WordpressTermsTree $b,
    int $defaultCategoryTermId
  ) {
    $aId = $a->getTermId();
    $bId = $b->getTermId();
    if ($aId === $defaultCategoryTermId) return -1;
    if ($bId === $defaultCategoryTermId) return 1;
    $nameSort = strcmp($a->getTermName(), $b->getTermName());
    if ($nameSort !== 0) return $nameSort;
    return $aId - $bId;
  }

  /**
   * @param WordpressTermsTree[] $trees
   *
   * @return WordpressTermsTree[]
   */
  private static function sortCategoryTrees(array $trees, int $defaultCategoryTermId) {
    return sortArrayCopy(
      $trees,
      function (WordpressTermsTree $a, WordpressTermsTree $b) use ($defaultCategoryTermId) {
        return self::compareSelectOptionsTerms($a, $b, $defaultCategoryTermId);
      }
    );
  }

  /** @return WordpressTermsTree[] */
  private static function findNonWinoProductCategoryTerms() {
    $metaQueries = [];
    $metaKeys = DataScope::findDistinctScopeWinoIdMetaKeys();
    foreach ($metaKeys as $metaKey) $metaQueries[] = metaQueryExists($metaKey, false);
    return WordpressTermsTree::fromTaxonomy(WinoCategory::TAXONOMY, ...$metaQueries);
  }

  /**
   * @param SelectOption[] $options
   * @param WordpressTermsTree[] $trees
   * @return void
   */
  private static function buildSelectOptionsFromTrees(
    array &$options,
    array $trees,
    int $defaultCategoryTermId,
    string $labelPrefix = ''
  ) {
    $sorted = self::sortCategoryTrees($trees, $defaultCategoryTermId);
    foreach ($sorted as $tree) {
      $options[] = new SelectOption($tree->getTermId(), trim($labelPrefix . ' ' . $tree->getTermName()));
      self::buildSelectOptionsFromTrees(
        $options,
        $tree->getChildren(),
        $defaultCategoryTermId,
        $labelPrefix . '-'
      );
    }
  }

  /** @return SelectOption[] */
  public static function selectOptions(
    ConfigurationRootCategoryTranslator $translator
  ) {
    $options = [
      new SelectOption(undefinedTermId(), $translator->noRootCategoryOptionLabel()),
      new SelectOption(self::DEFAULT, $translator->defaultWinoRootCategoryOptionLabel())
    ];
    self::buildSelectOptionsFromTrees(
      $options,
      self::findNonWinoProductCategoryTerms(),
      WinoCategory::defaultProductCategoryTermId(),
    );
    return $options;
  }

  /** @return Render */
  public static function select(
    ConfigurationRootCategoryTranslator $translator
  ) {
    $options = self::selectOptions($translator);
    $value = self::value();
    return configurationSelectRender(
      self::label($translator),
      self::OPTION,
      "$value",
      ConfigurationFieldMode::adminSave(),
      null,
      '',
      ...$options
    );
  }

  /** @return void */
  public static function verify(ConfigurationRootCategoryTranslator $translator) {
    CustomRootCategory::verify($translator, self::value());
  }
}
