<?php

namespace Wino;

interface ConfigurationShopIdTranslator {
  /** @return string */
  public function configurationShopIdLabel();

  /** @return string */
  public function missingRequiredShopIdConfiguration();

  /** @return string */
  public function apiKeyIsRequiredToSelectShop();
}

class ConfigurationShopId {
  public const OPTION = 'wino-shop-id';
  public const DEFAULT = '';

  /** @return string */
  public static function label(ConfigurationShopIdTranslator $translator) {
    return $translator->configurationShopIdLabel();
  }

  /** @return string */
  public static function missingSelectOptionsNotice(ConfigurationShopIdTranslator $translator) {
    return $translator->apiKeyIsRequiredToSelectShop();
  }

  public static ?string $autosaved = null;

  /** @return string */
  public static function value() {
    if (self::$autosaved !== null) return self::$autosaved;
    return textConfigurationValue(self::OPTION, self::DEFAULT);
  }

  /** @return WinoApiShopIdentity|null */
  public static function shopIdentity() {
    $id = trim(self::value());
    if ($id === "") return null;
    return new WinoApiShopIdentity($id);
  }

  /** @return WinoApiShopIdentity */
  public static function defaultShopIdentity() {
    return new WinoApiShopIdentity(self::DEFAULT);
  }

  /** @return void */
  public static function save() {
    saveTextConfigurationValue(self::OPTION);
  }

  /** @return SelectOption[] */
  public static function selectOptionsWithSingleOptionAutosave(
    Log $log,
    WinoApiClient $winoApi,
    bool $allowAutosave = true
  ) {
    self::$autosaved = null;
    $options = [];
    if (!$winoApi->isConfigured()) return $options;

    $shops = [];
    try {
      $shops = WinoApiShop::queryAll($winoApi);
    } catch (\Throwable $thrown) {
      debugLog($thrown);
      $log->error($thrown);
    }

    foreach ($shops as $shop) $options[] = new SelectOption($shop->id, $shop->name);

    if (!$allowAutosave) return $options;

    $singleOption = singleOptionFrom(...$options);
    if ($singleOption === null) return $options;

    self::$autosaved = ConfigurationOption::fromTextValue(
      self::OPTION,
      $singleOption->value,
      $allowAutosave
    )->savePostedText();
    return $options;
  }

  /** @return SelectOption[] */
  public static function selectOptions(Log $log, WinoApiClient $winoApi) {
    return self::selectOptionsWithSingleOptionAutosave($log, $winoApi, false);
  }

  /** @return ConfigurationFieldMode */
  public static function mode() {
    return ConfigurationFieldMode::adminSave();
  }

  /** @return string */
  public static function renderInput(
    Log $log,
    WinoApiClient $winoApi,
    ConfigurationShopIdTranslator $translator,
    bool $allowAutosave = false
  ) {
    $options = self::selectOptionsWithSingleOptionAutosave($log, $winoApi, $allowAutosave);
    return renderSelectInput(
      self::OPTION,
      self::value(),
      self::mode(),
      self::nothingSelectedOption(),
      self::missingSelectOptionsNotice($translator),
      ...$options
    );
  }

  /** @return SelectOption */
  private static function nothingSelectedOption() {
    return new SelectOption(self::DEFAULT, '');
  }

  /** @return string */
  public static function renderInputWithSingleOptionAutosave(
    Log $log,
    WinoApiClient $winoApi,
    ConfigurationShopIdTranslator $translator
  ) {
    $allowAutosave = true;
    return self::renderInput($log, $winoApi, $translator, $allowAutosave);
  }

  /** @return Render */
  public static function select(
    Log $log,
    WinoApiClient $winoApi,
    ConfigurationShopIdTranslator $translator,
    bool $allowAutosave = false
  ) {
    $options = self::selectOptionsWithSingleOptionAutosave($log, $winoApi, $allowAutosave);
    return configurationSelectRender(
      self::label($translator),
      self::OPTION,
      self::value(),
      self::mode(),
      self::nothingSelectedOption(),
      self::missingSelectOptionsNotice($translator),
      ...$options
    );
  }

  /** @return Render */
  public static function selectWithSingleOptionAutosave(
    Log $log,
    WinoApiClient $winoApi,
    ConfigurationShopIdTranslator $translator
  ) {
    $allowAutosave = true;
    return self::select($log, $winoApi, $translator, $allowAutosave);
  }

  /** @return void */
  public static function verify(ConfigurationShopIdTranslator $translator) {
    $trimed = trim(self::value());
    if ($trimed !== "") return;
    throw exception($translator->missingRequiredShopIdConfiguration());
  }
}
