<?php

namespace Wino;

class WinoCategory {
  public const TAXONOMY = "product_cat";
  private DataScope $scope;

  public \WP_Term $term;

  /** @return DataScope */
  public function getScope() {
    return $this->scope;
  }

  /** @return \WP_Term */
  public function getTerm() {
    return $this->term;
  }

  /** @return int */
  public function getTermId() {
    return $this->term->term_id;
  }

  /** @return WinoCategory|null */
  public function getParentCategory() {
    $parentId = $this->getParentTermId();
    if ($parentId === null) return null;
    $parentTerm = findWordpressTermById($parentId);
    $scope = $this->scope;
    return $parentTerm === null ? null : new self($scope, $parentTerm);
  }

  /**
   * @return WinoCategory[]
   */
  public function getChildren() {
    $scope = $this->scope;
    $children = [];
    $childTerms = findWordpressTermChildrenById($this->getTermId(), self::TAXONOMY);
    foreach ($childTerms as $childTerm) $children[] = new self($scope, $childTerm);
    return $children;
  }

  /** @return WinoCategory[] */
  public function getAncestors() {
    $rootId = $this->scope->winoApiRootCategoryId();
    $ancestors = [];
    $parent = $this->getParentCategory();
    while ($parent !== null) {
      $parentApiId = $parent->getWinoApiCategoryIdMeta();
      if ($parentApiId === null || $parentApiId === $rootId) break;
      $ancestors[] = $parent;
      $parent = $parent->getParentCategory();
    }
    return $ancestors;
  }

  /** @return int|null */
  public function getParentTermId() {
    return definedTermId($this->term->parent);
  }

  /** @return string */
  public function getTaxonomy() {
    return $this->term->taxonomy;
  }

  /** @return string */
  public function getName() {
    return $this->term->name;
  }

  /** @return string */
  public function getSlug() {
    return $this->term->slug;
  }

  public function __construct(
    DataScope $scope,
    \WP_Term $term
  ) {
    $this->scope = $scope;
    $this->term = $term;
  }

  /** @return null|string */
  public function getWinoApiCategoryIdMeta() {
    $termId = $this->getTermId();
    $metaKey = $this->scope->winoIdMetaKey();
    return findWordpressTermMetaString($termId, $metaKey);
  }

  /** @return bool */
  public function isWinoApiCategoryIdMetaValid() {
    return $this->getWinoApiCategoryIdMeta() !== null;
  }

  /** @return null|string */
  public function getWinoApiParentCategoryId() {
    $scope = $this->scope;

    $parentTermId = $this->getParentTermId();
    if ($parentTermId === null) return null;

    $parentCategoryId = findWordpressTermMetaString($parentTermId, $scope->winoIdMetaKey());
    if ($parentCategoryId !== $scope->winoApiRootCategoryId()) return $parentCategoryId;
    return findWordpressTermMetaString($this->getTermId(), $scope->winoParentCategoryIdMetaKey());
  }

  /** @return bool */
  public function isRoot() {
    $idMeta = $this->getWinoApiCategoryIdMeta();
    return $idMeta === $this->scope->winoApiRootCategoryId();
  }

  public const DEFAULT_PRODUCT_CATEGORY_TERM_ID_OPTION = 'default_product_cat';
  private static ?int $defaultProductCategoryTermIdCache = null;

  /** @return int */
  public static function defaultProductCategoryTermId() {
    if (self::$defaultProductCategoryTermIdCache !== null) return self::$defaultProductCategoryTermIdCache;
    $termId = assertedNullableStringOrInt(get_option(self::DEFAULT_PRODUCT_CATEGORY_TERM_ID_OPTION, null));
    if ($termId === null) throw exception("Default product category not found.");
    $termIdAsInt = parseInt($termId);
    self::$defaultProductCategoryTermIdCache = $termIdAsInt;
    return $termIdAsInt;
  }

  /** @return void */
  public static function updateDefaultProductCategoryTermId(int $termId) {
    self::$defaultProductCategoryTermIdCache = null;
    update_option(self::DEFAULT_PRODUCT_CATEGORY_TERM_ID_OPTION, $termId);
  }

  /**
   * Fetching the uncategorized product category term might
   * return null since the slug might be removed or different
   * according to the current wordpress instance locale.
   *
   * This function mostly has sense for developement purposes
   * and should not be used otherwise.
   *
   * @return \WP_Term|null
   */
  public static function uncategorizedProductCategoryTerm() {
    return findWordpressTerm(self::TAXONOMY, 'slug', 'uncategorized');
  }

  /**
   * Fetching the uncategorized product category term id might
   * return null since the slug might be removed or different
   * according to the current wordpress instance locale.
   *
   * This function mostly has sense for developement purposes
   * and should not be used otherwise.
   *
   * @return int|null
   */
  public static function uncategorizedProductCategoryTermId() {
    $term = self::uncategorizedProductCategoryTerm();
    return $term === null ? null : $term->term_id;
  }
}
