<?php

namespace Wino;

/** @return \WC_Order[] */
function getWcOrders(array $args = []) {
  $orders = \wc_get_orders($args);
  assertsArray($orders);
  return $orders;
}

class WooCommerceOrderProductItem {
  private DataScope $scope;
  private \WC_Order_Item_Product $productItem;

  public int $id;
  public int $quantity;
  public int $productId;

  private function __construct(
    DataScope $scope,
    \WC_Order_Item_Product $productItem
  ) {
    $this->scope = $scope;
    $this->productItem = $productItem;
    $this->id = $productItem->get_id();
    $this->quantity = $productItem->get_quantity();
    $this->productId = $productItem->get_product_id();
  }

  /** @return self */
  public static function from(
    DataScope $scope,
    \WC_Order_Item_Product $productItem
  ) {
    return new self($scope, $productItem);
  }

  /** @return self|null */
  public static function fromItem(
    DataScope $scope,
    \WC_Order_Item $item
  ) {
    return $item instanceof \WC_Order_Item_Product ? self::from($scope, $item) : null;
  }

  /** @return void */
  public function saveMeta(string $metaKey, string $metaValue, bool $unique = true) {
    $item = $this->productItem;
    $item->add_meta_data($metaKey, $metaValue, $unique);
    $item->save_meta_data();
  }

  /** @return mixed */
  public function getMeta(string $metaKey, bool $single = true) {
    $item = $this->productItem;
    return $item->get_meta($metaKey, $single);
  }

  /** @return void */
  public function saveWinoApiStockActivityIdMeta(string $id) {
    $key = $this->scope->winoIdMetaKey();
    $this->saveMeta($key, $id);
  }

  /** @return null|string */
  public function getWinoApiStockActivityIdMeta() {
    $key = $this->scope->winoIdMetaKey();
    /** @var mixed */
    $meta = $this->getMeta($key);
    $value = optionalStringData([$key => $meta], $key);
    if ($value === null) return null;
    return trim($value) === "" ? null : $value;
  }

  /** @return bool */
  public function isWinoApiStockActivityIdMetaValid() {
    return $this->getWinoApiStockActivityIdMeta() !== null;
  }

  /** @return array<string, mixed> */
  public function asArray() {
    return [
      'id' => $this->id,
      'quantity' => $this->quantity,
      'productId' => $this->productId
    ];
  }

  /** @return string */
  public function getWinoApiStockActivityId() {
    $meta = $this->getWinoApiStockActivityIdMeta();
    if ($meta !== null) return $meta;
    throw msgException("Missing Wino API stock activity id meta on:", jsonString($this->asArray()));
  }
}

class WinoStockMovement {
  private DataScope $scope;
  private WooCommerceOrderProductItem $orderProductItem;
  public string $variantId;
  public int $quantity;
  public string $comment;
  public string $kind;

  public function __construct(
    DataScope $scope,
    WooCommerceOrderProductItem $orderProductItem,
    string $variantId,
    int $quantity,
    string $comment,
    string $kind
  ) {
    $this->scope = $scope;
    $this->orderProductItem = $orderProductItem;
    $this->variantId = $variantId;
    $this->quantity = $quantity;
    $this->comment = $comment;
    $this->kind = $kind;
  }

  /** @return void */
  private function saveStockActivity(WinoApiStockActivity $stockActivity) {
    $item = $this->orderProductItem;
    $item->saveWinoApiStockActivityIdMeta($stockActivity->id);
  }

  /** @return bool */
  private function isStockActivitySaved() {
    $item = $this->orderProductItem;
    return $item->isWinoApiStockActivityIdMetaValid();
  }

  /** @return void */
  public function createStockActivity(
    WebDeviceCache $cache,
    WinoApiClient $apiClient,
    WinoApiShopIdentity $shop
  ) {
    if ($this->isStockActivitySaved()) return;
    $this->saveStockActivity(createStockActivity(
      $cache,
      $apiClient,
      $shop,
      $this->variantId,
      $this->quantity,
      $this->comment,
      $this->kind
    ));
  }
}

/**
 * The following woocommerce documentation section can be
 * helpful to understand what does each woocommerce order status:
 * https://woocommerce.com/document/managing-orders/#order-statuses
 *
 * The only status actually required by the plugin the completed one.
 * The processing status is currently only used by the test suite to
 * ensure that only the completion triggers the creation of a stock activity.
 */
class WooCommerceOrder {
  public const PROCESSING_STATUS = 'processing';
  public const COMPLETED_STATUS = 'completed';

  private DataScope $scope;

  public \WC_Order $order;

  private function __construct(
    DataScope $scope,
    \WC_Order $order
  ) {
    $this->scope = $scope;
    $this->order = $order;
  }

  /** @return self */
  public static function from(
    DataScope $scope,
    \WC_Order $order
  ) {
    return new self($scope, $order);
  }

  /** @return self|null */
  public static function fromId(DataScope $scope, int $orderId) {
    $order = wc_get_order($orderId);
    if ($order instanceof \WC_Order) return self::from($scope, $order);
    return null;
  }

  /** @return string */
  private static function normalizeStatus(string $status) {
    return strtolower(trim($status));
  }

  /** @return bool */
  private static function statusEqual(string $statusA, string $statusB) {
    return self::normalizeStatus($statusA) === self::normalizeStatus($statusB);
  }

  /** @return bool */
  private function isStatus(string $status) {
    return self::statusEqual($this->order->get_status(), $status);
  }

  /** @return bool */
  public function isCompleted() {
    return $this->isStatus(self::COMPLETED_STATUS);
  }

  /** @return int */
  public function getId() {
    return $this->order->get_id();
  }

  /** @return array */
  public function getOrderData() {
    return $this->order->get_data();
  }

  /** @return int|string */
  public function getItemsCount() {
    return $this->order->get_item_count();
  }

  /** @return bool */
  public function isEmpty() {
    return $this->getItemsCount() === 0;
  }

  /** @return \WC_Order_Item[] */
  public function getItems() {
    $array = [];
    $items = $this->order->get_items();
    foreach ($items as $item) $array[] = $item;
    return $array;
  }

  /** @return array[] */
  public function getItemsData() {
    $data = [];
    $items = $this->getItems();
    foreach ($items as $item) $data[] = $item->get_data();
    return $data;
  }

  /** @return WooCommerceOrderProductItem[] */
  public function getProductItems() {
    $products = [];
    $scope = $this->scope;
    $items = $this->getItems();
    foreach ($items as $item) {
      $product = WooCommerceOrderProductItem::fromItem($scope, $item);
      if ($product !== null) $products[] = $product;
    }
    return $products;
  }

  /** @return WinoStockMovement[] */
  public function getWinoSaleStockMovements() {
    $movements = [];
    if (!$this->isCompleted() || $this->isEmpty()) return $movements;

    $orderId = $this->getId();
    $scope = $this->scope;
    $productItems = $this->getProductItems();

    foreach ($productItems as $item) {
      $itemId = $item->id;
      $quantity = $item->quantity;
      $productId = $item->productId;
      $winoProduct = WinoProduct::fromPostId($scope, $productId);
      $variantId = $winoProduct === null ? null : $winoProduct->getWinoApiVariantId();
      if ($variantId === null) continue;

      $movements[] = new WinoStockMovement(
        $scope,
        $item,
        $variantId,
        $quantity,
        "WooCommerce order($orderId) item($itemId)",
        WinoApiStockActivity::SALE_KIND
      );
    }

    return $movements;
  }

  /** @return array<string, mixed> */
  public function asArray() {
    return [
      'id' => $this->getId(),
      'items' => $this->getItemsData()
    ];
  }

  public const STATUS_CHANGED_HOOK = 'woocommerce_order_status_changed';

  /**
   * @param Closure(int $orderId, string $oldStatus, string $newStatus): void $callback
   * @return void
   */
  public static function addStatusChangedAction(callable $callback, int $priority) {
    /** @psalm-suppress MissingClosureParamType */
    add_action(self::STATUS_CHANGED_HOOK, function ($orderId, $oldStatus, $newStatus) use ($callback) {
      if (!is_int($orderId)) throw msgError("Unexpected orderId value", jsonString($orderId));
      if (!is_string($oldStatus)) throw msgError("Unexpected oldStatus value", jsonString($oldStatus));
      if (!is_string($newStatus)) throw msgError("Unexpected newStatus value", jsonString($newStatus));
      if (self::statusEqual($newStatus, $oldStatus)) throw msgError(
        "Unexpected unchanged status value", jsonString($oldStatus)
      );
      call_user_func($callback, $orderId, $oldStatus, $newStatus);
    }, $priority, 3);
  }

  /**
   * @param Closure(WooCommerceOrder): void $callback
   * @return void
   */
  public static function addStatusChangedToAction(
    DataScope $scope,
    string $status,
    callable $callback,
    int $priority
  ) {
    self::addStatusChangedAction(
      function ($orderId, $oldStatus, $newStatus) use ($scope, $callback, $status) {
        if (!self::statusEqual($newStatus, $status)) return;
        $order = self::fromId($scope, $orderId);
        if ($order === null) throw exception("Order not found ($orderId)");
        call_user_func($callback, $order);
        noop($oldStatus);
      },
      $priority
    );
  }

  /**
   * @param Closure(WooCommerceOrder): void $callback
   * @return void
   */
  public static function addOrderCompletedAction(
    DataScope $scope,
    callable $callback,
    int $priority
  ) {
    self::addStatusChangedToAction($scope, self::COMPLETED_STATUS, $callback, $priority);
  }
}
