<?php

namespace Wino;

class WinoProductUpdateProxy extends UpdateProxy {
  private \WC_Product $product;
  private DataScope $scope;
  private ?string $skuUpdate;
  private ?string $taxClassSlugUpdate;
  private ?int $customRootCategoryId;

  protected function __construct(
    DataScope $scope,
    \WC_Product $product
  ) {
    parent::__construct();
    $this->scope = $scope;
    $this->product = $product;
    $this->taxClassSlugUpdate = null;
    $this->skuUpdate = null;
    $this->customRootCategoryId = null;
  }

  /** @return self */
  public static function from(
    DataScope $scope,
    \WC_Product $product
  ) {
    return new self($scope, $product);
  }

  /** @return self */
  public static function fromProduct(WinoProduct $product) {
    return self::from($product->getScope(), $product->wooCommerceProduct);
  }

  /** @return self */
  public static function updateOrCreate(
    DataScope $scope,
    ?WinoProduct $existingProduct,
    ?CustomRootCategory $customRoot
  ) {
    $product = $existingProduct === null
    ? WinoProduct::createWooCommerceProduct()
    : $existingProduct->wooCommerceProduct;

    return self::from($scope, $product)->setCustomRootCategory($customRoot);
  }

  /** @return self */
  public function setCustomRootCategory(?CustomRootCategory $customRoot) {
    $this->customRootCategoryId = $customRoot === null ? null : $customRoot->getTermId();
    return $this;
  }

  /** @return WinoProduct */
  public function getProduct() {
    return new WinoProduct($this->scope, $this->product);
  }

  /** @return void */
  public function setStatus(string $status) {
    if ($this->skip($status, $this->product->get_status())) return;
    $this->product->set_status($status);
  }

  /** @return void */
  public function setName(string $name) {
    if ($this->skip($name, $this->product->get_name())) return;
    $this->product->set_name($name);
  }

  /** @return void */
  public function setSku(string $sku) {
    $trim = trim($sku);
    $this->skuUpdate = null;
    if ($this->skip($trim, $this->product->get_sku())) return;
    $this->skuUpdate = $trim;
  }

  /** @return void */
  public function setDescription(string $description) {
    if ($this->skip($description, $this->product->get_description())) return;
    $this->product->set_description($description);
  }

  /** @return void */
  public function setManageStock(bool $manageStock) {
    if ($this->skip($manageStock, $this->product->get_manage_stock())) return;
    $this->product->set_manage_stock($manageStock);
  }

  /** @return void */
  public function setTaxStatus(string $taxStatus) {
    if ($this->skip($taxStatus, $this->product->get_tax_status())) return;
    $this->product->set_tax_status($taxStatus);
  }

  /** @return void */
  public function setTaxClass(WinoTaxClass $taxClass) {
    $slug = $taxClass->slug;
    $this->taxClassSlugUpdate = null;
    if ($this->skip($slug, $this->product->get_tax_class())) return;
    $this->taxClassSlugUpdate = $slug;
  }

  /** @return int[] */
  private function getProductCategoryIds() {
    $ids = [];

    $categoryTermIds = $this->product->get_category_ids();
    /** @var mixed */
    foreach ($categoryTermIds as $id)
    if (is_int($id)) $ids[] = $id;
    else throw msgException("Unexpected non int value", jsonString($id));

    return $ids;
  }

  /** @return bool */
  private function preserveCategoryTermId(
    int $termId,
    ?string $customRootMetaValue
  ) {
    $scope = $this->scope;
    $winoIdMetaKey = $scope->winoIdMetaKey();

    $winoId = findWordpressTermMetaString($termId, $winoIdMetaKey);
    $isCustomRoot = "$termId" === $customRootMetaValue;

    return $winoId === null && !$isCustomRoot;
  }

  /** @return void */
  public function setCategoryTermIds(int $termId, int ...$ancestorTermIds) {
    $scope = $this->scope;

    $defaultTermId = WinoCategory::defaultProductCategoryTermId();

    $newWinoTermIds = array_merge([$termId], $ancestorTermIds);
    $newCategoryTermIds = array_merge($newWinoTermIds);

    $oldWinoTermIds = [];
    $oldCategoryTermIds = $this->getProductCategoryIds();

    $customRootMetaKey = $scope->winoCustomRootCategoryMetaKey();
    $oldCustomRootMetaValue = $this->getProductMetaString($customRootMetaKey);

    foreach ($oldCategoryTermIds as $oldCategoryTermId) {
      if ($oldCategoryTermId === $defaultTermId) {
        $this->updateCount++;
        continue;
      }

      if ($this->preserveCategoryTermId($oldCategoryTermId, $oldCustomRootMetaValue)) {
        $newCategoryTermIds[] = $oldCategoryTermId;
        continue;
      }

      $oldWinoTermIds[] = $oldCategoryTermId;
    }

    sort($oldWinoTermIds);
    sort($newWinoTermIds);

    if ($this->skip($newWinoTermIds, $oldWinoTermIds)) return;
    $this->product->set_category_ids($newCategoryTermIds);

    $customRootTermId = $this->customRootCategoryId;
    $newCustomRootMetaValue = $customRootTermId === null ? null : "$customRootTermId";

    $useNewCustomRoot = false;
    foreach ($newCategoryTermIds as $newCategoryTermId)
    $useNewCustomRoot = $useNewCustomRoot || "$newCategoryTermId" === $newCustomRootMetaValue;

    $this->setMetaData($customRootMetaKey, $useNewCustomRoot ? $newCustomRootMetaValue : null);
  }

  /** @return void */
  public function unsetCategories() {
    $scope = $this->scope;

    $categoryTermIds = $this->getProductCategoryIds();
    $unsettingCount = 0;
    $newCategories = [];

    $customRootMetaKey = $scope->winoCustomRootCategoryMetaKey();
    $customRootMetaValue = $this->getProductMetaString($customRootMetaKey);

    foreach ($categoryTermIds as $categoryTermId) {
      if ($this->preserveCategoryTermId($categoryTermId, $customRootMetaValue)) {
        $newCategories[] = $categoryTermId;
        continue;
      }
      $unsettingCount++;
      $this->updateCount++;
    }
    if ($unsettingCount < 1) return;

    $this->product->set_category_ids($newCategories);
    $this->setMetaData($customRootMetaKey, null);
  }

  /** @return void */
  public function setCategories(WinoCategory $category, WinoCategory ...$ancestors) {
    $termIds = [];
    foreach ($ancestors as $ancestor) $termIds[] = $ancestor->getTermId();
    $this->setCategoryTermIds($category->getTermId(), ...$termIds);
  }

  /** @return void */
  public function setCategoryWithAncestors(WinoCategory $category) {
    $this->setCategories($category, ...$category->getAncestors());
  }

  /** @return void */
  public function setCategoriesHierarchy(bool $includesAncestors, WinoCategory $category) {
    if ($includesAncestors) $this->setCategoryWithAncestors($category);
    else $this->setCategories($category);
  }

  /** @return void */
  public function setPrice(float $price) {
    $value = "$price";
    if ($this->skip($value, $this->product->get_regular_price())) return;
    $this->product->set_regular_price($value);
  }

  /** @return void */
  public function setStockQuantity(int $quantity) {
    if ($this->skip($quantity, $this->product->get_stock_quantity())) return;
    $this->product->set_stock_quantity($quantity);
  }

  /** @return mixed */
  private function getProductMeta(string $key) {
    $meta = $this->product->get_meta($key);
    if (!is_string($meta)) return $meta;
    return trim($meta) === "" ? null : $meta;
  }

  /** @return string|null */
  private function getProductMetaString(string $key) {
    /** @var mixed */
    $meta = $this->getProductMeta($key);
    return is_string($meta) ? $meta : null;
  }

  /** @return void */
  public function setMetaData(string $key, ?string $value) {
    if ($this->skip($value, $this->getProductMeta($key))) return;

    $product = $this->product;
    if ($value === null) $product->delete_meta_data($key);
    else $product->update_meta_data($key, $value);
  }

  /** @return void */
  public function setWinoApiVariantId(string $id) {
    $this->setMetaData($this->scope->winoIdMetaKey(), $id);
  }

  /** @return void */
  public function setWinoApiTaxId(?string $id) {
    $this->setMetaData($this->scope->winoTaxIdMetaKey(), $id);
  }

  /** @return void */
  public function setWinoApiProductId(?string $id) {
    $this->setMetaData($this->scope->winoProductIdMetaKey(), $id);
  }

  /**
   * Due to the following line in woocommerce source code,
   * https://woocommerce.github.io/code-reference/files/woocommerce-includes-abstracts-abstract-wc-data.html#source-view.765
   *
   * The cached values in arrays properties like category_ids are not correct
   * if the previous value contained more elements than the new value.
   *
   * https://github.com/woocommerce/woocommerce/issues/35591
   *
   * @return void
   */
  private static function categoryIdsCacheFix(\WC_Product $product, int $maxLength) {
    $ids = $product->get_category_ids();
    $length = count($ids);
    if ($length <= $maxLength) return;
    $fixedIds = array_slice($ids, 0, $maxLength);
    $product->set_category_ids($fixedIds);
  }

  /**
   * For some reasons, tax class update need to be saved
   * independently, otherwise some bug will occur. For instance
   * woocommerce might not be able to find products by sku.
   *
   * @return void
   */
  private function commitTaxClassFix() {
    $update = $this->taxClassSlugUpdate;
    if ($update === null) return;
    $product = $this->product;
    $product->set_tax_class($update);
    $product->save();
    $this->taxClassSlugUpdate = null;
  }

  /** @return void */
  private function commitSkuFix() {
    $update = $this->skuUpdate;
    if ($update === null) return;
    $product = $this->product;
    $taxClass = $product->get_tax_class();
    /** @psalm-suppress NullArgument */
    $product->set_tax_class(null); // No idea why but sku is simply ignored if a tax class is set...
    $product->set_sku($update);
    $product->save();
    $product->set_tax_class($taxClass);
    $product->save();

    $this->skuUpdate = null;
  }

  protected function prepareCommit() {
    $product = $this->product;

    $categoriesCount = count($product->get_category_ids());
    if ($categoriesCount === 0) {
      $ids = [WinoCategory::defaultProductCategoryTermId()];
      $categoriesCount = count($ids);
      $product->set_category_ids($ids);
    }

    $this->commitCategoriesCount = $categoriesCount;
  }

  private int $commitCategoriesCount = 0;

  protected function applyCommit() {
    $product = $this->product;
    $product->save();
    $this->commitSkuFix();
    $this->commitTaxClassFix();

    self::categoryIdsCacheFix($product, $this->commitCategoriesCount);
  }
}
