<?php

namespace Wino;

class WinoProduct {
  public const POST_TYPE = 'product';
  public const NONE_POST_ID = 0;

  public const TAXABLE_TAX_STATUS = 'taxable';
  public const SHIPPING_TAX_STATUS = 'shipping';
  public const NONE_TAX_STATUS = 'none';

  private DataScope $scope;

  /** @return null|string */
  public function getEditionPageUrl() {
    $id = $this->getPostId();
    return $id === 0 ? null : site_url("wp-admin/post.php?post=$id&action=edit");
  }

  /** @return null|string */
  public function getWinoPageUrl() {
    $id = $this->getWinoApiVariantId();
    return $id === null ? null : winoAppUrl("catalog/redirect/$id");
  }

  /** @return DataScope */
  public function getScope() {
    return $this->scope;
  }
  public \WC_Product $wooCommerceProduct;

  /** @return \WC_Product */
  public static function createWooCommerceProduct() {
    return new \WC_Product_Simple();
  }

  public function __construct(
    DataScope $scope,
    \WC_Product $wooCommerceProduct
  ) {
    $this->scope = $scope;
    $this->wooCommerceProduct = $wooCommerceProduct;
  }

  /** @return WinoProduct[] */
  public static function array(WinoProduct ...$products) {
    return $products;
  }

  /** @return self */
  public static function fromWcProduct(DataScope $scope, \WC_Product $product) {
    return new self($scope, $product);
  }

  /**
   * @param \WP_Post|int $input
   * @return self|null
   */
  private static function fromWooCommerce(DataScope $scope, $input) {
    $product = wc_get_product($input);
    if ($product === false || $product === null) return null;
    return self::fromWcProduct($scope, $product);
  }

  /** @return self|null */
  public static function fromPost(DataScope $scope, \WP_Post $post) {
    return self::fromWooCommerce($scope, $post);
  }

  /** @return self|null */
  public static function fromPostId(DataScope $scope, int $postId) {
    if ($postId === self::NONE_POST_ID) return null;
    return self::fromWooCommerce($scope, $postId);
  }

  /** @return self|null */
  public static function fromSku(DataScope $scope, string $sku) {
    $postId = wc_get_product_id_by_sku(trim($sku));
    return self::fromPostId($scope, $postId);
  }

  /** @return string */
  public function getName() {
    return $this->wooCommerceProduct->get_name();
  }

  /** @return null|string */
  public function getSku() {
    $sku = $this->wooCommerceProduct->get_sku();
    return trim($sku) === "" ? null : $sku;
  }

  /** @return string */
  public function getDescription() {
    return $this->wooCommerceProduct->get_description();
  }

  /** @return bool */
  public function getManageStock() {
    return $this->wooCommerceProduct->get_manage_stock();
  }

  /** @return string */
  public function getStatus() {
    return $this->wooCommerceProduct->get_status();
  }

  /** @return string */
  public function getTaxClassSlug() {
    return $this->wooCommerceProduct->get_tax_class();
  }

  /** @return string */
  public function getTaxStatus() {
    return $this->wooCommerceProduct->get_tax_status();
  }

  /** @return bool */
  public function isTaxable() {
    return $this->wooCommerceProduct->is_taxable();
  }

  /** @return bool */
  public function isShippingTaxable() {
    return $this->wooCommerceProduct->is_shipping_taxable();
  }

  /** @return int */
  public function getPostId() {
    return $this->wooCommerceProduct->get_id();
  }

  /**
   * @return null|string
   */
  private function getWinoIdMeta(string $key) {
    /** @var mixed */
    $meta = $this->wooCommerceProduct->get_meta($key);
    $value = optionalStringData([$key => $meta], $key);
    if ($value === null) return null;
    return trim($value) === "" ? null : $value;
  }

  /** @return null|string */
  public function getWinoApiVariantId() {
    return $this->getWinoIdMeta($this->scope->winoIdMetaKey());
  }

  /** @return null|string */
  public function getWinoApiTaxId() {
    return $this->getWinoIdMeta($this->scope->winoTaxIdMetaKey());
  }

  /** @return null|string */
  public function getWinoApiProductId() {
    return $this->getWinoIdMeta($this->scope->winoProductIdMetaKey());
  }

  /** @return int[] */
  public function getCategoryIds() {
    return assertedIntsArray($this->wooCommerceProduct->get_category_ids());
  }

  /** @return bool */
  public function isOnSale() {
    return $this->wooCommerceProduct->is_on_sale();
  }

  /** @return int|null */
  public function getSaleStartTimestamp() {
    $date = $this->wooCommerceProduct->get_date_on_sale_from();
    if ($date === null) return null;
    return date_timestamp_get($date);
  }

  /** @return int|null */
  public function getSaleEndTimestamp() {
    $date = $this->wooCommerceProduct->get_date_on_sale_to();
    if ($date === null) return null;
    return date_timestamp_get($date);
  }

  /**
   * @param string|int|float $price
   * @return null|string
   */
  private static function priceFrom($price) {
    $string = "$price";
    return trim($string) === "" ? null : $string;
  }

  /** @return null|string */
  public function getSalePrice() {
    return self::priceFrom($this->wooCommerceProduct->get_sale_price());
  }

  /** @return null|string */
  public function getPrice() {
    return self::priceFrom($this->wooCommerceProduct->get_price());
  }

  /** @return null|string */
  public function getPriceIncludingTax() {
    return self::priceFrom(wc_get_price_including_tax($this->wooCommerceProduct));
  }

  /** @return null|string */
  public function getPriceExcludingTax() {
    return self::priceFrom(wc_get_price_excluding_tax($this->wooCommerceProduct));
  }

  /** @return string */
  public function getStockStatus() {
    return $this->wooCommerceProduct->get_stock_status();
  }

  /** @return int|null */
  public function getStockQuantity() {
    return $this->wooCommerceProduct->get_stock_quantity();
  }

  /** @return array<string, mixed> */
  public function asArray() {
    return [
      "postId" => $this->getPostId(),
      "name" => $this->getName(),
      "sku" => $this->getSku(),
      "description" => $this->getDescription(),
      "manageStock" => $this->getManageStock(),
      "status" => $this->getStatus(),
      "taxClassSlug" => $this->getTaxClassSlug(),
      "taxStatus" => $this->getTaxStatus(),
      "taxable" => $this->isTaxable(),
      "shippingTaxable" => $this->isShippingTaxable(),
      "winoApiVariantId" => $this->getWinoApiVariantId(),
      "categoryIds" => $this->getCategoryIds(),
      "onSale" => $this->isOnSale(),
      "price" => $this->getPrice(),
      "priceExcludingTax" => $this->getPriceExcludingTax(),
      "priceIncludingTax" => $this->getPriceIncludingTax(),
      "stockStatus" => $this->getStockStatus(),
      "stockQuantity" => $this->getStockQuantity(),
    ];
  }
}
