<?php

namespace Wino;

class WinoTaxKeys {
  public string $taxClassName;
  public string $taxClassSlug;
  public string $taxRateName;

  private function __construct(
    string $taxClassName,
    string $taxClassSlug,
    string $taxRateName
  ) {
    $this->taxClassName = $taxClassName;
    $this->taxClassSlug = $taxClassSlug;
    $this->taxRateName = $taxRateName;
  }

  private const NAME_PREFIX = "Wino - ";

  /** @return string */
  public static function namePrefix(DataScope $scope) {
    return $scope->scopedString(self::NAME_PREFIX);
  }

  /** @return string */
  public static function getTaxClassWinoName(DataScope $scope, string $taxClassName) {
    $prefix = self::namePrefix($scope);
    $suffix = self::extractValueSuffix($taxClassName);

    $withoutPrefix = strStartsWith($taxClassName, $prefix)
    ? substr($taxClassName, strlen($prefix))
    : $taxClassName;

    $withoutSuffix = is_string($suffix) && strEndsWith($withoutPrefix, $suffix)
    ? substr($withoutPrefix, 0, strlen($withoutPrefix) - strlen($suffix))
    : $withoutPrefix;

    return $withoutSuffix;
  }

  private const SLUG_PREFIX = "wino-";

  /** @return string */
  public static function slugPrefix(DataScope $scope) {
    return $scope->scopedSlug(self::SLUG_PREFIX);
  }

  /** @return bool */
  public static function isWinoTaxClassSlug(DataScope $scope, string $slug) {
    return strStartsWith($slug, self::slugPrefix($scope));
  }

  private const VALUE_SUFFIX_START = " (";
  private const VALUE_SUFFIX_END = "%)";

  /** @return string */
  private static function valueSuffix(float $value) {
    return self::VALUE_SUFFIX_START . "$value" . self::VALUE_SUFFIX_END;
  }

  /** @return bool */
  private static function isValueSuffix(string $string) {
    $start = self::VALUE_SUFFIX_START;
    $end = self::VALUE_SUFFIX_END;
    if (!strStartsWith($string, $start)) return false;
    if (!strEndsWith($string, $end)) return false;
    $value = str_replace([$start, $end], "", $string);
    $floatValue = floatval($value);
    return $string === self::valueSuffix($floatValue);
  }

  /** @return null|string */
  public static function extractValueSuffix(string $taxClassName) {
    if (strlen($taxClassName) < 1) return null;

    $split = explode(self::VALUE_SUFFIX_START, $taxClassName);

    $splitSize = count($split);
    if ($splitSize < 2) return null;

    $last = trim($split[$splitSize - 1]);
    if (!strEndsWith($last, self::VALUE_SUFFIX_END)) return null;

    $resplit = explode(self::VALUE_SUFFIX_END, $last);

    $resplitSize = count($resplit);
    if ($resplitSize !== 2) return null;

    $first = trim($resplit[0]);
    $expectedSuffix = self::valueSuffix(floatval($first));
    if (!strEndsWith($taxClassName, $expectedSuffix)) return null;

    return $expectedSuffix;
  }

  /** @return float */
  public static function extractValue(string $taxClassName) {
    $suffix = self::extractValueSuffix($taxClassName);
    if ($suffix === null || !self::isValueSuffix($suffix)) throw msgException(
      "Can't extract value from", jsonString($taxClassName)
    );
    $value = str_replace([self::VALUE_SUFFIX_START, self::VALUE_SUFFIX_END], "", $suffix);
    return floatval($value);
  }

  /** @return float */
  public function extractedValue() {
    return self::extractValue($this->taxClassName);
  }

  /** @return self */
  public static function fromApiTax(
    DataScope $scope,
    string $country,
    WinoApiTax $apiTax
  ) {
    $taxClassName = self::namePrefix($scope) . $apiTax->name . self::valueSuffix($apiTax->value);
    $taxClassSlug = sanitize_title($taxClassName);
    $countrySuffix = self::countrySuffix($country);
    return new self(
      $taxClassName,
      $taxClassSlug,
      $taxClassName . $countrySuffix
    );
  }

  private const COUNTRY_SUFFIX_SEPARATOR = " - ";

  /** @return string */
  private static function countrySuffix(string $country) {
    return self::COUNTRY_SUFFIX_SEPARATOR . strtoupper($country);
  }

  /** @return bool */
  private static function isCountrySuffix(string $string) {
    return strStartsWith($string, self::COUNTRY_SUFFIX_SEPARATOR);
  }

  /** @return null|string */
  public static function extractCountrySuffix(string $taxRateName) {
    if (strlen($taxRateName) < 1) return null;

    $split = explode(self::COUNTRY_SUFFIX_SEPARATOR, $taxRateName);

    $splitSize = count($split);
    if ($splitSize < 2) return null;

    $last = trim($split[$splitSize - 1]);
    $expectedSuffix = self::countrySuffix($last);
    if (!strEndsWith($taxRateName, $expectedSuffix)) return null;

    return $expectedSuffix;
  }

  /** @return string */
  public static function extractCountry(string $taxRateName) {
    $suffix = self::extractCountrySuffix($taxRateName);
    if ($suffix === null || !self::isCountrySuffix($suffix)) throw msgException(
      "Can't extract country from", jsonString($taxRateName)
    );
    return substr($suffix, strlen(self::COUNTRY_SUFFIX_SEPARATOR));
  }

  /** @return string */
  public function extractedCountry() {
    return self::extractCountry($this->taxRateName);
  }

  /** @return string */
  public static function getTaxRateWinoName(
    DataScope $scope,
    string $taxRateName
  ) {
    $countrySuffix = self::extractCountrySuffix($taxRateName);

    $withoutCountrySuffix = $countrySuffix !== null && strEndsWith($taxRateName, $countrySuffix)
    ? substr($taxRateName, 0, strlen($taxRateName) - strlen($countrySuffix))
    : $taxRateName;

    return self::getTaxClassWinoName($scope, $withoutCountrySuffix);
  }
}

class WinoTaxClass {
  private DataScope $scope;

  /** @return DataScope */
  public function getScope() {
    return $this->scope;
  }

  public string $name;
  public string $slug;

  public function __construct(
    DataScope $scope,
    string $name,
    string $slug
  ) {
    $this->scope = $scope;
    $this->name = $name;
    $this->slug = $slug;
  }

  /** @return self */
  public static function fromKeys(
    DataScope $scope,
    WinoTaxKeys $keys
  ) {
    return new self(
      $scope,
      $keys->taxClassName,
      $keys->taxClassSlug
    );
  }

  /** @return self */
  public static function from(DataScope $scope, array $data) {
    return new self(
      $scope,
      nonBlankStringData($data, 'name'),
      nonBlankStringData($data, 'slug')
    );
  }

  /** @return string */
  public function getWinoName() {
    return WinoTaxKeys::getTaxClassWinoName($this->scope, $this->name);
  }
}

class WinoTaxRate {
  private DataScope $scope;
  public int $id;
  public float $value;
  public string $name;
  public string $taxClassSlug;

  public const DEFAULT_COUNTRY = "*";
  public string $country;
  public int $priority;

  public function __construct(
    DataScope $scope,
    int $id,
    string $name,
    float $value,
    string $taxClassSlug,
    string $country,
    int $priority
  ) {
    $this->scope = $scope;
    $this->id = $id;
    $this->value = $value;
    $this->name = $name;
    $this->taxClassSlug = $taxClassSlug;
    $this->country = $country;
    $this->priority = $priority;
  }

  /** @return self */
  public static function fromKeys(
    DataScope $scope,
    int $id,
    WinoTaxKeys $keys,
    int $priority
  ) {
    return new self(
      $scope,
      $id,
      $keys->taxRateName,
      $keys->extractedValue(),
      $keys->taxClassSlug,
      $keys->extractedCountry(),
      $priority
    );
  }

  /** @return WinoTaxRate */
  public static function from(DataScope $scope, array $data) {
    $country = trim(stringData($data, "tax_rate_country"));
    return new self(
      $scope,
      intval(nonBlankStringData($data, 'tax_rate_id')),
      nonBlankStringData($data, "tax_rate_name"),
      floatval(nonBlankStringData($data, "tax_rate")),
      nonBlankStringData($data, "tax_rate_class"),
      $country === "" ? self::DEFAULT_COUNTRY : $country,
      stringifiedIntData($data, 'tax_rate_priority')
    );
  }

  /** @return string */
  public function getWinoName() {
    return WinoTaxKeys::getTaxRateWinoName(
      $this->scope,
      $this->name
    );
  }
}
