<?php

namespace Wino;

class CategoryImportResult extends DataImportResult {
  public const EVENT_TARGET_KIND = "Category";
  public const WORDPRESS_REF_KIND = "Product category";
  public const WINO_REF_KIND = "Category";

  protected function eventTargetKind() {
    return self::EVENT_TARGET_KIND;
  }

  protected function wordpressRefKind() {
    return self::WORDPRESS_REF_KIND;
  }

  protected function winoRefKind() {
    return self::WINO_REF_KIND;
  }

  public WinoCategory $category;

  /** @var CategoryImportResult[] */
  public array $children;

  /** @var ProductImportResult[] */
  public array $products;

  /**
   * @param CategoryImportResult[] $children
   * @param ProductImportResult[] $products
   */
  private function __construct(
    WinoCategory $category,
    DataImportEvent $event,
    array $children,
    array $products
  ) {
    $categoryId = $category->getTermId();
    parent::__construct(
      $event,
      $category->getName(),
      "$categoryId",
      getOrElse($category->getWinoApiCategoryIdMeta(), '-'),
      null,
      null
    );
    $this->category = $category;
    $this->children = $children;
    $this->products = $products;
  }

  /** @return self */
  public static function read(WinoCategory $category) {
    return new self($category, DataImportEvent::read(), [], []);
  }

  /** @return self */
  public static function created(WinoCategory $category) {
    return new self($category, DataImportEvent::created(), [], []);
  }

  /** @return self */
  public static function updated(WinoCategory $category) {
    return new self($category, DataImportEvent::updated(), [], []);
  }

  /** @return self */
  public static function archived(WinoCategory $category) {
    return new self($category, DataImportEvent::archived(), [], []);
  }

  /** @return self */
  public static function deleted(WinoCategory $category) {
    return new self($category, DataImportEvent::deleted(), [], []);
  }

  /** @return int */
  public function getTermId() {
    return $this->category->getTermId();
  }

  /** @return \WP_Term */
  public function getTerm() {
    return $this->category->getTerm();
  }

  /** @return void */
  public function reportChildImportResult(CategoryImportResult $child) {
    $this->children[] = $child;
  }

  /** @return void */
  public function reportProductImportResult(ProductImportResult $product) {
    $this->products[] = $product;
  }

  /** @return void */
  public function reportProductUpdate(WinoProductUpdateProxy $update) {
    $productImportResult = ProductImportResult::updated($update->getProduct(), null, null, null);
    $this->reportProductImportResult($productImportResult);
  }
}

/** @return void */
function importWinoCategoryProductsHierarchyUpdate(
  CategoryImportResult $result,
  ImportContext $context,
  WinoCategory $category
) {
  $explicit = $context->explicitCategoriesHierarchyEnabled;

  $products = findWinoCategoryProducts($context, $category);
  foreach ($products as $product) {
    $update = WinoProductUpdateProxy::fromProduct($product);
    $update->setCategoriesHierarchy($explicit, $category);
    if ($update->commit()) $result->reportProductUpdate($update);
  }
}

/** @return CategoryImportResult|null */
function importWinoCategory(
  ImportContext $context,
  WinoApiCategory $apiCategory
) {
  $translator = $context->getTranslator();
  $disabled = !$context->categoriesImportEnabled;
  $shop = $context->getShop();
  $shopId = $shop === null ? null : $shop->id;
  $invalidShop = $shopId === null || $apiCategory->shop->id !== $shopId;
  $archived = $apiCategory->isArchived();

  $scope = $context->getScope();
  $winoIdMetaKey = $scope->winoIdMetaKey();
  $winoParentCategoryIdMetaKey = $scope->winoParentCategoryIdMetaKey();
  $taxonomy = WinoCategory::TAXONOMY;
  $apiCategoryId = $apiCategory->id;
  $apiParentId = $apiCategory->getParentId();
  $termName = $apiCategory->name;

  $customRoot = $context->getCustomRootCategory();
  $apiRootCategoryId = $scope->winoApiRootCategoryId();
  $apiRootCategory = $scope->winoApiRootCategory($shopId);
  $defaultRoot = $apiCategoryId === $apiRootCategoryId;

  $existingCategory = $defaultRoot
  ? findDefaultRootCategory($context)
  : findWinoCategoryByApiId($context, $apiCategoryId);

  $categoryShouldNotExist = $disabled || $invalidShop || $archived;
  $defaultRootShouldNotExist = $defaultRoot && $customRoot !== null;
  $shouldNotExist = $categoryShouldNotExist || $defaultRootShouldNotExist;

  if ($shouldNotExist)
  if ($existingCategory === null) return null;
  else return deleteWinoCategory($context, $existingCategory);

  $termArgs = new UpdateTermArgs();
  $foundAncestorTermId = null;
  $missingParentCategoryApiId = null;

  if ($defaultRoot) $termArgs->data['slug'] = $scope->winoApiRootCategorySlug();
  else {
    $foundAncestorCategory = $apiParentId === null
    ? findRootCategoryWithDefaultAutoCreate($context)
    : findWinoCategoryByApiId($context, $apiParentId);

    if ($foundAncestorCategory === null) {
      $missingParentCategoryApiId = $apiParentId;
      $foundAncestorCategory = $missingParentCategoryApiId !== null
      ? findRootCategoryWithDefaultAutoCreate($context)
      : null;
    }

    if ($foundAncestorCategory !== null)
    $foundAncestorTermId = $foundAncestorCategory->getTermId();

    if ($foundAncestorTermId === null && $customRoot !== null) {
      $customRoot->assertsValidity($translator);
      $foundAncestorTermId = $customRoot->getTermId();
    }

    if ($foundAncestorTermId === null)
    throw apiRootCategoryImportError($translator, $apiRootCategory);

    $termArgs->data["parent"] = $foundAncestorTermId;
  }

  if ($existingCategory !== null) {
    $createMode = false;
    $termArgs->data["name"] = $termName;
    $update = WinoCategoryUpdateProxy::from($existingCategory);
    $update->setArgs($termArgs);
    $parentUpdate = $update->termArgs()->has('parent');
    $updateMode = $update->commit();
    $term = $update->getTerm();
  } else {
    $createMode = true;
    $updateMode = false;
    $parentUpdate = false;

    $termsWithSameNameAndParent = $foundAncestorTermId === null ? [] : findWordpressTermChildrenByName(
      $taxonomy,
      $foundAncestorTermId,
      $termName
    );

    $sameNameAndParentCount = count($termsWithSameNameAndParent);
    $duplicateName = $sameNameAndParentCount > 0;
    $duplicateNumber = $sameNameAndParentCount + 1;
    $bypassUniqNameArgs = new UpdateTermArgs(["name" => $termName]);
    if ($duplicateName) $termName = "$termName ($duplicateNumber)";

    $term = createWordpressTerm($termName, $taxonomy, $termArgs->asInsertArgs());
    if ($duplicateName) $term = updateWordpressTerm($term, $bypassUniqNameArgs);
  }

  updateWordpressTermMeta($term, $winoIdMetaKey, $apiCategoryId);
  if ($missingParentCategoryApiId !== null) updateWordpressTermMeta(
    $term, $winoParentCategoryIdMetaKey, $missingParentCategoryApiId
  );

  $childrenTermsWithMissingParent = findWordpressTermsByTaxonomy(
    WinoCategory::TAXONOMY,
    notNullMetaQuery($winoIdMetaKey),
    stringMetaQuery($winoParentCategoryIdMetaKey, '==', $apiCategoryId)
  );

  $termId = $term->term_id;
  $childrenCategoriesWithMissingParent = [];
  foreach ($childrenTermsWithMissingParent as $childTermWithMissingParent) {
    $childTermUpdate = new UpdateTermArgs(["parent" => $termId]);
    $updatedTerm = updateWordpressTerm($childTermWithMissingParent, $childTermUpdate);
    deleteWordpressTermMeta($updatedTerm, $winoParentCategoryIdMetaKey);
    $childCategoryWithMissingParent = new WinoCategory($scope, $updatedTerm);
    $context->cacheCategory($childCategoryWithMissingParent);
    $childrenCategoriesWithMissingParent[] = $childCategoryWithMissingParent;
  }

  $category = $context->cacheCategory(new WinoCategory($scope, $term));
  $result = $createMode ? CategoryImportResult::created($category) : ($updateMode
    ? CategoryImportResult::updated($category)
    : CategoryImportResult::read($category));

  if ($parentUpdate) importWinoCategoryProductsHierarchyUpdate($result, $context, $category);

  foreach ($childrenCategoriesWithMissingParent as $childCategoryWithMissingParent) {
    $result->reportChildImportResult(CategoryImportResult::updated($childCategoryWithMissingParent));
    importWinoCategoryProductsHierarchyUpdate($result, $context, $childCategoryWithMissingParent);
  }

  return $result;
}

/** @return WinoCategory|null */
function findDefaultRootCategoryWithAutoCreate(ImportContext $context) {
  $defaultRoot = findDefaultRootCategory($context);

  if ($defaultRoot !== null) return $defaultRoot;

  $scope = $context->getScope();

  $apiRootCategory = $scope->winoApiRootCategory($context->shopId);
  $rootCategoryImport = importWinoCategory($context, $apiRootCategory);

  return $rootCategoryImport === null ? null : $rootCategoryImport->category;
}

/** @return WinoCategory|null */
function findRootCategoryWithDefaultAutoCreate(ImportContext $context) {
  $customRoot = $context->getCustomRootCategory();

  $rootCategory = findRootCategory($context);
  if ($rootCategory !== null || $customRoot !== null) return $rootCategory;

  return findDefaultRootCategoryWithAutoCreate($context);
}

/** @return WinoCategory[] */
function findAllWinoCategories(
  ImportContext $context,
  bool $defaultRootCategoryAutoCreate
) {
  $scope = $context->getScope();
  $categories = [];

  $defaultRootCategory = $defaultRootCategoryAutoCreate
  ? findDefaultRootCategoryWithAutoCreate($context)
  : findDefaultRootCategory($context);

  if ($defaultRootCategory !== null) $categories[] = $defaultRootCategory;

  $winoIdMetaKey = $scope->winoIdMetaKey();

  $terms = findWordpressTermsByTaxonomy(
    WinoCategory::TAXONOMY,
    notNullMetaQuery($winoIdMetaKey),
    stringMetaQuery($winoIdMetaKey, '!=', $scope->winoApiRootCategoryId())
  );
  foreach ($terms as $term) {
    $category = new WinoCategory($scope, $term);
    $categories[] = $category;
    $context->cacheCategory($category);
  }
  return $categories;
}

/** @return WinoCategory|null */
function findWinoCategoryByApiId(ImportContext $context, string $winoApiId) {
  $translator = $context->getTranslator();
  $cachedCategory = $context->getCachedCategory($winoApiId);
  if ($cachedCategory !== null) return $cachedCategory;
  $scope = $context->getScope();
  $rootId = $scope->winoApiRootCategoryId();
  if ($rootId === $winoApiId) return findDefaultRootCategoryWithAutoCreate($context);
  $terms = findWordpressTermsByTaxonomy(
    WinoCategory::TAXONOMY,
    stringMetaQuery($scope->winoIdMetaKey(), '==', $winoApiId),
  );
  if (count($terms) > 1) throw duplicateCategoryException($translator, $winoApiId, $terms);
  $term = listElement($terms, 0);
  if ($term === null) return null;
  return $context->cacheCategory(new WinoCategory($scope, $term));
}

class CategoriesImportLog {
  public DataImportLog $log;

  /** @var int[] */
  public array $upToDateCategoryTermIds;

  public function __construct(DataImportLog $log) {
    $this->log = $log;
    $this->upToDateCategoryTermIds = [];
  }

  /** @return self */
  public static function from(DataImportLog $log) {
    return new self($log);
  }

  /** @return self */
  public static function empty() {
    return self::from(DataImportLog::empty());
  }

  /** @return void */
  public function logImportResult(?CategoryImportResult $result) {
    $log = $this->log;
    $log->logImportResult($result);
    if ($result === null) return;

    foreach ($result->children as $child) $this->logImportResult($child);
    foreach ($result->products as $product) $log->logImportResult($product);

    if ($result->isDeletion()) return;
    $this->upToDateCategoryTermIds[] = $result->category->getTermId();
  }

  /** @return bool */
  public function isUpToDate(WinoCategory $category) {
    return in_array($category->getTermId(), $this->upToDateCategoryTermIds);
  }
}

/** @return void */
function categoriesImportStep(
  ImportContext $context,
  CategoriesImportLog $log,
  WinoApiCategory $apiCategory
) {
  $log->logImportResult(importWinoCategory($context, $apiCategory));
}

/** @return bool */
function upToDateCategoriesImportNextStep(
  ImportContext $context,
  CategoriesImportLog $log,
  WinoApiCategoryQueryAllIterator $iterator
) {
  if (!$context->categoriesImportEnabled) return false;
  if (!$iterator->next()) return false;
  categoriesImportStep($context, $log, $iterator->currentObject());
  return true;
}

/** @return WinoApiCategoryQueryAllIterator */
function upToDateCategoriesImportIterator(
  WinoApiClient $winoApi,
  WinoApiShopIdentity $shop
) {
  return WinoApiCategory::queryAllIterator($winoApi, $shop);
}

/** @return void */
function importUpToDateCategories(
  WinoApiClient $winoApi,
  ImportContext $context,
  CategoriesImportLog $log
) {
  $shop = $context->getShop();
  if ($shop === null) return;

  $cache = wp_suspend_cache_addition(true);
  categoriesImportStep($context, $log, $context->winoApiRootCategory());
  $iterator = upToDateCategoriesImportIterator($winoApi, $shop);
  while (upToDateCategoriesImportNextStep($context, $log, $iterator));
  wp_suspend_cache_addition($cache);
}

/** @return CategoryImportResult|null */
function deleteWinoCategory(ImportContext $context, WinoCategory $category) {
  $result = CategoryImportResult::deleted($category);
  $scope = $context->getScope();
  $taxonomy = WinoCategory::TAXONOMY;
  $metaKey = $scope->winoIdMetaKey();

  $term = $category->getTerm();
  $defaultRoot = $category->isRoot();
  $parentCategory = $defaultRoot
  ? findRootCategory($context)
  : $category->getParentCategory();

  $updateCount = 0;
  $children = $category->getChildren();

  if (!$defaultRoot) foreach ($children as $child) {
    $childDeletion = deleteWinoCategory($context, $child);
    if ($childDeletion === null) continue;
    $updateCount++;
    $result->reportChildImportResult($childDeletion);
  }

  $products = findWinoCategoryProducts($context, $category);
  foreach ($products as $product) {
    $update = WinoProductUpdateProxy::fromProduct($product);
    productCategoryUpdate($context, $update, $parentCategory);
    if (!$update->commit()) continue;
    $updateCount++;
    $result->reportProductUpdate($update);
  }

  $context->unsetCategoryCache($category);

  $termMetaDeletion = deleteWordpressTermMeta($term, $metaKey);
  $termDeletion = deleteWordpressTerm($term, $taxonomy);
  if ($termMetaDeletion || $termDeletion) $updateCount++;

  return $updateCount > 0 ? $result : null;
}

/** @return void */
function categoriesDeletionStep(
  ImportContext $context,
  DataImportLog $log,
  WinoCategory $category
) {
  $log->logImportResult(deleteWinoCategory($context, $category));
}

/** @return void */
function deleteAllWinoCategories(ImportContext $context, DataImportLog $log) {
  $winoCategories = findAllWinoCategories($context, false);
  foreach ($winoCategories as $category) categoriesDeletionStep($context, $log, $category);
}

/** @return void */
function outOfDateCategoriesDeletionStep(
  ImportContext $context,
  CategoriesImportLog $log,
  WinoCategory $category
) {
  if ($log->isUpToDate($category)) return;
  $log->logImportResult(deleteWinoCategory($context, $category));
}

/** @return void */
function deleteOutOfDateCategories(ImportContext $context, CategoriesImportLog $log) {
  $categories = findAllWinoCategories($context, false);
  foreach ($categories as $category) outOfDateCategoriesDeletionStep($context, $log, $category);
}
