<?php

namespace Wino;

class DataImportEvent {
  public const READ = 'read';
  public const CREATED = 'created';
  public const UPDATED = 'updated';
  public const ARCHIVED = 'archived';
  public const DELETED = 'deleted';

  /** @var 'deleted'|'created'|'archived'|'updated'|'read' */
  public string $kind;

  /** @param 'deleted'|'created'|'archived'|'updated'|'read' $kind */
  private function __construct(string $kind) {
    $this->kind = $kind;
  }

  /** @return bool */
  private function is(string $kind) {
    return $this->kind === $kind;
  }

  /** @return bool */
  public function isRead() {
    return $this->is(self::READ);
  }

  /** @return bool */
  public function isCreated() {
    return $this->is(self::CREATED);
  }

  /** @return bool */
  public function isUpdated() {
    return $this->is(self::UPDATED);
  }

  /** @return bool */
  public function isArchived() {
    return $this->is(self::ARCHIVED);
  }

  /** @return bool */
  public function isDeleted() {
    return $this->is(self::DELETED);
  }

  /** @return self */
  public static function read() {
    return new self(self::READ);
  }

  /** @return self */
  public static function created() {
    return new self(self::CREATED);
  }

  /** @return self */
  public static function updated() {
    return new self(self::UPDATED);
  }

  /** @return self */
  public static function archived() {
    return new self(self::ARCHIVED);
  }

  /** @return self */
  public static function deleted() {
    return new self(self::DELETED);
  }
}

interface DataImportLogEntryTranslator {
  /** @return string */
  public function showAllImportedProductsOnDedicatedPage();

  /** @return string */
  public function theFollowingOperationsWereSuccessful();

  /** @return string */
  public function importReportTypeLabel();

  /** @return string */
  public function importReportNameLabel();

  /** @return string */
  public function importReportWordpressRefLabel();

  /** @return string */
  public function importReportWinoRefLabel();

  /** @return string */
  public function importReportType(string $eventKind, string $eventTargetKind);

  /** @return string */
  public function importReportWordpressRef(string $refKind, string $ref);

  /** @return string */
  public function importReportWinoRef(string $refKind, string $ref);
}

class DataImportLogEntry {
  public string $eventKind;
  public const EVENT_KIND_KEY = 'eventKind';

  public string $eventTargetKind;
  public const EVENT_TARGET_KIND_KEY = 'eventTargetKind';

  public string $name;
  public const NAME_KEY = 'name';

  public string $wordpressRefKind;
  public const WORDPRESS_REF_KIND_KEY = 'wordpressRefKind';

  public string $wordpressRef;
  public const WORDPRESS_REF_KEY = 'wordpressRef';

  public ?string $editionPageUrl;
  public const EDITION_PAGE_URL_KEY = 'editionPageUrl';

  public string $winoRefKind;
  public const WINO_REF_KIND_KEY = 'winoRefKind';

  public ?string $winoRef;
  public const WINO_REF_KEY = 'winoRef';

  public ?string $winoPageUrl;
  public const WINO_PAGE_URL_KEY = 'winoPageUrl';

  public function __construct(
    string $eventKind,
    string $eventTargetKind,
    string $name,
    string $wordpressRefKind,
    string $wordpressRef,
    string $winoRefKind,
    ?string $winoRef,
    ?string $editionPageUrl,
    ?string $winoPageUrl
  ) {
    $this->eventKind = $eventKind;
    $this->eventTargetKind = $eventTargetKind;
    $this->name = $name;
    $this->wordpressRefKind = $wordpressRefKind;
    $this->wordpressRef = $wordpressRef;
    $this->winoRefKind = $winoRefKind;
    $this->winoRef = $winoRef;
    $this->editionPageUrl = $editionPageUrl;
    $this->winoPageUrl = $winoPageUrl;
  }

  /** @return array<string, string|null> */
  public function asArray() {
    return [
      self::EVENT_KIND_KEY => $this->eventKind,
      self::EVENT_TARGET_KIND_KEY => $this->eventTargetKind,
      self::NAME_KEY => $this->name,
      self::WORDPRESS_REF_KIND_KEY => $this->wordpressRefKind,
      self::WORDPRESS_REF_KEY => $this->wordpressRef,
      self::WINO_REF_KIND_KEY => $this->winoRefKind,
      self::WINO_REF_KEY => $this->winoRef,
      self::EDITION_PAGE_URL_KEY => $this->editionPageUrl,
      self::WINO_PAGE_URL_KEY => $this->winoPageUrl,
    ];
  }

  /** @return self */
  public static function from(array $data) {
    return new self(
      stringData($data, self::EVENT_KIND_KEY),
      stringData($data, self::EVENT_TARGET_KIND_KEY),
      stringData($data, self::NAME_KEY),
      stringData($data, self::WORDPRESS_REF_KIND_KEY),
      stringData($data, self::WORDPRESS_REF_KEY),
      stringData($data, self::WINO_REF_KIND_KEY),
      stringData($data, self::WINO_REF_KEY),
      optionalStringData($data, self::EDITION_PAGE_URL_KEY),
      optionalStringData($data, self::WINO_PAGE_URL_KEY)
    );
  }

  /** @return non-empty-array<string, string> */
  public function htmlReportRow(DataImportLogEntryTranslator $translator) {
    $typeKey = $translator->importReportTypeLabel();
    $nameKey = $translator->importReportNameLabel();
    $wpRefKey = $translator->importReportWordpressRefLabel();
    $winoRefKey = $translator->importReportWinoRefLabel();

    $eventKind = $this->eventKind;
    $eventTargetKind = $this->eventTargetKind;
    $wpRefKind = $this->wordpressRefKind;
    $wpRef = $this->wordpressRef;
    $winoRefKind = $this->winoRefKind;
    $winoRef = $this->winoRef;
    $wpLink = $this->editionPageUrl;
    $winoLink = $this->winoPageUrl;

    $typeLine = $translator->importReportType($eventKind, $eventTargetKind);
    $wpRefLine = $translator->importReportWordpressRef($wpRefKind, $wpRef);
    $nameLine = $this->name;
    $winoRefLine = $winoRef === null ? '-' : $translator->importReportWinoRef($winoRefKind, $winoRef);

    return [
      $typeKey => $typeLine,
      $nameKey => renderHtmlLink($wpLink, $nameLine),
      $wpRefKey => renderHtmlLink($wpLink, $wpRefLine),
      $winoRefKey => renderHtmlLink($winoLink, $winoRefLine, '__blank')
    ];
  }
}

abstract class DataImportResult {
  public DataImportEvent $event;
  protected string $name;
  protected string $wordpressRef;
  protected ?string $editionPageUrl;
  protected ?string $winoRef;
  protected ?string $winoPageUrl;

  protected function __construct(
    DataImportEvent $event,
    string $name,
    string $wordpressRef,
    ?string $winoRef,
    ?string $editionPageUrl,
    ?string $winoPageUrl
  ) {
    $this->event = $event;
    $this->name = $name;
    $this->wordpressRef = $wordpressRef;
    $this->winoRef = $winoRef;
    $this->editionPageUrl = $editionPageUrl;
    $this->winoPageUrl = $winoPageUrl;
  }

  /** @return bool */
  public function isRead() {
    return $this->event->isRead();
  }

  /** @return bool */
  public function isCreation() {
    return $this->event->isCreated();
  }

  /** @return bool */
  public function isUpdate() {
    return $this->event->isUpdated();
  }

  /** @return bool */
  public function isArchive() {
    return $this->event->isArchived();
  }

  /** @return bool */
  public function isDeletion() {
    return $this->event->isDeleted();
  }

  /** @return 'deleted'|'created'|'archived'|'updated'|'read' */
  private function resultEventKind() {
    if ($this->isDeletion()) return DataImportEvent::DELETED;
    if ($this->isCreation()) return DataImportEvent::CREATED;
    if ($this->isArchive()) return DataImportEvent::ARCHIVED;
    if ($this->isUpdate()) return DataImportEvent::UPDATED;
    if ($this->isRead()) return DataImportEvent::READ;
    throw msgError("Unexpected event kind", jsonString($this->event));
  }

  /** @return string */
  abstract protected function eventTargetKind();

  /** @return string */
  abstract protected function wordpressRefKind();

  /** @return string */
  abstract protected function winoRefKind();

  /** @return DataImportLogEntry */
  public function logEntry() {
    $eventKind = $this->resultEventKind();
    $eventTargetKind = $this->eventTargetKind();
    $wpRefKind = $this->wordpressRefKind();
    $wpRef = $this->wordpressRef;
    $winoRefKind = $this->winoRefKind();
    $winoRef = $this->winoRef;

    return new DataImportLogEntry(
      $eventKind,
      $eventTargetKind,
      $this->name,
      $wpRefKind,
      $wpRef,
      $winoRefKind,
      $winoRef,
      $this->editionPageUrl,
      $this->winoPageUrl
    );
  }
}

class DataImportLog {

  /** @var DataImportLogEntry[] */
  public array $entries;

  /** @param DataImportLogEntry[] $entries */
  private function __construct(
    &$entries
  ) {
    $this->entries = $entries;
  }

  /** @return array<array<string, string|null>> */
  public function asArrays() {
    $entries = $this->entries;
    $arrays = [];
    foreach ($entries as $entry) $arrays[] = $entry->asArray();
    return $arrays;
  }

  /** @return self|null */
  public static function from(?array $data) {
    if ($data === null) return null;
    $entries = [];
    $arrays = assertedArraysArray($data);
    foreach ($arrays as $array) $entries[] = DataImportLogEntry::from($array);
    return new self($entries);
  }

  /** @return self */
  public static function empty() {
    $entries = [];
    return new self($entries);
  }

  /** @return void */
  public function logImportResult(?DataImportResult $result) {
    if ($result === null || $result->isRead()) return;
    $this->entries[] = $result->logEntry();
  }

  /** @return array<array<string, string>> */
  public function htmlReportRows(DataImportLogEntryTranslator $translator) {
    $entries = $this->entries;
    $rows = [];
    foreach ($entries as $entry) $rows[] = $entry->htmlReportRow($translator);
    return $rows;
  }
}

/** @return void */
function writeDataImportLogToAjaxResponse(
  DataImportLog $log,
  AjaxResponse $response,
  DataImportLogEntryTranslator $translator
) {
  $rows = $log->htmlReportRows($translator);
  $empty = count($rows) === 0;
  $p = '<p style="text-align: center">';
  $pEnd = '</p>';
  $notEmptyReport = $empty ? "" : str(
    $p . $translator->theFollowingOperationsWereSuccessful() . $pEnd,
    htmlTable($rows, "text-align:center; width: 100%")
  );
  $response->update(str(
    $p . $translator->showAllImportedProductsOnDedicatedPage() . $pEnd,
    $notEmptyReport
  ));
}
