<?php

namespace Wino;

interface ImportContextTranslator extends ConfigurationRootCategoryTranslator {
  /** @return string */
  public function apiRootCategoryImportErrorMessage(WinoApiCategory $apiRootCategory);

  /**
   * @param \WP_Term[] $duplicateTerms
   *
   * @return string
   */
  public function duplicateCategoryErrorMessage(string $winoApiId, $duplicateTerms);

  /** @return string */
  public function duplicateSkuErrorMessage(
    WinoApiVariant $importedApiVariant,
    WinoProduct $productWithSameSku,
    ?WinoProduct $productWithSameVariantId
  );
}

/** @return \Error */
function apiRootCategoryImportError(
  ImportContextTranslator $translator,
  WinoApiCategory $apiRootCategory
) {
  return error($translator->apiRootCategoryImportErrorMessage($apiRootCategory));
}

/**
 * @param \WP_Term[] $duplicateTerms
 *
 * @return \Exception
 */
function duplicateCategoryException(
  ImportContextTranslator $translator,
  string $winoApiId,
  $duplicateTerms
) {
  return exception($translator->duplicateCategoryErrorMessage($winoApiId, $duplicateTerms));
}

/** @return \Exception */
function duplicateSkuException(
  ImportContextTranslator $translator,
  WinoApiVariant $importedApiVariant,
  WinoProduct $productWithSameSku,
  ?WinoProduct $productWithSameVariantId
) {
  return exception($translator->duplicateSkuErrorMessage(
    $importedApiVariant,
    $productWithSameSku,
    $productWithSameVariantId
  ));
}

/**
 * The import context has several purposes:
 *
 * - Providing the translations for import operations log
 * - Providing the configuration values required during the import operations
 * - Providing a reference to the data scope
 * - Caching data during the import operations
 */
class ImportContext {
  private MonitoringContext $monitoringContext;
  private ImportContextTranslator $translator;
  private DataScope $scope;
  private WinoApiClient $apiClient;

  private bool $customRootCategoryCache;
  private ?CustomRootCategory $customRootCategoryCacheValue;

  /** @var DataCache<WinoCategory> */
  private DataCache $categoriesCache;

  /** @var DataCache<WinoTaxClass> */
  private DataCache $taxClassesCache;

  /** @var DataCache<WinoTaxRate> */
  private DataCache $taxRatesCache;

  public ?string $shopId;
  public bool $categoriesImportEnabled;
  private ?int $rootCategoryTermId;
  public bool $explicitCategoriesHierarchyEnabled;
  public string $priceListName;
  public string $productDescriptionFormat;
  public bool $productDescriptionUpdateEnabled;
  public bool $productNameUpdateEnabled;
  public bool $priceUpdateEnabled;
  public bool $skuBasedImportEnabled;
  public bool $stockManagementOverrideEnabled;

  /** @return bool */
  public function monitoringEnabled() {
    return $this->monitoringContext->monitoringEnabled;
  }

  /**
   * @param mixed $log
   * @param mixed[] $data
   * @return void
   */
  public function monitoringLog($log, ...$data) {
    $this->monitoringContext->log($log, ...$data);
  }

  /** @return ImportContextTranslator */
  public function getTranslator() {
    return $this->translator;
  }

  /** @return DataScope */
  public function getScope() {
    return $this->scope;
  }

  /** @return WinoApiClient */
  public function getApiClient() {
    return $this->apiClient;
  }

  /** @return $this */
  public function setApiClient(WinoApiClient $apiClient) {
    $this->apiClient = $apiClient;
    return $this;
  }

  /** @return self */
  public function setShopId(?string $shopId) {
    $this->shopId = $shopId === null ? null : self::getValidShopId(new WinoApiShopIdentity($shopId));
    return $this;
  }

  /** @return self */
  public function setShop(?WinoApiShopIdentity $shop) {
    return $this->setShopId(self::getValidShopId($shop));
  }

  /** @return WinoApiShopIdentity|null */
  private static function getValidShop(?WinoApiShopIdentity $shop) {
    if ($shop === null) return null;
    return $shop->isValid() ? $shop : null;
  }

  /** @return string|null */
  private static function getValidShopId(?WinoApiShopIdentity $shop) {
    $validShop = self::getValidShop($shop);
    return $validShop === null ? null : $validShop->id;
  }

  /** @return WinoApiShopIdentity|null */
  public function getShop() {
    $id = $this->shopId;
    return $id === null ? null : self::getValidShop(new WinoApiShopIdentity($id));
  }

  /** @return self */
  public function setPrice(WinoApiPrice $price) {
    $this->priceListName = $price->name;
    return $this;
  }

  /** @return self */
  public function disableStockManagementOverride() {
    $this->stockManagementOverrideEnabled = false;
    return $this;
  }

  /** @return self */
  public function enableStockManagementOverride() {
    $this->stockManagementOverrideEnabled = true;
    return $this;
  }

  /** @return self */
  public function disableCategoriesImport() {
    $this->categoriesImportEnabled = false;
    return $this;
  }

  /** @return self */
  public function enableCategoriesImport() {
    $this->categoriesImportEnabled = true;
    return $this;
  }

  /** @return CustomRootCategory|null */
  public function getCustomRootCategory() {
    if ($this->customRootCategoryCache)
    return $this->customRootCategoryCacheValue;

    $category = CustomRootCategory::fromTermId($this->rootCategoryTermId);

    $this->customRootCategoryCache = true;
    $this->customRootCategoryCacheValue = $category;
    return $category;
  }

  /** @return self */
  public function setRootCategoryTermId(?int $termId) {
    $this->customRootCategoryCache = false;
    $this->customRootCategoryCacheValue = null;
    $this->rootCategoryTermId = CustomRootCategory::termIdValue($termId);
    return $this;
  }

  /** @return self */
  public function disableRootCategory() {
    return $this->setRootCategoryTermId(undefinedTermId());
  }

  /** @return self */
  public function enableCustomRootCategory(int $termId) {
    return $this->setRootCategoryTermId($termId);
  }

  /** @return self */
  public function disableCustomRootCategory() {
    return $this->setRootCategoryTermId(null);
  }

  /** @return self */
  public function disableExplicitCategoriesHierarchy() {
    $this->explicitCategoriesHierarchyEnabled = false;
    return $this;
  }

  /** @return self */
  public function enableExplicitCategoriesHierarchy() {
    $this->explicitCategoriesHierarchyEnabled = true;
    return $this;
  }

  /** @return self */
  public function disableSkuBasedImport() {
    $this->skuBasedImportEnabled = false;
    return $this;
  }

  /** @return self */
  public function enableSkuBasedImport() {
    $this->skuBasedImportEnabled = true;
    return $this;
  }

  /** @return self */
  public function disableProductNameUpdate() {
    $this->productNameUpdateEnabled = false;
    return $this;
  }

  /** @return self */
  public function enableProductNameUpdate() {
    $this->productNameUpdateEnabled = true;
    return $this;
  }

  /** @return WinoApiCategory */
  public function winoApiRootCategory() {
    return $this->scope->winoApiRootCategory($this->shopId);
  }

  private function __construct(
    MonitoringContext $monitoringContext,
    ImportContextTranslator $translator,
    DataScope $scope,
    WinoApiClient $apiClient,
    ?WinoApiShopIdentity $shop,
    bool $categoriesImportEnabled,
    ?int $rootCategoryTermId,
    bool $explicitCategoriesHierarchyEnabled,
    string $priceList,
    string $productDescriptionFormat,
    bool $productDescriptionUpdateEnabled,
    bool $productNameUpdateEnabled,
    bool $priceUpdateEnabled,
    bool $skuUpdateEnabled,
    bool $stockManagementOverrideEnabled
  ) {
    $this->monitoringContext = $monitoringContext;
    $this->translator = $translator;
    $this->scope = $scope;
    $this->apiClient = $apiClient;

    $this->customRootCategoryCache = false;
    $this->customRootCategoryCacheValue = null;
    $this->categoriesCache = DataCache::empty();
    $this->taxClassesCache = DataCache::empty();
    $this->taxRatesCache = DataCache::empty();

    $this->shopId = self::getValidShopId($shop);
    $this->categoriesImportEnabled = $categoriesImportEnabled;
    $this->rootCategoryTermId = null;
    $this->setRootCategoryTermId($rootCategoryTermId);
    $this->explicitCategoriesHierarchyEnabled = $explicitCategoriesHierarchyEnabled;
    $this->priceListName = $priceList;
    $this->productDescriptionFormat = $productDescriptionFormat;
    $this->productDescriptionUpdateEnabled = $productDescriptionUpdateEnabled;
    $this->productNameUpdateEnabled = $productNameUpdateEnabled;
    $this->priceUpdateEnabled = $priceUpdateEnabled;
    $this->skuBasedImportEnabled = $skuUpdateEnabled;
    $this->stockManagementOverrideEnabled = $stockManagementOverrideEnabled;
  }

  /** @return self */
  public static function fromConfiguration(?int $userId = null) {
    return new self(
      MonitoringContext::fromConfiguration(),
      LocaleTranslator::fromConfiguration(),
      DataScope::none(),
      WinoApiClient::fromConfiguration($userId),
      ConfigurationShopId::shopIdentity(),
      ConfigurationCategoriesImportEnabled::value(),
      ConfigurationRootCategory::value(),
      ConfigurationExplicitCategoriesHierarchyEnabled::value(),
      ConfigurationPriceList::value(),
      ConfigurationProductDescriptionFormat::value(),
      ConfigurationProductsDescriptionsUpdateEnabled::value(),
      ConfigurationProductsNamesUpdateEnabled::value(),
      ConfigurationPricesUpdateEnabled::value(),
      ConfigurationSkuBasedImportEnabled::value(),
      ConfigurationStockManagementOverrideEnabled::value(),
    );
  }

  /** @return self */
  public static function fromConfigurationDefaults(
    DataScope $scope
  ) {
    return new self(
      MonitoringContext::fromConfigurationDefaults(),
      LocaleTranslator::fromConfigurationDefault(),
      $scope,
      WinoApiClient::fromConfigurationDefault(),
      ConfigurationShopId::defaultShopIdentity(),
      ConfigurationCategoriesImportEnabled::DEFAULT,
      ConfigurationRootCategory::DEFAULT,
      ConfigurationExplicitCategoriesHierarchyEnabled::DEFAULT,
      ConfigurationPriceList::DEFAULT,
      ConfigurationProductDescriptionFormat::defaultValue(),
      ConfigurationProductsDescriptionsUpdateEnabled::DEFAULT,
      ConfigurationProductsNamesUpdateEnabled::DEFAULT,
      ConfigurationPricesUpdateEnabled::DEFAULT,
      ConfigurationSkuBasedImportEnabled::DEFAULT,
      ConfigurationStockManagementOverrideEnabled::DEFAULT,
    );
  }

  /** @return self */
  public function emptyCacheCopy() {
    return new self(
      $this->monitoringContext->emptyCacheCopy(),
      $this->getTranslator(),
      $this->getScope(),
      $this->getApiClient(),
      $this->getShop(),
      $this->categoriesImportEnabled,
      $this->rootCategoryTermId,
      $this->explicitCategoriesHierarchyEnabled,
      $this->priceListName,
      $this->productDescriptionFormat,
      $this->productDescriptionUpdateEnabled,
      $this->productNameUpdateEnabled,
      $this->priceUpdateEnabled,
      $this->skuBasedImportEnabled,
      $this->stockManagementOverrideEnabled,
    );
  }

  /** @return array<string, WinoCategory> */
  public function cachedCategories() {
    return $this->categoriesCache->record;
  }

  /** @return bool */
  public function hasCachedCategory(string $winoApiCategoryId) {
    return $this->categoriesCache->has($winoApiCategoryId);
  }

  /** @return WinoCategory|null */
  public function getCachedCategory(string $winoApiCategoryId) {
    return $this->categoriesCache->get($winoApiCategoryId);
  }

  /** @return WinoCategory */
  public function cacheCategory(WinoCategory $category) {
    $idMeta = $category->getWinoApiCategoryIdMeta();
    if ($idMeta !== null) $this->categoriesCache->set($idMeta, $category);
    return $category;
  }

  /** @return void */
  public function unsetCategoryCache(WinoCategory $category) {
    $this->categoriesCache->unsetObject(
      $category,
      function (WinoCategory $cat) {
        return $cat->getWinoApiCategoryIdMeta();
      },
      function (WinoCategory $cat) {
        return $cat->getTermId();
      }
    );
  }

  /** @return bool */
  public function hasCachedTaxClass(string $taxClassSlug) {
    return $this->taxClassesCache->has($taxClassSlug);
  }

  /** @return WinoTaxClass|null */
  public function getCachedTaxClass(string $taxClassSlug) {
    return $this->taxClassesCache->get($taxClassSlug);
  }

  /** @return WinoTaxClass */
  public function cacheTaxClass(WinoTaxClass $taxClass) {
    $this->taxClassesCache->set($taxClass->slug, $taxClass);
    return $taxClass;
  }

  /** @return void */
  public function unsetTaxClassCache(WinoTaxClass $taxClass) {
    $this->taxClassesCache->unset($taxClass->slug);
  }

  /** @return bool */
  public function hasCachedTaxRate(string $taxRateName) {
    return $this->taxRatesCache->has($taxRateName);
  }

  /** @return WinoTaxRate|null */
  public function getCachedTaxRate(string $taxRateName) {
    return $this->taxRatesCache->get($taxRateName);
  }

  /** @return WinoTaxRate */
  public function cacheTaxRate(WinoTaxRate $taxRate) {
    $this->taxRatesCache->set($taxRate->name, $taxRate);
    return $taxRate;
  }

  /** @return void */
  public function unsetTaxRateCache(WinoTaxRate $taxRate) {
    $this->taxRatesCache->unset($taxRate->name);
  }
}
