<?php

namespace Wino;

interface ImportSessionTranslator {
  /** @return string */
  public function anImportIsRunningYet();

  /** @return string */
  public function nothingToImport();

  /** @return string */
  public function categoriesImportInProgress(int $importedCategories);

  /** @return string */
  public function taxesImportInProgress(int $importedTaxes);

  /** @return string */
  public function productsImportInProgress(int $importedProducts);

  /** @return string */
  public function unusedCategoriesDeletionInProgress();
}

class ImportSessionData {
  private DataScope $scope;

  /** @return DataScope */
  public function getScope() {
    return $this->scope;
  }

  public int $creationTimestamp;
  public int $updateTimestamp;
  public int $importedItemsCount;
  public bool $rootCategoryImported;
  public bool $upToDateCategoriesImported;
  public bool $upToDateTaxesImported;
  public bool $upToDateProductsImported;
  public bool $outOfDateCategoriesDeleted;
  public ?array $graphQLIteratorQueryItems;
  public ?int $graphQLIteratorQueryIndex;
  public ?string $graphQLIteratorQueryCursor;
  public ?string $errorMessage;

  /** @var int[] */
  public array $upToDateCategoryTermIds;

  /** @param int[] $upToDateCategoryTermIds */
  public function __construct(
    DataScope $scope,
    int $creationTimestamp,
    int $updateTimestamp,
    int $importedItemsCount,
    bool $rootCategoryImported,
    bool $upToDateCategoriesImported,
    bool $upToDateTaxesImported,
    bool $upToDateProductsImported,
    bool $outOfDateCategoriesDeleted,
    ?array $graphQLIteratorQueryItems,
    ?int $graphQLIteratorQueryIndex,
    ?string $graphQLIteratorQueryCursor,
    ?string $errorMessage,
    array $upToDateCategoryTermIds
  ) {
    $this->scope = $scope;
    $this->creationTimestamp = $creationTimestamp;
    $this->updateTimestamp = $updateTimestamp;
    $this->importedItemsCount = $importedItemsCount;
    $this->rootCategoryImported = $rootCategoryImported;
    $this->upToDateCategoriesImported = $upToDateCategoriesImported;
    $this->upToDateTaxesImported = $upToDateTaxesImported;
    $this->upToDateProductsImported = $upToDateProductsImported;
    $this->outOfDateCategoriesDeleted = $outOfDateCategoriesDeleted;
    $this->graphQLIteratorQueryItems = $graphQLIteratorQueryItems;
    $this->graphQLIteratorQueryIndex = $graphQLIteratorQueryIndex;
    $this->graphQLIteratorQueryCursor = $graphQLIteratorQueryCursor;
    $this->errorMessage = $errorMessage;
    $this->upToDateCategoryTermIds = $upToDateCategoryTermIds;
  }

  /** @return array|null */
  private function getItems() {
    return $this->graphQLIteratorQueryItems;
  }

  /** @return void */
  private function setItems(?array $items) {
    $this->graphQLIteratorQueryItems = $items;
  }

  /** @return int|null */
  private function getIndex() {
    return $this->graphQLIteratorQueryIndex;
  }

  /** @return void */
  private function setIndex(int $index) {
    $this->graphQLIteratorQueryIndex = $index;
  }

  /** @return string|null */
  private function getCursor() {
    return $this->graphQLIteratorQueryCursor;
  }

  /** @return void */
  private function setCursor(?string $cursor) {
    $this->graphQLIteratorQueryCursor = $cursor;
  }

  /** @return GraphQLPaginatedQueryEdgeNodeIteratorState */
  public function getIterator() {
    return new GraphQLPaginatedQueryEdgeNodeIteratorState(
      $this->getItems(),
      $this->getIndex(),
      $this->getCursor()
    );
  }

  /** @return void */
  public function setIterator(?GraphQLPaginatedQueryEdgeNodeIterator $iterator) {
    $state = $iterator !== null
    ? $iterator->getCurrentState()
    : new GraphQLPaginatedQueryEdgeNodeIteratorState(null, null, null);
    $this->setItems($state->currentPageItems);
    $this->setIndex($state->currentPageIndex);
    $this->setCursor($state->endCursor);
  }
}

class ImportSessionStorage {
  private const OPTION_KEY = "wino-running-import-session";

  public const CREATION_TIMESTAMP_KEY = "creationTimestamp";
  public const UPDATE_TIMESTAMP_KEY = "updateTimestamp";
  public const IMPORTED_ITEMS_COUNT_KEY = "importedItemsCount";
  public const ROOT_CATEGORY_IMPORTED_KEY = "rootCategoryImported";
  public const UP_TO_DATE_CATEGORIES_IMPORTED_KEY = "upToDateCategoriesImported";
  public const UP_TO_DATE_TAXES_IMPORTED_KEY = "upToDateTaxesImported";
  public const UP_TO_DATE_PRODUCTS_IMPORTED_KEY = "upToDateProductsImported";
  public const OUT_OF_DATE_CATEGORIES_DELETED_KEY = "outOfDateCategoriesDeleted";
  public const GRAPH_QL_ITERATOR_QUERY_ITEMS_KEY = "graphQLIteratorQueryItems";
  public const GRAPH_QL_ITERATOR_QUERY_INDEX_KEY = "graphQLIteratorQueryIndex";
  public const GRAPH_QL_ITERATOR_QUERY_CURSOR_KEY = "graphQLIteratorQueryCursor";
  public const ERROR_MESSAGE_KEY = "errorMessage";
  public const UP_TO_DATE_CATEGORY_TERM_IDS_KEY = "upToDateCategoryTermIds";

  /** @return string */
  public static function optionKey(DataScope $scope) {
    return $scope->scopedSlug(self::OPTION_KEY);
  }

  /**
   * @param int[] $upToDateCategoryTermIds
   *
   * @return void
   */
  public static function write(
    DataScope $scope,
    int $creationTimestamp,
    int $updateTimestamp,
    int $importedItemsCount,
    bool $rootCategoryImported,
    bool $upToDateCategoriesImported,
    bool $upToDateTaxesImported,
    bool $upToDateProductsImported,
    bool $outOfDateCategoriesDeleted,
    ?array $graphQLIteratorQueryItems,
    ?int $graphQLIteratorQueryIndex,
    ?string $graphQLIteratorQueryCursor,
    ?string $errorMessage,
    array $upToDateCategoryTermIds
  ) {
    $data = [
      self::CREATION_TIMESTAMP_KEY => $creationTimestamp,
      self::UPDATE_TIMESTAMP_KEY => $updateTimestamp,
      self::IMPORTED_ITEMS_COUNT_KEY => $importedItemsCount,
      self::ROOT_CATEGORY_IMPORTED_KEY => $rootCategoryImported,
      self::UP_TO_DATE_CATEGORIES_IMPORTED_KEY => $upToDateCategoriesImported,
      self::UP_TO_DATE_TAXES_IMPORTED_KEY => $upToDateTaxesImported,
      self::UP_TO_DATE_PRODUCTS_IMPORTED_KEY => $upToDateProductsImported,
      self::OUT_OF_DATE_CATEGORIES_DELETED_KEY => $outOfDateCategoriesDeleted,
      self::GRAPH_QL_ITERATOR_QUERY_ITEMS_KEY => $graphQLIteratorQueryItems,
      self::GRAPH_QL_ITERATOR_QUERY_INDEX_KEY => $graphQLIteratorQueryIndex,
      self::GRAPH_QL_ITERATOR_QUERY_CURSOR_KEY => $graphQLIteratorQueryCursor,
      self::ERROR_MESSAGE_KEY => $errorMessage,
      self::UP_TO_DATE_CATEGORY_TERM_IDS_KEY => $upToDateCategoryTermIds
    ];
    update_option(self::optionKey($scope), $data);
  }

  /** @return void */
  public static function writeData(ImportSessionData $data) {
    self::write(
      $data->getScope(),
      $data->creationTimestamp,
      $data->updateTimestamp,
      $data->importedItemsCount,
      $data->rootCategoryImported,
      $data->upToDateCategoriesImported,
      $data->upToDateTaxesImported,
      $data->upToDateProductsImported,
      $data->outOfDateCategoriesDeleted,
      $data->graphQLIteratorQueryItems,
      $data->graphQLIteratorQueryIndex,
      $data->graphQLIteratorQueryCursor,
      $data->errorMessage,
      $data->upToDateCategoryTermIds
    );
  }

  /** @return ImportSessionData|null */
  public static function readData(DataScope $scope) {
    $data = get_option(self::optionKey($scope), false);
    if (!is_array($data)) return null;
    return new ImportSessionData(
      $scope,
      intData($data, self::CREATION_TIMESTAMP_KEY),
      intData($data, self::UPDATE_TIMESTAMP_KEY),
      intData($data, self::IMPORTED_ITEMS_COUNT_KEY),
      boolData($data, self::ROOT_CATEGORY_IMPORTED_KEY),
      boolData($data, self::UP_TO_DATE_CATEGORIES_IMPORTED_KEY),
      boolData($data, self::UP_TO_DATE_TAXES_IMPORTED_KEY),
      boolData($data, self::UP_TO_DATE_PRODUCTS_IMPORTED_KEY),
      boolData($data, self::OUT_OF_DATE_CATEGORIES_DELETED_KEY),
      optionalArrayData($data, self::GRAPH_QL_ITERATOR_QUERY_ITEMS_KEY),
      optionalIntData($data, self::GRAPH_QL_ITERATOR_QUERY_INDEX_KEY),
      optionalStringData($data, self::GRAPH_QL_ITERATOR_QUERY_CURSOR_KEY),
      optionalStringData($data, self::ERROR_MESSAGE_KEY),
      intArrayData($data, self::UP_TO_DATE_CATEGORY_TERM_IDS_KEY)
    );
  }

  /** @return void */
  public static function deleteData(DataScope $scope) {
    delete_option(self::optionKey($scope));
  }
}

class ImportSession {
  public const TIME_IN_SECONDS_WITHOUT_UPDATE_BEFORE_CONSIDERING_THE_IMPORT_AS_PENDING = 30;
  public const TIME_IN_SECONDS_WITH_AN_ERROR_BEFORE_CONSIDERING_THE_PROGRESS_AS_BLOCKED = 120;

  private const REPORT_OPTION_KEY = "wino-running-import-session-report";
  private const ACTIVE_RUN_ID_KEY = "wino-active-import-session-run-id";
  private string $runId;
  private bool $flushed;

  public ImportSessionData $data;
  private ProductsImportLog $log;

  /** @return int */
  public function creationTimestamp() {
    return $this->data->creationTimestamp;
  }

  /** @return string|null */
  public function errorMessage() {
    return $this->data->errorMessage;
  }

  /** @return int[] */
  public function getUpToDateCategoryTermIds() {
    return $this->log->getUpToDateCategoryTermIds();
  }

  /**
   * @param int[] $ids
   * @return void
   */
  public function setUpToDateCategoryTermIds(array $ids) {
    $this->log->setUpToDateCategoryTermIds($ids);
  }

  private function __construct(
    ImportSessionData $data
  ) {
    $this->runId = uuid();
    $this->data = $data;
    $this->flushed = false;
    $this->log = ProductsImportLog::empty();
    $this->setUpToDateCategoryTermIds($data->upToDateCategoryTermIds);
  }

  /** @return void */
  public function persist() {
    if ($this->flushed) return;
    $this->data->upToDateCategoryTermIds = $this->getUpToDateCategoryTermIds();
    ImportSessionStorage::writeData($this->data);
  }

  /** @return void */
  public function saveReport() {
    if ($this->flushed) return;
    $autoload = false;
    update_option(self::REPORT_OPTION_KEY, $this->log->log->log->asArrays(), $autoload);
  }

  /** @return DataImportLog|null */
  public static function savedReport() {
    wordpressFlushOptionCache(self::REPORT_OPTION_KEY);
    $report = DataImportLog::from(assertedNullableArray(get_option(self::REPORT_OPTION_KEY, null)));
    return $report !== null && count($report->entries) > 0 ? $report : null;
  }

  /** @return void */
  public static function deleteReport() {
    delete_option(self::REPORT_OPTION_KEY);
  }

  /** @return self|null */
  public static function persisted(DataScope $scope) {
    $data = ImportSessionStorage::readData($scope);
    return $data === null ? null : new self($data);
  }

  /** @return void */
  public static function flushScope(DataScope $scope) {
    ImportSessionStorage::deleteData($scope);
  }

  /** @return void */
  public function flush() {
    self::flushScope($this->data->getScope());
    self::deactivateCurrentActiveRun();
    $this->flushed = true;
  }

  /** @return int */
  private static function currentTimestamp() {
    return timestampFrom(dateTimeNow(), TimeUnit::phpTimestampLike());
  }

  /** @return bool */
  public function errorOccured() {
    return $this->errorMessage() !== null;
  }

  /** @return bool */
  public function lastUpdateOlderThanTimeInSeconds(int $timeInSeconds) {
    $now = self::currentTimestamp();
    $lastUpdate = $this->data->updateTimestamp;
    return $now > $lastUpdate + $timeInSeconds;
  }

  /** @return bool */
  public function progressBlockedOnError() {
    return $this->errorOccured() && $this->lastUpdateOlderThanTimeInSeconds(
      self::TIME_IN_SECONDS_WITH_AN_ERROR_BEFORE_CONSIDERING_THE_PROGRESS_AS_BLOCKED
    );
  }

  /**
   * A running import session prevents new import session to be executed.
   * Once the import is done, new import sessions are not prevented.
   *
   * Also, new import session will NOT be prevented if:
   *
   * - The import session reported an error
   * - The import session is older than 12H
   *
   * Those cases would mostly be an error and
   * should not prevent the user to try running again a manual import.
   *
   * The 12H threshold is arbitrary but covers the following criterias:
   * - Long enough for any import to finish under normal condition
   * - Short enough to allow a user to retry at least 1 import per day if something goes wrong
   * - It also happens to be consistent with wordpress short validity duration for the nonces
   *
   * @return bool
   */
  public function preventConcurrentImports() {
    if ($this->isDone() || $this->progressBlockedOnError() || $this->flushed) return false;

    $creation = $this->creationTimestamp();
    $now = self::currentTimestamp();
    $minuteInSeconds = 60;
    $hourInSeconds = 60 * $minuteInSeconds;
    $thresholdInSeconds = 12 * $hourInSeconds;
    $tooSoonToIgnoreUndoneImportSession = $now < $creation + $thresholdInSeconds;

    return $tooSoonToIgnoreUndoneImportSession;
  }

  /** @return bool */
  public static function importPreventConcurrency(?ImportSession $import) {
    return $import === null ? false : $import->preventConcurrentImports();
  }

  /** @return self */
  public static function startAttempt(
    ImportContext $context,
    ImportSessionTranslator $translator
  ) {
    $scope = $context->getScope();
    $persisted = self::persisted($scope);
    $anImportIsRunningYet = $persisted !== null && $persisted->preventConcurrentImports();
    if ($anImportIsRunningYet) throw exception($translator->anImportIsRunningYet());

    self::deactivateCurrentActiveRun();

    $skipCategoriesImport = !$context->categoriesImportEnabled;
    $now = self::currentTimestamp();
    $data = new ImportSessionData(
      $scope,
      $now,
      $now,
      0,
      $skipCategoriesImport,
      $skipCategoriesImport,
      false,
      false,
      false,
      null,
      null,
      null,
      null,
      []
    );
    $session = new self($data);
    $session->persist();
    return $session;
  }

  /** @return string|null */
  public static function activeRunId() {
    $key = self::ACTIVE_RUN_ID_KEY;
    wordpressFlushOptionCache($key);

    $noneId = uuid();
    $id = assertedString(get_option($key, $noneId));

    return $id === $noneId ? null : $id;
  }

  /** @return bool */
  public function isActive() {
    return !$this->flushed && $this->runId === self::activeRunId();
  }

  /** @return void */
  private function activate(ImportSessionTranslator $translator) {
    $autoload = false;
    if (self::activeRunId() === null) update_option(self::ACTIVE_RUN_ID_KEY, $this->runId, $autoload);
    elseif (!$this->isActive()) throw exception($translator->anImportIsRunningYet());
  }

  /** @return void */
  private function deactivate() {
    if ($this->isActive()) self::deactivateCurrentActiveRun();
  }

  /** @return void */
  private static function deactivateCurrentActiveRun() {
    delete_option(self::ACTIVE_RUN_ID_KEY);
  }

  /** @return bool */
  public function isPending() {
    if (!$this->preventConcurrentImports()) return false;
    if (self::activeRunId() === null) return true;

    $tooMuchTimeSinceLastUpdate = $this->lastUpdateOlderThanTimeInSeconds(
      self::TIME_IN_SECONDS_WITHOUT_UPDATE_BEFORE_CONSIDERING_THE_IMPORT_AS_PENDING
    );
    if ($tooMuchTimeSinceLastUpdate) self::deactivateCurrentActiveRun();
    return $tooMuchTimeSinceLastUpdate;
  }

  /** @return void */
  public function run(
    ImportContext $context,
    ImportSessionTranslator $translator
  ) {
    $this->activate($translator);
    $this->runLoop($context);
    $this->deactivate();
  }

  /** @return void */
  private function runLoop(ImportContext $context) {
    $cache = wp_suspend_cache_addition(true);
    try {
      while ($this->nextStep($context)) $this->persist();
      wp_suspend_cache_addition($cache);
    } catch (\Throwable $thrown) {
      wp_suspend_cache_addition($cache);
      if (!$this->isActive()) throw $thrown;

      $this->data->errorMessage = $thrown->getMessage();
      $this->persist();
      $this->saveReport();
      throw $thrown;
    }
    if (!$this->isActive()) return;
    $this->persist();
    $this->saveReport();
  }

  /**
   * @template T of GraphQLPaginatedQueryEdgeNodeIterator
   * @param T $iterator
   * @return T
   */
  private static function resume(ImportSessionData $data, $iterator) {
    $iterator->resume($data->getIterator());
    return $iterator;
  }

  /** @return bool */
  private static function imported(
    ImportSessionData $data,
    GraphQLPaginatedQueryEdgeNodeIterator $iterator,
    bool $nextStepResult
  ) {
    $data->updateTimestamp = self::currentTimestamp();
    $data->importedItemsCount++;
    $data->setIterator($iterator);
    $lastStep = !$nextStepResult;
    if ($lastStep) {
      $data->setIterator(null);
      $data->importedItemsCount = 0;
    }
    return $lastStep;
  }

  /** @return bool */
  private function nextStep(ImportContext $context) {
    $data = $this->data;
    $error = $data->errorMessage;
    if ($error !== null) {
      if ($this->progressBlockedOnError()) throw error($error);
      $data->errorMessage = null;
    }

    $shop = $context->getShop();
    if ($shop === null) return false;

    if (!$this->isActive()) return false;
    $productsLog = $this->log;
    $categoriesLog = $productsLog->log;

    if (!$data->rootCategoryImported) {
      categoriesImportStep($context, $categoriesLog, $context->winoApiRootCategory());
      $data->rootCategoryImported = true;
      $data->setIterator(null);
      $data->importedItemsCount = 0;
      $data->updateTimestamp = self::currentTimestamp();
      return true;
    }

    $apiClient = $context->getApiClient();
    $dataLog = $categoriesLog->log;

    if (!$data->upToDateCategoriesImported) {
      $iterator = self::resume($data, upToDateCategoriesImportIterator($apiClient, $shop));
      $next = upToDateCategoriesImportNextStep($context, $categoriesLog, $iterator);
      $data->upToDateCategoriesImported = self::imported($data, $iterator, $next);
      return true;
    }

    if (!$data->upToDateTaxesImported) {
      $iterator = self::resume($data, upToDateTaxesImportIterator($apiClient, $shop));
      $next = upToDateTaxesImportNextStep($context, $dataLog, $iterator);
      $data->upToDateTaxesImported = self::imported($data, $iterator, $next);
      return true;
    }

    if (!$data->upToDateProductsImported) {
      $iterator = self::resume($data, upToDateProductsImportIterator($apiClient, $shop));
      $next = upToDateProductsImportNextStep($context, $productsLog, $iterator);
      $data->upToDateProductsImported = self::imported($data, $iterator, $next);
      return true;
    }

    deleteOutOfDateCategories($context, $categoriesLog);
    $data->outOfDateCategoriesDeleted = true;

    return false;
  }

  /** @return bool */
  private function categoriesImportInProgress() {
    return !$this->data->rootCategoryImported || !$this->data->upToDateCategoriesImported;
  }

  /** @return bool */
  private function taxesImportInProgress() {
    return !$this->data->upToDateTaxesImported;
  }

  /** @return bool */
  private function productsImportInProgress() {
    return !$this->data->upToDateProductsImported;
  }

  /** @return bool */
  private function unusedCategoriesDeletionInProgress() {
    return !$this->data->outOfDateCategoriesDeleted;
  }

  /** @return string|null */
  public function progressLabel(ImportSessionTranslator $translator) {
    if ($this->progressBlockedOnError()) return null;

    $progress = $this->data->importedItemsCount;
    if ($this->categoriesImportInProgress()) return $translator->categoriesImportInProgress($progress);
    if ($this->taxesImportInProgress()) return $translator->taxesImportInProgress($progress);
    if ($this->productsImportInProgress()) return $translator->productsImportInProgress($progress);
    if ($this->unusedCategoriesDeletionInProgress()) return $translator->unusedCategoriesDeletionInProgress();
    return null;
  }

  /** @return bool */
  private function isDone() {
    return truthy()
    && !$this->categoriesImportInProgress()
    && !$this->taxesImportInProgress()
    && !$this->productsImportInProgress()
    && !$this->unusedCategoriesDeletionInProgress();
  }
}
