<?php

namespace Wino;

class ProductImportResult extends DataImportResult {
  public const EVENT_TARGET_KIND = "Product";
  public const WORDPRESS_REF_KIND = "Product";
  public const WINO_REF_KIND = "Variant";

  protected function eventTargetKind() {
    return self::EVENT_TARGET_KIND;
  }

  protected function wordpressRefKind() {
    return self::WORDPRESS_REF_KIND;
  }

  protected function winoRefKind() {
    return self::WINO_REF_KIND;
  }

  public WinoProduct $product;

  public ?CategoryImportResult $categoryImport;

  /** @return DataImportEvent|null */
  public function categoryEvent() {
    $import = $this->categoryImport;
    return $import === null ? null : $import->event;
  }

  /** @return WinoCategory|null */
  public function importedCategory() {
    $import = $this->categoryImport;
    return $import === null ? null : $import->category;
  }

  /** @return int|null */
  public function importedCategoryTermId() {
    $category = $this->importedCategory();
    return $category === null ? null : $category->getTermId();
  }

  public ?TaxClassImportResult $taxClassImport;

  /** @return DataImportEvent|null */
  public function taxClassEvent() {
    $import = $this->taxClassImport;
    return $import === null ? null : $import->event;
  }

  /** @return WinoTaxClass|null */
  public function importedTaxClass() {
    $import = $this->taxClassImport;
    return $import === null ? null : $import->taxClass;
  }

  /** @return string|null */
  public function importedTaxClassSlug() {
    $taxClass = $this->importedTaxClass();
    return $taxClass === null ? null : $taxClass->slug;
  }

  public ?TaxRateImportResult $taxRateImport;

  /** @return DataImportEvent|null */
  public function taxRateEvent() {
    $import = $this->taxRateImport;
    return $import === null ? null : $import->event;
  }

  /** @return WinoTaxRate|null */
  public function importedTaxRate() {
    $import = $this->taxRateImport;
    return $import === null ? null : $import->taxRate;
  }

  public function __construct(
    WinoProduct $product,
    DataImportEvent $event,
    ?CategoryImportResult $categoryImport,
    ?TaxClassImportResult $taxClassImport,
    ?TaxRateImportResult $taxRateImport
  ) {
    $postId = $product->getPostId();
    parent::__construct(
      $event,
      $product->getName(),
      "$postId",
      $product->getWinoApiVariantId(),
      $event->isDeleted() ? null : $product->getEditionPageUrl(),
      $product->getWinoPageUrl()
    );
    $this->product = $product;
    $this->categoryImport = $categoryImport;
    $this->taxClassImport = $taxClassImport;
    $this->taxRateImport = $taxRateImport;
  }

  /** @return ProductImportResult */
  public static function read(
    WinoProduct $product,
    ?CategoryImportResult $categoryImport,
    ?TaxClassImportResult $taxClassImport,
    ?TaxRateImportResult $taxRateImport
  ) {
    return new self(
      $product,
      DataImportEvent::read(),
      $categoryImport,
      $taxClassImport,
      $taxRateImport
    );
  }

  /** @return ProductImportResult */
  public static function created(
    WinoProduct $product,
    ?CategoryImportResult $categoryImport,
    ?TaxClassImportResult $taxClassImport,
    ?TaxRateImportResult $taxRateImport
  ) {
    return new self(
      $product,
      DataImportEvent::created(),
      $categoryImport,
      $taxClassImport,
      $taxRateImport
    );
  }

  /** @return ProductImportResult */
  public static function updated(
    WinoProduct $product,
    ?CategoryImportResult $categoryImport,
    ?TaxClassImportResult $taxClassImport,
    ?TaxRateImportResult $taxRateImport
  ) {
    return new self(
      $product,
      DataImportEvent::updated(),
      $categoryImport,
      $taxClassImport,
      $taxRateImport
    );
  }

  /** @return ProductImportResult */
  public static function archived(
    WinoProduct $product,
    ?CategoryImportResult $categoryImport,
    ?TaxClassImportResult $taxClassImport,
    ?TaxRateImportResult $taxRateImport
  ) {
    return new self(
      $product,
      DataImportEvent::archived(),
      $categoryImport,
      $taxClassImport,
      $taxRateImport
    );
  }

  /** @return ProductImportResult */
  public static function deleted(
    WinoProduct $product,
    ?CategoryImportResult $categoryImport,
    ?TaxClassImportResult $taxClassImport,
    ?TaxRateImportResult $taxRateImport
  ) {
    return new self(
      $product,
      DataImportEvent::deleted(),
      $categoryImport,
      $taxClassImport,
      $taxRateImport
    );
  }
}

/** @return WinoProduct|null */
function productFromImport(?ProductImportResult $result) {
  return $result === null ? null : $result->product;
}

/**
 * It seems that woocommerce also has a notion of WC_Product_Variation.
 *
 * WC_Product is currently used to import variants but using the
 * WC_Product_Variation class instead might have sense. Not sure however
 * of what are supposed to be WC_Product_Variation in Woocommerce. Maybe
 * that option was yet considered but not appropriate.
 *
 * @return ProductImportResult|null
 */
function importWinoProduct(
  ImportContext $context,
  WinoApiVariant $apiVariant
) {
  $shop = $context->getShop();
  if ($shop === null || $apiVariant->shop->id !== $shop->id) return null;

  $translator = $context->getTranslator();
  $scope = $context->getScope();
  $customRoot = $context->getCustomRootCategory();
  $apiRootCategory = $scope->winoApiRootCategory($shop->id);

  $priceList = $context->priceListName;
  $categoryImportEnabled = $context->categoriesImportEnabled;
  $descriptionUpdateEnabled = $context->productDescriptionUpdateEnabled;
  $descriptionFormat = $context->productDescriptionFormat;
  $nameUpdateEnabled = $context->productNameUpdateEnabled;
  $priceUpdateEnabled = $context->priceUpdateEnabled;
  $skuImportEnabled = $context->skuBasedImportEnabled;
  $manageStockOverrideEnabled = $context->stockManagementOverrideEnabled;

  $apiVariantId = $apiVariant->id;
  $formattedName = $apiVariant->formattedName;
  $sku = $skuImportEnabled ? $apiVariant->stockKeepingUnit : null;
  $apiProduct = $apiVariant->product;
  $missingApiProduct = $apiProduct === null;
  $apiProductId = $apiProduct === null ? null : $apiProduct->id;
  $apiTax = $apiProduct === null ? null : $apiProduct->tax;
  $apiTaxId = $apiTax === null ? null : $apiTax->id;
  $apiCategory = $apiProduct === null ? null : $apiProduct->category;
  $active = $apiVariant->isActive();
  $archived = $apiVariant->isArchived();
  $bulk = $apiVariant->isBulk();
  $country = WinoTaxRate::DEFAULT_COUNTRY;
  $stockQuantity = $apiVariant->stock->getQuantity();

  $existingByVariantId = mergedWinoProductsByApiVariantId($context, $apiVariantId);

  $existingBySku = $sku === null ? null : WinoProduct::fromSku($scope, $sku);
  $foundBySku = $existingBySku !== null;
  $existingBySkuVariantId = $existingBySku === null ? null : $existingBySku->getWinoApiVariantId();

  $existingByVariantIdPostId = $existingByVariantId === null ? null : $existingByVariantId->getPostId();
  $existingBySkuPostId = $existingBySku === null ? null : $existingBySku->getPostId();

  $samePostId = $existingByVariantIdPostId !== null && $existingByVariantIdPostId === $existingBySkuPostId;

  $existingProduct = $existingBySku !== null ? $existingBySku : $existingByVariantId;
  $duplicateSku = $foundBySku && !$samePostId && $existingBySkuVariantId !== null;

  if ($duplicateSku && $existingBySku !== null)
  throw duplicateSkuException($translator, $apiVariant, $existingBySku, $existingByVariantId);

  $createMode = $existingProduct === null;

  if ($bulk && $existingProduct !== null) throw msgException(
    "A bulk variant was imported:", jsonString($apiVariant),
    'imported as', jsonString($existingProduct->asArray())
  );

  $price = null;
  $selectedPriceListName = PriceList::normalizeName($priceList);
  $apiVariantPricesEdges = $apiVariant->variantPrices->edges;
  foreach ($apiVariantPricesEdges as $apiVariantPricesEdge) {
    $node = $apiVariantPricesEdge->node;
    $nodePrice = $node->price;
    if ($nodePrice === null) continue;
    if (!PriceList::priceNameEqualsName($nodePrice, $selectedPriceListName)) continue;
    $price = $node->valueExcludingTax;
    break;
  }

  $importable = $active
  && !$archived
  && !$bulk
  && $price !== null
  && $apiProduct !== null
  && $apiTax !== null;
  if ($createMode && !$importable) return null;

  $taxClass = null;
  $taxClassImport = null;
  $taxRateImport = null;
  $existingTaxClassSlug = $existingProduct !== null ? $existingProduct->getTaxClassSlug() : null;

  $taxUpdateAllowed = $existingTaxClassSlug !== null
  ? WinoTaxKeys::isWinoTaxClassSlug($scope, $existingTaxClassSlug)
  : true;

  if ($taxUpdateAllowed) {
    $taxKeys = $apiTax !== null ? WinoTaxKeys::fromApiTax($scope, $country, $apiTax) : null;
    $taxClass = $taxKeys !== null ? findWinoTaxClassBySlug($context, $taxKeys->taxClassSlug) : null;

    if ($taxClass === null && $apiTax !== null) {
      $taxImport = importWinoTax($context, $apiTax);
      $taxClassImport = $taxImport === null ? null : $taxImport->taxClassImport;
      $taxRateImport = $taxImport === null ? null : $taxImport->taxRateImport;
      $taxClass = $taxClassImport === null ? null : $taxClassImport->taxClass;
    } elseif ($taxClass !== null) $taxClassImport = TaxClassImportResult::read($taxClass);
  }

  $taxable = $createMode && $taxClass !== null;
  $noneTaxStatus = WinoProduct::NONE_TAX_STATUS;
  $taxableTaxStatus = WinoProduct::TAXABLE_TAX_STATUS;
  $existingTaxStatus = $existingProduct !== null ? $existingProduct->getTaxStatus() : $noneTaxStatus;
  $taxStatusOverrideAllowed = $existingTaxStatus === $noneTaxStatus;
  $taxStatus = $taxable || $taxStatusOverrideAllowed ? $taxableTaxStatus : null;

  $category = null;
  $categoryImport = null;

  if ($categoryImportEnabled)
  if ($apiCategory === null) {
    $category = findRootCategory($context);

    if ($category === null)
    if ($customRoot !== null) $customRoot->assertsValidity($translator);
    else {
      $defaultRootImport = importWinoCategory($context, $apiRootCategory);
      $category = $defaultRootImport === null ? null : $defaultRootImport->category;
    }
  } else {
    $category = findWinoCategoryByApiId($context, $apiCategory->id);
    $categoryImport = $category === null
    ? importWinoCategory($context, $apiCategory)
    : CategoryImportResult::read($category);
    $category = $categoryImport === null ? null : $categoryImport->category;
  }

  $archiveMode = false;
  $pendingStatus = 'pending';
  $privateStatus = 'private';
  $existingArchive = $existingProduct === null ? false : $existingProduct->getStatus() === $privateStatus;
  $unarchive = $existingArchive && !$archived;
  $status = null;
  if ($archived && !$createMode) {
    $archiveMode = true; $status = $privateStatus;
  } elseif ($createMode || !$active || $unarchive) $status = $pendingStatus;

  $update = WinoProductUpdateProxy::updateOrCreate($scope, $existingProduct, $customRoot);
  $manageStock = $manageStockOverrideEnabled || $existingProduct === null ? true : $existingProduct->getManageStock();

  if ($createMode || $nameUpdateEnabled) $update->setName($formattedName);
  if ($createMode || $priceUpdateEnabled) if ($price !== null) $update->setPrice($price);

  if ($createMode || $descriptionUpdateEnabled) $update->setDescription(
    winoProductDescriptionFromApiVariant($apiVariant, $descriptionFormat)
  );

  if ($sku !== null) $update->setSku($sku);

  $updateCategory = !$missingApiProduct || !$categoryImportEnabled;
  if ($updateCategory) productCategoryUpdate($context, $update, $category);

  if ($taxClass !== null) $update->setTaxClass($taxClass);
  if ($taxStatus !== null) $update->setTaxStatus($taxStatus);
  if ($status !== null) $update->setStatus($status);

  if ($createMode || $manageStockOverrideEnabled) $update->setManageStock(true);
  if ($manageStock) $update->setStockQuantity($stockQuantity);

  $update->setWinoApiVariantId($apiVariantId);
  $update->setWinoApiTaxId($apiTaxId);
  $update->setWinoApiProductId($apiProductId);

  $updateMode = $update->commit();

  $event = DataImportEvent::read();
  if ($updateMode) $event = DataImportEvent::updated();
  if ($createMode) $event = DataImportEvent::created();
  if ($archiveMode) $event = DataImportEvent::archived();

  $updatedProduct = $update->getProduct();

  // The following line takes care of merging potential duplicate products
  // Such case might happen in case of concurrent import requests which is
  // not really testable through unit testing.
  // Therefore, while the mergedWinoProductsByApiVariantId function itself
  // is tested, its usage here can't be ensured through unit testing.
  // Just be aware that the following line is required to avoid product duplicates
  $mergedProduct = mergedWinoProductsByApiVariantId($context->emptyCacheCopy(), $apiVariantId);

  return new ProductImportResult(
    $mergedProduct !== null ? $mergedProduct : $updatedProduct,
    $event,
    $categoryImport,
    $taxClassImport,
    $taxRateImport
  );
}

class ProductsImportLog {
  public CategoriesImportLog $log;

  /** @return int[] */
  public function getUpToDateCategoryTermIds() {
    return $this->log->upToDateCategoryTermIds;
  }

  /**
   * @param int[] $ids
   * @return void
   */
  public function setUpToDateCategoryTermIds(array $ids) {
    $this->log->upToDateCategoryTermIds = $ids;
  }

  public function __construct(CategoriesImportLog $log) {
    $this->log = $log;
  }

  /** @return self */
  public static function from(CategoriesImportLog $log) {
    return new self($log);
  }

  /** @return self */
  public static function empty() {
    return self::from(CategoriesImportLog::empty());
  }

  /** @return void */
  public function logImportResult(?ProductImportResult $result) {
    $categoriesLog = $this->log;
    $log = $categoriesLog->log;

    $log->logImportResult($result);

    if ($result === null) return;
    $log->logImportResult($result->taxClassImport);
    $log->logImportResult($result->taxRateImport);
    $categoriesLog->logImportResult($result->categoryImport);
  }
}


/** @return void */
function productsImportStep(
  ImportContext $context,
  ProductsImportLog $log,
  WinoApiVariant $apiVariant
) {
  $log->logImportResult(importWinoProduct($context, $apiVariant));
}

/** @return bool */
function upToDateProductsImportNextStep(
  ImportContext $context,
  ProductsImportLog $log,
  WinoApiVariantQueryAllIterator $iterator
) {
  if (!$iterator->next()) return false;
  productsImportStep($context, $log, $iterator->currentObject());
  return true;
}

/** @return WinoApiVariantQueryAllIterator */
function upToDateProductsImportIterator(
  WinoApiClient $winoApi,
  WinoApiShopIdentity $shop
) {
  return WinoApiVariant::queryAllIterator($winoApi, $shop);
}

/** @return void */
function importUpToDateProducts(
  ImportContext $context,
  ProductsImportLog $log
) {
  $shop = $context->getShop();
  if ($shop === null) return;

  $cache = wp_suspend_cache_addition(true);
  $iterator = upToDateProductsImportIterator($context->getApiClient(), $shop);
  while (upToDateProductsImportNextStep($context, $log, $iterator));
  wp_suspend_cache_addition($cache);
}
