<?php

namespace Wino;

/** @return array<'draft' | 'pending' | 'private' | 'publish' | 'future' | 'trash' | 'auto-draft' | 'inherit'> */
function allWinoProductsPostStatus() {
  return ['draft', 'pending', 'private', 'publish', 'future', 'trash', 'auto-draft', 'inherit'];
}

/** @return GetPostsArgs */
function findWinoProductsGetPostsArgs(ImportContext $context, GetPostsArgs $args) {
  $scope = $context->getScope();
  $winoIdMetaKey = $scope->winoIdMetaKey();

  $getPostsArgs = new GetPostsArgs([
    'post_type' => WinoProduct::POST_TYPE,
    'post_status' => allWinoProductsPostStatus(),
    'numberposts' => -1,
    'meta_key' => $winoIdMetaKey
  ]);
  $getPostsArgs->set($args->data);
  return $getPostsArgs;
}

/** @return int[] */
function findWinoProductIds(ImportContext $context, GetPostsArgs $args) {
  return findWordpressPostIds(findWinoProductsGetPostsArgs($context, $args));
}

/** @return WinoProduct[] */
function findWinoProducts(ImportContext $context, GetPostsArgs $args) {
  $scope = $context->getScope();
  $productPosts = findWordpressPosts(findWinoProductsGetPostsArgs($context, $args));
  $products = [];

  foreach ($productPosts as $productPost) {
    $wcProduct = wc_get_product($productPost);
    if (!($wcProduct instanceof \WC_Product)) throw msgError(
      'Unexpected not found product', jsonString($productPost)
    );
    $winoProduct = new WinoProduct($scope, $wcProduct);
    $postIdFromPost = $productPost->ID;
    $postIdFromWinoProduct = $winoProduct->getPostId();
    if ($postIdFromPost !== $postIdFromWinoProduct) throw msgError(
      "Post id from post ($postIdFromPost) and",
      "post id from wino product ($postIdFromWinoProduct) not matching."
    );
    $products[] = $winoProduct;
  }

  return $products;
}

/** @return int[] */
function findAllWinoProductIds(ImportContext $context) {
  return findWinoProductIds($context, new GetPostsArgs());
}

/** @return WinoProduct[] */
function findAllWinoProducts(ImportContext $context) {
  return findWinoProducts($context, new GetPostsArgs());
}

/** @return WinoProduct[] */
function findWinoProductsByMeta(DataScope $scope, string $metaKey, string $metaValue) {
  $products = [];
  $productPosts = findWordpressPosts(new GetPostsArgs([
    'post_type' => WinoProduct::POST_TYPE,
    'post_status' => allWinoProductsPostStatus(),
    'numberposts' => -1,
    'meta_key' => $metaKey,
    'meta_value' => $metaValue
  ]));
  foreach ($productPosts as $post) {
    $product = WinoProduct::fromPost($scope, $post);
    if ($product !== null) $products[] = $product;
  }
  return $products;
}

class DuplicateWinoProductMergingData {
  public WinoProduct $product;
  public int $postId;
  public int $statusPriority;
  public int $updateTimestamp;

  private function __construct(
    WinoProduct $product,
    int $postId,
    int $statusPriority,
    int $updateTimestamp
  ) {
    $this->product = $product;
    $this->postId = $postId;
    $this->statusPriority = $statusPriority;
    $this->updateTimestamp = $updateTimestamp;
  }

  /** @return self */
  public static function from(WinoProduct $product) {
    $postId = $product->getPostId();

    return new self(
      $product,
      $postId,
      self::statusPriority($product->getStatus()),
      getWordpressPostUpdateTimestamp($postId)
    );
  }

  /** @return string|null */
  public function getSku() {
    return $this->product->getSku();
  }

  /** @return void */
  public function mergeSku(?string $sku) {
    if ($sku === null) return;
    $update = WinoProductUpdateProxy::fromProduct($this->product);
    $update->setSku($sku);
    $update->commit();
  }

  /** @return int */
  public static function statusPriority(string $status) {
    $max = 25;
    switch ($status) {
      case 'publish':
      return $max;

      case 'future':
      return $max - 1;

      case 'pending':
      case 'private':
      return $max - 2;

      case 'draft':
      return $max - 3;

      case 'auto-draft':
      return $max - 4;

      case 'trash':
      return $max - 5;

      default:
      return $max - 6;
    }
  }

  /** @return int */
  public static function compare(DuplicateWinoProductMergingData $a, DuplicateWinoProductMergingData $b) {
    $idA = $a->postId;
    $idB = $b->postId;
    $priorityA = $a->statusPriority;
    $priorityB = $b->statusPriority;
    $updateA = $a->updateTimestamp;
    $updateB = $b->updateTimestamp;

    // For intance:
    // Published ones are prioritized over pending ones,
    // which themself are prioritized over trash ones...
    $priorityDiff = $priorityB - $priorityA;
    if ($priorityDiff !== 0) return $priorityDiff;

    // Both have the same status priority, choose the last one modified
    $updateDiff = $updateB - $updateA;
    if ($updateDiff !== 0) return $updateDiff;

    // Both are published & were modified at same time,
    // fallback to the last that was created based on the post id
    return $idB - $idA;
  }
}

class DuplicateWinoProduct {

  /** @var WinoProduct[] */
  public array $products;

  /**
   * @param WinoProduct[] $products
   */
  public function __construct(
    $products
  ) {
    $this->products = $products;
  }

  /** @return WinoProduct */
  public function merge() {
    return self::mergeProducts($this->products);
  }

  /**
   * @param WinoProduct[] $products
   * @return WinoProduct
   */
  public static function mergeProducts($products) {
    $data = [];
    foreach ($products as $product) $data[] = DuplicateWinoProductMergingData::from($product);
    $sorted = sortArrayCopy($data, function ($a, $b) {
      return DuplicateWinoProductMergingData::compare($a, $b);
    });

    $lastKnownSku = null;
    foreach ($sorted as $duplicate) {
      $sku = $duplicate->getSku();
      if ($sku !== null) $lastKnownSku = $sku;
    }

    $latestData = assertedNotNull(listElement($sorted, 0), $sorted);
    $olderData = array_slice($sorted, 1);
    foreach ($olderData as $duplicate) $duplicate->product->wooCommerceProduct->delete(true);

    $latestData->mergeSku($lastKnownSku);
    return $latestData->product;
  }

  /** @return void */
  public function sortByPostId() {
    $this->products = sortArrayCopyByInt($this->products, function ($product) {
      return $product->getPostId();
    });
  }
}

/** @return WinoProduct|DuplicateWinoProduct|null */
function findWinoProductByApiVariantId(ImportContext $context, string $id) {
  $scope = $context->getScope();
  $metaKey = $scope->winoIdMetaKey();

  $products = findWinoProductsByMeta($scope, $metaKey, $id);

  $a = listElement($products, 0);
  $b = listElement($products, 1);

  if ($a === null) return null;
  if ($b === null) return $a;
  return new DuplicateWinoProduct($products);
}

/** @return DuplicateWinoProduct|null */
function findDuplicateWinoProductByApiVariantId(ImportContext $context, string $id) {
  $found = findWinoProductByApiVariantId($context, $id);
  return $found instanceof DuplicateWinoProduct ? $found : null;
}

/** @return WinoProduct|null */
function findUniqueWinoProductByApiVariantId(ImportContext $context, string $id) {
  $found = findWinoProductByApiVariantId($context, $id);
  return $found instanceof DuplicateWinoProduct ? null : $found;
}

/** @return WinoProduct|null */
function mergedWinoProductsByApiVariantId(ImportContext $context, string $id) {
  $found = findWinoProductByApiVariantId($context, $id);
  return $found instanceof DuplicateWinoProduct ? $found->merge() : $found;
}

/** @return WinoProduct[] */
function findWinoProductsByApiProductId(ImportContext $context, string $id) {
  $scope = $context->getScope();
  $metaKey = $scope->winoProductIdMetaKey();
  return findWinoProductsByMeta($scope, $metaKey, $id);
}

/** @return WinoProduct[] */
function findWinoProductsByApiTaxId(ImportContext $context, string $id) {
  $scope = $context->getScope();
  $metaKey = $scope->winoTaxIdMetaKey();
  return findWinoProductsByMeta($scope, $metaKey, $id);
}

/** @return string[] */
function winoVariantIdsFrom(WinoProduct ...$products) {
  $ids = [];
  foreach ($products as $product) {
    if (!($product instanceof WinoProduct)) continue;
    $id = $product->getWinoApiVariantId();
    if ($id !== null) $ids[] = $id;
  }
  return $ids;
}

/** @return string[] */
function findWinoVariantIdsByApiProductId(ImportContext $context, string $id) {
  $products = findWinoProductsByApiProductId($context, $id);
  return winoVariantIdsFrom(...$products);
}

/** @return string[] */
function findWinoVariantIdsByApiTaxId(ImportContext $context, string $id) {
  $products = findWinoProductsByApiTaxId($context, $id);
  return winoVariantIdsFrom(...$products);
}

/** @return string */
function winoProductDescriptionFromApiVariant(
  WinoApiVariant $apiVariant,
  string $format
) {
  $description = $format;
  $placeholders = ConfigurationProductDescriptionFormat::PLACEHOLDERS;
  foreach ($placeholders as $placeholder) $description = str_replace(
    $placeholder,
    ConfigurationProductDescriptionFormat::placeholderValue($apiVariant, $placeholder),
    $description
  );
  return $description;
}

/** @return ProductImportResult */
function deleteWinoProduct(ImportContext $context, WinoProduct $product) {
  noop($context);
  $result = ProductImportResult::deleted($product, null, null, null);
  $product->wooCommerceProduct->delete(true);
  return $result;
}

/** @return void */
function deleteAllWinoProducts(ImportContext $context, DataImportLog $log) {
  $scope = $context->getScope();
  $ids = findAllWinoProductIds($context);
  foreach ($ids as $id) {
    $product = WinoProduct::fromPostId($scope, $id);
    if ($product === null) continue;
    $log->logImportResult(deleteWinoProduct($context, $product));
  }
}

/** @return WinoProduct[] */
function findWinoCategoryProducts(ImportContext $context, WinoCategory $category) {
  return findWinoProducts($context, new GetPostsArgs([
    'tax_query' => [[
      'taxonomy' => $category->getTaxonomy(),
      'field'    => 'term_id',
      'terms'    => $category->getTermId(),
      'include_children' => false
    ]]
  ]));
}

/** @return WinoProduct[] */
function findAllMergedWinoProducts(ImportContext $context) {
  $all = findAllWinoProducts($context);

  $duplicateIds = [];
  $refsMap = [];
  $refsIds = [];
  foreach ($all as $product) {
    $variantId = $product->getWinoApiVariantId();
    if ($variantId === null) continue;
    $duplicate = key_exists($variantId, $refsMap);
    $firstDuplicate = $duplicate && !in_array($variantId, $duplicateIds);
    if ($firstDuplicate) $duplicateIds[] = $variantId;
    if ($duplicate) continue;
    $refsMap[$variantId] = $product;
    $refsIds[] = $variantId;
  }
  if (count($duplicateIds) === 0) return $all;

  foreach ($duplicateIds as $variantId) {
    $merged = mergedWinoProductsByApiVariantId($context, $variantId);
    if ($merged === null) continue;
    $refsMap[$variantId] = $merged;
  }

  $allMerged = [];
  foreach ($refsIds as $variantId) {
    $product = arrayElement($refsMap, $variantId);
    if ($product === null) continue;
    $allMerged[] = $product;
  }

  return $allMerged;
}

/** @return void */
function productCategoryUpdate(
  ImportContext $context,
  WinoProductUpdateProxy $update,
  ?WinoCategory $category
) {
  $categoryImportEnabled = $context->categoriesImportEnabled;
  $explicitCategoriesHierarchyEnabled = $context->explicitCategoriesHierarchyEnabled;
  $customRoot = $context->getCustomRootCategory();

  if (!$categoryImportEnabled) $update->unsetCategories();

  if ($category !== null) $update->setCategoriesHierarchy(
    $explicitCategoriesHierarchyEnabled, $category
  ); elseif ($customRoot !== null) $update->unsetCategories();
}
