<?php

namespace Wino;

class TaxClassImportResult extends DataImportResult {
  public const EVENT_TARGET_KIND = "Tax";
  public const WORDPRESS_REF_KIND = "Tax class";
  public const WINO_REF_KIND = "Tax";

  protected function eventTargetKind() {
    return self::EVENT_TARGET_KIND;
  }

  protected function wordpressRefKind() {
    return self::WORDPRESS_REF_KIND;
  }

  protected function winoRefKind() {
    return self::WINO_REF_KIND;
  }

  public WinoTaxClass $taxClass;

  public function __construct(
    WinoTaxClass $taxClass,
    DataImportEvent $event
  ) {
    parent::__construct(
      $event,
      $taxClass->name,
      $taxClass->slug,
      $taxClass->getWinoName(),
      null,
      null
    );
    $this->taxClass = $taxClass;
  }

  /** @return TaxClassImportResult */
  public static function read(WinoTaxClass $taxClass) {
    return new self($taxClass, DataImportEvent::read());
  }

  /** @return TaxClassImportResult */
  public static function created(WinoTaxClass $taxClass) {
    return new self($taxClass, DataImportEvent::created());
  }

  /** @return TaxClassImportResult */
  public static function updated(WinoTaxClass $taxClass) {
    return new self($taxClass, DataImportEvent::updated());
  }

  /** @return TaxClassImportResult */
  public static function archived(WinoTaxClass $taxClass) {
    return new self($taxClass, DataImportEvent::archived());
  }

  /** @return TaxClassImportResult */
  public static function deleted(WinoTaxClass $taxClass) {
    return new self($taxClass, DataImportEvent::deleted());
  }
}

class TaxRateImportResult extends DataImportResult {
  public const EVENT_TARGET_KIND = TaxClassImportResult::EVENT_TARGET_KIND;
  public const WORDPRESS_REF_KIND = "Tax rate";
  public const WINO_REF_KIND = TaxClassImportResult::WINO_REF_KIND;

  protected function eventTargetKind() {
    return self::EVENT_TARGET_KIND;
  }

  protected function wordpressRefKind() {
    return self::WORDPRESS_REF_KIND;
  }

  protected function winoRefKind() {
    return self::WINO_REF_KIND;
  }

  public WinoTaxRate $taxRate;

  public function __construct(
    WinoTaxRate $taxRate,
    DataImportEvent $event
  ) {
    $rateId = $taxRate->id;
    parent::__construct(
      $event,
      $taxRate->name,
      "$rateId",
      $taxRate->getWinoName(),
      null,
      null
    );
    $this->taxRate = $taxRate;
  }

  /** @return TaxRateImportResult */
  public static function read(WinoTaxRate $taxRate) {
    return new self($taxRate, DataImportEvent::read());
  }

  /** @return TaxRateImportResult */
  public static function created(WinoTaxRate $taxRate) {
    return new self($taxRate, DataImportEvent::created());
  }

  /** @return TaxRateImportResult */
  public static function updated(WinoTaxRate $taxRate) {
    return new self($taxRate, DataImportEvent::updated());
  }

  /** @return TaxRateImportResult */
  public static function archived(WinoTaxRate $taxRate) {
    return new self($taxRate, DataImportEvent::archived());
  }

  /** @return TaxRateImportResult */
  public static function deleted(WinoTaxRate $taxRate) {
    return new self($taxRate, DataImportEvent::deleted());
  }
}

class TaxImportResult {
  public TaxClassImportResult $taxClassImport;
  public TaxRateImportResult $taxRateImport;

  public function __construct(
    TaxClassImportResult $taxClassImport,
    TaxRateImportResult $taxRateImport
  ) {
    $this->taxClassImport = $taxClassImport;
    $this->taxRateImport = $taxRateImport;
  }
}

/** @return TaxImportResult|null */
function importWinoTax(ImportContext $context, WinoApiTax $apiTax) {
  $shop = $context->getShop();
  if ($shop === null || $apiTax->shop->id !== $shop->id) return null;
  if ($apiTax->isArchived()) return null;

  $scope = $context->getScope();
  $country = WinoTaxRate::DEFAULT_COUNTRY;
  $taxKeys = WinoTaxKeys::fromApiTax($scope, $country, $apiTax);
  $taxClassName = $taxKeys->taxClassName;
  $taxClassSlug = $taxKeys->taxClassSlug;
  $taxRateName = $taxKeys->taxRateName;
  $value = $apiTax->value;

  $createTaxClassMode = false;
  $createTaxRateMode = false;

  $existingTaxClass = findWinoTaxClassBySlug($context, $taxClassSlug);
  if ($existingTaxClass === null) {
    $createTaxClassMode = true;
    $createdTaxClass = \WC_Tax::create_tax_class($taxClassName, $taxClassSlug);
    if ($createdTaxClass instanceof \WP_Error) throw wordpressErrorException(
      $createdTaxClass,
      "Happened while trying to create $taxClassName ($taxClassSlug)."
    );
  }

  $taxClass = WinoTaxClass::fromKeys($scope, $taxKeys);
  $taxRate = $createTaxClassMode ? null : findWinoTaxClassRateByName(
    $context,
    $taxClass,
    $taxRateName
  );

  if ($taxRate === null) {
    $createTaxRateMode = true;
    $ratePriority = 0;
    $rateInput = [
      'tax_rate_priority' => $ratePriority,
      'tax_rate_country' => $country,
      'tax_rate' => $value,
      'tax_rate_name' => $taxRateName,
      'tax_rate_class' => $taxClassSlug
    ];
    $createdRateId = \WC_Tax::_insert_tax_rate($rateInput);
    \WC_Tax::_update_tax_rate_postcodes($createdRateId, '*');
    \WC_Tax::_update_tax_rate_cities($createdRateId, '*');
    $taxRate = new WinoTaxRate(
      $scope,
      $createdRateId,
      $taxRateName,
      $value,
      $taxClassSlug,
      $country,
      $ratePriority
    );
  }

  $taxClassEvent = $createTaxClassMode ? DataImportEvent::created() : DataImportEvent::read();
  $taxRateEvent = $createTaxRateMode ? DataImportEvent::created() : DataImportEvent::read();

  return new TaxImportResult(
    new TaxClassImportResult(
      $context->cacheTaxClass($taxClass),
      $taxClassEvent
    ),
    new TaxRateImportResult(
      $context->cacheTaxRate($taxRate),
      $taxRateEvent
    )
  );
}

/** @return void */
function taxesImportStep(
  ImportContext $context,
  DataImportLog $log,
  WinoApiTax $apiTax
) {
  $import = importWinoTax($context, $apiTax);
  if ($import === null) return;
  $log->logImportResult($import->taxClassImport);
  $log->logImportResult($import->taxRateImport);
}

/** @return bool */
function upToDateTaxesImportNextStep(
  ImportContext $context,
  DataImportLog $log,
  WinoApiTaxQueryAllIterator $iterator
) {
  if (!$iterator->next()) return false;
  taxesImportStep($context, $log, $iterator->currentObject());
  return true;
}

/** @return WinoApiTaxQueryAllIterator */
function upToDateTaxesImportIterator(
  WinoApiClient $winoApi,
  WinoApiShopIdentity $shop
) {
  return WinoApiTax::queryAllIterator($winoApi, $shop);
}

/** @return void */
function importUpToDateTaxes(
  WinoApiClient $winoApi,
  ImportContext $context,
  DataImportLog $log
) {
  $shop = $context->getShop();
  if ($shop === null) return;

  $cache = wp_suspend_cache_addition(true);
  $iterator = upToDateTaxesImportIterator($winoApi, $shop);
  while (upToDateTaxesImportNextStep($context, $log, $iterator));
  wp_suspend_cache_addition($cache);
}

/** @return TaxClassImportResult|null */
function deleteWinoTaxClass(ImportContext $context, WinoTaxClass $taxClass) {
  $scope = $context->getScope();
  $result = TaxClassImportResult::deleted($taxClass);

  $taxClassSlug = $taxClass->slug;
  if (!WinoTaxKeys::isWinoTaxClassSlug($scope, $taxClassSlug)) return null;

  $deletion = \WC_Tax::delete_tax_class_by('slug', $taxClassSlug);
  if ($deletion instanceof \WP_Error) throw wordpressErrorException($deletion);

  $context->unsetTaxClassCache($taxClass);
  return $result;
}

/** @return TaxRateImportResult|null */
function deleteWinoTaxRate(ImportContext $context, WinoTaxRate $taxRate) {
  $scope = $context->getScope();
  $result = TaxRateImportResult::deleted($taxRate);

  $taxClassSlug = $taxRate->taxClassSlug;
  if (!WinoTaxKeys::isWinoTaxClassSlug($scope, $taxClassSlug)) return null;
  \WC_Tax::_delete_tax_rate($taxRate->id);

  $context->unsetTaxRateCache($taxRate);
  return $result;
}

/** @return void */
function deleteAllWinoTaxClasses(
  ImportContext $context,
  DataImportLog $log
) {
  $winoTaxClasses = findAllWinoTaxClasses($context);
  foreach ($winoTaxClasses as $winoTaxClass) $log->logImportResult(
    deleteWinoTaxClass($context, $winoTaxClass)
  );
}
