<?php

namespace Wino;

class WinoApiSynchronizationContext {
  public WebDeviceCache $webDeviceCache;
  private ?ImportContext $importContext;

  private function __construct(?ImportContext $importContext) {
    $this->webDeviceCache = WebDeviceCache::empty();
    $this->importContext = $importContext;
  }

  /** @return self */
  public static function fromConfiguration() {
    return new self(null);
  }

  /** @return self */
  public static function fromImportContext(ImportContext $importContext) {
    return new self($importContext);
  }

  /** @return ImportContext */
  public function getImportContext() {
    $importContext = $this->importContext;
    if ($importContext !== null) return $importContext;

    $importContext = ImportContext::fromConfiguration();
    $this->importContext = $importContext;
    return $importContext;
  }

  /** @return WinoApiClient */
  public function getApiClient() {
    return $this->getImportContext()->getApiClient();
  }

  /** @return DataScope */
  public function getScope() {
    return $this->getImportContext()->getScope();
  }

  /** @return bool */
  public function monitoringEnabled() {
    return $this->getImportContext()->monitoringEnabled();
  }

  /**
   * @param mixed $log
   * @param mixed[] $data
   * @return void
   */
  public function monitoringLog($log, ...$data) {
    $this->getImportContext()->monitoringLog($log, ...$data);
  }

  /** @return WinoApiShopIdentity|null */
  public function getShop() {
    return $this->getImportContext()->getShop();
  }
}

class WinoApiSynchronization {
  public const HOOK_PRIORITY = 50;

  /** @return RestRoute */
  public static function winoApiNotificationHandlingRoute(DataScope $scope) {
    return RestRoute::from($scope->scopedSlug('webhooks-notifications'));
  }

  /** @return void */
  public static function enable(WinoApiSynchronizationContext $context) {
    $shop = $context->getShop();
    if ($shop !== null) WinoApiWebhook::activateWebhook(
      $context->getApiClient(),
      $shop,
      self::winoApiNotificationHandlingRoute($context->getScope())
    );
  }

  /** @return void */
  public static function disable(WinoApiSynchronizationContext $context) {
    WinoApiWebhook::deactivateWebhook(
      $context->getApiClient(),
      self::winoApiNotificationHandlingRoute($context->getScope())
    );
  }

  /** @return void */
  public static function update(WinoApiSynchronizationContext $context) {
    if (!$context->getApiClient()->isConfigured()) return;

    $enabled = ConfigurationSynchronizationEnabled::value();
    if ($enabled) self::enable($context);
    else self::disable($context);
  }

  /** @return void */
  public static function initialize() {
    $context = WinoApiSynchronizationContext::fromConfiguration();
    $scope = $context->getScope();

    self::initializeOrderCompletionAction($context);

    WinoApiWebhook::registerNotificationHandler(
      self::winoApiNotificationHandlingRoute($scope),
      function (WinoApiWebhookNotification $notification) {
        return self::handleWinoNotification($notification);
      }
    );
  }

  /** @return void */
  public static function initializeOrderCompletionAction(WinoApiSynchronizationContext $context) {
    $scope = $context->getScope();
    if ($context->monitoringEnabled()) WooCommerceOrder::addStatusChangedAction(
      function ($orderId, $oldStatus, $newStatus) use ($context) {
        $context->monitoringLog("Order $orderId status changed from $oldStatus to $newStatus");
      },
      self::HOOK_PRIORITY
    );

    WooCommerceOrder::addOrderCompletedAction(
      $scope,
      function (WooCommerceOrder $order) use ($context) {
        self::handleCompletedOrder($context, $order);
      },
      self::HOOK_PRIORITY
    );
  }

  /** @return RestResponse */
  public static function handleWinoNotification(WinoApiWebhookNotification $notification) {
    // return a 410 Gone if sync is disabled
    // the webhook should be automatically deleted if successive 410 are detected...

    $log = InMemoryLog::empty();
    $response = RestResponse::from($log);

    $category = $notification->getCategoryData();
    if ($category !== null) return self::handleWinoCategoryNotification($response, $category);

    $stockActivity = $notification->getStockActivityData();
    if ($stockActivity !== null) return self::handleWinoStockActivityNotification($response, $stockActivity);

    $price = $notification->getPriceData();
    if ($price !== null) return self::handleWinoPriceNotification($response, $price);

    $product = $notification->getProductData();
    if ($product !== null) return self::handleWinoProductNotification($response, $product);

    $tax = $notification->getTaxData();
    if ($tax !== null) return self::handleWinoTaxNotification($response, $tax);

    $variant = $notification->getVariantData();
    if ($variant !== null) return self::handleWinoVariantNotification($response, $variant);

    $variantPrice = $notification->getVariantPriceData();
    if ($variantPrice !== null) return self::handleWinoVariantPriceNotification($response, $variantPrice);

    return $response;
  }

  /** @return RestResponse */
  public static function handleWinoCategoryNotification(
    RestResponse $response,
    WinoApiCategoryStateRecordData $category
  ) {
    $context = ImportContext::fromConfiguration();
    $api = $context->getApiClient();
    $upToDateCategory = WinoApiCategory::queryById($api, $category->id);
    importWinoCategory($context, $upToDateCategory);
    return $response;
  }

  /** @return RestResponse */
  private static function updateWinoVariant(
    RestResponse $response,
    ImportContext $context,
    string $variantId
  ) {
    $api = $context->getApiClient();
    $upToDateVariant = WinoApiVariant::queryById($api, $variantId);
    importWinoProduct($context, $upToDateVariant);
    return $response;
  }

  /** @return RestResponse */
  private static function updateWinoVariants(
    RestResponse $response,
    ImportContext $context,
    string ...$variantIds
  ) {
    foreach ($variantIds as $variantId) self::updateWinoVariant($response, $context, $variantId);
    return $response;
  }

  /** @return RestResponse */
  public static function handleWinoVariantNotification(
    RestResponse $response,
    WinoApiVariantStateRecordData $variant
  ) {
    $context = ImportContext::fromConfiguration();
    return self::updateWinoVariant($response, $context, $variant->id);
  }

  /** @return RestResponse */
  public static function handleWinoVariantPriceNotification(
    RestResponse $response,
    WinoApiVariantPriceStateRecordData $variantPrice
  ) {
    $context = ImportContext::fromConfiguration();
    return self::updateWinoVariant($response, $context, $variantPrice->variantId);
  }

  /** @return RestResponse */
  public static function handleWinoStockActivityNotification(
    RestResponse $response,
    WinoApiStockActivityStateRecordData $stockActivity
  ) {
    $context = ImportContext::fromConfiguration();
    return self::updateWinoVariant($response, $context, $stockActivity->variantId);
  }

  /** @return RestResponse */
  public static function handleWinoTaxNotification(
    RestResponse $response,
    WinoApiTaxStateRecordData $tax
  ) {
    $context = ImportContext::fromConfiguration();
    $variantIds = findWinoVariantIdsByApiTaxId($context, $tax->id);
    return self::updateWinoVariants($response, $context, ...$variantIds);
  }

  /** @return RestResponse */
  public static function handleWinoProductNotification(
    RestResponse $response,
    WinoApiProductStateRecordData $product
  ) {
    $context = ImportContext::fromConfiguration();
    $variantIds = findWinoVariantIdsByApiProductId($context, $product->id);
    return self::updateWinoVariants($response, $context, ...$variantIds);
  }

  /** @return RestResponse */
  public static function handleWinoPriceNotification(
    RestResponse $response,
    WinoApiPriceStateRecordData $price
  ) {
    $currentPriceListId = ConfigurationPriceList::id();
    if ($price->id !== $currentPriceListId) return $response;

    $currentPriceListName = ConfigurationPriceList::value();
    $newName = PriceList::normalizeName($price->name);
    if (PriceList::nameEquals($newName, $currentPriceListName)) return $response;

    ConfigurationPriceList::updateValue($newName);

    $context = ImportContext::fromConfiguration();
    $translator = LocaleTranslator::fromConfiguration();

    // Here we only want to create an import session.
    // Running it might be a long process and is so
    // delayed to when the admin will request an import session check.
    //
    // At some point, we might also use a WP cron to
    // check and run any pending import session at
    // regular interval. So accessing the admin would
    // not be necessary.
    ImportSession::startAttempt($context, $translator);

    return $response;
  }

  /** @return void */
  public static function handleCompletedOrder(
    WinoApiSynchronizationContext $context,
    WooCommerceOrder $order
  ) {
    $stockMovements = $order->getWinoSaleStockMovements();

    $cache = $context->webDeviceCache;
    $api = $context->getApiClient();
    $shop = $context->getShop();
    if ($shop !== null) foreach ($stockMovements as $move) $move->createStockActivity($cache, $api, $shop);
  }
}
