<?php

namespace Wino;

/** @template T */
class DataCache {
  /** @var array<string, T> */
  public array $record;

  private function __construct() {
    $this->record = [];
  }

  /**
   * @template E
   * @return DataCache<E>
   */
  public static function empty() {
    return new self();
  }

  /** @return bool */
  public function has(?string $key) {
    if ($key === null) return false;
    return key_exists($key, $this->record);
  }

  /** @return T|null */
  public function get(?string $key) {
    if ($key === null) return null;
    return $this->has($key) ? $this->record[$key] : null;
  }

  /**
   * @param T $value
   *
   * @return void
   */
  public function set(?string $key, $value) {
    if ($key === null) return;
    $cached = $this->get($key);
    $cachedClass = is_object($cached) ? get_class($cached) : null;
    $valueClass = is_object($value) ? get_class($value) : null;
    if ($valueClass === null || $valueClass !== $cachedClass) {
      $this->record[$key] = $value;
      return;
    }

    /** @psalm-suppress MixedArgumentTypeCoercion */
    $objectVars = get_object_vars($value);

    /** @psalm-suppress MixedAssignment */
    foreach ($objectVars as $objectKey => $objectValue) $cached->$objectKey = $objectValue;
  }

  /** @return void */
  public function unset(?string $key) {
    if ($key === null) return;
    if ($this->has($key)) unset($this->record[$key]);
  }

  /**
   * @param T $object
   * @param Closure(T): ?string $getCacheKey
   * @param Closure(T): ?string|int $getIdentity
   *
   * @return void
   */
  public function unsetObject(
    object $object,
    callable $getCacheKey,
    callable $getIdentity
  ) {
    $cacheKey = call_user_func($getCacheKey, $object);
    $hasKey = $cacheKey !== null && $this->has($cacheKey);
    $this->unset($cacheKey);
    if ($hasKey) return;

    $record = $this->record;
    $identity = call_user_func($getIdentity, $object);
    foreach ($record as $key => $cached) {
      $cachedIdentity = call_user_func($getIdentity, $cached);
      if ($cachedIdentity !== $identity) continue;
      $this->unset($key);
    }
  }
}
