<?php

namespace Wino;

class Render {
  /** @var null|(Closure(): string) $closure */
  private ?\Closure $closure;

  /** @param null|(Closure(): string) $render */
  private function __construct(?callable $render) {
    $this->closure = $render === null ? $render : \Closure::fromCallable($render);
  }

  /**
   * @param null|(Closure(): string) $render
   * @return self
   */
  public static function fromCallable(?callable $render) {
    return new self($render);
  }

  /** @return self */
  public static function fromString(?string $render) {
    return self::fromCallable($render === null ? null : function () use ($render) {
      return $render;
    });
  }

  /** @return self */
  public static function none() {
    return self::fromCallable(null);
  }

  /** @return self */
  public static function join(string $separator, ?Render ...$renders) {
    return self::fromCallable(function () use ($separator, $renders) {
      $content = [];
      foreach ($renders as $render) $content[] = Render::renderContent($render);
      return join($separator, $content);
    });
  }

  /** @return self */
  public static function concat(?Render ...$render) {
    return self::join("", ...$render);
  }

  /** @return string */
  public function content() {
    $render = $this->closure;
    return $render === null ? '' : call_user_func($render);
  }

  /** @return string */
  public static function renderContent(?Render $render) {
    return $render === null ? '' : $render->content();
  }

  /** @return void */
  public static function echo(?Render $render) {
    echo self::renderContent($render);
  }
}

/** @return string */
function renderHtmlPage(string $head, string $body) {
  return str(
    '<!DOCTYPE html>',
    '<html lang="en">',
    '<head>',
    '<meta charset="UTF-8" />',
    $head,
    '</head>',
    '<body>',
    $body,
    '</body>',
    '</html>'
  );
}


/** @return string */
function renderNoIndexHtmlPageHead() {
  return str(
    '<meta name="Googlebot" content="noindex" />',
    '<meta name="robots" content="noindex" />'
  );
}

/**
 * @param string|bool|int|float|null $value
 * @return string
 */
function renderHtmlTagAttribute(string $key, $value, string $prefix = "") {
  $escaped = esc_attr("$value");
  if ($value === null) return "";
  if (!is_bool($value)) return "$prefix$key=\"$escaped\"";
  elseif ($value) return "$prefix$key";
  return "";
}

/**
 * @param array<string, string|bool|int|float|null> $attributes
 *
 * @return string
 */
function renderHtmlTag(string $tag, array $attributes = [], string $content = "", bool $selfClosing = false) {
  $escaped = tag_escape($tag);
  $html = "<$escaped";
  foreach ($attributes as $key => $value) $html .= renderHtmlTagAttribute($key, $value, " ");

  $html .= $selfClosing ? "/>" : ">$content</$escaped>";
  return $html;
}

/**
 * @param array<string, string|bool|int|float|null> $attributes
 *
 * @return string
 */
function renderHtmlLink(?string $href, string $text, ?string $target = null, array $attributes = []) {
  $attributes['target'] = $target;
  $attributes['href'] = $href;
  return $href === null ? $text : renderHtmlTag('a', $attributes, $text);
}

/**
 * @param array<string, string|bool|int|float|null> $attributes
 * @return Render
 */
function htmlTagRender(string $tag, array $attributes = [], ?Render ...$content) {
  $render = Render::concat(...$content);
  return Render::fromCallable(function () use ($tag, $attributes, $render) {
    return renderHtmlTag($tag, $attributes, $render->content());
  });
}

/**
 * @param array<string, string|bool|int|float|null> $attributes
 * @return Render
 */
function spanRender(array $attributes = [], ?Render ...$content) {
  return htmlTagRender('span', $attributes, ...$content);
}

/**
 * @param array<string, string|bool|int|float|null> $attributes
 * @return Render
 */
function divRender(array $attributes = [], ?Render ...$content) {
  return htmlTagRender('div', $attributes, ...$content);
}

/**
 * @param array<string, string|bool|int|float|null> $attributes
 * @return Render
 */
function emRender(array $attributes = [], ?Render ...$content) {
  return htmlTagRender('em', $attributes, ...$content);
}

/**
 * @param array<string, string|bool|int|float|null> $attributes
 * @return Render
 */
function ulRender(array $attributes = [], ?Render ...$content) {
  return htmlTagRender('ul', $attributes, ...$content);
}

/**
 * @param array<string, string|bool|int|float|null> $attributes
 * @return Render
 */
function liRender(array $attributes = [], ?Render ...$content) {
  return htmlTagRender('li', $attributes, ...$content);
}

/**
 * @param array<string, string|bool|int|float|null> $attributes
 * @return Render
 */
function htmlLinkRender(?string $href, string $text, ?string $target = null, array $attributes = []) {
  return Render::fromCallable(function () use ($href, $text, $target, $attributes) {
    return renderHtmlLink($href, $text, $target, $attributes);
  });
}

/**
 * @param array<string, string|bool|int|float|null> $attributes
 * @return string
 */
function renderHtmlSelfClosingTag(string $tag, array $attributes = []) {
  return renderHtmlTag($tag, $attributes, "", true);
}

/** @return string */
function renderHtmlLineBreak() {
  return renderHtmlSelfClosingTag('br');
}

/**
 * @param array<string, string|bool|int|float|null> $attributes
 * @return Render
 */
function htmlTagRenderSelfClosing(string $tag, array $attributes = []) {
  return Render::fromCallable(function () use ($tag, $attributes) {
    return renderHtmlSelfClosingTag($tag, $attributes);
  });
}

/**
 * @param array<array<string, string>> $lines
 *
 * @return string
 */
function htmlTable(
  array $lines,
  string $style = ""
) {
  $headers = [];
  $table = [];
  foreach ($lines as $line) {
    $lineKeys = array_keys($line);
    foreach ($lineKeys as $key) if (!isset($headers[$key])) $headers[$key] = $key;
    $table[] = '<tr><td>' . implode('</td><td>', $line) . '</td></tr>';
  }
  return str(
    "<table style=\"" . esc_attr($style) . "\">",
    "<thead><tr><th>" . implode('</th><th>', $headers) . '</th></tr></thead>',
    implode('', $table) . '</table>'
  );
}

/**
 * @param string[] $tags
 *
 * @return \DOMNode[]
 */
function domNodesByTagNames(\DOMDocument $dom, array $tags) {
  $nodes = [];
  foreach ($tags as $tag) {
    $elements = $dom->getElementsByTagName($tag);
    foreach ($elements as $element) $nodes[] = $element;
  }
  return $nodes;
}

/** @return string|null */
function htmlEncoding(string $string, string $fromEncoding = 'UTF-8') {
  $mbBasedEncodingDeprecated = PHP_MAJOR_VERSION >= 8 && PHP_MINOR_VERSION > 0;
  $htmlEncoding = !$mbBasedEncodingDeprecated
  ? mb_convert_encoding($string, 'HTML', $fromEncoding)
  // The following alternative to mb_convert_encoding comes from here:
  // https://github.com/symfony/symfony/issues/44281
  : mb_encode_numericentity(
    htmlspecialchars_decode(
      htmlentities($string, ENT_NOQUOTES, $fromEncoding, false),
      ENT_NOQUOTES
    ), [0x80, 0x10FFFF, 0, ~0],
    $fromEncoding
  );
  return is_string($htmlEncoding) ? $htmlEncoding : null;
}

/**
 * @param string[] $tags
 *
 * @return string
 */
function removeHtmlTags(
  array $tags,
  string $html
) {
  $dom = new \DOMDocument();
  $uniqueId = uuid();
  $start = '<div id="' . $uniqueId . '">';
  $end = '</div>';
  $htmlWrap = htmlEncoding($start . $html . $end);
  if ($htmlWrap === null) throw exception("Failed to safely remove tags from the following input: $html");
  $dom->loadHTML($htmlWrap, LIBXML_NOERROR | LIBXML_NOWARNING);

  $nodes = domNodesByTagNames($dom, $tags);
  foreach ($nodes as $node) {
    $parent = $node->parentNode;
    if ($parent === null) throw error('Unexpected missing html parent.');
    $parent->removeChild($node);
  }

  $output = $dom->saveHTML($dom->getElementById($uniqueId));
  if (!strStartsWith($output, $start)) throw msgError("Unexpected html start:", jsonString($output));
  if (!strEndsWith($output, $end)) throw msgError("Unexpected html end: " . jsonString($output));
  return substr(substr($output, 0, strlen($output) - strlen($end)), strlen($start));
}
