<?php

namespace Wino;

/**
 * @param array $args
 * @psalm-suppress UnusedParam
 * @return void
 */
// phpcs:ignore Generic.CodeAnalysis.UnusedFunctionParameter.Found
// phpcs:ignore PEAR.WhiteSpace.ScopeClosingBrace.Line
function noop(...$args) {}

/**
 * @template T
 * @param T $value
 * @return T
 */
function identity($value) {
  return $value;
}

/** @return bool */
function truthy(bool $value = true) {
  return $value;
}

/** @return bool */
function falsy(bool $value = true) {
  return !$value;
}

/** @return string */
function str(string ...$slices) {
  return join("", $slices);
}

/** @return string */
function msg(string ...$slices) {
  return join(" ", $slices);
}

/**
 * @param mixed $value
 * @psalm-assert-if-true \Error $value
 *
 * @return bool
 */
function isNativeError($value) {
  return $value instanceof \Error;
}

/**
 * @param mixed $value
 * @psalm-assert-if-true \Exception $value
 *
 * @return bool
 */
function isNativeException($value) {
  return $value instanceof \Exception;
}

class WinoError extends \Error {
  /** @var int|string */
  protected $code;

  /**
   * @param int|string $code
   */
  public function __construct(string $message, $code = 0, \Throwable $previous = null) {
    parent::__construct($message, is_int($code) ? $code : 0, $previous);
    $this->code = $code;
  }
}

class WinoException extends \Exception {
  /** @var int|string */
  protected $code;

  /**
   * @param int|string $code
   */
  public function __construct(string $message, $code = 0, \Throwable $previous = null) {
    parent::__construct($message, is_int($code) ? $code : 0, $previous);
    $this->code = $code;
  }
}

/**
 * @param mixed $value
 * @psalm-assert-if-true WinoError $value
 *
 * @return bool
 */
function isWinoError($value) {
  return $value instanceof WinoError;
}

/**
 * @param mixed $value
 * @psalm-assert-if-true WinoException $value
 *
 * @return bool
 */
function isWinoException($value) {
  return $value instanceof WinoException;
}

/**
 * @param mixed $value
 * @psalm-assert-if-true \Error|WinoError $value
 *
 * @return bool
 */
function isError($value) {
  return isWinoError($value) || isNativeError($value);
}

/**
 * @param mixed $value
 * @psalm-assert-if-true \Exception|WinoException $value
 *
 * @return bool
 */
function isException($value) {
  return isWinoException($value) || isNativeException($value);
}

/**
 * @param mixed $value
 * @psalm-assert-if-true \Throwable $value
 *
 * @return bool
 */
function isThrowable($value) {
  return isError($value) || isException($value);
}

/**
 * @param int|string $code
 * @return \Error
 */
function error(string $message, $code = 0, ?\Throwable $previous = null) {
  return new WinoError($message, $code, $previous);
}

/**
 * @param int|string $code
 * @return \Exception
 */
function exception(string $message, $code = 0, ?\Throwable $previous = null) {
  return new WinoException($message, $code, $previous);
}

/** @return \Error */
function strError(string ...$slices) {
  return error(str(...$slices));
}

/** @return \Error */
function msgError(string ...$slices) {
  return error(msg(...$slices));
}

/** @return \Exception */
function strException(string ...$slices) {
  return exception(str(...$slices));
}

/** @return \Exception */
function msgException(string ...$slices) {
  return exception(msg(...$slices));
}

class ErrorCode {
  /** @var int|string */
  public $code;

  /** @param int|string $code */
  public function __construct($code) {
    $this->code = $code;
  }

  /** @return \Error */
  public function error(string $message, ?\Throwable $previous = null) {
    return error($message, $this->code, $previous);
  }

  /** @return \Exception */
  public function exception(string $message, ?\Throwable $previous = null) {
    return exception($message, $this->code, $previous);
  }

  /** @return \Error */
  public function strError(string ...$slices) {
    return $this->error(str(...$slices));
  }

  /** @return \Exception */
  public function strException(string ...$slices) {
    return $this->exception(str(...$slices));
  }

  /** @return \Error */
  public function msgError(string ...$slices) {
    return $this->error(msg(...$slices));
  }

  /** @return \Exception */
  public function msgException(string ...$slices) {
    return $this->exception(msg(...$slices));
  }
}

/**
 * @param int|string $code
 * @return ErrorCode
 */
function errorCode($code) {
  return new ErrorCode($code);
}

/**
 * @param mixed $value
 * @param int<1, 2147483647> $depth
 * @return string
 */
function jsonString($value, int $flags = 0, int $depth = 512) {
  $encoded = json_encode($value, $flags, $depth);
  if (is_string($encoded)) return $encoded;
  throw exception("Value can't be turned into a json string: $value");
}

/**
 * @param mixed $value
 * @param int<1, 2147483647> $depth
 * @return string
 */
function prettyJson($value, int $depth = 512) {
  return jsonString($value, JSON_PRETTY_PRINT, $depth);
}

/** @return string */
function definedStringOrDefault(?string $value, string $default) {
  return is_string($value) ? $value : $default;
}


/**
 * @param mixed $value
 * @psalm-assert array $value
 * @return void
 */
function assertsArray($value) {
  if (!is_array($value)) throw msgError("Unexpected non array value", jsonString($value));
}

/**
 * @param mixed $value
 * @psalm-assert array|null $value
 * @return void
 */
function assertsNullableArray($value) {
  if (!is_array($value) && $value !== null) throw msgError("Unexpected non nullable array value", jsonString($value));
}

/**
 * @param mixed $value
 * @return array
 */
function assertedArray($value) {
  assertsArray($value);
  return $value;
}

/**
 * @param mixed $value
 * @return array|null
 */
function assertedNullableArray($value) {
  assertsNullableArray($value);
  return $value;
}

/**
 * @param mixed $value
 * @return array[]
 */
function assertedArraysArray($value) {
  $values = [];
  $items = assertedArray($value);
  /** @var mixed */
  foreach ($items as $item) $values[] = assertedArray($item);
  return $values;
}

/**
 * @param mixed $value
 * @psalm-assert object $value
 * @return void
 */
function assertsObject($value) {
  if (!is_object($value)) throw msgError("Unexpected non object value", jsonString($value));
}

/**
 * @param mixed $value
 * @return object
 */
function assertedObject($value) {
  assertsObject($value);
  return $value;
}

/**
 * @param mixed $value
 * @return object[]
 */
function assertedObjectsArray($value) {
  $values = [];
  $items = assertedArray($value);
  /** @var mixed */
  foreach ($items as $item) $values[] = assertedObject($item);
  return $values;
}

/**
 * @param mixed $value
 * @psalm-assert null $value
 * @return void
 */
function assertsNull($value) {
  if ($value !== null) throw msgError("Unexpected non null value", jsonString($value));
}

/**
 * @param mixed $value
 * @return null
 */
function assertedNull($value) {
  assertsNull($value);
  return $value;
}

/**
 * @param mixed $value
 * @param mixed $from
 * @psalm-assert !null $value
 * @return void
 */
function assertsNotNull($value, $from = null) {
  if ($value === null) throw msgError("Unexpected null value", jsonString($value), "from", jsonString($from));
}

/**
 * @template T
 * @param T|null $value
 * @param mixed $from
 * @psalm-assert !null $value
 * @return T
 */
function assertedNotNull($value, $from = null) {
  assertsNotNull($value, $from);
  return $value;
}

/**
 * @param mixed $value
 * @psalm-assert bool $value
 * @return void
 */
function assertsBool($value) {
  if (!is_bool($value)) throw msgError("Unexpected non bool value", jsonString($value));
}

/**
 * @param mixed $value
 * @return bool
 */
function assertedBool($value) {
  assertsBool($value);
  return $value;
}

/**
 * @param mixed $value
 * @return bool[]
 */
function assertedBoolsArray($value) {
  $values = [];
  $items = assertedArray($value);
  /** @var mixed */
  foreach ($items as $item) $values[] = assertedBool($item);
  return $values;
}

/**
 * @param mixed $value
 * @psalm-assert string $value
 * @return void
 */
function assertsString($value) {
  if (!is_string($value)) throw msgError("Unexpected non string value", jsonString($value));
}

/**
 * @param mixed $value
 * @return string
 */
function assertedString($value) {
  assertsString($value);
  return $value;
}

/**
 * @param mixed $value
 * @psalm-assert string|null $value
 * @return void
 */
function assertsNullableString($value) {
  if (!is_string($value) && $value !== null) throw msgError("Unexpected non nullable string value", jsonString($value));
}

/**
 * @param mixed $value
 * @return string|null
 */
function assertedNullableString($value) {
  assertsNullableString($value);
  return $value;
}

/**
 * @param mixed $value
 * @return string[]
 */
function assertedStringsArray($value) {
  $values = [];
  $items = assertedArray($value);
  /** @var mixed */
  foreach ($items as $item) $values[] = assertedString($item);
  return $values;
}

/**
 * @param mixed $value
 * @psalm-assert int $value
 * @return void
 */
function assertsInt($value) {
  if (!is_int($value)) throw msgError("Unexpected non int value", jsonString($value));
}

/**
 * @param mixed $value
 * @return int
 */
function assertedInt($value) {
  assertsInt($value);
  return $value;
}

/**
 * @param mixed $value
 * @psalm-assert int|null $value
 * @return void
 */
function assertsNullableInt($value) {
  if (!is_int($value) && $value !== null) throw msgError("Unexpected non nullable int value", jsonString($value));
}

/**
 * @param mixed $value
 * @return int|null
 */
function assertedNullableInt($value) {
  assertsNullableInt($value);
  return $value;
}

/**
 * @param mixed $value
 * @psalm-assert string|int|null $value
 * @return void
 */
function assertsNullableStringOrInt($value) {
  if (!is_int($value) && !is_string($value) && $value !== null) throw msgError(
    "Unexpected non nullable string or int value", jsonString($value)
  );
}

/**
 * @param mixed $value
 * @return string|int|null
 */
function assertedNullableStringOrInt($value) {
  assertsNullableStringOrInt($value);
  return $value;
}

/**
 * @param mixed $value
 * @return int[]
 */
function assertedIntsArray($value) {
  $values = [];
  $items = assertedArray($value);
  /** @var mixed */
  foreach ($items as $item) $values[] = assertedInt($item);
  return $values;
}

/**
 * @param mixed $value
 * @psalm-assert float $value
 * @return void
 */
function assertsFloat($value) {
  if (!is_float($value)) throw msgError("Unexpected non int value", jsonString($value));
}

/**
 * @param mixed $value
 * @return float
 */
function assertedFloat($value) {
  assertsFloat($value);
  return $value;
}

/**
 * @param mixed $value
 * @return float[]
 */
function assertedFloatsArray($value) {
  $values = [];
  $items = assertedArray($value);
  /** @psalm-suppress MixedAssignment */
  foreach ($items as $item) $values[] = assertedFloat($item);
  return $values;
}

/**
 * @template T
 * @template D
 *
 * @param T|null $maybeNotDefined
 * @param D $defaultValue
 * @return T|D
 */
function getOrElse($maybeNotDefined, $defaultValue) {
  return $maybeNotDefined !== null ? $maybeNotDefined : $defaultValue;
}

/**
 * @param string|int $value
 * @return int
 */
function parseInt($value) {
  if (is_int($value)) return $value;
  $intVal = intval(trim($value));
  if ("$intVal" === $value) return $intVal;
  throw msgException("Parsing the following value as int is not possible. Not an int:", jsonString($value));
}
