<?php

namespace Wino;

class AllocatedSize {
  private int $start;

  private function __construct() {
    $this->start = memory_get_usage();
  }

  /** @return self */
  public static function current() {
    return new self();
  }

  /** @return int */
  public function diffValue() {
    return memory_get_usage() - $this->start;
  }

  /** @return array<string, int> */
  public function diff(?string $label = null) {
    $suffix = $label === null ? "" : " ($label)";
    return ["allocated$suffix" => $this->diffValue()];
  }

  /** @return void */
  public function logDiff(?string $label = null) {
    prettyLog($this->diff($label));
  }
}

/**
 * @param mixed $log
 * @return void
 */
function writeLog($log) {
  // phpcs:ignore Squiz.PHP.DiscouragedFunctions.Discouraged
  if (is_array($log) || is_object($log)) error_log(print_r($log, true));
  // phpcs:ignore Squiz.PHP.DiscouragedFunctions.Discouraged
  else error_log("$log");
}

/**
 * @param mixed $log
 * @param mixed[] $data
 * @return string
 */
function prettyLogString($log, ...$data) {
  $logString = is_string($log) ? $log : prettyJson($log);
  $dataString = join("", array_map(function ($value) {
    return " " . prettyJson($value);
  }, $data));
  return "$logString$dataString";
}

/**
 * @param mixed $log
 * @param mixed[] $data
 * @return void
 */
function prettyLog($log, ...$data) {
  writeLog(prettyLogString($log, ...$data));
}

interface Log {
  /**
   * @param mixed $value
   * @return $this
   */
  public function info($value);
  /**
   * @param mixed $value
   * @return $this
   */
  public function error($value);
}

/**
 * @template L
 * @param L[] $logs
 *
 * @return mixed[]
 */
function filterDuplicateLogs(array $logs) {
  $map = [];
  $filter = [];
  foreach ($logs as $log) {
    $key = null;
    if (is_string($log)) $key = $log;
    if ($log instanceof \Error || $log instanceof \Exception) $key = $log->getMessage();
    if ($key === null) {
      $filter[] = $log;
      continue;
    }
    if (key_exists($key, $map)) continue;
    $map[$key] = true;
    $filter[] = $log;
  }
  return $filter;
}

class StdLog implements Log {
  /**
   * @param mixed $value
   * @return $this
   */
  public function info($value) {
    writeLog($value);
    return $this;
  }

  /**
   * @param mixed $value
   * @return never
   */
  public function error($value) {
    throw $value;
  }
}

class InMemoryLog implements Log {
  /** @var mixed[] */
  protected array $infos;

  /** @var mixed[] */
  protected array $errors;

  /**
   * @param mixed $value
   * @return $this
   */
  public function info($value) {
    $this->infos[] = $value;
    return $this;
  }

  /**
   * @param mixed $value
   * @return $this
   */
  public function error($value) {
    $this->errors[] = $value;
    return $this;
  }

  /**
   * @param mixed[] $infos
   * @param mixed[] $errors
   */
  private function __construct(
    array &$infos,
    array &$errors
  ) {
    $this->infos = $infos;
    $this->errors = $errors;
  }

  /** @return self */
  public static function empty() {
    $infos = [];
    $errors = [];
    return new self($infos, $errors);
  }

  /** @return void */
  public function flush() {
    $this->infos = [];
    $this->errors = [];
  }

  /** @return mixed[] */
  public function getInfos() {
    return filterDuplicateLogs($this->infos);
  }

  /** @return mixed[] */
  public function getErrors() {
    return filterDuplicateLogs($this->errors);
  }
}
