<?php

namespace Wino;

/**
 * @param mixed[]|object $value
 *
 * @return array-key[]
 */
function sortedKeys($value) {
  $array = is_array($value) ? $value : get_object_vars($value);
  $keys = [];
  foreach ($array as $key => $_) $keys[] = $key;
  sort($keys);
  return $keys;
}

/**
 * @return array<string, mixed>
 */
function objectAsArray(object $object) {
  $vars = get_object_vars($object);
  return $vars;
}

/**
 * @template T
 * @param T $value
 * @return (T is object ? array<string, mixed> : T)
 */
function objectToArray($value) {
  return is_object($value) ? objectAsArray($value) : $value;
}

/**
 * @template T
 * @param T $value
 * @return (T is object|array ? array : T)
 */
function recursiveObjectToArray($value) {
  if (!is_object($value) && !is_array($value)) return $value;
  /** @var array */
  $array = objectToArray($value);
  /** @psalm-suppress MixedAssignment */
  foreach ($array as $key => $keyValue) $array[$key] = recursiveObjectToArray($keyValue);
  return $array;
}

/**
 * @param mixed $object
 * @param mixed[] $params
 * @return array{
 *  object: mixed,
 *  method: string,
 *  params: mixed[],
 *  result?: mixed
 * }
 */
function mixedMethodCallLogData(
  $object,
  string $method,
  $params
) {
  return ['object' => $object, 'method' => $method, 'params' => $params];
}

/**
 * @param mixed $object
 * @param mixed[] $params
 * @return array{
 *  object: mixed,
 *  property: string,
 *  result?: mixed
 * }
 */
function mixedPropertyAccessLogData(
  $object,
  string $property
) {
  return ['object' => $object, 'property' => $property];
}

/**
 * @param mixed $object
 * @param mixed[] $params
 * @return mixed
 */
function mixedMethodCall(
  bool $debug,
  $object,
  string $method,
  ...$params
) {
  $logData = mixedMethodCallLogData($object, $method, $params);
  if (!is_object($object)) {
    $message = "Unexpected non object method call:";
    if ($debug) throw msgError($message, jsonString($logData));
    prettyLog($message, $logData);
    return null;
  }
  if (!method_exists($object, $method)) {
    $message = "Unexpected undefined method call:";
    if ($debug) throw msgError($message, jsonString($logData));
    prettyLog($message, $logData);
    return null;
  }
  try {
    return $object->$method(...$params);
  } catch (\Throwable $thrown) {
    if ($debug) throw $thrown;
    prettyLog("Error thrown on method call:", $logData);
    writeLog($thrown->getTraceAsString());
    return null;
  }
}

/**
 * @param mixed $object
 * @return mixed
 */
function mixedPropertyAccess(
  bool $debug,
  $object,
  string $property
) {
  $logData = mixedPropertyAccessLogData($object, $property);
  if (!is_object($object)) {
    $message = "Unexpected non object property access:";
    if ($debug) throw msgError($message, jsonString($logData));
    prettyLog($message, $logData);
    return null;
  }
  if (!property_exists($object, $property)) {
    $message = "Unexpected undefined property access:";
    if ($debug) throw msgError($message, jsonString($logData));
    prettyLog($message, $logData);
    return null;
  }
  try {
    return $object->$property;
  } catch (\Throwable $thrown) {
    if ($debug) throw $thrown;
    prettyLog("Error thrown on property access:", $logData);
    writeLog($thrown->getTraceAsString());
    return null;
  }
}

/**
 * @template T
 * @param T $result
 * @return null
 */
function unexpectedMixedAccessResult(
  bool $debug,
  array $logData,
  string $expected,
  $result
) {
  $logData['result'] = $result;
  $message = "Unexpected non $expected result:";
  if ($debug) throw msgError($message, jsonString($logData));
  prettyLog($message, $logData);
  return null;
}

/**
 * @param mixed $object
 * @param mixed[] $params
 * @return string|null
 */
function mixedStringMethodCall(
  bool $debug,
  $object,
  string $method,
  ...$params
) {
  $logData = mixedMethodCallLogData($object, $method, $params);

  /** @var mixed */
  $result = mixedMethodCall($debug, $object, $method, ...$params);

  return is_string($result) ? $result : unexpectedMixedAccessResult(
    $debug,
    $logData,
    'string method call',
    $result
  );
}

/**
 * @param mixed $object
 * @param mixed[] $params
 * @return array|null
 */
function mixedArrayMethodCall(
  bool $debug,
  $object,
  string $method,
  ...$params
) {
  $logData = mixedMethodCallLogData($object, $method, $params);

  /** @var mixed */
  $result = mixedMethodCall($debug, $object, $method, ...$params);

  return is_array($result) ? $result : unexpectedMixedAccessResult(
    $debug,
    $logData,
    'array method call',
    $result
  );
}


/**
 * @param mixed $object
 * @param mixed[] $params
 * @return array|object|null
 */
function mixedNullableArrayOrObjectMethodCall(
  bool $debug,
  $object,
  string $method,
  ...$params
) {
  $logData = mixedMethodCallLogData($object, $method, $params);

  /** @var mixed */
  $result = mixedMethodCall($debug, $object, $method, ...$params);

  if ($result === null || is_array($result) || is_object($result)) return $result;

  return unexpectedMixedAccessResult($debug, $logData, 'array or object method call', $result);
}

/**
 * @param mixed $object
 * @return string|null
 */
function mixedStringPropertyAccess(
  bool $debug,
  $object,
  string $property
) {
  $logData = mixedPropertyAccessLogData($object, $property);

  /** @var mixed */
  $result = mixedPropertyAccess($debug, $object, $property);

  return is_string($result) ? $result : unexpectedMixedAccessResult(
    $debug,
    $logData,
    'string property access',
    $result
  );
}
