<?php

namespace Wino;

/** @return int */
function stringIndexOf(string $haystack, string $needle) {
  $position = strpos($haystack, $needle);
  if ($position === false) return -1;
  if ($position >= 0) return $position;
  throw msgError("Unexpected string position", jsonString($position));
}

/** @return string|null */
function stringHead(string $separator, string $string) {
  $separatorPosition = stringIndexOf($string, $separator);
  if ($separatorPosition < 0) return null;
  return substr($string, 0, $separatorPosition);
}

/** @return string|null */
function stringTail(string $separator, string $string) {
  $separatorPosition = stringIndexOf($string, $separator);
  if ($separatorPosition < 0) return null;
  return substr($string, $separatorPosition);
}

/** @return \Exception */
function dataException(string $message, array $data, string $key) {
  return msgException("$message => $key in", jsonString($data));
}

/** @return bool */
function missingData(array $data, string $key) {
  if (!key_exists($key, $data)) return true;
  return $data[$key] === null;
}

/** @return mixed */
function requiredData(array $data, string $key) {
  if (missingData($data, $key)) throw dataException("Missing property", $data, $key);
  return $data[$key];
}

/** @return mixed */
function optionalData(array $data, string $key) {
  if (missingData($data, $key)) return null;
  return requiredData($data, $key);
}

/** @return array */
function arrayData(array $data, string $key) {
  $value = requiredData($data, $key);
  if (!is_array($value)) throw dataException("Not an array", $data, $key);
  return $value;
}

/** @return array|null */
function optionalArrayData(array $data, string $key) {
  if (missingData($data, $key)) return null;
  return arrayData($data, $key);
}

/** @return array[] */
function arrayOfArraysData(array $data, string $key) {
  $arrayOfArrays = [];
  $value = arrayData($data, $key);

  /** @psalm-suppress MixedAssignment */
  foreach ($value as $index => $item)
  if (is_array($item)) $arrayOfArrays[] = $item;
  else throw dataException("Not an array", $data, "$key.$index");

  return $arrayOfArrays;
}

/** @return int[] */
function intArrayData(array $data, string $key) {
  $intArray = [];
  $value = arrayData($data, $key);

  /** @psalm-suppress MixedAssignment */
  foreach ($value as $index => $item)
  if (is_int($item)) $intArray[] = $item;
  else throw dataException("Not an int", $data, "$key.$index");

  return $intArray;
}

/** @return string */
function stringData(array $data, string $key) {
  $value = requiredData($data, $key);
  if (!is_string($value)) throw dataException("Not a string", $data, $key);
  return $value;
}

/** @return string|null */
function optionalStringData(array $data, string $key) {
  if (missingData($data, $key)) return null;
  return stringData($data, $key);
}

/** @return non-empty-string */
function nonEmptyStringData(array $data, string $key) {
  $value = stringData($data, $key);
  if ($value === "") throw dataException("Empty string", $data, $key);
  return $value;
}

/** @return non-empty-string|null */
function optionalNonEmptyStringData(array $data, string $key) {
  if (missingData($data, $key)) return null;
  return nonEmptyStringData($data, $key);
}

/** @return non-empty-string */
function nonBlankStringData(array $data, string $key) {
  $value = nonEmptyStringData($data, $key);
  if (trim($value) === "") throw dataException("Blank string", $data, $key);
  return $value;
}

/** @return non-empty-string|null */
function optionalNonBlankStringData(array $data, string $key) {
  if (missingData($data, $key)) return null;
  return nonBlankStringData($data, $key);
}

/** @return string|null */
function blankStringToNull(?string $string) {
  if ($string === null) return null;
  $trim = trim($string);
  return $trim === "" ? null : $string;
}

/** @return float */
function floatData(array $data, string $key) {
  $value = requiredData($data, $key);
  if (!is_float($value) && !is_int($value)) throw dataException("Not a float", $data, $key);
  return $value;
}

/** @return float|null */
function optionalFloatData(array $data, string $key) {
  if (missingData($data, $key)) return null;
  return floatData($data, $key);
}

/** @return int */
function intData(array $data, string $key) {
  $value = requiredData($data, $key);
  if (!is_int($value)) throw dataException("Not an int", $data, $key);
  return $value;
}

/** @return int|null */
function optionalIntData(array $data, string $key) {
  if (missingData($data, $key)) return null;
  return intData($data, $key);
}

/** @return int */
function stringifiedIntData(array $data, string $key) {
  $value = trim(stringData($data, $key));
  $intValue = intval($value);
  if ("$intValue" !== $value) throw dataException("Not a stringified int", $data, $key);
  return $intValue;
}

/** @return int|null */
function optionalStringifiedIntData(array $data, string $key) {
  if (missingData($data, $key)) return null;
  return stringifiedIntData($data, $key);
}

/** @return \DateTime */
function dateData(array $data, string $key, TimeUnit $timestampUnit) {
  /** @var mixed $value */
  $value = requiredData($data, $key);
  if ($value instanceof \DateTime) return $value;
  if (is_string($value)) return new \DateTime($value);
  if (is_int($value)) return dateTimeFromTimestamp($value, $timestampUnit);
  throw dataException("Not a valid date input", $data, $key);
}

/** @return \DateTime|null */
function optionalDateData(array $data, string $key, TimeUnit $timestampUnit) {
  if (missingData($data, $key)) return null;
  return dateData($data, $key, $timestampUnit);
}

/** @return bool */
function boolData(array $data, string $key) {
  $value = requiredData($data, $key);
  if (!is_bool($value)) throw dataException("Not a bool", $data, $key);
  return $value;
}

/** @return bool|null */
function optionalBoolData(array $data, string $key) {
  if (missingData($data, $key)) return null;
  return boolData($data, $key);
}
