<?php

namespace Wino;

require_once __DIR__ . '/arrays.php';
require_once __DIR__ . '/objects.php';
require_once __DIR__ . '/logging.php';
require_once __DIR__ . '/parsing.php';
require_once __DIR__ . '/uuid.php';

class TestCase {
  public const TODO = 'todo';
  public const SKIP = 'skip';
  public const TEST = 'test';
  public const ONLY = 'only';

  public string $describeLabel;
  public string $label;
  public string $state;
  public \Closure $callback;

  private function __construct(
    string $describeLabel,
    string $label,
    string $state,
    callable $callback
  ) {
    $this->describeLabel = $describeLabel;
    $this->label = $label;
    $this->state = $state;
    $this->callback = \Closure::fromCallable($callback);

  }

  /** @return bool */
  private function is(string $state) {
    return $this->state === $state;
  }

  /** @return bool */
  public function isTodo() {
    return $this->is(self::TODO);
  }

  /** @return bool */
  public function isSkip() {
    return $this->is(self::SKIP);
  }

  /** @return bool */
  public function isOnly() {
    return $this->is(self::ONLY);
  }

  /** @return bool */
  public function isTest() {
    return $this->is(self::TEST);
  }

  /** @return self */
  public static function todo(string $describeLabel, string $label) {
    // phpcs:ignore PEAR.WhiteSpace.ScopeClosingBrace.Line
    return new self($describeLabel, $label, self::TODO, function () {});
  }

  /** @return self */
  public static function skip(string $describeLabel, string $label, callable $callback) {
    return new self($describeLabel, $label, self::SKIP, $callback);
  }

  /** @return self */
  public static function test(string $describeLabel, string $label, callable $callback) {
    return new self($describeLabel, $label, self::TEST, $callback);
  }

  /** @return self */
  public static function only(string $describeLabel, string $label, callable $callback) {
    return new self($describeLabel, $label, self::ONLY, $callback);
  }
}

class TestEnvContext {
  private static bool $runningInTestSuite = false;
  private static string $describeLabel = '';
  private static bool $hideSuccess = false;
  private static bool $hideSkipped = false;
  private static bool $hideTodo = false;
  private static bool $hideErrorDetails = false;

  /** @return bool */
  public static function areDetailsHidden() {
    return self::$hideErrorDetails;
  }
  private static bool $stopOnFirstError = false;
  private static int $executedCount = 0;
  private static int $failedCount = 0;
  private static int $successCount = 0;
  private static int $skippedCount = 0;
  private static int $todoCount = 0;

  /** @var TestCase[] */
  private static array $testCases = [];

  /** @return void */
  public static function hideTestSuccess(bool $enable = true) {
    self::$hideSuccess = $enable;
  }

  /** @return void */
  public static function hideTestSkipped(bool $enable = true) {
    self::$hideSkipped = $enable;
  }

  /** @return void */
  public static function hideTestTodo(bool $enable = true) {
    self::$hideTodo = $enable;
  }

  /** @return void */
  public static function hideTestErrorDetails(bool $enable = true) {
    self::$hideErrorDetails = $enable;
  }

  /** @return void */
  public static function stopTestOnFirstError(bool $enable = true) {
    self::$stopOnFirstError = $enable;
  }

  /** @return string */
  private static function testCaseLabel(TestCase $case) {
    return self::label($case->describeLabel, $case->label);
  }

  /** @return string */
  private static function label(string $describeLabel, string $label) {
    return $describeLabel === '' ? $label : $describeLabel . ' > ' . $label;
  }

  private static int $fakeStringIndex = 0;
  /** @return int */
  public static function fakeStringIndex() {
    return self::$fakeStringIndex++;
  }

  private static float $fakeFloatValue = 0.0;
  /** @return float */
  public static function fakeFloatValueIncrementedBy(
    float $increment,
    float $max,
    float $min = 0.0
  ) {
    $value = self::$fakeFloatValue + $increment;
    if ($value > $max) $value = $min;
    if ($value < $min) $value = $max;
    self::$fakeFloatValue = $value;
    return $value;
  }

  private static int $fakeFloatValueIncrementIndex = 0;
  private const FAKE_FLOAT_VALUE_INCREMENTS = [6.53, 0.33, 10.12, 3.17, 1.49, 5.76, 3.60, 9.67, 9.18, 4.02, 0.023];

  /** @return float */
  public static function fakeFloatValueIncrement() {
    $increments = self::FAKE_FLOAT_VALUE_INCREMENTS;
    $index = self::$fakeFloatValueIncrementIndex + 1;
    $max = count($increments) - 1;
    $index = $index < 0 ? $max : $index;
    $index = $index > $max ? 0 : $index;
    self::$fakeFloatValueIncrementIndex = $index;
    return $increments[$index];
  }

  private const TEST_SUITE_CONTEXT_ERROR_MESSAGE = 'Tests functions must be executed inside a test suite.';
  /** @return void */
  private static function checkTestSuiteContext() {
    if (self::$runningInTestSuite) return;
    throw error(self::TEST_SUITE_CONTEXT_ERROR_MESSAGE);
  }

  /** @return void */
  public static function describe(string $label, callable $callback) {
    self::checkTestSuiteContext();
    $previous = self::$describeLabel;
    self::$describeLabel = self::label($previous, $label);
    call_user_func($callback);
    self::$describeLabel = $previous;
  }

  /** @return void */
  public static function todo(string $label) {
    self::checkTestSuiteContext();
    self::$testCases[] = TestCase::todo(self::$describeLabel, $label);
  }

  /** @return void */
  public static function executeTodo(TestCase $case) {
    self::$todoCount++;
    if (self::$hideTodo) return;
    writeLog('TODO 🟡 ' . self::testCaseLabel($case));
  }

  /** @return void */
  public static function skip(string $label, callable $callback) {
    self::checkTestSuiteContext();
    self::$testCases[] = TestCase::skip(self::$describeLabel, $label, $callback);
  }

  /** @return void */
  public static function executeSkip(TestCase $case) {
    self::$skippedCount++;
    if (self::$hideSkipped) return;
    writeLog('SKIPPED ⚪ ' . self::testCaseLabel($case));
  }

  /** @return void */
  public static function only(string $label, callable $callback) {
    self::checkTestSuiteContext();
    self::$testCases[] = TestCase::only(self::$describeLabel, $label, $callback);
  }

  /** @return void */
  public static function test(string $label, callable $callback) {
    self::checkTestSuiteContext();
    self::$testCases[] = TestCase::test(self::$describeLabel, $label, $callback);
  }

  /** @return never */
  public static function wip(string $label, ?callable $callback = null) {
    noop($callback);
    $fullLabel = self::label(self::$describeLabel, $label);
    throw error("\n\n 🟠🟠🟠 WORK IN PROGRESS: $fullLabel 🟠🟠🟠 \n\n");
  }

  private static ?string $currentTestLabelValue = null;

  /** @return string */
  private static function currentTestLabel() {
    $label = self::$currentTestLabelValue;
    if ($label !== null) return $label;
    throw error(self::TEST_SUITE_CONTEXT_ERROR_MESSAGE);
  }

  /**
   * @param mixed $log
   * @param array $data
   * @return void
   */
  public static function currentTestLog($log, ...$data) {
    prettyLog("\n\n⚪ " . self::currentTestLabel() . " ⚪\n", $log, ...$data);
  }

  /**
   * @param string[] $paths
   * @return void
   */
  private static function executeTest(
    TestCase $case,
    array $paths,
    callable $beforeEach,
    callable $afterEach,
    callable $teardown
  ) {
    $oldReporting = error_reporting(E_ALL);
    set_error_handler(function ($severity, $message, $file, $line) use ($paths) {
      foreach ($paths as $path)
      if (strStartsWith($file, $path)) throw new \ErrorException($message, 0, $severity, $file, $line);
      return true;
    });
    call_user_func($beforeEach);

    self::$executedCount++;

    $success = true;
    $error = null;
    $fullLabel = self::testCaseLabel($case);
    self::$currentTestLabelValue = $fullLabel;

    try {
      call_user_func($case->callback);
    } catch (\Throwable $err) {
      $success = false;
      $error = $err;
    }

    call_user_func($afterEach);
    restore_error_handler();
    error_reporting($oldReporting);

    if ($success) {
      self::$successCount++;
      if (self::$hideSuccess) return;
      writeLog('✅ ' . $fullLabel . ' ✅');
      return;
    }
    self::$failedCount++;
    $summary = '❌ ' . $fullLabel . ' ❌';
    if (self::$hideErrorDetails) {
      writeLog($summary);
      return;
    }

    if ($error === null) throw error("Unexpected null error");

    $stop = self::$stopOnFirstError;
    $details = $stop ? testErrorTrace($error) . PHP_EOL . PHP_EOL . "$error" : "$error";
    $fullLog = PHP_EOL . PHP_EOL . $summary . PHP_EOL . PHP_EOL . $details . PHP_EOL;

    if (!$stop || strlen($fullLog) > 1000) writeLog($fullLog);
    if (!$stop) return;

    self::unsetTestRunningOption($teardown);
    throw error("\n\nTest interruption after the first error$fullLog");
  }

  /** @return string */
  private static function countReportLine(string $label, int $count) {
    return " - $label: $count   ";
  }

  /** @return void */
  private static function logTestReport() {
    $labelLine = '== Tests report ==';
    $countLines = [
      self::countReportLine('executed', self::$executedCount),
      self::countReportLine('success', self::$successCount),
      self::countReportLine('failed', self::$failedCount),
      self::countReportLine('todo', self::$todoCount),
      self::countReportLine('skipped', self::$skippedCount),
    ];
    $maxLength = strlen($labelLine);
    foreach ($countLines as $line) {
      $length = strlen($line);
      if ($length > $maxLength) {
        $maxLength = $length;
      }
    }
    $sepLine = str_repeat('=', $maxLength);
    $arrayLines = array_merge([
      $sepLine,
      $labelLine,
      $sepLine,
    ], $countLines, [$sepLine]);
    $report = "\n";
    foreach ($arrayLines as $line) {
      $offset = str_repeat(' ', $maxLength - strlen($line));
      $report .= "\n|$line$offset|";
    }
    writeLog($report . "\n");
  }

  /** @return void */
  private static function collectTestSuite() {
    self::$runningInTestSuite = true;
    self::$testCases = [];
  }

  private const TEST_RUNNING_OPTION = 'wino_test_running';

  /** @return bool */
  private static function testRunning() {
    return get_option(self::TEST_RUNNING_OPTION) === 'yes';
  }

  /** @return void */
  private static function setTestRunningOption() {
    update_option(self::TEST_RUNNING_OPTION, 'yes');
  }

  /** @return void */
  private static function unsetTestRunningOption(
    callable $teardown
  ) {
    try {
      call_user_func($teardown);
    } catch (\Throwable $error) {
      delete_option(self::TEST_RUNNING_OPTION);
      throw $error;
    }
    delete_option(self::TEST_RUNNING_OPTION);
  }

  /**
   * @param string[] $paths
   * @return void
   */
  private static function executeTestSuite(
    array $paths,
    callable $beforeEach,
    callable $afterEach,
    callable $teardown
  ) {
    self::$runningInTestSuite = false;

    $onlyTestCases = [];
    $testTestCases = [];
    foreach (self::$testCases as $testCase) {
      if ($testCase->isTodo()) self::executeTodo($testCase);
      if ($testCase->isSkip()) self::executeSkip($testCase);
      if ($testCase->isOnly()) $onlyTestCases[] = $testCase;
      if ($testCase->isTest()) $testTestCases[] = $testCase;
    }

    $foundOnly = count($onlyTestCases) > 0;
    foreach ($testTestCases as $testCase)
    if ($foundOnly) self::executeSkip($testCase);
    else self::executeTest($testCase, $paths, $beforeEach, $afterEach, $teardown);

    foreach ($onlyTestCases as $testCase) self::executeTest(
      $testCase, $paths, $beforeEach, $afterEach, $teardown
    );
    self::$currentTestLabelValue = null;

    self::logTestReport();
    self::$testCases = [];
  }

  /** @return void */
  public static function disableTestRunningYetDetection() {
    delete_option(self::TEST_RUNNING_OPTION);
  }

  /**
   * @param string|string[]|null $targetPath
   * @return void
   */
  public static function testSuite(
    callable $setup,
    string $path,
    $targetPath,
    callable $beforeEach,
    callable $afterEach,
    callable $teardown
  ) {
    if (self::testRunning()) {
      writeLog('🟠🟠🟠 Test suite cancelled, since a test suite is running yet 🟠🟠🟠');
      return;
    }
    self::setTestRunningOption();
    try {
      call_user_func($setup);
    } catch (\Throwable $error) {
      self::unsetTestRunningOption($teardown);
      throw $error;
    }

    $pattern = path_join($path, '*.spec.php');
    $testFiles = glob($pattern);
    if ($testFiles === false || count($testFiles) === 0) throw msgError(
      'No test files found at path', $path
    );

    self::collectTestSuite();

    writeLog("\n\n🔵🔵🔵 Running tests 🔵🔵🔵\n");
    try {
      /** @psalm-suppress UnresolvableInclude */
      foreach ($testFiles as $testFile) include_once $testFile;
    } catch (\Throwable $err) {
      self::unsetTestRunningOption($teardown);
      throw $err;
    }

    $pathArray = [$path];
    if (is_string($targetPath)) $pathArray[] = $targetPath;
    if (is_array($targetPath)) $pathArray = array_merge($pathArray, $targetPath);
    self::executeTestSuite($pathArray, $beforeEach, $afterEach, $teardown);
    self::unsetTestRunningOption($teardown);
  }

  /** @return bool */
  public static function configBool(string $testConfigPath, string $key, ?bool $defaultValue = null) {
    $config = self::config($testConfigPath, $key, $defaultValue === null ? null : "$defaultValue");
    return strtolower(trim($config)) === 'true';
  }

  /** @return string */
  public static function config(string $testConfigPath, string $key, ?string $defaultValue = null) {
    try {
      if (!file_exists($testConfigPath)) throw msgError(
        "Missing config file at path:", $testConfigPath
      );

      $configContent = file_get_contents($testConfigPath);
      if (!is_string($configContent)) throw msgError(
        "Failure reading config file at path", $testConfigPath
      );

      $configLines = explode("\n", $configContent);

      $config = [];
      foreach ($configLines as $line) {
        $keyAndValues = explode('=', $line);
        if (count($keyAndValues) < 2) continue;
        $configKey = array_shift($keyAndValues);
        $value = implode('=', $keyAndValues);
        $config[$configKey] = $value;
      }
      return nonBlankStringData($config, $key);
    } catch (\Throwable $thrown) {
      if ($defaultValue !== null) return $defaultValue;
      throw $thrown;
    }
  }
}

class TestSuite {
  public string $path;

  /** @var string|string[]|null $targetPath */
  public $targetPath;

  /** @var Closure(): void */
  public \Closure $setup;

  /** @var Closure(): void */
  public \Closure $beforeEach;

  /** @var Closure(): void */
  public \Closure $afterEach;

  /** @var Closure(): void */
  public \Closure $teardown;

  /**
   * @param string|string[]|null $targetPath
   * @param Closure(): void $setup
   * @param Closure(): void $beforeEach
   * @param Closure(): void $afterEach
   * @param Closure(): void $teardown
   */
  private function __construct(
    string $path,
    $targetPath,
    callable $setup,
    callable $beforeEach,
    callable $afterEach,
    callable $teardown
  ) {
    $this->path = $path;
    $this->targetPath = $targetPath;
    $this->setup = \Closure::fromCallable($setup);
    $this->beforeEach = \Closure::fromCallable($beforeEach);
    $this->afterEach = \Closure::fromCallable($afterEach);
    $this->teardown = \Closure::fromCallable($teardown);
  }

  /**
   * @param Closure(): void $setup
   * @param string|string[]|null $targetPath
   * @param Closure(): void $beforeEach
   * @param Closure(): void $afterEach
   * @param Closure(): void $teardown
   *
   * @return self
   */
  public static function from(
    callable $setup,
    string $path,
    $targetPath,
    callable $beforeEach,
    callable $afterEach,
    callable $teardown
  ) {
    return new self(
      $path,
      $targetPath,
      $setup,
      $beforeEach,
      $afterEach,
      $teardown
    );
  }

  /** @return void */
  public function run() {
    if (!is_favicon()) TestEnvContext::testSuite(
      $this->setup,
      $this->path,
      $this->targetPath,
      $this->beforeEach,
      $this->afterEach,
      $this->teardown
    );
  }

  /** @return void */
  public function skip() {
    if (!is_favicon()) writeLog('🟠🟠🟠 Test suite disabled 🟠🟠🟠');
  }
}

/** @return void */
function describe(string $label, callable $callback) {
  TestEnvContext::describe($label, $callback);
}

/** @return void */
function todo(string $label) {
  TestEnvContext::todo($label);
}

/** @return void */
function skip(string $label, callable $callback) {
  TestEnvContext::skip($label, $callback);
}

/** @return void */
function skipTodo(string $label) {
  // phpcs:ignore PEAR.WhiteSpace.ScopeClosingBrace.Line
  skip($label, function () {});
}

/** @return void */
function only(string $label, callable $callback) {
  TestEnvContext::only($label, $callback);
}

/** @return void */
function test(string $label, callable $callback) {
  TestEnvContext::test($label, $callback);
}

/** @return never */
function wip(string $label, ?callable $callback = null) {
  TestEnvContext::wip($label, $callback);
}

/** @return void */
function hideTestSuccess() {
  TestEnvContext::hideTestSuccess();
}

/** @return void */
function hideTestSkipped() {
  TestEnvContext::hideTestSkipped();
}

/** @return void */
function hideTestTodo() {
  TestEnvContext::hideTestTodo();
}

/** @return void */
function hideTestErrorDetails() {
  TestEnvContext::hideTestErrorDetails();
}

/** @return void */
function stopTestOnFirstError() {
  TestEnvContext::stopTestOnFirstError();
}

/**
 * @param mixed $log
 * @param array $data
 * @return void
 */
function testLog($log, ...$data) {
  TestEnvContext::currentTestLog($log, ...$data);
}

/** @return float */
function fakeFloat(float $max = 100.0, float $min = 0.5) {
  return TestEnvContext::fakeFloatValueIncrementedBy(
    TestEnvContext::fakeFloatValueIncrement(), $max, $min
  );
}

/** @return int */
function fakeInt(int $max = 100, int $min = 1) {
  return intval(fakeFloat(floatval($max), floatval($min)));
}

/** @return string */
function fakeString(string $value = 'test') {
  $index = TestEnvContext::fakeStringIndex();
  $id = uuidName();
  return "Fake string $index: $value $id";
}

/** @return string */
function fakeName(string $kind = 'test') {
  $index = TestEnvContext::fakeStringIndex();
  $id = uuidName();
  return "Fake $kind name $index $id";
}

/** @return string */
function fakeSlug(string $kind = 'test') {
  $index = TestEnvContext::fakeStringIndex();
  $id = uuidName();
  return sanitize_title("fake-$kind-slug-$index-$id");
}

/** @return string */
function testErrorTrace(?\Throwable $throwable = null) {
  $error = getOrElse($throwable, new \Error());
  $trace = explode('#', $error->getTraceAsString());
  foreach ($trace as $line) if (strContains($line, '.spec.php(')) return '#' . $line;
  return '';
}

class AssertionError extends \Error {
  /** @return self */
  public static function expectedVsActual(string $label, string $expected, string $actual) {
    if (TestEnvContext::areDetailsHidden()) return new self($label);
    return new self(str(
      $label,
      PHP_EOL,
      testErrorTrace(),
      PHP_EOL,
      '---------',
      PHP_EOL,
      '(expected)',
      PHP_EOL,
      $expected,
      PHP_EOL,
      PHP_EOL,
      '(actual)',
      PHP_EOL,
      $actual,
      PHP_EOL,
      '________',
      PHP_EOL,
      PHP_EOL
    ));
  }

  /** @return self */
  public static function notExpected(string $label, string $notExpected) {
    if (TestEnvContext::areDetailsHidden()) return new self($label);
    return new self(str(
      $label,
      PHP_EOL,
      testErrorTrace(),
      PHP_EOL,
      '---------',
      PHP_EOL,
      '(not expected)',
      PHP_EOL,
      $notExpected,
      PHP_EOL,
      '________',
      PHP_EOL,
      PHP_EOL
    ));
  }
}

class EquatableObject {
  private object $source;
  private array $equatable;

  public function __construct(
    object $source
  ) {
    $this->source = $source;
    $this->equatable = [];
  }

  /** @return bool */
  public function matchSource(object $source) {
    return $source === $this->source;
  }

  /** @return object */
  public function getSource() {
    return $this->source;
  }

  /** @return array */
  public function getEquatable() {
    return $this->equatable;
  }

  /** @return EquatableObject|null */
  public static function check(EquatableObject $equatable, object $source) {
    return $equatable->matchSource($source) ? $equatable : null;
  }

  /**
   * @param array-key $key
   * @param mixed $value
   * @return void
   */
  public function set($key, $value) {
    $this->equatable[$key] = $value;
  }
}

class EquatableObjectsRecord {
  /** @var object[] */
  private array $sources;

  /** @var EquatableObject[] */
  private array $equatables;

  public function __construct() {
    $this->sources = [];
    $this->equatables = [];
  }

  /** @return void */
  public function add(EquatableObject $equatable) {
    $this->equatables[] = $equatable;
    $this->sources[] = $equatable->getSource();
  }

  /** @return EquatableObject|null */
  public function get(object $source) {
    foreach ($this->equatables as $equatable) {
      $checked = EquatableObject::check($equatable, $source);
      if ($checked === null) continue;
      return $checked;
    }
    return null;
  }
}

/** @return array */
function equatableObject(object $object, EquatableObjectsRecord $record) {
  $fromRecord = $record->get($object);
  if ($fromRecord !== null) return $fromRecord->getEquatable();
  $equatableObject = new EquatableObject($object);
  $record->add($equatableObject);
  $keys = sortedKeys($object);
  foreach ($keys as $key) $equatableObject->set($key, equatableRoot($object->$key, $record));
  return $equatableObject->getEquatable();
}

/** @return array */
function equatableArray(array $array, EquatableObjectsRecord $record) {
  $output = [];
  $keys = sortedKeys($array);
  /** @psalm-suppress MixedAssignment */
  foreach ($keys as $key) $output[$key] = equatableRoot($array[$key], $record);
  return $output;
}

/**
 * @param mixed $root
 * @return mixed
 */
function equatableRoot($root, EquatableObjectsRecord $record) {
  if ($root instanceof \DateTime) return $root->format(DATE_ISO8601);
  if (is_object($root)) return equatableObject($root, $record);
  if (is_array($root)) return equatableArray($root, $record);
  return $root;
}

/**
 * @param mixed $input
 * @return mixed
 */
function equatable($input) {
  $record = new EquatableObjectsRecord();
  return equatableRoot($input, $record);
}

/**
 * @param int|string $key
 * @return string
 */
function collectEquatableDiffKeyPathFormat(string $path, $key) {
  if (is_int($key)) return $path . "[$key]";
  return $path === '' ? $key : $path . ".$key";
}

/**
 * @param string[] $diffs
 * @param mixed $actual
 * @param mixed $expected
 * @return void
 */
function collectEquatableDiff(&$diffs, $actual, $expected, string $path = '') {
  if (!is_array($actual) || !is_array($expected)) {
    if ($actual === $expected) return;
    $diffs[] = "- Unexpected $path";
    return;
  }

  /** @psalm-suppress MixedAssignment */
  foreach ($expected as $key => $value) {
    $keyPath = collectEquatableDiffKeyPathFormat($path, $key);
    if (key_exists($key, $actual)) collectEquatableDiff($diffs, $actual[$key], $value, $keyPath);
    else $diffs[] = "- Missing $keyPath";
  }

  /** @psalm-suppress MixedAssignment */
  foreach ($actual as $key => $value) {
    noop($value);
    if (key_exists($key, $expected)) continue;
    $keyType = is_int($key) ? 'index' : 'property';
    $keyPath = collectEquatableDiffKeyPathFormat($path, $key);
    $diffs[] = "- Unexpected $keyType $keyPath";
  }
}

/**
 * @param mixed $actual
 * @param mixed $expected
 * @return void
 */
function expectEqual($actual, $expected, string $label = 'expect equal') {
  $actualType = gettype($actual);
  $expectedType = gettype($expected);

  $actualClass = is_object($actual) ? get_class($actual) : '';
  $expectedClass = is_object($expected) ? get_class($expected) : '';

  /** @var mixed */
  $actualEquatable = equatable($actual);
  /** @var mixed */
  $expectedEquatable = equatable($expected);
  $sameValue = $actualEquatable === $expectedEquatable;
  $sameClass = $actualClass === $expectedClass;
  $objectFieldsExpectation = is_object($actual) && is_array($expected);

  if ($sameValue && ($sameClass || $objectFieldsExpectation)) return;

  $expectedType = $objectFieldsExpectation ? $actualType : $expectedType;
  $expectedClass = $objectFieldsExpectation ? $actualClass : $expectedClass;
  $actualString = trim($actualType . ' ' . $actualClass) . ' ' . prettyJson($actual);
  $expectedString = trim($expectedType . ' ' . $expectedClass) . ' ' . prettyJson($expected);
  $diff = [];
  collectEquatableDiff($diff, $actualEquatable, $expectedEquatable);
  $diffString = implode("\n", $diff);

  throw AssertionError::expectedVsActual(
    $label,
    $expectedString,
    "\n" . $diffString . "\n\n" . $actualString
  );
}

/**
 * @param mixed $actual
 * @param mixed $expected
 * @return void
 */
function expectEqualAsArray($actual, $expected, string $label = 'expect equal as array') {
  $asArray = 'asArray';
  /** @var mixed */
  $actualAsArray = is_object($actual) && method_exists($actual, $asArray) ? $actual->$asArray() : $actual;
  /** @var mixed */
  $expectedAsArray = is_object($expected) && method_exists($expected, $asArray) ? $expected->$asArray() : $expected;
  expectEqual($actualAsArray, $expectedAsArray, $label);
}

/**
 * @param mixed $actual
 * @psalm-assert null $actual
 * @return void
 */
function expectNull($actual) {
  expectEqual($actual, null, 'expect null');
}

/**
 * @psalm-assert true $actual
 * @return void
 */
function expectTrue(bool $actual) {
  expectEqual($actual, true, 'expect true');
}

/**
 * @psalm-assert false $actual
 * @return void
 */
function expectFalse(bool $actual) {
  expectEqual($actual, false, 'expect false');
}

/**
 * @template TA
 * @template TE
 * @param TA[] $actual
 * @param TE[] $expected
 * @param Closure(TA|TE, TA|TE): int $compareFunction
 *
 * @return void
 */
function expectSortedEqual(array $actual, array $expected, callable $compareFunction) {
  expectEqual(
    sortArrayCopy($actual, $compareFunction),
    sortArrayCopy($expected, $compareFunction)
  );
}

/**
 * @template TA
 * @template TE
 * @param TA[] $actual
 * @param TE[] $expected
 * @param Closure(TA|TE): int $getInt
 *
 * @return void
 */
function expectSortedByIntEqual(array $actual, array $expected, callable $getInt) {
  expectEqual(
    sortArrayCopyByInt($actual, $getInt),
    sortArrayCopyByInt($expected, $getInt)
  );
}

/**
 * @param int[] $actual
 * @param int[] $expected
 *
 * @return void
 */
function expectSortedIntsEqual(array $actual, array $expected) {
  expectSortedByIntEqual($actual, $expected, function (int $int) {
    return $int;
  });
}

/**
 * @template T
 * @param T[] $actual
 * @param T[] $expected
 * @param Closure(T): string $getString
 *
 * @return void
 */
function expectSortedByStringEqual(array $actual, array $expected, callable $getString) {
  expectEqual(
    sortArrayCopyByString($actual, $getString),
    sortArrayCopyByString($expected, $getString)
  );
}

/**
 * @param string[] $actual
 * @param string[] $expected
 *
 * @return void
 */
function expectSortedStringsEqual(array $actual, array $expected) {
  expectSortedByStringEqual($actual, $expected, function ($str) {
    return $str;
  });
}

/**
 * @param mixed $value
 * @return void
 */
function expectType(string $expectedType, $value) {
  $valueType = gettype($value);
  $valueClass = is_object($value) ? get_class($value) : '';

  if ($valueType === $expectedType) return;

  $valueString = trim($valueType . ' ' . $valueClass) . ' ' . prettyJson($value);
  throw AssertionError::expectedVsActual(
    'expect type',
    $expectedType,
    $valueString
  );
}

/** @return void */
function expectClass(string $expectedClass, object $object) {
  $expectedClass = str_replace('/', '\\', $expectedClass);
  $objectClass = get_class($object);

  if ($objectClass === $expectedClass) return;

  $objectString = trim(gettype($object) . ' ' . $objectClass) . ' ' . prettyJson($object);
  throw AssertionError::expectedVsActual(
    'expect class',
    $expectedClass,
    $objectString
  );
}

/**
 * @param mixed $actual
 * @param mixed $notExpected
 * @return void
 */
function expectNotEqual($actual, $notExpected, string $label = 'expect not equal') {
  $sameValue = equatable($actual) === equatable($notExpected);
  if (!$sameValue) return;

  $notExpectedType = gettype($notExpected);
  $notExpectedClass = is_object($notExpected) ? get_class($notExpected) : '';
  $notExpectedString = trim($notExpectedType . ' ' . $notExpectedClass) . ' ' . prettyJson($notExpected);

  throw AssertionError::notExpected(
    $label,
    $notExpectedString,
  );
}

/**
 * @param mixed $actual
 * @psalm-assert !null $actual
 * @return void
 */
function expectNotNull($actual) {
  expectNotEqual($actual, null, 'expect not null');
}

/**
 * @param mixed $a
 * @param mixed $b
 * @psalm-assert !null $a
 * @psalm-assert !null $b
 * @return void
 */
function expectNotNulls($a, $b) {
  expectNotNull($a);
  expectNotNull($b);
}

/**
 * @param mixed $a
 * @param mixed $b
 * @param mixed $c
 * @psalm-assert !null $a
 * @psalm-assert !null $b
 * @psalm-assert !null $c
 * @return void
 */
function expectNotNull3($a, $b, $c) {
  expectNotNull($a);
  expectNotNull($b);
  expectNotNull($c);
}


/**
 * @param mixed $actual
 * @param mixed $expected
 * @psalm-assert !null $actual
 * @return void
 */
function expectNotNullEqual($actual, $expected) {
  expectEqual($actual, $expected);
  expectNotNull($actual);
}

/** @return void */
function expectNotEmpty(array $actual) {
  if (count($actual) > 0) return;

  $actualType = gettype($actual);
  $actualString = trim($actualType) . ' ' . prettyJson($actual);

  throw AssertionError::notExpected(
    'expect not empty',
    $actualString
  );
}

/**
 * @param mixed $actual
 * @param mixed $expected
 * @return void
 */
function expectGreater($actual, $expected) {
  $greaterThan = equatable($actual) > equatable($expected);
  if ($greaterThan) return;

  $actualType = gettype($actual);
  $expectedType = gettype($expected);

  $actualClass = is_object($actual) ? get_class($actual) : '';
  $expectedClass = is_object($expected) ? get_class($expected) : '';
  $objectFieldsExpectation = is_object($actual) && is_array($expected);
  $expectedType = $objectFieldsExpectation ? $actualType : $expectedType;
  $expectedClass = $objectFieldsExpectation ? $actualClass : $expectedClass;
  $actualString = trim($actualType . ' ' . $actualClass) . ' ' . prettyJson($actual);
  $expectedString = trim($expectedType . ' ' . $expectedClass) . ' ' . prettyJson($expected);

  throw AssertionError::expectedVsActual(
    'expect greater',
    'greater than ' . $expectedString,
    $actualString
  );
}

/**
 * @param mixed $actual
 * @param mixed $expected
 * @return void
 */
function expectLower($actual, $expected) {
  $lowerThan = equatable($actual) < equatable($expected);
  if ($lowerThan) return;

  $actualType = gettype($actual);
  $expectedType = gettype($expected);

  $actualClass = is_object($actual) ? get_class($actual) : '';
  $expectedClass = is_object($expected) ? get_class($expected) : '';
  $objectFieldsExpectation = is_object($actual) && is_array($expected);
  $expectedType = $objectFieldsExpectation ? $actualType : $expectedType;
  $expectedClass = $objectFieldsExpectation ? $actualClass : $expectedClass;
  $actualString = trim($actualType . ' ' . $actualClass) . ' ' . prettyJson($actual);
  $expectedString = trim($expectedType . ' ' . $expectedClass) . ' ' . prettyJson($expected);

  throw AssertionError::expectedVsActual(
    'expect lower',
    'lower than ' . $expectedString,
    $actualString
  );
}

/**
 * @param mixed $actual
 * @param mixed $expected
 * @return void
 */
function expectGreaterOrEqual($actual, $expected) {
  $greaterThan = equatable($actual) >= equatable($expected);
  if ($greaterThan) return;

  $actualType = gettype($actual);
  $expectedType = gettype($expected);

  $actualClass = is_object($actual) ? get_class($actual) : '';
  $expectedClass = is_object($expected) ? get_class($expected) : '';
  $objectFieldsExpectation = is_object($actual) && is_array($expected);
  $expectedType = $objectFieldsExpectation ? $actualType : $expectedType;
  $expectedClass = $objectFieldsExpectation ? $actualClass : $expectedClass;
  $actualString = trim($actualType . ' ' . $actualClass) . ' ' . prettyJson($actual);
  $expectedString = trim($expectedType . ' ' . $expectedClass) . ' ' . prettyJson($expected);

  throw AssertionError::expectedVsActual(
    'expect greater or equal',
    'greater or equal to ' . $expectedString,
    $actualString
  );
}

/**
 * @param mixed $actual
 * @param mixed $expected
 * @return void
 */
function expectLowerOrEqual($actual, $expected) {
  $lowerThan = equatable($actual) <= equatable($expected);
  if ($lowerThan) return;

  $actualType = gettype($actual);
  $expectedType = gettype($expected);

  $actualClass = is_object($actual) ? get_class($actual) : '';
  $expectedClass = is_object($expected) ? get_class($expected) : '';
  $objectFieldsExpectation = is_object($actual) && is_array($expected);
  $expectedType = $objectFieldsExpectation ? $actualType : $expectedType;
  $expectedClass = $objectFieldsExpectation ? $actualClass : $expectedClass;
  $actualString = trim($actualType . ' ' . $actualClass) . ' ' . prettyJson($actual);
  $expectedString = trim($expectedType . ' ' . $expectedClass) . ' ' . prettyJson($expected);

  throw AssertionError::expectedVsActual(
    'expect lower or equal',
    'lower or equal to ' . $expectedString,
    $actualString
  );
}

/**
 * @param mixed $actual
 * @param mixed $beforeExpected
 * @param mixed $afterExpected
 * @return void
 */
function expectBetween($actual, $beforeExpected, $afterExpected) {
  expectGreaterOrEqual($actual, $beforeExpected);
  expectLowerOrEqual($actual, $afterExpected);
}

/**
 * @param mixed $actual
 * @param mixed $beforeExpected
 * @param mixed $afterExpected
 * @return void
 */
function expectStrictBetween($actual, $beforeExpected, $afterExpected) {
  expectGreater($actual, $beforeExpected);
  expectLower($actual, $afterExpected);
}


/**
 * @param mixed $actual
 * @return void
 */
function expectStringStartingWith(string $expectedStart, $actual) {
  $actualAsString = jsonString($actual);
  if (!is_string($actual) || !strStartsWith($actual, $expectedStart)) throw AssertionError::expectedVsActual(
    'expected string starting with',
    $expectedStart,
    is_string($actual) ? $actual : "Not a string ($actualAsString)"
  );
}


/** @return void */
function expectToThrow(callable $function, string ...$message) {
  $thrown = null;
  $returned = null;
  $expectedMessage = trim(implode(' ', $message));
  $expectMessage = $expectedMessage !== "";

  try {
    /** @var mixed */
    $returned = call_user_func($function);
  } catch (\Throwable $actualThrown) {
    $thrown = $actualThrown;
  }

  if ($thrown === null) {
    $returnedText = $returned === null ? '' : ' => returned ' . jsonString($returned);
    throw AssertionError::expectedVsActual(
      'expect to throw',
      'Throwable with message ( ' . $expectedMessage . ' )',
      "Nothing thrown$returnedText",
    );
  }

  if (!$expectMessage) return;
  $actualMessage = $thrown->getMessage();
  if ($actualMessage !== $expectedMessage) throw AssertionError::expectedVsActual(
    'expected to throw',
    'Throwable with message ( ' . $expectedMessage . ' )',
    'Throwable with message ( ' . $actualMessage . ' )'
  );
}

/** @return void */
function expectToThrowErrorStartingWith(string $expectedStart, callable $function) {
  $thrown = null;
  $returned = null;

  try {
    /** @var mixed */
    $returned = call_user_func($function);
  } catch (\Throwable $actualThrown) {
    $thrown = $actualThrown;
  }

  if ($thrown === null) {
    $returnedText = $returned === null ? '' : ' => returned ' . jsonString($returned);
    throw AssertionError::expectedVsActual(
      'expect to throw',
      'Throwable with message start ( ' . $expectedStart . ' )',
      "Nothing thrown$returnedText",
    );
  }

  $actualMessage = $thrown->getMessage();
  if (!strStartsWith($actualMessage, $expectedStart)) throw AssertionError::expectedVsActual(
    'expected to throw',
    'Throwable with message start ( ' . $expectedStart . ' )',
    'Throwable with message ( ' . $actualMessage . ' )'
  );
}
