<?php

namespace Wino;

/**
 * @template T
 * @param T $object
 * @return (T is \DateTime ? string : T)
 */
function formatDateTime($object, string $format = \DateTime::ATOM) {
  if ($object instanceof \DateTime) return $object->format($format);
  return $object;
}

class TimeUnit {
  public const SECONDS = 'sec';
  public const SECONDS_TO_MILLISECONDS_FACTOR = 1000;
  public const MILLISECONDS = 'ms';
  public string $value;

  private function __construct(string $value) {
    $this->value = $value;
  }

  /** @return self */
  public static function seconds() {
    return new self(self::SECONDS);
  }

  /** @return self */
  public static function milliseconds() {
    return new self(self::MILLISECONDS);
  }

  /** @return self */
  public static function phpTimestampLike() {
    return self::seconds();
  }

  /** @return self */
  public static function javascriptTimestampLike() {
    return self::milliseconds();
  }

  /** @return string */
  private static function trim(string $value) {
    return trim(strtolower($value));
  }

  /** @return bool */
  private function is(string $value) {
    return self::trim($value) === self::trim($this->value);
  }

  /** @return bool */
  public function isSeconds() {
    return $this->is(self::SECONDS);
  }

  /** @return bool */
  public function isMilliseconds() {
    return $this->is(self::MILLISECONDS);
  }
}

/** @return \DateTime */
function dateTimeFromTimestampSeconds(int $timestamp) {
  return date_timestamp_set(new \DateTime(), $timestamp);
}

/** @return int */
function timestampSecondsFrom(\DateTime $dateTime) {
  return $dateTime->getTimestamp();
}

/** @return \DateTime */
function dateTimeFromTimestampMilliseconds(int $timestamp) {
  return dateTimeFromTimestampSeconds(intval($timestamp / TimeUnit::SECONDS_TO_MILLISECONDS_FACTOR));
}

/** @return int */
function timestampMillisecondsFrom(\DateTime $dateTime) {
  return $dateTime->getTimestamp() * TimeUnit::SECONDS_TO_MILLISECONDS_FACTOR;
}

/** @return \Exception */
function unexpectedTimeUnitException(TimeUnit $unit) {
  return msgException("Unexpected time unit", $unit->value);
}

/** @return \DateTime */
function dateTimeFromTimestamp(int $timestamp, TimeUnit $unit) {
  if ($unit->isSeconds()) return dateTimeFromTimestampSeconds($timestamp);
  if ($unit->isMilliseconds()) return dateTimeFromTimestampMilliseconds($timestamp);
  throw unexpectedTimeUnitException($unit);
}

/** @return int */
function timestampFrom(\DateTime $dateTime, TimeUnit $unit) {
  if ($unit->isSeconds()) return timestampSecondsFrom($dateTime);
  if ($unit->isMilliseconds()) return timestampMillisecondsFrom($dateTime);
  throw unexpectedTimeUnitException($unit);
}

/** @return \DateTime */
function dateTimeNow() {
  return new \DateTime();
}
