<?php

namespace Wino;

/** @extends GraphQLPaginatedQueryEdgeNodeIterator<WinoApiDevice> */
class WinoApiDeviceQueryAllIterator extends GraphQLPaginatedQueryEdgeNodeIterator {
  /** @return WinoApiDevice */
  public function currentObject() {
    return WinoApiDevice::from($this->currentNode());
  }
}

class WinoApiDevice {
  public WinoApiShopIdentity $shop;
  public string $id;
  public ?\DateTime $archivedAt;
  public string $name;
  public string $slug;

  public function __construct(
    WinoApiShopIdentity $shop,
    string $id,
    ?\DateTime $archivedAt,
    string $name,
    string $slug
  ) {
    $this->id = $id;
    $this->archivedAt = $archivedAt;
    $this->name = $name;
    $this->slug = $slug;
    $this->shop = $shop;
  }

  public const FIELDS_QUERY = "id, archivedAt, name, slug, shop { " . WinoApiShopIdentity::FIELDS_QUERY . " }";

  /** @return self */
  public static function from(array $data) {
    return new self(
      WinoApiShopIdentity::from(arrayData($data, "shop")),
      nonBlankStringData($data, "id"),
      optionalDateData($data, "archivedAt", TimeUnit::javascriptTimestampLike()),
      nonBlankStringData($data, "name"),
      nonBlankStringData($data, "slug")
    );
  }

  public const WEB_DEVICE_SLUG = "W";
  public const NODE_QUERY_NAME = "Device";
  public const NODE_QUERY_RESOLVER = "device";

  /** @return self */
  public static function queryById(WinoApiClient $winoApi, string $deviceId) {
    return self::from($winoApi->postGraphQLNodeByIdQuery(
      self::NODE_QUERY_NAME,
      "",
      self::NODE_QUERY_RESOLVER,
      "",
      $deviceId,
      self::FIELDS_QUERY,
      []
    ));
  }

  public const COLLECTION_QUERY_NAME = "Devices";
  public const COLLECTION_QUERY_RESOLVER = "devices";

  /**
   * It should be possible to use the shop identity but the API doesn't allow it
   *
   * @param array<string, mixed> $queryVars
   *
   * @return WinoApiDeviceQueryAllIterator
   */
  public static function queryAllIterator(
    WinoApiClient $winoApi,
    string $queryName = self::COLLECTION_QUERY_NAME,
    string $queryArgs = "",
    string $filterByArgs = "",
    string $resolverArgs = "",
    array $queryVars = []
  ) {
    return new WinoApiDeviceQueryAllIterator(
      $winoApi,
      GraphQLPaginatedQueryRoot::none(),
      $queryName,
      WinoApiShopIdentity::queryArgs(null, $queryArgs),
      self::COLLECTION_QUERY_RESOLVER,
      WinoApiShopIdentity::resolverArgs(null, $filterByArgs, $resolverArgs),
      self::FIELDS_QUERY,
      WinoApiShopIdentity::queryVars(null, $queryVars)
    );
  }

  /**
   * Using `queryAllIterator` might be prefered in order to reduce
   * memory footprint. This is especially true if you expect lot of
   * items in the array which could lead to reach the php memory limit.
   *
   * @param array<string, mixed> $queryVars
   *
   * @return WinoApiDevice[]
   */
  public static function queryAll(
    WinoApiClient $winoApi,
    ?WinoApiShopIdentity $shop,
    string $queryName = self::COLLECTION_QUERY_NAME,
    string $queryArgs = "",
    string $filterByArgs = "",
    string $resolverArgs = "",
    array $queryVars = []
  ) {
    $devices = [];
    $iterator = self::queryAllIterator(
      $winoApi,
      $queryName,
      $queryArgs,
      $filterByArgs,
      $resolverArgs,
      $queryVars
    );

    while ($iterator->next()) {
      $device = $iterator->currentObject();
      if ($shop !== null && $device->shop->id !== $shop->id) continue;
      $devices[] = $device;
    }
    return $devices;
  }

  /** @return self */
  public static function queryWebDevice(
    WinoApiClient $winoApi,
    WinoApiShopIdentity $shop
  ) {
    $devices = self::queryAll(
      $winoApi,
      $shop,
      "DevicesBySlug",
      "\$slug: String!",
      "slug: { _equals: \$slug }",
      "",
      ['slug' => self::WEB_DEVICE_SLUG]
    );
    if (count($devices) > 1) throw msgException(
      "Duplicate web devices found:", jsonString($devices)
    );
    $webDevice = listElement($devices, 0);
    if ($webDevice === null) throw exception("Web device not found.");
    return $webDevice;
  }

  /** @return bool */
  public function isArchived() {
    return $this->archivedAt !== null;
  }
}

class WebDeviceCache {
  public DataCache $cache;

  public function __construct(DataCache $cache) {
    $this->cache = $cache;
  }

  /** @return self */
  public static function empty() {
    return new self(DataCache::empty());
  }

  /** @return WinoApiDevice */
  public function cacheWebDevice(
    WinoApiShopIdentity $shop,
    WinoApiDevice $webDevice
  ) {
    $this->cache->set($shop->id, $webDevice);
    return $webDevice;
  }

  /** @return WinoApiDevice|null */
  public function getCachedWebDevice(WinoApiShopIdentity $shop) {
    /** @var mixed */
    $cached = $this->cache->get($shop->id);
    return $cached instanceof WinoApiDevice ? $cached : null;
  }

  /** @return WinoApiDevice */
  public function queryWebDevice(
    WinoApiClient $api,
    WinoApiShopIdentity $shop
  ) {
    $cached = $this->getCachedWebDevice($shop);
    if ($cached !== null) return $cached;
    $cached = WinoApiDevice::queryWebDevice($api, $shop);
    return $this->cacheWebDevice($shop, $cached);
  }
}
