<?php

namespace Wino;

/**
 * All WinoApi entities other than Shop should includes a shop property,
 * typed with this class, allowing to check that all data is actually coming from
 * the configured shop.
 *
 * The class also provides static methods to create queries (args and vars) which
 * only retrieve data from a specific shop.
 */
class WinoApiShopIdentity {
  public string $id;

  public function __construct(string $id) {
    $this->id = $id;
  }

  /** @return bool */
  public function isValid() {
    return trim($this->id) !== "";
  }

  public const FIELDS_QUERY = "id";

  /** @return self */
  public static function from(array $data) {
    return new self(nonBlankStringData($data, "id"));
  }

  /** @return string */
  public static function queryArgs(?WinoApiShopIdentity $shop, string ...$args) {
    if ($shop === null) return WinoApiClient::joinArgs('', ...$args);
    return WinoApiClient::joinArgs("\$shopId: String!", ...$args);
  }

  /** @return string */
  public static function filterByArgs(?WinoApiShopIdentity $shop, string ...$args) {
    return WinoApiClient::filterByArgs(
      $shop === null ? "" : "shopIds: { _in: [\$shopId] }",
      ...$args
    );
  }

  /** @return string */
  public static function resolverArgs(
    ?WinoApiShopIdentity $shop,
    string $filterByArgs,
    string ...$resolverArgs
  ) {
    return WinoApiClient::joinArgs(
      self::filterByArgs($shop, $filterByArgs),
      ...$resolverArgs
    );
  }

  /**
   * @param array<string, mixed> $vars
   * @return array<string, mixed>
   */
  public static function queryVars(
    ?WinoApiShopIdentity $shop,
    array $vars
  ) {
    if ($shop === null) return $vars;
    return WinoApiClient::joinVars(['shopId' => $shop->id], $vars);
  }
}

/** @extends GraphQLPaginatedQueryEdgeNodeIterator<WinoApiShop> */
class WinoApiShopQueryAllIterator extends GraphQLPaginatedQueryEdgeNodeIterator {
  /** @return WinoApiShop */
  public function currentObject() {
    return WinoApiShop::from($this->currentNode());
  }
}

/**
 * If you only need the id, `WinoApiShopIdentity` should be prefered.
 *
 * This `WinoApiShop` class might contain more and more data
 * over time, so we don't want to use it in every WinoApi entities.
 */
class WinoApiShop {
  public string $id;
  public ?\DateTime $archivedAt;
  public string $name;

  public function __construct(
    string $id,
    ?\DateTime $archivedAt,
    string $name
  ) {
    $this->id = $id;
    $this->archivedAt = $archivedAt;
    $this->name = $name;
  }

  public const FIELDS_QUERY = "id, name, archivedAt";

  /** @return self */
  public static function from(array $data) {
    return new self(
      nonBlankStringData($data, "id"),
      optionalDateData($data, "archivedAt", TimeUnit::javascriptTimestampLike()),
      nonBlankStringData($data, "name")
    );
  }

  public const NODE_QUERY_NAME = "Shop";
  public const NODE_QUERY_RESOLVER = "shop";

  /** @return self */
  public static function queryById(WinoApiClient $winoApi, string $shopId) {
    return self::from($winoApi->postGraphQLNodeByIdQuery(
      self::NODE_QUERY_NAME,
      "",
      self::NODE_QUERY_RESOLVER,
      "",
      $shopId,
      self::FIELDS_QUERY,
      []
    ));
  }

  public const COLLECTION_QUERY_NAME = "Shops";
  public const COLLECTION_QUERY_RESOLVER = "shops";

  /**
   * @param array<string, mixed> $queryVars
   * @return WinoApiShopQueryAllIterator
   */
  public static function queryAllIterator(
    WinoApiClient $winoApi,
    string $queryName = self::COLLECTION_QUERY_NAME,
    string $queryArgs = "",
    string $resolverArgs = "",
    array $queryVars = []
  ) {
    return new WinoApiShopQueryAllIterator(
      $winoApi,
      GraphQLPaginatedQueryRoot::none(),
      $queryName,
      $queryArgs,
      self::COLLECTION_QUERY_RESOLVER,
      $resolverArgs,
      self::FIELDS_QUERY,
      $queryVars
    );
  }

  /**
   * Using `queryAllIterator` might be prefered in order to reduce
   * memory footprint. This is especially true if you expect lot of
   * items in the array which could lead to reach the php memory limit.
   *
   * @return WinoApiShop[]
   */
  public static function queryAll(WinoApiClient $winoApi) {
    return self::queryAllIterator($winoApi)->queryAll();
  }

  /** @return WinoApiShopIdentity */
  public function identity() {
    return new WinoApiShopIdentity($this->id);
  }

  /** @return bool */
  public function isArchived() {
    return $this->archivedAt !== null;
  }
}
