<?php

namespace Wino;

class WinoApiVariantPrice {
  public float $valueExcludingTax;
  public ?WinoApiPrice $price;

  public function __construct(
    float $valueExcludingTax,
    ?WinoApiPrice $price
  ) {
    $this->valueExcludingTax = $valueExcludingTax;
    $this->price = $price;
  }

  public const FIELDS_QUERY = 'valueExcludingTax, price { ' . WinoApiPrice::FIELDS_QUERY . ' }';

  /** @return self */
  public static function from(array $data) {
    $price = optionalArrayData($data, 'price');
    return new self(
      floatData($data, 'valueExcludingTax'),
      $price === null ? null : WinoApiPrice::from($price)
    );
  }
}

class WinoApiVariantPriceEdge {
  public WinoApiVariantPrice $node;

  public function __construct(WinoApiVariantPrice $node) {
    $this->node = $node;
  }

  public const FIELDS_QUERY = 'node { ' . WinoApiVariantPrice::FIELDS_QUERY . ' }';

  /** @return self */
  public static function from(array $data) {
    return new self(
      WinoApiVariantPrice::from(arrayData($data, 'node'))
    );
  }
}

class WinoApiVariantPriceConnection {
  /** @var WinoApiVariantPriceEdge[] $edges */
  public array $edges;

  /**
   * @param WinoApiVariantPriceEdge[] $edges
   */
  public function __construct($edges) {
    $this->edges = $edges;
  }

  public const FIELDS_QUERY = 'edges { ' . WinoApiVariantPriceEdge::FIELDS_QUERY . ' }';

  /** @return self */
  public static function from(array $data) {
    $edges = [];
    $rawEdges = arrayOfArraysData($data, 'edges');
    foreach ($rawEdges as $edge) $edges[] = WinoApiVariantPriceEdge::from($edge);
    return new self($edges);
  }
}

class WinoApiVariantStock {
  public ?int $rawQuantity;

  public function __construct(
    ?int $rawQuantity
  ) {
    $this->rawQuantity = $rawQuantity;
  }

  public const FIELDS_QUERY = 'rawQuantity';

  /** @return self */
  public static function from(array $data) {
    return new self(
      optionalIntData($data, 'rawQuantity')
    );
  }

  /** @return int */
  public function getQuantity() {
    $rawQuantity = $this->rawQuantity;
    return $rawQuantity === null ? 0 : $rawQuantity;
  }
}

class WinoApiVariantProduct {
  public string $id;
  public const ID_KEY = 'id';
  public ?WinoApiTax $tax;
  public const TAX_KEY = 'tax';
  public ?WinoApiCategory $category;
  public const CATEGORY_KEY = 'category';

  public function __construct(
    string $id,
    ?WinoApiTax $tax,
    ?WinoApiCategory $category
  ) {
    $this->id = $id;
    $this->tax = $tax;
    $this->category = $category;
  }

  /** @return string */
  public static function fieldsQuery() {
    return WinoApiClient::graphQLFieldsQuery(
      self::ID_KEY,
      WinoApiClient::graphQLNestedFieldsQuery(
        self::TAX_KEY,
        WinoApiTax::fieldsQuery()
      ),
      WinoApiClient::graphQLNestedFieldsQuery(
        self::CATEGORY_KEY,
        WinoApiCategory::fieldsQuery()
      )
    );
  }

  /** @return self */
  public static function from(array $data) {
    $tax = optionalArrayData($data, self::TAX_KEY);
    $category = optionalArrayData($data, self::CATEGORY_KEY);
    return new self(
      nonBlankStringData($data, self::ID_KEY),
      $tax === null ? null : WinoApiTax::from($tax),
      $category === null ? null : WinoApiCategory::from($category),
    );
  }
}

/** @extends GraphQLPaginatedQueryEdgeNodeIterator<WinoApiVariant> */
class WinoApiVariantQueryAllIterator extends GraphQLPaginatedQueryEdgeNodeIterator {
  /** @return WinoApiVariant */
  public function currentObject() {
    return WinoApiVariant::from($this->currentNode());
  }
}

class WinoApiVariantIdentity {
  public string $id;
  public const ID_KEY = 'id';

  public function __construct(string $id) {
    $this->id = $id;
  }

  /** @return string */
  public static function fieldsQuery() {
    return WinoApiClient::graphQLFieldsQuery(self::ID_KEY);
  }

  /** @return self */
  public static function from(array $data) {
    return new self(nonBlankStringData($data, self::ID_KEY));
  }

  /** @return self */
  public function identity() {
    return new self($this->id);
  }

  /** @return WinoApiVariant|null */
  public function variant() {
    return $this instanceof WinoApiVariant ? $this : null;
  }
}

class WinoApiVariant extends WinoApiVariantIdentity {
  public WinoApiShopIdentity $shop;
  public const SHOP_KEY = 'shop';
  public string $id;
  public const ID_KEY = 'id';
  public ?\DateTime $archivedAt;
  public const ARCHIVED_AT_KEY = 'archivedAt';

  public string $formattedName;
  public const FORMATTED_NAME_KEY = 'formattedName';
  public string $formattedDescription;
  public const FORMATTED_DESCRIPTION_KEY = 'formattedDescription';
  public ?string $stockKeepingUnit;
  public const STOCK_KEEPING_UNIT_KEY = 'stockKeepingUnit';
  public ?string $tastingNote;
  public const TASTING_NOTE_KEY = 'tastingNote';
  public ?bool $active;
  public const ACTIVE_KEY = 'active';
  public ?bool $bulk;
  public const BULK_KEY = 'bulk';

  public ?WinoApiVariantProduct $product;
  public const PRODUCT_KEY = 'product';
  public WinoApiVariantStock $stock;
  public const STOCK_KEY = 'stock';
  public WinoApiVariantPriceConnection $variantPrices;
  public const VARIANT_PRICES_KEY = 'variantPrices';

  public function __construct(
    WinoApiShopIdentity $shop,
    string $id,
    ?\DateTime $archivedAt,
    string $formattedName,
    string $formattedDescription,
    ?string $stockKeepingUnit,
    ?string $tastingNote,
    ?bool $active,
    ?bool $bulk,
    ?WinoApiVariantProduct $product,
    WinoApiVariantStock $stock,
    WinoApiVariantPriceConnection $variantPrices
  ) {
    $this->shop = $shop;
    $this->id = $id;
    $this->archivedAt = $archivedAt;
    $this->formattedName = $formattedName;
    $this->stockKeepingUnit = $stockKeepingUnit;
    $this->formattedDescription = $formattedDescription;
    $this->tastingNote = $tastingNote;
    $this->active = $active;
    $this->bulk = $bulk;
    $this->stock = $stock;
    $this->product = $product;
    $this->variantPrices = $variantPrices;
  }

  /** @return string */
  public static function fieldsQuery() {
    return WinoApiClient::graphQLFieldsQuery(
      self::ID_KEY,
      self::ARCHIVED_AT_KEY,
      self::FORMATTED_NAME_KEY,
      self::STOCK_KEEPING_UNIT_KEY,
      self::FORMATTED_DESCRIPTION_KEY,
      self::TASTING_NOTE_KEY,
      self::ACTIVE_KEY,
      self::BULK_KEY,
      WinoApiClient::graphQLNestedFieldsQuery(
        self::SHOP_KEY,
        WinoApiShopIdentity::FIELDS_QUERY
      ),
      WinoApiClient::graphQLNestedFieldsQuery(
        self::STOCK_KEY,
        WinoApiVariantStock::FIELDS_QUERY
      ),
      WinoApiClient::graphQLNestedFieldsQuery(
        self::PRODUCT_KEY,
        WinoApiVariantProduct::fieldsQuery()
      ),
      WinoApiClient::graphQLNestedFieldsQuery(
        self::VARIANT_PRICES_KEY,
        WinoApiVariantPriceConnection::FIELDS_QUERY
      )
    );
  }

  /** @return self */
  public static function from(array $data) {
    $product = optionalArrayData($data, self::PRODUCT_KEY);
    return new self(
      WinoApiShopIdentity::from(arrayData($data, self::SHOP_KEY)),
      nonBlankStringData($data, self::ID_KEY),
      optionalDateData($data, self::ARCHIVED_AT_KEY, TimeUnit::javascriptTimestampLike()),
      nonBlankStringData($data, self::FORMATTED_NAME_KEY),
      stringData($data, self::FORMATTED_DESCRIPTION_KEY),
      blankStringToNull(optionalStringData($data, self::STOCK_KEEPING_UNIT_KEY)),
      optionalStringData($data, self::TASTING_NOTE_KEY),
      optionalBoolData($data, self::ACTIVE_KEY),
      optionalBoolData($data, self::BULK_KEY),
      $product === null ? null : WinoApiVariantProduct::from($product),
      WinoApiVariantStock::from(arrayData($data, self::STOCK_KEY)),
      WinoApiVariantPriceConnection::from(arrayData($data, self::VARIANT_PRICES_KEY))
    );
  }

  public const NODE_QUERY_NAME = "Variant";
  public const NODE_QUERY_RESOLVER = "variant";

  /** @return self */
  public static function queryById(WinoApiClient $winoApi, string $variantId) {
    return self::from($winoApi->postGraphQLNodeByIdQuery(
      self::NODE_QUERY_NAME,
      "",
      self::NODE_QUERY_RESOLVER,
      "",
      $variantId,
      self::fieldsQuery(),
      []
    ));
  }

  public const COLLECTION_QUERY_NAME = "Variants";
  public const COLLECTION_QUERY_RESOLVER = "variants";

  /**
   * @param array<string, mixed> $queryVars
   * @return WinoApiVariantQueryAllIterator
   */
  public static function queryAllIterator(
    WinoApiClient $winoApi,
    ?WinoApiShopIdentity $shop,
    string $queryName = self::COLLECTION_QUERY_NAME,
    string $queryArgs = "",
    string $filterByArgs = "",
    string $resolverArgs = "",
    array $queryVars = []
  ) {
    return new WinoApiVariantQueryAllIterator(
      $winoApi,
      GraphQLPaginatedQueryRoot::none(),
      $queryName,
      WinoApiShopIdentity::queryArgs($shop, $queryArgs),
      self::COLLECTION_QUERY_RESOLVER,
      WinoApiShopIdentity::resolverArgs($shop, $filterByArgs, $resolverArgs),
      self::fieldsQuery(),
      WinoApiShopIdentity::queryVars($shop, $queryVars)
    );
  }

  /**
   * Using `queryAllIterator` might be prefered in order to reduce
   * memory footprint. This is especially true if you expect lot of
   * items in the array which could lead to reach the php memory limit.
   *
   * @return WinoApiVariant[]
   */
  public static function queryAll(
    WinoApiClient $winoApi,
    ?WinoApiShopIdentity $shop
  ) {
    return self::queryAllIterator($winoApi, $shop)->queryAll();
  }

  /** @return bool */
  public function isBulk() {
    return $this->bulk ?? false;
  }

  /** @return bool */
  public function isActive() {
    return $this->active ?? true;
  }

  /** @return bool */
  public function isArchived() {
    return $this->archivedAt !== null;
  }
}
