<?php

namespace Wino;

interface WinoApiClientTranslator extends ConfigurationApiKeyTranslator {
  /** @return string */
  public function invalidApiUrl();


  /** @return string */
  public function impossibleToReachApiUrl(string $url);

  /** @return string */
  public function connectionRefusedToApiUrl(string $url);
}

class WinoApiClient {
  private const PRODUCTION_URL = "https://production.api.v2.wino.fr";
  private const STAGING_URL = "https://staging.api.v2.wino.fr";

  private WinoApiClientTranslator $translator;
  private string $baseUrl;
  private string $token;

  /** @return string */
  public function getToken() {
    return $this->token;
  }

  public function __construct(
    WinoApiClientTranslator $translator,
    string $url,
    string $token
  ) {
    $baseUrl = self::PRODUCTION_URL;
    if ($url === 'staging') $baseUrl = self::STAGING_URL;

    $customUrlPrefix = "custom:";
    if (strStartsWith($url, $customUrlPrefix)) $baseUrl = substr($url, strlen($customUrlPrefix));

    $this->translator = $translator;
    $this->baseUrl = rtrim($baseUrl, '/') . '/';
    $this->token = $token;
  }

  /** @return self */
  public static function production(
    WinoApiClientTranslator $translator,
    string $token
  ) {
    return new self($translator, 'production', $token);
  }

  /** @return self */
  public static function staging(
    WinoApiClientTranslator $translator,
    string $token
  ) {
    return new self($translator, 'staging', $token);
  }

  /** @return self */
  public static function custom(
    WinoApiClientTranslator $translator,
    string $url,
    string $token
  ) {
    return new self($translator, "custom:$url", $token);
  }

  /** @return bool */
  public function isConfigured() {
    $emptyBaseUrl = strlen(trim($this->baseUrl)) === 0;
    $emptyToken = strlen(trim($this->token)) === 0;
    return !$emptyBaseUrl && !$emptyToken;
  }

  /** @return void */
  public static function verifyConfiguration(WinoApiClientTranslator $translator) {
    ConfigurationApiKey::verify($translator);
  }

  /** @return self */
  public static function fromConfiguration(?int $userId = null) {
    $translator = LocaleTranslator::fromConfiguration($userId);
    $url = ConfigurationApiUrl::value();
    $bearer = ConfigurationApiKey::value();

    return new self($translator, $url, $bearer);
  }

  /** @return self */
  public static function fromConfigurationDefault() {
    $translator = LocaleTranslator::fromConfigurationDefault();
    $url = ConfigurationApiUrl::DEFAULT;
    $bearer = ConfigurationApiKey::DEFAULT;

    return new self($translator, $url, $bearer);
  }

  public const ERROR_CODE = "wino-api-client-error";

  /**
   * @param mixed $data
   * @return mixed
   */
  public function httpRequest(string $method, string $endpoint, $data = null) {
    $code = errorCode(self::ERROR_CODE);

    $translator = $this->translator;
    $endpointUrl = $this->baseUrl . $endpoint;
    if (!isUrlValid($endpointUrl)) throw $code->error($translator->invalidApiUrl());

    $body = jsonString($data);
    $method = trim(strtolower($method));
    $requestLog = msg($method, $endpointUrl, $body);

    $ch = curl_init();
    if ($ch === false) throw $code->msgException("Failed curl initialization", $requestLog);
    curl_setopt($ch, CURLOPT_URL, $endpointUrl);
    curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);

    if ($method === 'post') curl_setopt($ch, CURLOPT_POST, 1);

    if ($method === 'patch' || $method === 'put' || $method === 'delete') curl_setopt(
      $ch, CURLOPT_CUSTOMREQUEST, strtoupper($method)
    );

    if ($data !== null) curl_setopt($ch, CURLOPT_POSTFIELDS, $body);

    curl_setopt($ch, CURLOPT_HTTPHEADER, [
      'Content-Type: application/json',
      'Authorization: Bearer ' . $this->token,
    ]);

    $response = curl_exec($ch);
    $debugMode = isDebugModeEnabled();
    if ($response === false) {
      $errorLog = curlErrorLog($ch);
      if (strContains($errorLog, 'Could not resolve host')) throw $code->exception(
        $translator->impossibleToReachApiUrl($endpointUrl)
      );

      throw $code->msgException(
        'cURL session failed while sending http request =>',
        $requestLog,
        $errorLog
      );
    }

    $info = self::arrayResponseData(curl_getinfo($ch));
    $httpCode = intData($info, 'http_code');
    if ($httpCode < 200 || $httpCode > 299) {
      if ($httpCode === 401) throw $code->exception(
        $translator->connectionRefusedToApiUrl($endpointUrl)
      );

      $debugMessageEnd = " Query => " . $requestLog;
      throw $code->strException(
        "Unexpected http status code ", jsonString($httpCode), " returned by the API. ",
        "Response => ", jsonString($response),
        $debugMode ? $debugMessageEnd : ""
      );
    }

    /** @var mixed */
    $responseData = $response === true ? null : json_decode($response, true);
    $hasError = false;

    if (truthy()
      && is_array($responseData)
      && key_exists('errors', $responseData)
      && is_array(arrayElement($responseData, 'errors'))
    ) $hasError = true;
    elseif (truthy()
      && is_object($responseData)
      && property_exists($responseData, 'errors')
      && is_array($responseData->errors)
    ) $hasError = true;

    if ($hasError) throw $code->msgException(
      'Errors found in http request response:',
      $method, $endpointUrl, $body, " => response($response)"
    );

    return $responseData;
  }

  /**
   * @param mixed $data
   * @return mixed
   */
  public function postHttpRequest(string $endpoint, $data) {
    return $this->httpRequest('post', $endpoint, $data);
  }

  /**
   * @param mixed $data
   * @return mixed
   */
  public function putHttpRequest(string $endpoint, $data) {
    return $this->httpRequest('put', $endpoint, $data);
  }

  /**
   * @param mixed $data
   * @return mixed
   */
  public function patchHttpRequest(string $endpoint, $data) {
    return $this->httpRequest('patch', $endpoint, $data);
  }

  /** @return mixed */
  public function getHttpRequest(string $endpoint) {
    return $this->httpRequest('GET', $endpoint);
  }

  /** @return mixed */
  public function deleteHttpRequest(string $endpoint) {
    return $this->httpRequest('delete', $endpoint);
  }

  private const WEBHOOKS_SUBSCRIPTIONS_ENDPOINT = 'webhooks/subscriptions';

  /**
   * @param mixed $response
   * @return array
   */
  public static function arrayResponseData($response) {
    if (is_array($response)) return $response;
    throw msgException("Unexpected non array response data", jsonString($response));
  }

  /** @return array */
  public function getWebhooksSubscriptions() {
    return self::arrayResponseData(
      $this->getHttpRequest(self::WEBHOOKS_SUBSCRIPTIONS_ENDPOINT)
    );
  }

  /** @return array */
  public function getWebhookSubscription(string $subscriptionId) {
    return self::arrayResponseData(
      $this->getHttpRequest(self::WEBHOOKS_SUBSCRIPTIONS_ENDPOINT . '/' . $subscriptionId)
    );
  }

  /** @return array */
  public function postWebhookSubscription(array $payload) {
    return self::arrayResponseData(
      $this->postHttpRequest(self::WEBHOOKS_SUBSCRIPTIONS_ENDPOINT, $payload)
    );
  }

  /** @return array */
  public function putWebhookSubscription(string $subscriptionId, array $payload) {
    return self::arrayResponseData(
      $this->putHttpRequest(self::WEBHOOKS_SUBSCRIPTIONS_ENDPOINT . '/' . $subscriptionId, $payload)
    );
  }

  /** @return array */
  public function patchWebhookSubscription(string $subscriptionId, array $payload) {
    return self::arrayResponseData(
      $this->patchHttpRequest(self::WEBHOOKS_SUBSCRIPTIONS_ENDPOINT . '/' . $subscriptionId, $payload)
    );
  }

  /** @return void */
  public function deleteWebhookSubscription(string $subscriptionId) {
    $this->deleteHttpRequest(self::WEBHOOKS_SUBSCRIPTIONS_ENDPOINT . '/' . $subscriptionId);
  }

  /** @return string */
  public static function joinArgs(string $arg, string ...$args) {
    $trim = trim($arg);
    $validArgs = $trim === '' ? [] : [$trim];
    foreach ($args as $additionalArg) {
      $trim = trim($additionalArg);
      if ($trim === '') continue;
      $validArgs[] = $trim;
    }
    return implode(", ", $validArgs);
  }

  /** @return string */
  public static function filterByArgs(string ...$args) {
    $filterByArgs = trim(self::joinArgs(...$args));
    return $filterByArgs === "" ? "" : "filterBy: { $filterByArgs }";
  }

  /**
   * @param array<string, mixed> $vars
   * @param array<string, mixed> $additionalVars
   *
   * @return array<string, mixed>
   */
  public static function joinVars(array $vars, array $additionalVars) {
    /** @psalm-suppress MixedAssignment */
    foreach ($additionalVars as $varName => $varValue) $vars[$varName] = $varValue;
    return $vars;
  }

  /** @return string */
  public static function graphQLFieldsQuery(string ...$fields) {
    return join(', ', $fields);
  }

  /** @return string */
  public static function graphQLResolverQuery(
    string $name,
    string $args,
    string $fields
  ) {
    $argsString = trim($args) === "" ? "" : "($args)";
    return "$name$argsString { $fields }";
  }

  /** @return string */
  public static function graphQLNestedFieldsQuery(string $key, string $fieldsQuery) {
    return self::graphQLResolverQuery($key, '', $fieldsQuery);
  }

  /** @return string */
  public static function graphQLQuery(
    string $name,
    string $args,
    string $fields
  ) {
    $query = self::graphQLResolverQuery($name, $args, $fields);
    return "query $query";
  }

  /** @return string */
  public static function graphQLMutationQuery(
    string $name,
    string $args,
    string $fields
  ) {
    $query = self::graphQLResolverQuery($name, $args, $fields);
    return "mutation $query";
  }

  /** @return string */
  public static function graphQLNodeQuery(
    string $name,
    string $args,
    string $resolverName,
    string $resolverArgs,
    string $nodeFieldsQuery
  ) {
    return self::graphQLQuery(
      $name,
      $args,
      self::graphQLResolverQuery($resolverName, $resolverArgs, $nodeFieldsQuery)
    );
  }

  /** @return string */
  public static function graphQLNodeMutationQuery(
    string $name,
    string $args,
    string $resolverName,
    string $resolverArgs,
    string $nodeFieldsQuery
  ) {
    return self::graphQLMutationQuery(
      $name,
      $args,
      self::graphQLResolverQuery($resolverName, $resolverArgs, $nodeFieldsQuery)
    );
  }

  /** @return string */
  public static function graphQLNodeByIdQueryArgs(
    string $queryArgs,
    string $queryArgsNodeIdVarName = "nodeId"
  ) {
    return self::joinArgs("\$$queryArgsNodeIdVarName: ID!", $queryArgs);
  }

  /** @return string */
  public static function graphQLNodeByIdResolverArgs(
    string $resolverArgs,
    string $queryArgsNodeIdVarName = "nodeId",
    string $resolverArgsIdVarName = "id"
  ) {
    return self::joinArgs("$resolverArgsIdVarName:\$$queryArgsNodeIdVarName", $resolverArgs);
  }

  /**
   * @template V
   * @param array<string, V> $variables
   *
   * @return array<string, V|string>
   */
  public static function graphQLNodeByIdQueryVars(
    string $nodeId,
    array $variables,
    string $queryArgsNodeIdVarName = "nodeId"
  ) {
    return self::joinVars([$queryArgsNodeIdVarName => $nodeId], $variables);
  }

  /** @return string */
  public static function graphQLPaginatedQueryArgs(
    string $queryArgs,
    string $queryArgsFirstVarName = "first",
    string $queryArgsAfterVarName = "after"
  ) {
    return self::joinArgs(
      "\$$queryArgsFirstVarName: Int",
      "\$$queryArgsAfterVarName: String",
      $queryArgs
    );
  }

  /** @return string */
  public static function graphQLPaginatedQueryFields(
    string $resolverName,
    string $resolverArgs,
    string $nodeFieldsQuery,
    string $queryArgsFirstVarName = "first",
    string $queryArgsAfterVarName = "after"
  ) {
    $resolverArgsInCall = self::joinArgs(
      "first:\$$queryArgsFirstVarName",
      "after:\$$queryArgsAfterVarName",
      $resolverArgs
    );

    $fields = <<<GRAPHQL
      $resolverName($resolverArgsInCall) {
        pageInfo {
          hasNextPage
          endCursor
        }
        edges { node { $nodeFieldsQuery } }
      }
    GRAPHQL;
    return $fields;
  }

  /**
   * @param array<string, mixed> $variables
   * @return array<string, mixed>
   */
  public static function graphQLPaginatedQueryVars(
    array $variables,
    string $queryArgsFirstVarName = "first",
    string $queryArgsAfterVarName = "after"
  ) {
    return self::joinVars([
      $queryArgsFirstVarName => self::MAX_ITEMS_PER_PAGE,
      $queryArgsAfterVarName => '',
    ], $variables);
  }

  private const GRAPHQL_ENDPOINT = 'graphql';

  /**
   * @param array<string, mixed> $vars
   * @return mixed
   */
  public function postGraphQLQuery(string $query, array $vars = []) {
    return $this->postHttpRequest(self::GRAPHQL_ENDPOINT, ['query' => $query, 'variables' => $vars]);
  }

  /**
   * @param mixed $response
   * @return array
   */
  public static function graphQLNodeData($response, string $resolverName) {
    return self::arrayResponseData(
      requiredData(arrayData(self::arrayResponseData($response), 'data'), $resolverName)
    );
  }

  /**
   * @param array<string, mixed> $variables
   *
   * @return array
   */
  public function postGraphQLNodeQuery(
    string $queryName,
    string $queryArgs,
    string $resolverName,
    string $resolverArgs,
    string $nodeFieldsQuery,
    array $variables
  ) {
    $query = self::graphQLNodeQuery($queryName, $queryArgs, $resolverName, $resolverArgs, $nodeFieldsQuery);
    return self::graphQLNodeData($this->postGraphQLQuery($query, $variables), $resolverName);
  }

  /**
   * @param array<string, mixed> $variables
   * @return array
   */
  public function postGraphQLNodeMutationQuery(
    string $queryName,
    string $queryArgs,
    string $resolverName,
    string $resolverArgs,
    string $nodeFieldsQuery,
    array $variables
  ) {
    $query = self::graphQLNodeMutationQuery($queryName, $queryArgs, $resolverName, $resolverArgs, $nodeFieldsQuery);
    return self::graphQLNodeData(
      $this->postGraphQLQuery($query, $variables),
      $resolverName
    );
  }

  /**
   * @param array<string, mixed> $variables
   * @return array
   */
  public function postGraphQLNodeByIdQuery(
    string $queryName,
    string $queryArgs,
    string $resolverName,
    string $resolverArgs,
    string $nodeId,
    string $nodeFieldsQuery,
    array $variables
  ) {
    return $this->postGraphQLNodeQuery(
      $queryName,
      self::graphQLNodeByIdQueryArgs($queryArgs),
      $resolverName,
      self::graphQLNodeByIdResolverArgs($resolverArgs),
      $nodeFieldsQuery,
      self::graphQLNodeByIdQueryVars($nodeId, $variables),
    );
  }

  private const MAX_ITEMS_PER_PAGE = 50;

  /**
   * @param array<string, mixed> $variables
   * @return mixed
   */
  public function postGraphQLPaginatedQuery(
    string $queryName,
    string $queryArgs,
    string $resolverName,
    string $resolverArgs,
    string $nodeFieldsQuery,
    array $variables
  ) {
    $query = self::graphQLQuery(
      $queryName,
      self::graphQLPaginatedQueryArgs($queryArgs),
      self::graphQLPaginatedQueryFields(
        $resolverName,
        $resolverArgs,
        $nodeFieldsQuery
      )
    );
    $vars = self::graphQLPaginatedQueryVars($variables);
    return $this->postGraphQLQuery($query, $vars);
  }

  /**
   * @param array<string, mixed> $variables
   * @return mixed
   */
  public function postGraphQLPaginatedByRootNodeIdQuery(
    string $queryName,
    string $queryArgs,
    string $rootResolverName,
    string $rootResolverArgs,
    string $rootNodeId,
    string $resolverName,
    string $resolverArgs,
    string $nodeFieldsQuery,
    array $variables
  ) {
    $query = self::graphQLQuery(
      $queryName,
      self::graphQLPaginatedQueryArgs(self::graphQLNodeByIdQueryArgs($queryArgs)),
      self::graphQLResolverQuery(
        $rootResolverName,
        self::graphQLNodeByIdResolverArgs($rootResolverArgs),
        self::graphQLPaginatedQueryFields(
          $resolverName,
          $resolverArgs,
          $nodeFieldsQuery
        )
      ),
    );
    $vars = self::graphQLPaginatedQueryVars(
      self::graphQLNodeByIdQueryVars($rootNodeId, $variables)
    );
    return $this->postGraphQLQuery($query, $vars);
  }
}

class GraphQLPaginatedQueryRoot {
  private string $resolverName;

  /** @return string */
  public function getResolverName() {
    return $this->resolverName;
  }

  private string $resolverArgs;

  /** @return string */
  public function getResolverArgs() {
    return $this->resolverArgs;
  }

  private string $nodeId;

  /** @return string */
  public function getNodeId() {
    return $this->nodeId;
  }

  private function __construct(
    string $resolverName,
    string $resolverArgs,
    string $nodeId
  ) {
    $this->resolverName = $resolverName;
    $this->resolverArgs = $resolverArgs;
    $this->nodeId = $nodeId;
  }

  /** @return null */
  public static function none() {
    return null;
  }

  /** @return self */
  public static function nodeById(
    string $resolverName,
    string $resolverArgs,
    string $nodeId
  ) {
    return new self($resolverName, $resolverArgs, $nodeId);
  }
}

class GraphQLPaginatedQueryEdgeNodeIteratorState extends PaginatedDataIteratorState {
  public ?string $endCursor;

  public function __construct(
    ?array $currentPageItems,
    ?int $currentPageIndex,
    ?string $endCursor
  ) {
    $hasNextPage = $endCursor !== null;
    parent::__construct($currentPageItems, $currentPageIndex, $hasNextPage);
    $this->endCursor = $endCursor;
  }

  /** @return self */
  public static function from(PaginatedDataIteratorState $state, ?string $endCursor) {
    return new self(
      $state->currentPageItems,
      $state->currentPageIndex,
      $endCursor
    );
  }
}

/**
 * @template T
 *
 * @extends PaginatedDataIterator<array{node: array}>
 */
abstract class GraphQLPaginatedQueryEdgeNodeIterator extends PaginatedDataIterator {
  protected ?string $endCursor;

  /** @return GraphQLPaginatedQueryEdgeNodeIteratorState */
  public function getCurrentState() {
    return GraphQLPaginatedQueryEdgeNodeIteratorState::from(
      parent::getCurrentState(),
      $this->endCursor
    );
  }

  /** @return void */
  public function resume(GraphQLPaginatedQueryEdgeNodeIteratorState $state) {
    parent::resumeState($state);
    $this->endCursor = $state->endCursor;
  }

  private WinoApiClient $winoApi;
  private ?GraphQLPaginatedQueryRoot $root;
  private string $queryName;
  private string $queryArgs;
  private string $resolverName;
  private string $resolverArgs;
  private string $nodeFieldsQuery;

  /** @var array<string, mixed> */
  private array $variables;

  /**
   * @param array<string, mixed> $variables
   */
  public function __construct(
    WinoApiClient $winoApi,
    ?GraphQLPaginatedQueryRoot $root,
    string $queryName,
    string $queryArgs,
    string $resolverName,
    string $resolverArgs,
    string $nodeFieldsQuery,
    array $variables
  ) {
    if (key_exists('after', $variables)) throw error('Unexpected "after" variable.');

    parent::__construct();
    $this->winoApi = $winoApi;
    $this->root = $root;
    $this->queryName = $queryName;
    $this->queryArgs = $queryArgs;
    $this->resolverName = $resolverName;
    $this->resolverArgs = $resolverArgs;
    $this->nodeFieldsQuery = $nodeFieldsQuery;
    $this->variables = $variables;
    $this->endCursor = null;
  }

  /** @return mixed */
  private function queryResponse() {
    $api = $this->winoApi;
    $root = $this->root;
    $variables = $this->variables;
    $queryName = $this->queryName;
    $queryArgs = $this->queryArgs;
    $resolverName = $this->resolverName;
    $resolverArgs = $this->resolverArgs;
    $nodeFieldsQuery = $this->nodeFieldsQuery;
    $endCursor = $this->endCursor;

    $vars = $endCursor === null ? [] : ['after' => $endCursor];
    /** @psalm-suppress MixedAssignment */
    foreach ($variables as $varName => $varValue) $vars[$varName] = $varValue;

    if ($root === null) return $api->postGraphQLPaginatedQuery(
      $queryName,
      $queryArgs,
      $resolverName,
      $resolverArgs,
      $nodeFieldsQuery,
      $vars
    );

    return $api->postGraphQLPaginatedByRootNodeIdQuery(
      $queryName,
      $queryArgs,
      $root->getResolverName(),
      $root->getResolverArgs(),
      $root->getNodeId(),
      $resolverName,
      $resolverArgs,
      $nodeFieldsQuery,
      $vars
    );
  }

  /** @return array{node: array}[] */
  protected function nextPageItems() {
    $resolverName = $this->resolverName;
    $response = WinoApiClient::arrayResponseData($this->queryResponse());
    $root = $this->root;
    $data = arrayData($response, 'data');
    $rootData = $root === null ? $data : arrayData($data, $root->getResolverName());
    $resolverData = arrayData($rootData, $resolverName);
    $pageInfo = arrayData($resolverData, 'pageInfo');
    $this->state->hasNextPage = boolData($pageInfo, 'hasNextPage');
    $this->endCursor = optionalNonBlankStringData($pageInfo, 'endCursor');

    /** @var array{node: array}[] */
    $nodes = [];
    $edges = arrayData($resolverData, 'edges');
    foreach ($edges as $edge) {
      if (!is_array($edge)) throw msgException('Unexpected non array edge', jsonString($edge));
      $nodes[] = ['node' => arrayData($edge, 'node')];
    }
    return $nodes;
  }

  /** @return array */
  protected function currentNode() {
    return $this->currentPageItem()['node'];
  }

  /** @return T */
  abstract protected function currentObject();

  /** @return T[] */
  public function queryAll() {
    $allObjects = [];
    while ($this->next()) $allObjects[] = $this->currentObject();
    return $allObjects;
  }
}
