<?php

namespace Wino;

class WinoApiStateRecord {
  private string $kind;
  private array $data;

  /** @return string */
  private static function trim(string $string) {
    return trim(strtolower($string));
  }

  /** @return bool */
  public function is(string $kind) {
    return self::trim($this->kind) === self::trim($kind);
  }

  /**
   * @param mixed $kind
   * @param mixed $data
   */
  public function __construct($kind, $data) {
    $this->kind = is_string($kind) ? $kind : '';
    $this->data = is_array($data) ? $data : [];
  }

  /** @return mixed */
  private function getData(string $key) {
    return optionalData($this->data, $key);
  }

  /** @return string|null */
  public function getString(string $key) {
    /** @var mixed */
    $data = $this->getData($key);
    return is_string($data) ? $data : null;
  }

  /** @return non-empty-string|null */
  public function getNonEmptyString(string $key) {
    $data = $this->getString($key);
    if ($data === null) return null;
    return $data === "" ? null : $data;
  }

  /** @return non-empty-string|null */
  public function getNonBlankString(string $key) {
    $data = $this->getNonEmptyString($key);
    if ($data === null) return null;
    return trim($data) === "" ? null : $data;
  }
}

class WinoApiCategoryStateRecordData {
  public const KIND = "Category";

  public string $id;

  public function __construct(string $id) {
    $this->id = $id;
  }

  /** @return self|null */
  public static function from(WinoApiStateRecord $record) {
    if (!$record->is(self::KIND)) return null;
    $id = $record->getNonBlankString('id');
    return $id === null ? null : new self($id);
  }
}

class WinoApiTaxStateRecordData {
  public const KIND = "Tax";

  public string $id;

  public function __construct(string $id) {
    $this->id = $id;
  }

  /** @return self|null */
  public static function from(WinoApiStateRecord $record) {
    if (!$record->is(self::KIND)) return null;
    $id = $record->getNonBlankString('id');
    return $id === null ? null : new self($id);
  }
}

class WinoApiPriceStateRecordData {
  public const KIND = "Price";

  public string $id;
  public const ID_KEY = 'id';
  public string $name;
  public const NAME_KEY = 'name';

  public function __construct(
    string $id,
    string $name
  ) {
    $this->id = $id;
    $this->name = $name;
  }

  /** @return self|null */
  public static function from(WinoApiStateRecord $record) {
    if (!$record->is(self::KIND)) return null;

    $id = $record->getNonBlankString(self::ID_KEY);
    $name = $record->getString(self::NAME_KEY);
    if ($id === null || $name === null) return null;

    return new self($id, $name);
  }
}

class WinoApiVariantPriceStateRecordData {
  public const KIND = "VariantPrice";

  public string $variantId;

  public function __construct(string $variantId) {
    $this->variantId = $variantId;
  }

  /** @return self|null */
  public static function from(WinoApiStateRecord $record) {
    if (!$record->is(self::KIND)) return null;

    $variantId = $record->getNonBlankString('variantId');
    return $variantId === null ? null : new self($variantId);
  }
}

class WinoApiProductStateRecordData {
  public const KIND = "Product";

  public string $id;

  public function __construct(string $id) {
    $this->id = $id;
  }

  /** @return self|null */
  public static function from(WinoApiStateRecord $record) {
    if (!$record->is(self::KIND)) return null;

    $id = $record->getNonBlankString('id');
    return $id === null ? null : new self($id);
  }
}

class WinoApiVariantStateRecordData {
  public const KIND = "Variant";

  public string $id;

  public function __construct(string $id) {
    $this->id = $id;
  }

  /** @return self|null */
  public static function from(WinoApiStateRecord $record) {
    if (!$record->is(self::KIND)) return null;

    $id = $record->getNonBlankString('id');
    return $id === null ? null : new self($id);
  }
}

class WinoApiStockActivityStateRecordData {
  public const KIND = "StockActivity";

  public string $variantId;

  public function __construct(string $variantId) {
    $this->variantId = $variantId;
  }

  /** @return self|null */
  public static function from(WinoApiStateRecord $record) {
    if (!$record->is(self::KIND)) return null;

    $variantId = $record->getNonBlankString('variantId');
    return $variantId === null ? null : new self($variantId);
  }
}

class WinoApiWebhookNotification {
  private \WP_REST_Request $request;
  private WinoApiStateRecord $data;

  private function __construct(\WP_REST_Request $request) {
    $this->request = $request;

    $body = getWordpressRestRequestJsonParams($request);
    /** @var mixed */
    $kind = optionalData($body, 'kind');
    /** @var mixed */
    $data = optionalData($body, 'data');

    $this->data = new WinoApiStateRecord($kind, $data);
  }

  /** @return self */
  public static function from(\WP_REST_Request $request) {
    return new self($request);
  }

  /** @return WinoApiCategoryStateRecordData|null */
  public function getCategoryData() {
    return WinoApiCategoryStateRecordData::from($this->data);
  }

  /** @return WinoApiTaxStateRecordData|null */
  public function getTaxData() {
    return WinoApiTaxStateRecordData::from($this->data);
  }

  /** @return WinoApiPriceStateRecordData|null */
  public function getPriceData() {
    return WinoApiPriceStateRecordData::from($this->data);
  }

  /** @return WinoApiVariantPriceStateRecordData|null */
  public function getVariantPriceData() {
    return WinoApiVariantPriceStateRecordData::from($this->data);
  }

  /** @return WinoApiProductStateRecordData|null */
  public function getProductData() {
    return WinoApiProductStateRecordData::from($this->data);
  }

  /** @return WinoApiVariantStateRecordData|null */
  public function getVariantData() {
    return WinoApiVariantStateRecordData::from($this->data);
  }

  /** @return WinoApiStockActivityStateRecordData|null */
  public function getStockActivityData() {
    return WinoApiStockActivityStateRecordData::from($this->data);
  }
}

class WinoApiWebhookSubscription {
  public string $id;
  public \DateTime $createdAt;
  public \DateTime $updatedAt;
  public string $userId;
  public string $url;
  public ?string $clientState;
  public string $shopId;
  public bool $disabled;

  public function __construct(
    string $id,
    \DateTime $createdAt,
    \DateTime $updatedAt,
    string $userId,
    string $url,
    ?string $clientState,
    string $shopId,
    bool $disabled
  ) {
    $this->id = $id;
    $this->createdAt = $createdAt;
    $this->updatedAt = $updatedAt;
    $this->userId = $userId;
    $this->url = $url;
    $this->clientState = $clientState;
    $this->shopId = $shopId;
    $this->disabled = $disabled;
  }

  public const ID_KEY = 'id';
  public const CREATED_AT_KEY = 'createdAt';
  public const UPDATED_AT_KEY = 'updatedAt';
  public const USER_ID_KEY = 'userId';
  public const URL_KEY = 'url';
  public const CLIENT_STATE_KEY = 'clientState';
  public const SHOP_ID_KEY = 'shopId';
  public const DISABLED_KEY = 'disabled';

  /** @return self */
  public static function from(array $data) {
    return new self(
      nonBlankStringData($data, self::ID_KEY),
      dateData($data, self::CREATED_AT_KEY, TimeUnit::javascriptTimestampLike()),
      dateData($data, self::UPDATED_AT_KEY, TimeUnit::javascriptTimestampLike()),
      nonBlankStringData($data, self::USER_ID_KEY),
      nonBlankStringData($data, self::URL_KEY),
      optionalStringData($data, self::CLIENT_STATE_KEY),
      nonBlankStringData($data, self::SHOP_ID_KEY),
      boolData($data, self::DISABLED_KEY)
    );
  }

  /** @return WinoApiWebhookSubscription[] */
  public static function queryAll(WinoApiClient $api) {
    $data = assertedArraysArray($api->getWebhooksSubscriptions());
    $all = [];
    foreach ($data as $subscriptionData) $all[] = self::from($subscriptionData);
    return $all;
  }

  /** @return WinoApiWebhookSubscription */
  public static function queryById(WinoApiClient $api, string $id) {
    $data = $api->getWebhookSubscription($id);
    return self::from($data);
  }

  /** @return self */
  public static function updateById(WinoApiClient $api, string $id, array $patchPayload) {
    $patch = $api->patchWebhookSubscription($id, $patchPayload);
    return self::from($patch);
  }

  /** @return void */
  public static function deleteById(WinoApiClient $api, string $id) {
    $api->deleteWebhookSubscription($id);
  }

  /** @return self */
  public static function createOne(
    WinoApiClient $api,
    string $shopId,
    string $url,
    bool $disabled,
    ?string $clientState
  ) {
    return self::from($api->postWebhookSubscription([
      self::SHOP_ID_KEY => $shopId,
      self::URL_KEY => $url,
      self::DISABLED_KEY => $disabled,
      self::CLIENT_STATE_KEY => $clientState
    ]));
  }
}

class WinoApiWebhook {
  /** @return WinoApiWebhookSubscription[] */
  public static function queryAllSubscriptions(WinoApiClient $api) {
    return WinoApiWebhookSubscription::queryAll($api);
  }

  /** @return WinoApiWebhookSubscription */
  public static function createSubscription(
    WinoApiClient $api,
    WinoApiShopIdentity $shop,
    string $url,
    bool $disabled,
    ?string $clientState
  ) {
    return WinoApiWebhookSubscription::createOne(
      $api,
      $shop->id,
      $url,
      $disabled,
      $clientState
    );
  }

  /** @return WinoApiWebhookSubscription */
  public static function updateSubscriptionById(
    WinoApiClient $api,
    string $id,
    array $patch
  ) {
    return WinoApiWebhookSubscription::updateById($api, $id, $patch);
  }

  /** @return WinoApiWebhookSubscription */
  public static function querySubscriptionById(WinoApiClient $api, string $id) {
    return WinoApiWebhookSubscription::queryById($api, $id);
  }

  /** @return void */
  public static function deleteSubscriptionById(WinoApiClient $api, string $id) {
    WinoApiWebhookSubscription::deleteById($api, $id);
  }

  /**
   * @param Closure(WinoApiWebhookNotification): (null|RestResponse|\WP_REST_Response) $callback
   *
   * @return void
   */
  public static function registerNotificationHandler(RestRoute $route, callable $callback) {
    RestRouter::registerWebhookNotificationHandler(
      $route,
      function (\WP_REST_Request $request) use ($callback) {
        $notification = WinoApiWebhookNotification::from($request);
        return call_user_func($callback, $notification);
      }
    );
  }

  /** @return WinoApiWebhookSubscription|null */
  private static function queryExisting(WinoApiClient $api, string $url) {
    $existingSubscriptions = WinoApiWebhook::queryAllSubscriptions($api);
    foreach ($existingSubscriptions as $subscription) {
      if (trim($subscription->url) === trim($url)) return $subscription;
    }
    return null;
  }

  /**
   * The client state is generated from the api token hashed with bcrypt.
   *
   * This allows to get a client state without needing additonnal configuration
   * from the user, while ensuring the token is not readable, since hashed.
   *
   * The wino webhook api should also probably support a signature
   * header at some point to also ensure the response integrity.
   *
   * @return string
   */
  public static function secretClientState(WinoApiClient $api) {
    return password_hash($api->getToken(), PASSWORD_BCRYPT);
  }

  /** @return bool */
  public static function isSecretClientStateValid(WinoApiClient $api, string $clientState) {
    return password_verify($api->getToken(), $clientState);
  }

  /** @return WinoApiWebhookSubscription */
  public static function activateWebhook(
    WinoApiClient $api,
    WinoApiShopIdentity $shop,
    RestRoute $route
  ) {
    $url = $route->url();
    $disabled = false;
    $existing = self::queryExisting($api, $url);
    $clientState = self::secretClientState($api);
    if ($existing === null) return self::createSubscription(
      $api, $shop, $url, $disabled, $clientState
    );

    $patchPayload = [];
    if ($existing->disabled !== $disabled) {
      $patchPayload[WinoApiWebhookSubscription::DISABLED_KEY] = $disabled;
    }

    if (count(array_keys($patchPayload)) === 0) return $existing;
    return self::updateSubscriptionById($api, $existing->id, $patchPayload);
  }

  /** @return void */
  public static function deactivateWebhook(WinoApiClient $api, RestRoute $route) {
    $existing = self::queryExisting($api, $route->url());
    if ($existing === null) return;
    self::deleteSubscriptionById($api, $existing->id);
  }
}
