<?php

namespace Wino;

/** @return void */
function includeAdminStyles() {
  include __DIR__ . '/wp.admin.css.php';
}

/**
 * @param Closure(): mixed $callback
 *
 * @return void
 */
function addAdminPostAction(WordpressTransientLog $log, callable $callback) {
  add_action('admin_post', function () use ($callback, $log) {
    try {
      call_user_func($callback);
    } catch (\Throwable $thrown) {
      debugLog($thrown);
      $log->error($thrown);
    }
    refererRedirect();
  });
}

/**
 * Redirect to the page from which we came (which should always be the
 * admin page. If the referred isn't set, then we redirect the user to
 * the login page.
 *
 * @return never
 */
function refererRedirect() {
  $key = '_wp_http_referer';
  $postRefererUrl = isset($_POST[$key]) ? $_POST[$key] : null;
  $redirectUrl = is_string($postRefererUrl) ? $postRefererUrl : wp_login_url();
  $url = sanitize_text_field(wp_unslash($redirectUrl));
  wp_safe_redirect(urldecode($url));
  exit;
}

class Auth {
  public const ADMIN_CAPABILITY = 'manage_options';
  public const WOOCOMMERCE_ADMIN_CAPABILITY = 'manage_woocommerce';

  /** @return void */
  public static function verifyCurrentUserCapability(string $capability) {
    if (current_user_can($capability)) return;
    throw error("Operation not permitted: missing user capability ($capability)");
  }

  /** @return void */
  public static function verifyAdminCapability() {
    Auth::verifyCurrentUserCapability(self::ADMIN_CAPABILITY);
  }

  /** @return void */
  public static function verifyDebugMode() {
    if (!isDebugModeEnabled()) throw error("Debug mode is disabled.");
  }

  /** @return bool */
  public static function isCurrentUserAdmin() {
    return current_user_can(self::ADMIN_CAPABILITY);
  }
}

class ConfigurationOption {
  private array $post;
  private string $key;
  private bool $posted;

  /** @var mixed */
  private $value;

  /**
   * @param mixed $value
   */
  private function __construct(
    array $post,
    string $key,
    bool $posted,
    $value
  ) {
    $this->post = $post;
    $this->key = $key;
    $this->posted = $posted;
    $this->value = $value;
  }

  /** @return string */
  public static function readPostedValueNonceName(string $key) {
    return FormHandling::nonceName("read-posted-configuration-option-$key");
  }

  /** @return string */
  public static function readPostedValueNonce(string $key) {
    return FormHandling::nonce(self::readPostedValueNonceName($key));
  }

  /** @return string */
  public static function savePostedValueNonceName(string $key) {
    return FormHandling::nonceName("save-posted-configuration-option-$key");
  }

  /** @return string */
  public static function savePostedValueNonce(string $key) {
    return FormHandling::nonce(self::savePostedValueNonceName($key));
  }

  /** @return self */
  public static function fromPost(array $post, string $key) {
    $validSavePostedNonce = FormHandling::isNonceValid($post, self::savePostedValueNonceName($key));
    $validReadPostedNonce = FormHandling::isNonceValid($post, self::readPostedValueNonceName($key));
    $validNonce = $validSavePostedNonce || $validReadPostedNonce;
    $posted = $validNonce && key_exists($key, $post);

    /** @var mixed */
    $value = $posted ? wp_unslash($post[$key]) : get_option($key, null);

    return new self($post, $key, $posted, $value);
  }

  /** @return self */
  public static function from(string $key) {
    return self::fromPost($_POST, $key);
  }

  /**
   * @template V
   * @param V $value
   * @return self
   */
  public static function fromValue(
    string $key,
    $value,
    bool $allowSave
  ) {
    $post = [];
    $post[$key] = $value;
    $nonceName = $allowSave ? self::savePostedValueNonceName($key) : self::readPostedValueNonceName($key);
    $nonce = $allowSave ? self::savePostedValueNonce($key) : self::readPostedValueNonce($key);
    $post[$nonceName] = $nonce;

    return self::fromPost($post, $key);
  }

  /** @return self */
  public static function fromTextValue(
    string $key,
    string $value,
    bool $allowSave
  ) {
    return self::fromValue($key, $value, $allowSave);
  }

  /** @return self */
  public static function fromBooleanValue(
    string $key,
    bool $value,
    bool $allowSave
  ) {
    return self::fromValue($key, $value, $allowSave);
  }

  /**
   * @param mixed $default
   * @return mixed
   */
  public function valueWithDefault($default) {
    /** @var mixed */
    $value = $this->value;
    return $value === null ? $default : $value;
  }

  /**
   * @param mixed $value
   * @return string
   */
  public static function unexpectedOptionErrorMessage(string $key, $value) {
    return "Unexpected " . jsonString($key) . " option value => " . jsonString($value);
  }

  /**
   * @param mixed $value
   * @return \Exception
   */
  public static function unexpectedOptionErrorFrom(string $key, $value) {
    return exception(self::unexpectedOptionErrorMessage($key, $value));
  }

  /** @return \Exception */
  private function unexpectedOptionError() {
    return self::unexpectedOptionErrorFrom($this->key, $this->value);
  }

  /** @return void */
  private function handleUnexpectedValue() {
    if ($this->posted) throw $this->unexpectedOptionError();
  }

  /** @return string */
  public function stringValueWithDefault(string $default) {
    /** @var mixed */
    $value = $this->valueWithDefault($default);
    if ($value === null) return $default;
    if ($value === true) return "true";
    if ($value === false) return "false";
    if (is_string($value) || is_int($value) || is_float($value)) return "$value";
    $this->handleUnexpectedValue();
    return $default;
  }

  /** @return string */
  public static function saneText(string $text) {
    return sanitize_text_field($text);
  }

  /** @return string */
  public static function saneHtml(string $html) {
    return trim(removeHtmlTags(['script'], $html));
  }

  /** @return string */
  public function textValueWithDefault(string $default) {
    $sanitized = self::saneText($this->stringValueWithDefault($default));
    return strlen($sanitized) > 0 ? $sanitized : self::saneText($default);
  }

  /** @return string */
  public function htmlValueWithDefault(string $default) {
    $sanitized = self::saneHtml($this->stringValueWithDefault($default));
    return strlen($sanitized) > 0 ? $sanitized : self::saneHtml($default);
  }

  /**
   * @param mixed $value
   * @psalm-assert-if-true bool|string|int|float $value
   * @return bool
   */
  public static function isValidBooleanLike($value) {
    return is_bool($value) || is_string($value) || is_int($value) || is_float($value);
  }

  /**
   * @param mixed $value
   * @return bool
   */
  private static function isTruthyBooleanLike($value) {
    if (!self::isValidBooleanLike($value)) return false;
    if ($value === true) return true;
    $str = strtolower(trim("$value"));
    return $str === "1" || $str === "true" || $str === "yes" || $str === "on";
  }

  /**
   * @param mixed $value
   * @return bool
   */
  private static function isFalsyBooleanLike($value) {
    if (!self::isValidBooleanLike($value)) return false;
    if ($value === false) return true;
    $str = strtolower(trim("$value"));
    return $str === "" || $str === "0" || $str === "false" || $str === "no" || $str === "off";
  }

  /**
   * @param mixed $value
   * @return bool|null
   */
  public static function saneBooleanLike($value) {
    if (self::isTruthyBooleanLike($value)) return true;
    if (self::isFalsyBooleanLike($value)) return false;
    return null;
  }

  /** @return bool */
  public function booleanValueWithDefault(bool $default) {
    $sane = self::saneBooleanLike($this->valueWithDefault($default));
    if (is_bool($sane)) return $sane;
    $this->handleUnexpectedValue();
    return $default;
  }

  /**
   * @param mixed $value
   * @return int|null
   */
  public static function saneIntegerLike($value) {
    if (is_int($value)) return $value;
    if (!is_string($value)) return null;
    $trim = trim($value);
    $parsedInt = intval($trim);
    return "$parsedInt" === $trim ? $parsedInt : null;
  }

  /** @return int */
  public function integerValueWithDefault(int $default) {
    /** @var mixed */
    $valueWithDefault = $this->valueWithDefault($default);
    if (is_string($valueWithDefault) && trim($valueWithDefault) === "") return $default;
    $sane = self::saneIntegerLike($valueWithDefault);
    if (is_int($sane)) return $sane;
    $this->handleUnexpectedValue();
    return $default;
  }

  /** @return bool|null */
  public function booleanValue() {
    /** @var mixed */
    $value = $this->valueWithDefault(null);
    if ($value === null) return null;
    $sane = self::saneBooleanLike($value);
    if (is_bool($sane)) return $sane;
    $this->handleUnexpectedValue();
    return null;
  }

  /** @return integer|null */
  public function integerValue() {
    /** @var mixed */
    $value = $this->valueWithDefault(null);
    if ($value === null) return null;
    $sane = self::saneIntegerLike($value);
    if (is_int($sane)) return $sane;
    $this->handleUnexpectedValue();
    return null;
  }

  /**
   * @template T
   * @param T $value
   * @return T
   */
  private function save($value) {
    $key = $this->key;
    $value === null ? delete_option($key) : update_option($key, $value);
    return $value;
  }

  /**
   * @template T
   * @param T $value
   * @return T|null
   */
  private function savePosted($value) {
    Auth::verifyAdminCapability();
    FormHandling::verifyNonce(
      $this->post,
      self::savePostedValueNonceName($this->key)
    );
    return $this->posted ? $this->save($value) : null;
  }

  /** @return string|null */
  private function savePostedStringValue(?string $value) {
    return $this->savePosted($value);
  }

  /** @return bool|null */
  private function savePostedBooleanValue(?bool $value) {
    return $this->savePosted($value);
  }

  /** @return int|null */
  private function savePostedIntegerValue(?int $value) {
    return $this->savePosted($value);
  }

  /** @return string|null */
  private function savePostedNonEmptyString(string $value) {
    return $this->savePostedStringValue($value === "" ? null : $value);
  }

  /** @return string|null */
  public function savePostedText() {
    return $this->savePostedNonEmptyString($this->textValueWithDefault(''));
  }

  /** @return string|null */
  public function savePostedHtml() {
    return $this->savePostedNonEmptyString($this->htmlValueWithDefault(''));
  }

  /** @return bool|null */
  public function savePostedBoolean() {
    return $this->savePostedBooleanValue($this->booleanValue());
  }

  /** @return int|null */
  public function savePostedInteger() {
    return $this->savePostedIntegerValue($this->integerValue());
  }
}

/**
 * Get value from a $_POST key,
 * or from the wp options if the $_POST key doesn't exist,
 * or the default value if the wp option is not defined.
 *
 * @param mixed $default
 * @return mixed
 */
function configurationValue(string $key, $default) {
  return ConfigurationOption::from($key)->valueWithDefault($default);
}

/** @return string */
function textConfigurationValue(string $key, string $default) {
  return ConfigurationOption::from($key)->textValueWithDefault($default);
}

/** @return string */
function htmlConfigurationValue(string $key, string $default) {
  return ConfigurationOption::from($key)->htmlValueWithDefault($default);
}

/** @return bool */
function booleanConfigurationValue(string $key, bool $default) {
  return ConfigurationOption::from($key)->booleanValueWithDefault($default);
}

/** @return int */
function integerConfigurationValue(string $key, int $default) {
  return ConfigurationOption::from($key)->integerValueWithDefault($default);
}

/** @return void */
function saveTextConfigurationValue(string $key) {
  ConfigurationOption::from($key)->savePostedText();
}

/** @return void */
function saveHtmlConfigurationValue(string $key) {
  ConfigurationOption::from($key)->savePostedHtml();
}

/** @return void */
function saveBooleanConfigurationValue(string $key) {
  ConfigurationOption::from($key)->savePostedBoolean();
}

/** @return void */
function saveIntegerConfigurationValue(string $key) {
  ConfigurationOption::from($key)->savePostedInteger();
}

class ConfigurationFieldMode extends FormRenderMode {
  public const HIDDEN = "hidden";
  public const SHOW = "show";
  public const READ = "read";
  public const SAVE = "save";

  private string $name;

  protected function __construct(
    string $name
  ) {
    parent::__construct();
    $this->name = $name;

    $this->protect(!$this->saveAllowed());
  }

  /** @return self */
  public static function hidden() {
    return new self(self::HIDDEN);
  }

  /** @return self */
  public static function show() {
    return new self(self::SHOW);
  }

  /** @return self */
  public static function read() {
    return new self(self::READ);
  }

  /** @return self */
  public static function save() {
    return new self(self::SAVE);
  }

  /** @return bool */
  private static function admin() {
    return Auth::isCurrentUserAdmin();
  }

  /** @return self */
  public static function adminShow() {
    return self::admin() ? self::show() : self::hidden();
  }

  /** @return self */
  public static function adminRead() {
    return self::admin() ? self::read() : self::hidden();
  }

  /** @return self */
  public static function adminSave() {
    return self::admin() ? self::save() : self::hidden();
  }

  /** @return self */
  public static function adminDebug() {
    return isDebugModeEnabled() ? self::adminSave() : self::hidden();
  }

  /** @return self */
  public static function adminReadElseShow() {
    return self::admin() ? self::read() : self::show();
  }

  /** @return self */
  public static function adminSaveElseRead() {
    return self::admin() ? self::save() : self::read();
  }

  /** @return self */
  public static function adminSaveElseShow() {
    return self::admin() ? self::save() : self::show();
  }

  /** @return bool */
  private function is(string $modeName) {
    return $this->name === $modeName;
  }

  /** @return bool */
  public function isShow() {
    return $this->is(self::SHOW);
  }

  /** @return bool */
  public function isRead() {
    return $this->is(self::READ);
  }

  /** @return bool */
  public function isSave() {
    return $this->is(self::SAVE);
  }

  /** @return bool */
  public function showAllowed() {
    return $this->isShow() || $this->readAllowed();
  }

  /** @return bool */
  public function readAllowed() {
    return $this->isRead() || $this->saveAllowed();
  }

  /** @return bool */
  public function saveAllowed() {
    return $this->isSave();
  }
}

/** @return string */
function renderConfiguration(
  string $key,
  ConfigurationFieldMode $mode,
  string $inputField
) {
  if (!$mode->showAllowed()) return "";

  $nonceKey = null;
  if ($mode->saveAllowed()) $nonceKey = ConfigurationOption::savePostedValueNonceName($key);
  elseif ($mode->readAllowed()) $nonceKey = ConfigurationOption::readPostedValueNonceName($key);

  $nonceField = $nonceKey === null ? "" : FormHandling::renderNonceField($nonceKey);
  return $nonceField . $inputField;
}

/** @return string */
function renderConfigurationTextField(
  string $label,
  string $key,
  string $value,
  ConfigurationFieldMode $mode
) {
  return renderConfiguration(
    $key,
    $mode,
    renderTextFormField($label, $key, $value, $mode)
  );
}

/** @return Render */
function configurationTextFieldRender(
  string $label,
  string $key,
  string $value,
  ConfigurationFieldMode $mode
) {
  return Render::fromCallable(function () use ($label, $mode, $key, $value) {
    return renderConfigurationTextField($label, $key, $value, $mode);
  });
}

/** @return string */
function renderConfigurationSelect(
  string $label,
  string $key,
  string $value,
  ConfigurationFieldMode $mode,
  ?SelectOption $nothingSelectedOption,
  string $missingOptionsNotice,
  SelectOption ...$options
) {
  return renderConfiguration(
    $key,
    $mode,
    renderSelectFormField(
      $label,
      $key,
      $value,
      $mode,
      $nothingSelectedOption,
      $missingOptionsNotice,
      ...$options
    )
  );
}

/** @return Render */
function configurationSelectRender(
  string $label,
  string $key,
  string $value,
  ConfigurationFieldMode $mode,
  ?SelectOption $nothingSelectedOption,
  string $missingOptionsNotice,
  SelectOption ...$options
) {
  return Render::fromCallable(function () use (
    $label,
    $mode,
    $nothingSelectedOption,
    $key,
    $value,
    $missingOptionsNotice,
    $options
  ) {
    return renderConfigurationSelect(
      $label,
      $key,
      $value,
      $mode,
      $nothingSelectedOption,
      $missingOptionsNotice,
      ...$options
    );
  });
}

/** @return string */
function renderConfigurationTextArea(
  string $label,
  string $key,
  string $value,
  ConfigurationFieldMode $mode,
  string $notice = ""
) {
  return renderConfiguration(
    $key,
    $mode,
    renderTextAreaFormField(
      $label,
      $key,
      $value,
      $mode,
      $notice
    )
  );
}

/** @return Render */
function configurationTextAreaRender(
  string $label,
  string $key,
  string $value,
  ConfigurationFieldMode $mode,
  string $notice = ""
) {
  return Render::fromCallable(function () use ($label, $mode, $key, $value, $notice) {
    return renderConfigurationTextArea($label, $key, $value, $mode, $notice);
  });
}

/** @return string */
function renderConfigurationCheckbox(
  string $label,
  string $key,
  bool $value,
  ConfigurationFieldMode $mode
) {
  return renderConfiguration(
    $key,
    $mode,
    renderCheckboxFormField(
      $label,
      $key,
      $value,
      $mode
    )
  );
}


/** @return string */
function renderConfigurationCheckboxWithConditionalBlock(
  string $label,
  string $key,
  bool $value,
  ConfigurationFieldMode $mode,
  Render $conditionalBlock
) {
  return renderConfiguration(
    $key,
    $mode,
    renderCheckboxFormFieldWithConditionalBlock(
      $label,
      $key,
      $value,
      $mode,
      $conditionalBlock
    )
  );
}

/** @return Render */
function configurationCheckboxRender(
  string $label,
  string $key,
  bool $value,
  ConfigurationFieldMode $mode
) {
  return Render::fromCallable(function () use ($label, $mode, $key, $value) {
    return renderConfigurationCheckbox($label, $key, $value, $mode);
  });
}

/** @return Render */
function configurationCheckboxRenderWithConditionalBlock(
  string $label,
  string $key,
  bool $value,
  ConfigurationFieldMode $mode,
  Render $conditionalBlock
) {
  return Render::fromCallable(function () use ($label, $mode, $key, $value, $conditionalBlock) {
    return renderConfigurationCheckboxWithConditionalBlock(
      $label,
      $key,
      $value,
      $mode,
      $conditionalBlock
    );
  });
}
