<?php

namespace Wino;

/** @return void */
function includeAjaxScript() {
  include __DIR__ . '/wp.ajax.js.php';
}

/** @return string */
function ajaxFunctionCall(string $functionName, string $params) {
  return Constant::vendorPrefixed("Ajax$functionName($params);");
}

/** @return string */
function ajaxOnChangeFunctionCall(string $params) {
  return ajaxFunctionCall('OnChange', $params);
}

/** @return string */
function ajaxCallActionOnClickFunctionCall(string $params) {
  return ajaxFunctionCall('CallActionOnClick', $params);
}

/** @return string */
function ajaxActionClass(string $action = "") {
  return Constant::vendorPrefixedSlug("action-$action");
}

/** @return string */
function ajaxActionFormClass(string $action = "") {
  return Constant::vendorPrefixedSlug("action-form-$action");
}

/** @return string */
function renderAjaxStartUpAction(string $action) {
  $nonce = FormHandling::renderNonceField($action);
  $startUpDiv = renderHtmlTag("div", [
    "class" => Constant::vendorPrefixedSlug("start-up-action"),
    "style" => "display:none;"
  ], $action);
  return $nonce . $startUpDiv;
}

/** @return string */
function renderAjaxAction(string $action, bool $callOnStartUp = false) {
  return $callOnStartUp ? renderAjaxStartUpAction($action) : FormHandling::renderNonceField($action);
}

/** @return Render */
function ajaxActionRender(string $action, bool $callOnStartUp = false) {
  return Render::fromCallable(function () use ($action, $callOnStartUp) {
    return renderAjaxAction($action, $callOnStartUp);
  });
}

/** @return string */
function renderAjaxButton(
  string $label,
  string $action,
  ?FormRenderMode $mode = null
) {
  $nonce = FormHandling::renderNonceField($action);
  $definedMode = $mode !== null ? $mode : FormRenderMode::create();
  $button = renderHtmlTag('button', $definedMode->buttonAttributes($action), $label);
  return $nonce . $button;
}

/** @return Render */
function ajaxButtonRender(
  string $label,
  string $action,
  ?FormRenderMode $mode = null
) {
  return Render::fromCallable(function () use ($label, $action, $mode) {
    return renderAjaxButton($label, $action, $mode);
  });
}

class AjaxErrorHandler {
  public const THROW_ERROR_HANDLER = "throw";
  public const NOTICE_ERROR_HANDLER = "notice";
  public const CONSOLE_ERROR_HANDLER = "console";

  private ?string $name;

  /** @return string|null */
  public function getName() {
    return $this->name;
  }

  private function __construct(?string $name) {
    $this->name = $name;
  }

  /** @return self */
  public static function inherit() {
    return new self(null);
  }

  /** @return self */
  public static function throw() {
    return new self(self::THROW_ERROR_HANDLER);
  }

  /** @return self */
  public static function notice() {
    return new self(self::NOTICE_ERROR_HANDLER);
  }

  /** @return self */
  public static function console() {
    return new self(self::CONSOLE_ERROR_HANDLER);
  }
}

/**
 * @template T
 * @param Closure(): T $callback
 * @return void
 */
function addAjaxAction(string $action, AjaxErrorHandler $errorHandler, callable $callback) {
  add_action('wp_ajax_' . $action, function () use ($callback, $errorHandler) {
    try {
      $response = call_user_func($callback);
      if ($response instanceof AjaxResponse) $response = $response->data();
      echo json_encode($response);
    } catch (\Throwable $thrown) {
      debugLog($thrown);
      echo json_encode([ajaxErrorData($thrown, $errorHandler)]);
    }
    wp_die();
  });
}

/**
 * @template T
 * @param T $type
 * @return array{msg: string, type: T}
 */
function ajaxData(string $type, string $message) {
  return ['type' => $type, 'msg' => $message];
}

class AjaxError {
  public const TRACE_PREFIX = ' in ';

  /**
   * @param mixed $error
   * @return string
   */
  public static function errorMessageFrom($error) {
    return isThrowable($error) ? $error->getMessage() : "$error";
  }

  /**
   * @param mixed $error
   * @return string
   */
  public static function ajaxTraceFrom($error) {
    return isThrowable($error) ? self::TRACE_PREFIX . $error->getTraceAsString() : '';
  }

  /**
   * @param mixed $error
   * @return string|int|null
   */
  public static function ajaxCodeFrom($error) {
    return isThrowable($error) ? $error->getCode() : null;
  }
}

/**
 * @param mixed $error
 * @return array{msg: string, type: 'error', errorHandlerName?: string, errorCode?: string|int}
 */
function ajaxErrorData($error, AjaxErrorHandler $errorHandler) {
  $msg = AjaxError::errorMessageFrom($error);
  $trace = AjaxError::ajaxTraceFrom($error);
  $code = AjaxError::ajaxCodeFrom($error);

  $data = ajaxData('error', $msg . $trace);
  $errorHandlerName = $errorHandler->getName();
  if ($errorHandlerName !== null) $data['errorHandlerName'] = $errorHandlerName;
  if ($code !== null) $data['errorCode'] = $code;
  return $data;
}

/**
 * @param mixed $message
 * @return array{msg: string, type: 'info'}
 */
function ajaxInfo($message) {
  return ajaxData('info', "$message");
}

/** @return array{msg: string, type: 'updated'} */
function ajaxUpdate(string $message) {
  return ajaxData('updated', $message);
}

/** @return array{id: string, type: 'dom-replacement-by-id', html: string} */
function ajaxDomReplacementById(string $id, string $html) {
  return ['type' => 'dom-replacement-by-id', 'id' => $id, 'html' => $html];
}

/** @return array{type: 'enable-action', action: string, label: null|string} */
function ajaxEnableAction(string $action, ?string $label) {
  return ['type' => 'enable-action', 'action' => $action, 'label' => $label];
}

/** @return array{type: 'disable-action', action: string, label: null|string} */
function ajaxDisableAction(string $action, ?string $label) {
  return ['type' => 'disable-action', 'action' => $action, 'label' => $label];
}

/** @return array{type: 'call-action', action: string, errorHandlerName?: string} */
function ajaxCallAction(string $action, AjaxErrorHandler $errorHandler) {
  $data = ['type' => 'call-action', 'action' => $action];
  $errorHandlerName = $errorHandler->getName();
  if ($errorHandlerName !== null) $data['errorHandlerName'] = $errorHandlerName;
  return $data;
}

/**
 * @param int|string $code
 * @return array{type: 'dismiss-notice', code: int|string}
 */
function ajaxDismissNotice($code) {
  $data = ['type' => 'dismiss-notice', 'code' => $code];
  return $data;
}

class AjaxResponse implements Log {

  private AjaxErrorHandler $errorHandler;

  /** @return $this */
  public function setErrorHandler(AjaxErrorHandler $errorHandler) {
    $this->errorHandler = $errorHandler;
    return $this;
  }

  private InMemoryLog $log;

  /** @var string[] */
  private array $updates;

  /** @var array<array{html: string, id: string, type: string}> */
  private array $domReplacementsById;

  /** @var array<array{type: string, action: string, label: null|string}> */
  private array $actionToggles;

  /** @var array<array{type: string, action: string, errorHandlerName?: string}> */
  private array $actionCalls;

  /** @var array<int|string> */
  private array $noticeDismissalCodes;

  /**
   * Summary of __construct
   * @param InMemoryLog $log
   * @param string[] $updates
   * @param array<array{html: string, id: string, type: string}> $domReplacementsById
   * @param array<array{type: string, action: string, label: null|string}> $actionToggles
   * @param array<array{type: string, action: string, errorHandlerName?: string}> $actionCalls
   * @param array<int|string> $noticeDismissalCodes
   */
  private function __construct(
    InMemoryLog $log,
    array &$updates,
    array &$domReplacementsById,
    array &$actionToggles,
    array &$actionCalls,
    array &$noticeDismissalCodes
  ) {
    $this->log = $log;
    $this->errorHandler = AjaxErrorHandler::inherit();
    $this->updates = $updates;
    $this->domReplacementsById = $domReplacementsById;
    $this->actionToggles = $actionToggles;
    $this->actionCalls = $actionCalls;
    $this->noticeDismissalCodes = $noticeDismissalCodes;
  }

  /** @return self */
  public static function from(InMemoryLog $log) {
    $updates = [];
    $domReplacementsById = [];
    $actionToggles = [];
    $actionCalls = [];
    $noticeDismissalCodes = [];
    return new self($log, $updates, $domReplacementsById, $actionToggles, $actionCalls, $noticeDismissalCodes);
  }

  /** @return mixed[] */
  public function data() {
    $data = [];
    $errors = $this->log->getErrors();
    $infos = $this->log->getInfos();
    $updates = $this->updates;
    $domReplacementsById = $this->domReplacementsById;
    $actionToggles = $this->actionToggles;
    $actionCalls = $this->actionCalls;
    $noticeDismissalCodes = $this->noticeDismissalCodes;

    foreach ($noticeDismissalCodes as $dismissalCode) $data[] = ajaxDismissNotice($dismissalCode);
    /** @psalm-suppress MixedAssignment */
    foreach ($errors as $error) $data[] = ajaxErrorData($error, $this->errorHandler);
    /** @psalm-suppress MixedAssignment */
    foreach ($infos as $info) $data[] = ajaxInfo($info);
    foreach ($updates as $update) $data[] = ajaxUpdate($update);
    foreach ($domReplacementsById as $domReplacementById) $data[] = $domReplacementById;
    foreach ($actionToggles as $actionToggle) $data[] = $actionToggle;
    foreach ($actionCalls as $actionCall) $data[] = $actionCall;

    return $data;
  }

  /**
   * @param mixed $value
   * @return $this
   */
  public function info($value) {
    $this->log->info($value);
    return $this;
  }

  /**
   * @param mixed $value
   * @return $this
   */
  public function error($value) {
    $this->log->error($value);
    return $this;
  }

  /** @return void */
  public function update(string $value) {
    $this->updates[] = $value;
  }

  /** @return void */
  public function replaceDomById(string $id, string $html) {
    $this->domReplacementsById[] = ajaxDomReplacementById($id, $html);
  }

  /** @return void */
  public function enableAction(string $action, ?string $label) {
    $this->actionToggles[] = ajaxEnableAction($action, $label);
  }

  /** @return void */
  public function disableAction(string $action, ?string $label) {
    $this->actionToggles[] = ajaxDisableAction($action, $label);
  }

  /** @return void */
  public function toggleAction(bool $enabled, string $action, ?string $label) {
    if ($enabled) $this->enableAction($action, $label);
    else $this->disableAction($action, $label);
  }

  /** @return void */
  public function callAction(string $action, AjaxErrorHandler $errorHandler) {
    $this->actionCalls[] = ajaxCallAction($action, $errorHandler);
  }

  /**
   * @param int|string $code
   * @return void
   */
  public function dismissNotice($code) {
    $this->noticeDismissalCodes[] = $code;
  }
}
