<?php

namespace Wino;

interface FormHandlingTranslator {
  /** @return string */
  public function operationNotPermittedMissingNonceErrorMessage(string $action);

  /** @return string */
  public function operationNotPermittedTooOldNonceErrorMessage(string $action);

  /** @return string */
  public function operationNotPermittedInvalidNonceErrorMessage(string $action);

  /** @return string */
  public function operationNotPermittedUnexpectedNonceVerificationResulrErrorMessage(string $action);

  /** @return string */
  public function unexpectedPluginVersionErrorMessage(
    string $plugin,
    string $postVersion,
    string $expectedVersion
  );

  /** @return string */
  public function unexpectedPluginVersionFromUnknownErrorMessage(
    string $plugin,
    string $expectedVersion
  );
}

class DefaultFormHandlingTranslator implements FormHandlingTranslator {
  /** @return string */
  public static function tryToRefresh() {
    return "Try to refresh the page";
  }

  /** @return string */
  public static function operationNotPermitted(string $action, string $reason) {
    $tryRefresh = self::tryToRefresh();
    return "Operation not permitted ($action): $reason. $tryRefresh.";
  }

  public function operationNotPermittedMissingNonceErrorMessage(string $action) {
    return self::operationNotPermitted($action, "missing nonce");
  }

  public function operationNotPermittedTooOldNonceErrorMessage(string $action) {
    return self::operationNotPermitted($action, "too old nonce");
  }

  public function operationNotPermittedInvalidNonceErrorMessage(string $action) {
    return self::operationNotPermitted($action, "invalid nonce");
  }

  public function operationNotPermittedUnexpectedNonceVerificationResulrErrorMessage(string $action) {
    return self::operationNotPermitted($action, "unexpected nonce verification result");
  }

  /** @return string */
  public static function itSeemsThatThePluginVersionWasUpdated(
    string $plugin,
    string $updateDetails
  ) {
    $pluginName = jsonString($plugin);
    $tryRefresh = self::tryToRefresh();
    return msg(
      "It seems that the plugin $pluginName version was updated",
      "$updateDetails. $tryRefresh."
    );
  }

  public function unexpectedPluginVersionErrorMessage(
    string $plugin,
    string $postVersion,
    string $expectedVersion
  ) {
    $oldVersion = jsonString($postVersion);
    $newVersion = jsonString($expectedVersion);
    return self::itSeemsThatThePluginVersionWasUpdated($plugin, "from $oldVersion to $newVersion");
  }

  public function unexpectedPluginVersionFromUnknownErrorMessage(
    string $plugin,
    string $expectedVersion
  ) {
    $newVersion = jsonString($expectedVersion);
    return self::itSeemsThatThePluginVersionWasUpdated($plugin, "to $newVersion");
  }
}

class FrFormHandlingTranslator implements FormHandlingTranslator {
  /** @return string */
  public static function tryToRefresh() {
    return "Essayez de rafraîchir la page";
  }

  /** @return string */
  public static function nonce() {
    return "jeton de securité";
  }

  /** @return string */
  public static function operationNotPermitted(string $action, string $reason) {
    $tryRefresh = self::tryToRefresh();
    return "Opération non autorisée ($action): $reason. $tryRefresh.";
  }

  public function operationNotPermittedMissingNonceErrorMessage(string $action) {
    $nonce = self::nonce();
    return self::operationNotPermitted($action, "$nonce manquant");
  }

  public function operationNotPermittedTooOldNonceErrorMessage(string $action) {
    $nonce = self::nonce();
    return self::operationNotPermitted($action, "$nonce expiré");
  }

  public function operationNotPermittedInvalidNonceErrorMessage(string $action) {
    $nonce = self::nonce();
    return self::operationNotPermitted($action, "$nonce invalide");
  }

  public function operationNotPermittedUnexpectedNonceVerificationResulrErrorMessage(string $action) {
    $nonce = self::nonce();
    return self::operationNotPermitted($action, "Résultat inattendu lors de la vérification du $nonce");
  }

  /** @return string */
  public static function itSeemsThatThePluginVersionWasUpdated(
    string $plugin,
    string $updateDetails
  ) {
    $pluginName = jsonString($plugin);
    $tryRefresh = self::tryToRefresh();
    return msg(
      "Il semble que le plugin $pluginName a été mis à jour.",
      "Il est passé",
      "$updateDetails. $tryRefresh."
    );
  }

  public function unexpectedPluginVersionErrorMessage(
    string $plugin,
    string $postVersion,
    string $expectedVersion
  ) {
    $oldVersion = jsonString($postVersion);
    $newVersion = jsonString($expectedVersion);
    return self::itSeemsThatThePluginVersionWasUpdated(
      $plugin, "de la version $oldVersion à la version $newVersion"
    );
  }

  public function unexpectedPluginVersionFromUnknownErrorMessage(
    string $plugin,
    string $expectedVersion
  ) {
    $newVersion = jsonString($expectedVersion);
    return self::itSeemsThatThePluginVersionWasUpdated($plugin, "à la version $newVersion");
  }
}

class LocaleFormHandlingTranslator {
  /**
   * @param FormHandlingTranslator|string|null $locale
   *
   * @return FormHandlingTranslator
   */
  public static function from($locale) {
    $default = new DefaultFormHandlingTranslator();
    if ($locale === null) return $default;
    if (!is_string($locale)) return $locale;
    switch (strtolower(trim($locale))) {
      case 'fr':
      return new FrFormHandlingTranslator();

      default:
      return $default;
    }
  }

  /**
   * @return FormHandlingTranslator
   */
  public static function fromDefault() {
    return self::from(null);
  }
}

class FormHandling {
  private static bool $formRenderDone = false;

  /** @var string[] */
  private static array $inputNames = [];

  public static bool $formDisabled = false;
  public static bool $formTable = false;
  public static bool $fieldsGroup = false;

  public static int $nonceCount = 0;

  private const NONCE_PREFIX = 'nonce-';

  /** @return void */
  public static function endsFormRendering() {
    self::$formRenderDone = true;
  }

  /** @return void */
  public static function startsFormRendering() {
    self::$formRenderDone = false;
    self::$inputNames = [];
    self::$nonceCount = 0;
  }

  /** @return void */
  public static function collectInputName(string $name) {
    if (self::$formRenderDone) self::startsFormRendering();
    self::$inputNames[] = $name;
  }

  /** @return string[] */
  public static function getCurrentFormInputNames() {
    if (self::$formRenderDone) return self::$inputNames;
    throw error("Form not correctly rendered with the render form function.");
  }

  /** @return string */
  public static function nonceName(string $key) {
    $prefix = self::NONCE_PREFIX;
    $prefixed = trim(strStartsWith($key, $prefix) ? $key : $prefix . $key);
    if ($prefixed !== $prefix) return $prefixed;
    throw exception("Unexpected empty nonce field names");
  }

  /** @return string */
  public static function nonceAction(string $key) {
    $prefix = self::NONCE_PREFIX;
    $trim = trim($key);
    return strStartsWith($trim, $prefix) ? trim(substr($trim, strlen($prefix))) : $trim;
  }

  /** @return string */
  public static function nonce(string $key) {
    return wp_create_nonce(self::nonceAction($key));
  }

  /** @return string */
  public static function renderNonceField(string $key) {
    $referer = self::$nonceCount === 0;
    self::$nonceCount++;
    $echo = false;
    $name = self::nonceName($key);
    self::collectInputName($name);
    return wp_nonce_field(self::nonceAction($key), $name, $referer, $echo);
  }

  /** @return string|null */
  public static function nonceValue(array $post, string $key) {
    return optionalStringData($post, self::nonceName($key));
  }

  /** @return false|int */
  private static function wpVerifyNonce(string $key, string $nonce) {
    return wp_verify_nonce(wp_unslash($nonce), self::nonceAction($key));
  }

  private const VALID_NONCE = 1;
  private const TOO_OLD_NONCE = 2;
  private const INVALID_NONCE = false;

  /** @return bool */
  private static function validNonce(string $key, string $nonce) {
    return self::wpVerifyNonce($key, $nonce) === self::VALID_NONCE;
  }

  /** @return bool */
  public static function isNonceValid(array $post, string $key) {
    $nonce = self::nonceValue($post, $key);
    return isset($nonce) ? self::validNonce($key, $nonce) : false;
  }

  /**
   * @param FormHandlingTranslator|string|null $locale
   *
   * @return void
   */
  public static function verifyNonce(
    array $post,
    string $key,
    $locale = null
  ) {
    $translator = LocaleFormHandlingTranslator::from($locale);
    $action = self::nonceAction($key);
    $nonce = self::nonceValue($post, $key);
    if (!isset($nonce)) throw exception(
      $translator->operationNotPermittedMissingNonceErrorMessage($action)
    );

    $nonceValidation = self::wpVerifyNonce($key, $nonce);
    if ($nonceValidation === self::VALID_NONCE) return;
    if ($nonceValidation === self::TOO_OLD_NONCE) throw exception(
      $translator->operationNotPermittedTooOldNonceErrorMessage($action)
    );
    if ($nonceValidation === self::INVALID_NONCE) throw exception(
      $translator->operationNotPermittedInvalidNonceErrorMessage($action)
    );
    throw exception(
      $translator->operationNotPermittedUnexpectedNonceVerificationResulrErrorMessage($action)
    );
  }

  /**
   * @param FormHandlingTranslator|string|null $locale
   *
   * @return void
   */
  public static function verifyPostNonce(string $key, $locale = null) {
    self::verifyNonce($_POST, $key, $locale);
  }

  /** @return string */
  public static function stringifyInputNamesArray() {
    return jsonString(self::getCurrentFormInputNames());
  }

  /** @return string */
  public static function pluginVersionCheckKey(string $plugin = 'Wino') {
    return "--$plugin-plugin-version-check";
  }

  /** @return string|null */
  public static function pluginVersionCheckValue(string $plugin = 'Wino') {
    return wordpressPluginVersion($plugin);
  }

  /**
   * @param FormHandlingTranslator|string|null $locale
   *
   * @return void
   */
  public static function verifyPluginVersion(
    $locale = null,
    string $plugin = 'Wino'
  ) {
    $translator = LocaleFormHandlingTranslator::from($locale);
    $currentVersionKey = self::pluginVersionCheckKey($plugin);
    $expectedVersion = self::pluginVersionCheckValue($plugin);

    if ($expectedVersion === null) return;
    $postValue = optionalData($_POST, $currentVersionKey);
    if (!is_string($postValue)) throw exception(
      $translator->unexpectedPluginVersionFromUnknownErrorMessage($plugin, $expectedVersion)
    );
    if ($postValue !== $expectedVersion) throw exception(
      $translator->unexpectedPluginVersionErrorMessage($plugin, $postValue, $expectedVersion)
    );
  }
}

class FormRenderMode {
  private bool $isProtected;

  /** @return bool */
  public function protected() {
    return $this->isProtected;
  }

  /** @return $this */
  public function protect(bool $protected = true) {
    $this->isProtected = $protected;
    return $this;
  }

  private bool $isDisablingPrevented;

  /** @return $this */
  public function preventDisabling(bool $disablingPrevented = true) {
    $this->isDisablingPrevented = $disablingPrevented;
    return $this;
  }

  private bool $isDisabled;

  /** @return bool */
  public function disabled() {
    $formDisabled = FormHandling::$formDisabled;
    return !$this->isDisablingPrevented && ($this->isDisabled || $this->protected() || $formDisabled);
  }

  /** @return bool */
  public function enabled() {
    return !$this->disabled();
  }

  /** @return $this */
  public function disable(bool $disabled = true) {
    $this->isDisabled = $disabled;
    return $this;
  }

  private bool $ajax;

  /** @return $this */
  public function disableAjax(bool $disabled = true) {
    $this->ajax = !$disabled;
    return $this;
  }

  private bool $disableAfterFirstUsage;

  /** @return $this */
  public function singleUse(bool $singleUse = true) {
    $this->disableAfterFirstUsage = $singleUse;
    return $this;
  }

  private string $customClasses;

  /** @return $this */
  public function addClass(string $class) {
    $classes = $this->customClasses . " $class";
    $this->customClasses = trim($classes);
    return $this;
  }

  private bool $isPrimary;

  /** @return $this */
  public function primary(bool $primary = true) {
    $this->isPrimary = $primary;
    return $this;
  }

  private ?string $confirmationMessage;

  /** @return $this */
  public function confirm(string $message) {
    $this->confirmationMessage = $message;
    return $this;
  }

  protected function __construct() {
    $this->isProtected = false;
    $this->isDisabled = false;
    $this->isDisablingPrevented = false;
    $this->disableAfterFirstUsage = false;
    $this->confirmationMessage = null;
    $this->customClasses = "";
    $this->isPrimary = false;
    $this->ajax = true;
  }

  /** @return self */
  public static function create() {
    return new self();
  }

  /** @return string */
  private function classAttribute(
    ?string $tagName,
    ?string $action
  ) {
    $classes = [];
    if ($tagName !== null) {
      $classes[] = $tagName;
      if ($this->isPrimary) $classes[] = " $tagName-primary";
    }
    if ($action !== null) {
      $classes[] = ajaxActionClass();
      $classes[] = ajaxActionClass($action);
    }
    if ($this->disabled()) $classes[] = 'disabled';
    if ($this->disableAfterFirstUsage) $classes[] = 'disable-after-first-usage';
    if ($this->isDisablingPrevented) $classes[] = 'disabling-prevented';
    if ($this->protected()) $classes[] = 'protected';

    return trim(implode(" ", $classes) . " " . $this->customClasses);
  }

  /** @return void */
  public function defineFormHandlingProperties() {
    FormHandling::$formDisabled = $this->disabled();
  }

  /**
   * @param array<string, string|bool|int|float|null> $attributes
   * @return array<string, string|bool|int|float|null>
   */
  public function formAttributes(string $action, $attributes = []) {
    $attributes['method'] = 'post';
    $attributes['action'] = $action;
    $attributes['disabled'] = $this->disabled();
    $attributes['class'] = $this->classAttribute('form', $action);
    return $attributes;
  }

  /** @return string|null */
  private function buttonOnClickAttribute(?string $action) {
    if ($action === null) return null;
    $call = ajaxCallActionOnClickFunctionCall(jsonString($action));
    $confirm = $this->confirmationMessage;
    if ($confirm === null) return $call;
    $message = jsString($confirm);
    return "if (confirm($message)) $call";
  }

  /**
   * @param array<string, string|bool|int|float|null> $attributes
   * @return array<string, string|bool|int|float|null>
   */
  public function buttonAttributes(?string $action = null, $attributes = []) {
    $attributes['type'] = 'button';
    $attributes['onclick'] = $this->ajax ? $this->buttonOnClickAttribute($action) : null;
    $attributes['disabled'] = $this->disabled();
    $attributes['class'] = $this->classAttribute('button', $action);

    return $attributes;
  }

  /**
   * @param array<string, string|bool|int|float|null> $attributes
   * @return array<string, string|bool|int|float|null>
   */
  public function inputAttributes(string $key, $attributes = []) {
    $attributes['id'] = $key;
    $attributes['name'] = $key;
    $attributes['onchange'] = $this->ajax ? ajaxOnChangeFunctionCall(jsonString($key)) : null;
    $attributes['disabled'] = $this->disabled();
    $attributes['class'] = $this->classAttribute(null, null);
    return $attributes;
  }
}

/** @return string */
function renderForm(
  string $action,
  string $content,
  FormRenderMode $mode
) {
  $nonce = FormHandling::renderNonceField($action);
  FormHandling::endsFormRendering();
  return renderHtmlTag('form', $mode->formAttributes($action), $nonce . $content);
}

/** @return Render */
function formRender(
  string $action,
  Render $content,
  FormRenderMode $mode
) {
  return Render::fromCallable(function () use ($action, $mode, $content) {
    $mode->defineFormHandlingProperties();
    return renderForm($action, Render::renderContent($content), $mode);
  });
}

/** @return string */
function renderFormTableStart() {
  FormHandling::$formTable = true;
  return '<table class="form-table" role="presentation"><tbody>';
}

/** @return string */
function renderFormTableEnd() {
  FormHandling::$formTable = false;
  return '</tbody></table>';
}

/** @return Render */
function formTableRender(?Render ...$content) {
  return Render::fromCallable(function () use ($content) {
    $start = renderFormTableStart();
    $body = Render::concat(...$content)->content();
    $end = renderFormTableEnd();
    return $start . $body . $end;
  });
}

/** @return string */
function renderFormFieldsGroupStart() {
  if (!FormHandling::$formTable) return "";
  FormHandling::$formTable = false;
  FormHandling::$fieldsGroup = true;
  return '<tr><td style="padding-top:0;" colspan="2">';
}

/** @return string */
function renderFormFieldsGroupStartWithLabel(string $label) {
  if (!FormHandling::$formTable) return $label === "" ? "" : '<div><label>' . $label . '</label></div>';
  FormHandling::$formTable = false;
  FormHandling::$fieldsGroup = true;
  return str(
    '<tr>',
    '<th colspan="2"><label>',
    $label,
    '</label></th>',
    '</tr>',
    '<tr><td style="padding-top:0;" colspan="2">'
  );
}

/** @return string */
function renderFormFieldsGroupEnd() {
  if (!FormHandling::$fieldsGroup) return "";
  FormHandling::$formTable = true;
  FormHandling::$fieldsGroup = false;
  return "</td></tr>";
}

/** @return Render */
function formFieldsGroupRenderWithLabel(string $label, ?Render ...$content) {
  return Render::fromCallable(function () use ($label, $content) {
    $start = renderFormFieldsGroupStartWithLabel($label);
    $body = Render::concat(...$content)->content();
    $end = renderFormFieldsGroupEnd();
    return $start . $body . $end;
  });
}

/** @return Render */
function formFieldsGroupRender(?Render ...$content) {
  return Render::fromCallable(function () use ($content) {
    $start = renderFormFieldsGroupStart();
    $body = Render::concat(...$content)->content();
    $end = renderFormFieldsGroupEnd();
    return $start . $body . $end;
  });
}

/** @return string */
function renderFormLabel(
  string $label,
  string $for
) {
  return '<label for="' . $for . '">' . $label . '</label>';
}

/** @return string */
function renderFormTableRow(
  string $label,
  string $for,
  string $content
) {
  return str(
    '<tr><th scope="row">',
    renderFormLabel($label, $for),
    '</th><td>' . $content . '</td></tr>'
  );
}

/** @return string */
function renderFormDiv(
  string $label,
  string $for,
  string $content
) {
  return str(
    '<div style="display:flex;gap:0.5em;align-items:center;line-height:1.32em;">',
    renderFormLabel($label, $for),
    $content,
    '</div>'
  );
}

/** @return string */
function renderFormField(
  string $label,
  string $for,
  string $content
) {
  return FormHandling::$formTable
  ? renderFormTableRow($label, $for, $content)
  : renderFormDiv($label, $for, $content);
}

/** @return string */
function renderFormSpacing() {
  return FormHandling::$formTable
  ? '<tr><td colspan="2"><br/></td></tr>'
  : '<div><br/></div>';
}

/** @return Render */
function formSpacingRender() {
  return Render::fromCallable(function () {
    return renderFormSpacing();
  });
}

/** @return string */
function renderHiddenInput(
  string $key,
  string $value
) {
  FormHandling::collectInputName($key);
  $mode = FormRenderMode::create()->protect();
  return renderHtmlSelfClosingTag('input', $mode->inputAttributes($key, [
    'type' => 'hidden',
    'value' => $value
  ]));
}

/** @return string */
function renderTextInput(
  string $key,
  string $value,
  FormRenderMode $mode
) {
  FormHandling::collectInputName($key);
  return renderHtmlSelfClosingTag('input', $mode->addClass('regular-text')->inputAttributes($key, [
    'type' => 'text',
    'value' => $value
  ]));
}

/** @return string */
function renderTextFormField(
  string $label,
  string $key,
  string $value,
  FormRenderMode $mode
) {
  return renderFormField(
    $label,
    $key,
    renderTextInput($key, $value, $mode)
  );
}

/** @return string */
function renderTextAreaInput(
  string $key,
  string $value,
  FormRenderMode $mode,
  string $notice = ""
) {
  FormHandling::collectInputName($key);
  $trimNotice = trim($notice);
  $noticeHtml = $trimNotice === "" ? "" : '<em>' . $trimNotice . '</em>';
  $textArea = renderHtmlTag('textarea', $mode->addClass('widefat')->inputAttributes($key), $value);
  return $textArea . $noticeHtml;
}

/** @return string */
function renderTextAreaFormField(
  string $label,
  string $key,
  string $value,
  FormRenderMode $mode,
  string $notice = ""
) {
  return renderFormField(
    $label,
    $key,
    renderTextAreaInput($key, $value, $mode, $notice)
  );
}

class SelectOption {
  public string $value;
  public string $label;

  /**
   * @param string|int $value
   */
  public function __construct(
    $value,
    string $label
  ) {
    $this->value = is_string($value) ? $value : "$value";
    $this->label = $label;
  }

  /** @return self */
  public static function from(string $value) {
    return new self($value, $value);
  }
}

/** @return SelectOption|null */
function singleOptionFrom(SelectOption ...$options) {
  if (count($options) !== 1) return null;
  return listElement($options, 0);
}

/** @return string */
function renderSelectInputOption(
  SelectOption $option,
  bool $selected
) {
  return str(
    '<option ',
    'value="' . esc_attr($option->value) . '" ',
    ($selected ? 'selected="selected"' : '') . '>',
    $option->label,
    '</option>'
  );
}

/** @return string */
function renderSelectInputOptions(
  ?SelectOption $nothingSelectedOption,
  string $value,
  SelectOption ...$options
) {
  $html = '';
  $nothingSelected = true;
  foreach ($options as $option) {
    $selected = $option->value === $value;
    $html .= renderSelectInputOption($option, $selected);
    $nothingSelected = $nothingSelected && !$selected;
  }
  $nothing = $nothingSelected && $nothingSelectedOption !== null;
  $nothingSelectedHtml = $nothing ? renderSelectInputOption($nothingSelectedOption, true) : "";
  return $nothingSelectedHtml . $html;
}

/** @return string */
function renderSelectInput(
  string $key,
  string $value,
  FormRenderMode $mode,
  ?SelectOption $nothingSelectedOption,
  string $missingOptionsNotice,
  SelectOption ...$options
) {
  FormHandling::collectInputName($key);

  if (count($options) < 1) return '<em id="' . $key . '">' . $missingOptionsNotice . '</em>';

  $hasSingleOption = singleOptionFrom(...$options) !== null;
  if ($hasSingleOption) $mode->protect();

  return renderHtmlTag(
    'select',
    $mode->inputAttributes($key),
    renderSelectInputOptions($nothingSelectedOption, $value, ...$options)
  );
}

/** @return string */
function renderSelectFormField(
  string $label,
  string $key,
  string $value,
  FormRenderMode $mode,
  ?SelectOption $nothingSelectedOption,
  string $missingOptionsNotice,
  SelectOption ...$options
) {
  return renderFormField(
    $label,
    $key,
    renderSelectInput(
      $key,
      $value,
      $mode,
      $nothingSelectedOption,
      $missingOptionsNotice,
      ...$options
    )
  );
}

/** @return string */
function renderCheckboxInput(
  string $key,
  bool $value,
  FormRenderMode $mode
) {
  FormHandling::collectInputName($key);
  return renderHtmlSelfClosingTag('input', $mode->inputAttributes($key, [
    'type' => 'checkbox',
    'checked' => $value ? 'checked' : null,
    'style' => 'margin: 0;'
  ]));
}

/** @return string */
function renderCheckboxInputWithConditionalBlock(
  string $key,
  bool $value,
  FormRenderMode $mode,
  Render $conditionalBlock
) {
  $input = renderCheckboxInput($key, $value, $mode);
  $initialFormTable = FormHandling::$formTable;
  FormHandling::$formTable = false;
  $block = $conditionalBlock->content();
  FormHandling::$formTable = $initialFormTable;
  return str(
    '<div style="display: flex; gap: 23pt; align-items: center;">',
    $input,
    '<div class="wino-conditional-block">',
    $block,
    '</div>',
    '</div>'
  );
}

/** @return string */
function renderCheckboxFormField(
  string $label,
  string $key,
  bool $value,
  FormRenderMode $mode
) {
  return renderFormField(
    $label,
    $key,
    renderCheckboxInput($key, $value, $mode)
  );
}

/** @return string */
function renderCheckboxFormFieldWithConditionalBlock(
  string $label,
  string $key,
  bool $value,
  FormRenderMode $mode,
  Render $conditionalBlock
) {
  return renderFormField(
    $label,
    $key,
    renderCheckboxInputWithConditionalBlock($key, $value, $mode, $conditionalBlock)
  );
}

/** @return string */
function renderRequiredFieldsNotice() {
  return '<em style="font-weight:normal;font-size: 0.82em;color:gray;">* champs requis</em>';
}
