<?php

namespace Wino;

/**
 * @param mixed $log
 * @param array $logs
 * @return void
 */
function debugLog($log, ...$logs) {
  if (isDebugModeEnabled()) prettyLog($log, ...$logs);
}

/** @return \Exception */
function wordpressErrorException(\WP_Error $error, string ...$meta) {
  $errors = $error->get_error_messages();
  return exception(implode("\n", array_merge($errors, $meta)));
}

class WordpressTransientThrowable {
  public string $asString;
  public string $getMessageValue;

  /** @var int|string */
  public $code;

  public function __construct(\Throwable $throwable) {
    $this->asString = "$throwable";
    $this->getMessageValue = $throwable->getMessage();
    $this->code = $throwable->getCode();
  }

  /** @return string */
  public function getMessage() {
    return $this->getMessageValue;
  }

  /** @return int|string */
  public function getCode() {
    return $this->code;
  }
}

class WordpressTransientLog implements Log {
  public string $key;
  public int $maxExpirationTimeInSeconds;

  /** @var array */
  private array $infoLogs;

  /** @var array */
  private array $errorLogs;

  public function __construct(
    string $key,
    int $maxExpirationTimeInSeconds
  ) {
    $this->key = $key;
    $this->maxExpirationTimeInSeconds = $maxExpirationTimeInSeconds;
    $infoKey = $this->infoLogsKey();
    $errorKey = $this->errorLogsKey();
    /** @var mixed */
    $infoTransient = get_transient($infoKey);
    /** @var mixed */
    $errorTransient = get_transient($errorKey);
    delete_transient($infoKey);
    delete_transient($errorKey);

    /** @var mixed */
    $infoLogs = $infoTransient === false ? [] : $infoTransient;
    /** @var mixed */
    $errorLogs = $errorTransient === false ? [] : $errorTransient;

    $this->infoLogs = is_array($infoLogs) ? $infoLogs : [$infoLogs];
    $this->errorLogs = is_array($errorLogs) ? $errorLogs : [$errorLogs];
  }

  /** @return string */
  private function infoLogsKey() {
    $key = $this->key;
    return "$key-transient-info-logs";
  }

  /** @return string */
  private function errorLogsKey() {
    $key = $this->key;
    return "$key-transient-error-logs";
  }

  /**
   * @template T
   * @param T $value
   *
   * @return (T is \Error|\Exception ? WordpressTransientThrowable : T)
   */
  private static function transientValue($value) {
    return isThrowable($value) ? new WordpressTransientThrowable($value) : $value;
  }

  /** @return array */
  public function getInfoLogs() {
    return filterDuplicateLogs($this->infoLogs);
  }

  /** @return array */
  public function getErrorLogs() {
    return filterDuplicateLogs($this->errorLogs);
  }

  /**
   * @param mixed $value
   * @return $this
   */
  public function info($value) {
    $this->infoLogs[] = self::transientValue($value);
    set_transient($this->infoLogsKey(), $this->infoLogs, $this->maxExpirationTimeInSeconds);
    return $this;
  }

  /**
   * @param mixed $value
   * @return $this
   */
  public function error($value) {
    $this->errorLogs[] = self::transientValue($value);
    set_transient($this->errorLogsKey(), $this->errorLogs, $this->maxExpirationTimeInSeconds);
    return $this;
  }

  /** @return void */
  public function clear() {
    $this->infoLogs = [];
    $this->errorLogs = [];
    delete_transient($this->infoLogsKey());
    delete_transient($this->errorLogsKey());
  }

  /** @return void */
  public function writeTo(Log $log) {
    $infoLogs = $this->getInfoLogs();
    $errorLogs = $this->getErrorLogs();
    /** @psalm-suppress MixedAssignment */
    foreach ($infoLogs as $infoLog) $log->info($infoLog);
    /** @psalm-suppress MixedAssignment */
    foreach ($errorLogs as $errorLog) $log->error($errorLog);
  }
}

/** @return string */
function vendorLogKey() {
  return Constant::vendorPrefixed('-log');
}

/** @return WordpressTransientLog */
function transientLog(int $maxExpirationTimeInSeconds = 60) {
  return new WordpressTransientLog(vendorLogKey(), $maxExpirationTimeInSeconds);
}

class WordpressSettingsErrorLog implements Log {
  public string $setting;
  public string $codePrefix;
  public bool $filterDuplicate;

  private array $duplicates;

  public function __construct(
    string $setting,
    string $codePrefix,
    bool $filterDuplicate
  ) {
    $this->setting = $setting;
    $this->codePrefix = $codePrefix;
    $this->filterDuplicate = $filterDuplicate;
    $this->duplicates = [];
  }

  /**
   * @param mixed $value
   * @param int|string|null $errorCode
   * @return string
   */
  private static function htmlErrorNotice($value, $errorCode) {
    $asString = "$value";
    $splitIndex = strpos($asString, " in ");
    if (!is_int($splitIndex)) return $asString;
    $errorMessage = substr($asString, 0, $splitIndex);
    $errorStack = substr($asString, $splitIndex);
    $message = renderHtmlTag('details', [
      'class' => $errorCode === null ? null : "details-notice-code-$errorCode"
    ], '<summary>' . $errorMessage . '</summary><i>' . $errorStack . '</i>');
    return str_replace('\\n', '<br />', '<p>' . $message . '</p>');
  }

  /**
   * @template T
   * @param T $value
   *
   * @return $this
   */
  protected function log(string $type, $value, string $code = '') {
    $fullCode = $this->codePrefix . ($code === '' ? '' : '-') . $code;

    if ($this->filterDuplicate) {
      $duplicateValue = $value;
      if (isThrowable($value)) $duplicateValue = $value->getMessage();
      if ($value instanceof WordpressTransientThrowable) $duplicateValue = $value->getMessage();
      if (in_array($duplicateValue, $this->duplicates)) return $this;
      $this->duplicates[] = $duplicateValue;
    }

    $loggedValue = $value;
    if (isThrowable($value)) $loggedValue = self::htmlErrorNotice($value, $value->getCode());
    if ($value instanceof WordpressTransientThrowable) $loggedValue = self::htmlErrorNotice(
      $value->asString,
      $value->getCode()
    );
    add_settings_error($this->setting, $fullCode, "$loggedValue", $type);
    return $this;
  }

  /**
   * @param mixed $value
   * @return $this
   */
  public function info($value, string $code = '') {
    return $this->log('info', $value, $code);
  }

  /**
   * @param mixed $value
   * @return $this
   */
  public function error($value, string $code = '') {
    return $this->log('error', $value, $code);
  }
}

/** @return WordpressSettingsErrorLog */
function settingsErrorLog(string $menuSlug) {
  $filterDuplicate = true;
  return new WordpressSettingsErrorLog($menuSlug, vendorLogKey(), $filterDuplicate);
}

/** @return WordpressSettingsErrorLog */
function settingsErrorLogFromTransient(string $menuSlug, WordpressTransientLog $transientLog) {
  $log = settingsErrorLog($menuSlug);
  $transientLog->writeTo($log);
  $transientLog->clear();
  return $log;
}
