<?php

namespace Wino;

/** @template T of array<string, mixed> */
class WordpressQueryArgs {
  public const TERM_ID = 'term_id';
  public const META_QUERY = 'meta_query';

  /** @var T */
  public array $data;

  /** @param T $data */
  public function __construct(array $data = []) {
    $this->data = $data;
  }

  /**
   * @param T $data
   * @return void
   */
  public function set($data) {
    /**
     * @psalm-suppress MixedAssignment
     * @psalm-suppress PossiblyInvalidPropertyAssignmentValue
     */
    foreach ($data as $key => $value) $this->data[$key] = $value;
  }

  /**
   * @param key-of<T> $key
   * @return bool
   */
  public function has(string $key) {
    return key_exists($key, $this->data);
  }
}

/**
 * @extends WordpressQueryArgs<array{
 *  attachment_id?: int,
 *  author?: int|string,
 *  author__in?: array<array-key, int>,
 *  author__not_in?: array<array-key, int>,
 *  author_name?: string,
 *  cache_results?: bool,
 *  cat?: int|string,
 *  category?: int|string,
 *  category__and?: array<array-key, int>,
 *  category__in?: array<array-key, int>,
 *  category__not_in?: array<array-key, int>,
 *  category_name?: string,
 *  comment_count?: array<array-key, mixed>|int,
 *  comment_status?: string,
 *  comments_per_page?: int,
 *  date_query?: array<array-key, mixed>,
 *  day?: int,
 *  exact?: bool,
 *  exclude?: array<array-key, int>,
 *  fields?: string,
 *  hour?: int,
 *  ignore_sticky_posts?: bool|int,
 *  include?: array<array-key, int>,
 *  lazy_load_term_meta?: bool,
 *  m?: int,
 *  menu_order?: int,
 *  meta_compare?: string,
 *  meta_compare_key?: string,
 *  meta_key?: array<array-key, string>|string,
 *  meta_query?: array<array-key, mixed>,
 *  meta_type?: string,
 *  meta_type_key?: string,
 *  meta_value?: array<array-key, string>|string,
 *  minute?: int,
 *  monthnum?: int,
 *  name?: string,
 *  no_found_rows?: bool,
 *  nopaging?: bool,
 *  numberposts?: int,
 *  offset?: int,
 *  order?: string,
 *  orderby?: array<array-key, mixed>|string,
 *  p?: int,
 *  page?: int,
 *  page_id?: int,
 *  paged?: int,
 *  pagename?: string,
 *  perm?: string,
 *  ping_status?: string,
 *  post__in?: array<array-key, int>,
 *  post__not_in?: array<array-key, int>,
 *  post_mime_type?: string,
 *  post_name__in?: array<array-key, string>,
 *  post_parent?: int,
 *  post_parent__in?: array<array-key, int>,
 *  post_parent__not_in?: array<array-key, int>,
 *  post_status?: array<array-key, string>|string,
 *  post_type?: array<array-key, string>|string,
 *  posts_per_archive_page?: int,
 *  posts_per_page?: int,
 *  s?: string,
 *  second?: int,
 *  sentence?: bool,
 *  suppress_filters?: bool,
 *  tag?: string,
 *  tag__and?: array<array-key, int>,
 *  tag__in?: array<array-key, int>,
 *  tag__not_in?: array<array-key, int>,
 *  tag_id?: int,
 *  tag_slug__and?: array<array-key, string>,
 *  tag_slug__in?: array<array-key, string>,
 *  tax_query?: array<array-key, mixed>,
 *  title?: string,
 *  update_post_meta_cache?: bool,
 *  update_post_term_cache?: bool,
 *  w?: int,
 *  year?: int
 * }>
 */
class GetPostsArgs extends WordpressQueryArgs {
  //
}

/**
 * @extends WordpressQueryArgs<array{
 *  comment_status?: string,
 *  guid?: string,
 *  import_id?: int,
 *  menu_order?: int,
 *  meta_input?: array<array-key, mixed>,
 *  ping_status?: string,
 *  pinged?: string,
 *  post_author?: int,
 *  post_category?: array<array-key, int>,
 *  post_content?: string,
 *  post_content_filtered?: string,
 *  post_date?: string,
 *  post_date_gmt?: string,
 *  post_excerpt?: string,
 *  post_mime_type?: string,
 *  post_name?: string,
 *  post_parent?: int,
 *  post_password?: string,
 *  post_status?: string,
 *  post_title?: string,
 *  post_type?: string,
 *  tags_input?: array<array-key, mixed>,
 *  tax_input?: array<array-key, mixed>,
 *  to_ping?: string
 * }>
 */
class UpdatePostArgs extends WordpressQueryArgs {
  /**
   * @param 'publish'|'future'|'pending'|'draft'|'auto-draft'|'private'|'trash'|'inherit' $status
   * @return UpdatePostArgs
   */
  public static function status($status, ?UpdatePostArgs $args = null) {
    $data = $args === null ? [] : $args->data;
    $data['post_status'] = $status;
    return new self($data);
  }
}

/**
 * @extends WordpressQueryArgs<array{
 *  name?: string,
 *  alias_of?: string,
 *  description?: string,
 *  parent?: int,
 *  slug?: string
 * }>
 */
class UpdateTermArgs extends WordpressQueryArgs {
  /** @return InsertTermArgs */
  public function asInsertArgs() {
    $data = $this->data;
    unset($data['name']);
    return new InsertTermArgs($data);
  }

  /** @return string|null */
  public function getName() {
    return optionalStringData($this->data, 'name');
  }

  /** @return string|null */
  public function getAliasOf() {
    return optionalStringData($this->data, 'alias_of');
  }

  /** @return string|null */
  public function getDescription() {
    return optionalStringData($this->data, 'description');
  }

  /** @return int|null */
  public function getParent() {
    return optionalIntData($this->data, 'parent');
  }

  /** @return string|null */
  public function getSlug() {
    return optionalStringData($this->data, 'slug');
  }
}

/**
 * @extends WordpressQueryArgs<array{
 *  alias_of?: string,
 *  description?: string,
 *  parent?: int,
 *  slug?: string
 * }>
 */
class InsertTermArgs extends WordpressQueryArgs {
  /** @return UpdateTermArgs */
  public function asUpdateArgs() {
    return new UpdateTermArgs($this->data);
  }
}

/**
 * @extends WordpressQueryArgs<array<string, mixed>>
 */
class MetaQuery extends WordpressQueryArgs {
  //
}

/**
 * @param "==" | "!=" | "<=" | ">=" | "<" | ">" $compare
 * @param mixed $value
 * @return MetaQuery
 */
function mixedMetaQuery(string $key, string $compare, $value) {
  return new MetaQuery(['key' => $key, 'compare' => $compare, 'value' => $value]);
}

/**
 * @param "==" | "!=" $compare
 * @return MetaQuery
 */
function nullMetaQuery(string $key, string $compare = '==') {
  return mixedMetaQuery($key, $compare, null);
}

/** @return MetaQuery */
function notNullMetaQuery(string $key) {
  return nullMetaQuery($key, '!=');
}

/**
 * @param "==" | "!=" $compare
 * @return MetaQuery
 */
function boolMetaQuery(string $key, bool $value, string $compare = '==') {
  return mixedMetaQuery($key, $compare, $value);
}

/**
 * @param "==" | "!=" | "<=" | ">=" | "<" | ">" $compare
 * @return MetaQuery
 */
function intMetaQuery(string $key, string $compare, int $value) {
  return mixedMetaQuery($key, $compare, $value);
}

/**
 * @param "==" | "!=" | "<=" | ">=" | "<" | ">" $compare
 * @return MetaQuery
 */
function floatMetaQuery(string $key, string $compare, float $value) {
  return mixedMetaQuery($key, $compare, $value);
}

/**
 * @param "==" | "!=" $compare
 * @return MetaQuery
 */
function stringMetaQuery(string $key, string $compare, string $value) {
  return mixedMetaQuery($key, $compare, $value);
}

/** @return MetaQuery */
function metaQueryExists(string $key, bool $exist = true) {
  return new MetaQuery(['key' => $key, 'compare' => $exist ? 'EXISTS' : 'NOT EXISTS']);
}

/**
 * @extends WordpressQueryArgs<array{default?: int, force_default?: bool}>
 */
class DeleteTermsArgs extends WordpressQueryArgs {
  //
}

/**
 * @extends WordpressQueryArgs<array{
 *  cache_domain?: string,
 *  child_of?: int,
 *  childless?: bool,
 *  count?: bool,
 *  description__like?: string,
 *  exclude?: array<array-key, int>|string,
 *  exclude_tree?: array<array-key, int>|string,
 *  fields?: string,
 *  get?: string,
 *  hide_empty?: bool|int,
 *  hierarchical?: bool,
 *  include?: array<array-key, int>|string,
 *  meta_compare?: string,
 *  meta_compare_key?: string,
 *  meta_key?: array<array-key, string>|string,
 *  meta_query?: array<array-key, mixed>,
 *  meta_type?: string,
 *  meta_type_key?: string,
 *  meta_value?: array<array-key, string>|string,
 *  name?: array<array-key, string>|string,
 *  name__like?: string,
 *  number?: int|string,
 *  object_ids?: array<array-key, int>|int,
 *  offset?: int, order?: string,
 *  orderby?: string,
 *  pad_counts?: bool,
 *  parent?: int,
 *  search?: string,
 *  slug?: array<array-key, string>|string,
 *  taxonomy?: array<array-key, string>|string,
 *  term_taxonomy_id?: array<array-key, int>|int,
 *  update_term_meta_cache?: bool
 * }>
 */
class GetTermsArgs extends WordpressQueryArgs {
  /** @return int|null */
  public function termId() {
    return optionalIntData($this->data, WordpressQueryArgs::TERM_ID);
  }

  /** @return void */
  public function setMetaQuery(MetaQuery ...$metaQueries) {
    $queries = [];
    foreach ($metaQueries as $query) $queries[] = $query->data;
    if (count($queries) > 0) $this->data[self::META_QUERY] = $queries;
    else unset($this->data[self::META_QUERY]);
  }

  /** @return void */
  public function addMetaQueries(MetaQuery ...$metaQueries) {
    foreach ($metaQueries as $query) $this->data[self::META_QUERY][] = $query->data;
  }
}
