<?php

namespace Wino;

/**
 * global `$wpdb` variable is not correctly typed.
 *
 * While it is preferable to avoid, it is sometimes
 * necessary to accomplish some tasks.
 *
 * This class provides a typed access to some $wpdb
 * methods. Use it with caution.
 */
abstract class WordpressDatabase {
  /** @return mixed */
  private static function instance() {
    return arrayElement($GLOBALS, 'wpdb');
  }

  /** @return bool */
  private static function debug() {
    return isDebugModeEnabled();
  }

  /**
   * @param mixed[] $params
   * @return mixed
   */
  private static function mixedMethod(string $method, ...$params) {
    return mixedMethodCall(self::debug(), self::instance(), $method, ...$params);
  }

  /**
   * @param mixed[] $params
   * @return string|null
   */
  private static function stringMethod(string $method, ...$params) {
    return mixedStringMethodCall(self::debug(), self::instance(), $method, ...$params);
  }

  /**
   * @param mixed[] $params
   * @return array|null
   */
  private static function arrayMethod(string $method, ...$params) {
    return mixedArrayMethodCall(self::debug(), self::instance(), $method, ...$params);
  }

  /**
   * @param mixed[] $params
   * @return array|object|null
   */
  private static function nullableArrayOrObjectMethod(string $method, ...$params) {
    return mixedNullableArrayOrObjectMethodCall(self::debug(), self::instance(), $method, ...$params);
  }

  /** @return string|null */
  private static function stringProperty(string $property) {
    return mixedStringPropertyAccess(self::debug(), self::instance(), $property);
  }

  /** @return string|null */
  public static function dbVersion() {
    return self::stringMethod('db_version');
  }

  /** @return string|null */
  public static function dbServerInfo() {
    return self::stringMethod('db_server_info');
  }

  /** @return string|null */
  public static function prefix() {
    return self::stringProperty('prefix');
  }

  /** @return string|null */
  public static function prefixed(string $name) {
    $prefix = self::prefix();
    return $prefix === null ? null : $prefix . $name;
  }

  /** @return string|null */
  public static function termMetaTableName() {
    return self::prefixed('termmeta');
  }

  /** @return string|null */
  public static function prepare(string $sqlQuery, ?array $args = null) {
    $argsArray = is_array($args) ? $args : [];
    return self::stringMethod('prepare', $sqlQuery, ...$argsArray);
  }

  /** @return array|object|null */
  public static function getResults(string $sqlQuery) {
    return self::nullableArrayOrObjectMethod('get_results', $sqlQuery);
  }

  /**
   * @param string|null $query
   * @return array|null
   */
  public static function getCol($query, int $x = 0) {
    return self::arrayMethod('get_col', $query, $x);
  }
}

/** @return void */
function wordpressFlushOptionCache(string $option) {
  $all = wp_load_alloptions();
  if (key_exists($option, $all)) unset($all[$option]);
  $group = 'options';
  wp_cache_set('alloptions', $all, $group);
  wp_cache_delete($option, $group);
}

/**
 * @template T
 * @param Closure(): T $getProperty
 *
 * @return T | string
 */
function unsafeSetupDescriptionProperty(callable $getProperty) {
  try {
    return call_user_func($getProperty);
  } catch (\Throwable $thrown) {
    return $thrown->getMessage();
  }
}

/** @return string */
function wordpressVersion() {
  return unsafeSetupDescriptionProperty(function () {
    /** @psalm-suppress UndefinedConstant */
    $path = ABSPATH . WPINC . '/version.php';

    /** @psalm-suppress UnresolvableInclude */
    include $path;
    /** @psalm-suppress UndefinedVariable */
    return assertedString($wp_version);
  });
}

/** @return string|null */
function mysqlVersion() {
  return unsafeSetupDescriptionProperty(function () {
    $version = WordpressDatabase::dbVersion();
    $info = WordpressDatabase::dbServerInfo();
    $versionString = $version === null ? '' : $version;
    $infoString = $info === null ? '' : $info;
    $shortInfoString = str_replace($versionString, '', strtolower($infoString));
    $shortInfoString = str_replace('-', '', $shortInfoString);
    $shortInfoString = preg_replace('/[0-9]/', '', $shortInfoString);
    $cleanInfoString = is_string($shortInfoString) ? trim($shortInfoString) : '';
    $cleanInfoSuffix = $cleanInfoString === '' || $cleanInfoString === 'mysql' ? '' : "($cleanInfoString)";
    $fullVersion = trim("$versionString $cleanInfoSuffix");
    return $fullVersion === '' ? null : $fullVersion;
  });
}

/** @return string */
function serverSoftware() {
  return unsafeSetupDescriptionProperty(function () {
    return stringData($_SERVER, 'SERVER_SOFTWARE');
  });
}

/** @return string */
function apacheVersion() {
  return unsafeSetupDescriptionProperty(function () {
    return assertedString(apache_get_version());
  });
}

/** @return string[] | string */
function loadedExtensions() {
  return unsafeSetupDescriptionProperty(function () {
    return assertedStringsArray(get_loaded_extensions());
  });
}

class WordpressSetupDescriptionCache {
  /** @var null|array<string, mixed> */
  public static $cache = null;
}

/** @return string|null */
function wordpressPluginVersion(string $plugin) {
  $pluginsFiles = wp_get_active_and_valid_plugins();
  foreach ($pluginsFiles as $file) {
    $data = get_file_data($file, [
      'name' => 'plugin name',
      'version' => 'version'
    ]);
    $name = arrayElement($data, 'name');
    if ($name === $plugin) return arrayElement($data, 'version');
  }
  return null;
}

/** @return array<string, mixed> */
function wordpressSetupDescription() {
  $cache = WordpressSetupDescriptionCache::$cache;
  if ($cache !== null) return $cache;

  $plugins = [];
  $pluginsFiles = wp_get_active_and_valid_plugins();
  foreach ($pluginsFiles as $file) {
    $data = get_file_data($file, [
      'name' => 'plugin name',
      'version' => 'version'
    ]);
    $name = arrayElement($data, 'name');
    $version = arrayElement($data, 'version');
    if ($name === null) continue;
    $plugins[$name] = $version === null ? 'N/A' : $version;
  }
  $description = [
    'os' => PHP_OS,
    'bits' => PHP_INT_SIZE === 4 ? 32 : 64,
    'apache' => apacheVersion(),
    'server' => serverSoftware(),
    'memory' => Ini::getMemoryLimit(),
    'php' => PHP_VERSION,
    'extensions' => loadedExtensions(),
    'mysql' => mysqlVersion(),
    'wordpress' => wordpressVersion(),
    'initial_db_version' => get_site_option('initial_db_version'),
    'multisite' => is_multisite() ? get_blog_count() : 0,
    'locale' => get_locale(),
    'plugins' => $plugins
  ];
  WordpressSetupDescriptionCache::$cache = $description;
  return $description;
}

/**
 * @param int|null $parent
 *
 * @return bool
 */
function wordpressTermExists(string $term, string $taxonomy, $parent = null) {
  return term_exists($term, $taxonomy, $parent) !== null;
}

/**
 * @return int
 */
function insertWordpressTerm(string $term, string $taxonomy, ?InsertTermArgs $args = null) {
  $insertArgs = $args === null ? [] : $args->data;
  $insertion = wp_insert_term($term, $taxonomy, $insertArgs);
  if ($insertion instanceof \WP_Error) throw wordpressErrorException($insertion);
  return intData($insertion, WordpressQueryArgs::TERM_ID);
}

/** @return \WP_Post|null */
function findWordpressPostById(int $id) {
  $post = get_post($id);
  if ($post === null || $post instanceof \WP_Post) return $post;
  throw msgError("Post not being an instance of \WP_Post.", jsonString($post));
}

/**
 * @param GetPostsArgs $args
 *
 * @return int[]
 */
function findWordpressPostIds(GetPostsArgs $args) {
  $posts = [];
  $idsArgs = new GetPostsArgs(["fields" => "ids"]);
  $idsArgs->set($args->data);
  $postsOrIds = get_posts($idsArgs->data);
  foreach ($postsOrIds as $postOrId)
  if (is_int($postOrId)) $posts[] = $postOrId;
  else throw error("Found a post id not being an int.");
  return $posts;
}

/**
 * @param GetPostsArgs $args
 *
 * @return \WP_Post[]
 */
function findWordpressPosts(GetPostsArgs $args) {
  $posts = [];
  $postsOrIds = get_posts($args->data);
  foreach ($postsOrIds as $postOrId)
  if ($postOrId instanceof \WP_Post) $posts[] = $postOrId;
  else throw error("Found a post not being an instance of \WP_Post.");
  return $posts;
}

class WordpressFilter {
  public string $hook;
  public string $function;
  public ?int $priority;
  public ?int $acceptedArgs;

  private function __construct(
    string $hook,
    string $function,
    ?int $priority,
    ?int $acceptedArgs
  ) {
    $this->hook = $hook;
    $this->function = $function;
    $this->priority = $priority;
    $this->acceptedArgs = $acceptedArgs;
  }

  /** @return self */
  public static function disabled(string $hook, string $function) {
    return new self($hook, $function, null, null);
  }

  /** @return self */
  public static function enabled(
    string $hook,
    string $function,
    int $priority,
    int $acceptedArgs
  ) {
    return new self($hook, $function, $priority, $acceptedArgs);
  }

  /** @return self */
  public static function from(string $hook, string $function) {
    $disabled = self::disabled($hook, $function);
    $filters = arrayElement($GLOBALS, 'wp_filter');
    if (!is_array($filters)) return $disabled;
    $filter = arrayElement($filters, $hook);
    if (!($filter instanceof \WP_Hook)) return $disabled;
    $callbacks = $filter->callbacks;

    foreach ($callbacks as $key => $callback) {
      if (!is_int($key)) continue;
      $priority = $key;
      if (!is_array($callback)) continue;
      $functionCallback = arrayElement($callback, $function);
      if (!is_array($functionCallback)) continue;
      $acceptedArgs = arrayElement($functionCallback, 'accepted_args');
      if (!is_int($acceptedArgs)) continue;
      return self::enabled($hook, $function, $priority, $acceptedArgs);
    }
    return $disabled;
  }

  /** @return bool */
  public function isEnabled() {
    return $this->priority !== null && $this->acceptedArgs !== null;
  }

  /** @return void */
  public function enable(int $defaultPriority = 10, int $defaultAcceptedArgs = 0) {
    $hook = $this->hook;
    $function = $this->function;
    if (!function_exists($function)) return;

    $priority = $this->priority;
    $acceptedArgs = $this->acceptedArgs;

    $current = self::from($hook, $function);
    if ($current->isEnabled()) return;

    $definedPriority = $priority === null ? $defaultPriority : $priority;
    $definedAcceptedArgs = $acceptedArgs === null ? $defaultAcceptedArgs : $acceptedArgs;

    $this->priority = $definedPriority;
    $this->acceptedArgs = $definedAcceptedArgs;

    add_action($hook, $function, $definedPriority, $definedAcceptedArgs);
  }

  /** @return void */
  public function disable(int $defaultPriority = 10) {
    $hook = $this->hook;
    $function = $this->function;
    $current = self::from($hook, $function);
    if (!$current->isEnabled()) return;
    $priority = $current->priority;
    $definedPriority = $priority === null ? $defaultPriority : $priority;
    $this->priority = $definedPriority;
    $this->acceptedArgs = $current->acceptedArgs;
    remove_action($hook, $function, $definedPriority);
  }
}

/**
 * WooCommerce defines a custom filter `wc_change_pre_get_terms`
 * registered on the `pre_get_terms` hook. This filter causes
 * some `get_terms` queries to not behave as expected.
 *
 * `findWordpressTerms` should so be prefered to avoid the side
 * effects caused by WooCommerce. It takes care of disabling then
 * enabling again the filters causing issues.
 *
 * However, you can still use this function in case you need
 * the WooCommerce filters to be applied.
 *
 * @return \WP_Term[]
 */
function findWordpressTermsWithFilters(GetTermsArgs $args) {
  $terms = [];

  $termsOr = get_terms($args->data);
  if ($termsOr instanceof \WP_Error) throw wordpressErrorException($termsOr);
  assertsArray($termsOr);

  foreach ($termsOr as $termOr)
  if ($termOr instanceof \WP_Term) $terms[] = $termOr;
  else throw error("Found a term not being an instance of \WP_Term.");

  return $terms;
}

/** @return \WP_Term[] */
function findWordpressTerms(GetTermsArgs $args) {
  $filter = WordpressFilter::from('pre_get_terms', 'wc_change_pre_get_terms');
  $filter->disable();

  try {
    $terms = findWordpressTermsWithFilters($args);
    $filter->enable();
    return $terms;
  } catch (\Throwable $thrown) {
    $filter->enable();
    throw $thrown;
  }
}

/** @return \WP_Term[] */
function findWordpressTermsBy(GetTermsArgs $args, MetaQuery ...$metaQueries) {
  $allArgs = new GetTermsArgs(array_merge(
    ['hide_empty' => false],
    $args->data
  ));
  $allArgs->addMetaQueries(...$metaQueries);

  return findWordpressTerms($allArgs);
}

/** @return \WP_Term[] */
function findWordpressTermsByName(string $name, MetaQuery ...$metaQueries) {
  $args = new GetTermsArgs(['name' => $name]);
  return findWordpressTermsBy($args, ...$metaQueries);
}

/**
 * Don't know why but without the taxonomy here the query doesn't work.
 * On the contrary, the `findWordpressTermsByName` function doesn't work
 * if we use $taxonomy (internally).
 *
 * Both internally use the same get_terms functions so this is very disturbing.
 * Maybe just not possible to mix taxonomy and meta queries in the same get_terms query ?
 * But still why the taxonomy is required here ?
 * The parent and termName termId should be enough...
 *
 * @return \WP_Term[]
 */
function findWordpressTermChildrenByName(string $taxonomy, int $parent, string $name, MetaQuery ...$metaQueries) {
  $args = new GetTermsArgs(['taxonomy' => $taxonomy, 'parent' => $parent, 'name' => $name]);
  return findWordpressTermsBy($args, ...$metaQueries);
}

/** @return \WP_Term[] */
function findWordpressTermsByTaxonomy(string $taxonomy, MetaQuery ...$metaQueries) {
  $args = new GetTermsArgs(['taxonomy' => $taxonomy]);
  return findWordpressTermsBy($args, ...$metaQueries);
}

class WordpressTermsTree {
  private \WP_Term $root;

  /** @var WordpressTermsTree[] */
  private array $children;

  /** @var int[] */
  private array $childrenTermIds;

  private function __construct(\WP_Term $root) {
    $this->root = $root;
    $this->children = [];
    $this->childrenTermIds = [];
  }

  /** @return int */
  public function getTermId() {
    return $this->root->term_id;
  }

  /** @return string */
  public function getTermName() {
    return $this->root->name;
  }

  /** @return WordpressTermsTree[] */
  public function getChildren() {
    return $this->children;
  }

  /**
   * @param \WP_Term[] $descendantTerms
   *
   * @return  self[]
   */
  private static function buildTree(array $descendantTerms) {
    $rootTree = [];
    $treesByRootTermId = [];

    foreach ($descendantTerms as $term) {
      $termId = $term->term_id;
      $termTree = new self($term);
      $treesByRootTermId[$termId] = $termTree;
    }

    foreach ($descendantTerms as $term) {
      $tree = arrayElement($treesByRootTermId, $term->term_id);
      if ($tree === null) continue;

      $parentTermId = $term->parent;
      $parentTree = arrayElement($treesByRootTermId, $parentTermId);

      if ($parentTermId === undefinedTermId()) $rootTree[] = $tree;
      elseif ($parentTree !== null) $parentTree->children[] = $tree;
    }

    return $rootTree;
  }

  /** @return self[] */
  public static function fromTaxonomy(string $taxonomy, MetaQuery ...$metaQueries) {
    $terms = findWordpressTermsByTaxonomy($taxonomy, ...$metaQueries);
    return self::buildTree($terms);
  }
}

/**
 * @param int|string $value
 * @return \WP_Term|null
 */
function findWordpressTerm(string $taxonomy, string $field, $value) {
  $term = get_term_by($field, $value, $taxonomy);
  if ($term instanceof \WP_Term) return $term;
  if ($term === false) return null;
  throw msgException("Unexpected term returned as an array", jsonString($term));
}

/**
 * @param int|string $value
 * @psalm-suppress MixedReturnTypeCoercion
 * @return \WP_Term[]
 */
function findAllWordpressTerms(string $taxonomy, string $field, $value) {
  $term = get_term_by($field, $value, $taxonomy, Constant::arrayN());
  if ($term === false) return [];
  if (is_array($term)) return $term;
  throw exception("Unexpected terms not returned as an array.");
}

/** @return \WP_Term|null */
function findWordpressTermById(int $termId) {
  $termId = definedTermId($termId);
  if ($termId === null) return null;
  $term = get_term($termId);
  if ($term instanceof \WP_Error) throw wordpressErrorException($term);
  if (is_array($term)) throw exception("Unexpected term returned as an array.");
  return $term;
}


/** @return int[] */
function findWordpressTermChildrenIdsById(
  int $termId,
  string $taxonomy
) {
  $termId = definedTermId($termId);
  if ($termId === null) return [];

  $children = [];
  $terms = get_term_children($termId, $taxonomy);
  if ($terms instanceof \WP_Error) throw wordpressErrorException($terms);

  /** @psalm-suppress MixedAssignment */
  foreach ($terms as $term)
  if (is_int($term)) $children[] = $term;
  else throw msgError("Unexpected non int found in array.", jsonString($terms));

  return $children;
}


/** @return \WP_Term[] */
function findWordpressTermChildrenById(
  int $termId,
  string $taxonomy
) {
  $children = [];
  $childrenIds = findWordpressTermChildrenIdsById($termId, $taxonomy);

  foreach ($childrenIds as $childId) {
    $child = findWordpressTermById($childId);
    if ($child === null) continue;
    $children[] = $child;
  }

  return $children;
}

/** @return \WP_Term */
function createWordpressTerm(string $term, string $taxonomy, ?InsertTermArgs $args = null) {
  $termId = insertWordpressTerm($term, $taxonomy, $args);
  $insertedTerm = findWordpressTermById($termId);
  if ($insertedTerm !== null) return $insertedTerm;
  throw exception("Unexpected null term while it was just created.");
}

/**
 * wp_insert_category or wp_create_category might be preferred,
 * but are actually not available outside of the admin_init hook.
 *
 * @return \WP_Term
 */
function createWordpressCategory(string $category, ?InsertTermArgs $args = null) {
  return createWordpressTerm($category, 'category', $args);
}

/**
 * @return \WP_Term
 */
function updateWordpressTerm(\WP_Term $term, ?UpdateTermArgs $args = null) {
  $termId = $term->term_id;
  $updateArgs = $args === null ? [] : $args->data;

  /**
   * @psalm-suppress InvalidArgument
   * According to the worpress codex, updating name is actually valid
   */
  $update = wp_update_term($termId, $term->taxonomy, $updateArgs);

  if ($update instanceof \WP_Error) throw wordpressErrorException($update);
  $updatedTerm = findWordpressTermById($termId);
  if ($updatedTerm !== null) return $updatedTerm;
  throw exception("Null term found while it was actually updated.");
}

/**
 * @return bool
 */
function deleteWordpressTerm(\WP_Term $term, string $taxonomy, ?DeleteTermsArgs $args = null) {
  $deleteArgs = $args === null ? [] : $args->data;
  $deletion = wp_delete_term($term->term_id, $taxonomy, $deleteArgs);
  if ($deletion instanceof \WP_Error) throw wordpressErrorException($deletion);
  if (is_int($deletion)) throw exception("Unexpected int returned while deleting a term.");
  return $deletion;
}

/**
 * @param mixed $value
 * @return bool|int
 */
function updateWordpressTermMeta(\WP_Term $term, string $metaKey, $value) {
  $update = update_term_meta($term->term_id, $metaKey, $value);
  if ($update instanceof \WP_Error) throw wordpressErrorException($update);
  return $update;
}

/** @return bool */
function deleteWordpressTermMeta(\WP_Term $term, string $metaKey) {
  $deletion = delete_term_meta($term->term_id, $metaKey);
  return $deletion;
}

/** @return string[] */
function findWordpressTermMetaKeys(int $termId) {
  $metas = assertedArray(get_term_meta($termId));
  $keys = [];
  /** @psalm-suppress MixedAssignment */
  foreach ($metas as $key => $value)
  if (is_string($key)) {
    noop($value);
    $keys[] = $key;
  }

  return $keys;
}

/** @return string|null */
function findWordpressTermMetaString(int $termId, string $metaKey) {
  /** @var mixed */
  $value = get_term_meta($termId, $metaKey, true);
  if ($value === false || $value === "") return null;
  if (is_string($value)) return $value;
  throw exception("Non string term ($termId) meta $metaKey.");
}

/**
 * @param mixed $value
 * @return bool|int
 */
function updateWordpressPostMeta(int $postId, string $metaKey, $value) {
  return update_post_meta($postId, $metaKey, $value);
}

/** @return bool */
function deleteWordpressPostMeta(int $postId, string $metaKey) {
  return delete_post_meta($postId, $metaKey);
}

/** @return string|null */
function optionalWordpressPostMetaString(int $postId, string $metaKey) {
  /** @var mixed */
  $value = get_post_meta($postId, $metaKey, true);
  if ($value === false) return null;
  if (is_string($value)) return trim($value) === "" ? null : $value;
  throw exception("Non string post ($postId) meta $metaKey.");
}

/** @return string */
function requiredWordpressPostMetaString(int $postId, string $metaKey) {
  $value = optionalWordpressPostMetaString($postId, $metaKey);
  if ($value !== null) return $value;
  throw exception("Missing required post ($postId) meta $metaKey.");
}

/** @return string[] */
function wordpressTermsMetaFromQuery(array ...$metaQuery) {
  $metaKeys = [];
  foreach ($metaQuery as $query) $metaKeys[] = nonBlankStringData($query, 'key');
  return $metaKeys;
}

/** @return void */
function deleteWordpressTermsByTaxonomy(string $taxonomy, MetaQuery ...$metaQueries) {
  $terms = findWordpressTermsByTaxonomy($taxonomy, ...$metaQueries);

  foreach ($terms as $term) {
    if (deleteWordpressTerm($term, $term->taxonomy)) continue;
    throw msgException("Term not correctly deleted", jsonString($term));
  }
}

/** @return void */
function updateWordpressPost(int $postId, ?UpdatePostArgs $args = null) {
  $data = $args === null ? [] : $args->data;
  $data['ID'] = $postId;
  $result = identity(wp_update_post($data, true));
  if ($result instanceof \WP_Error) throw wordpressErrorException($result);
}

/** @return void */
function publishWordpressPost(int $postId, ?UpdatePostArgs $args = null) {
  updateWordpressPost($postId, $args);
  wp_publish_post($postId);
}

/** @return int */
function getWordpressPostUpdateTimestamp(int $postId) {
  $modified = get_post_timestamp($postId, 'modified');
  if (is_int($modified)) return $modified;
  $created = get_post_timestamp($postId);
  if (is_int($created)) return $created;
  return 0;
}

/**
 * @param int|\WP_Post $post
 * @return \WP_Post|false|null
 */
function deleteWordpressPost($post) {
  return wp_delete_post(is_int($post) ? $post : $post->ID, true);
}

/** @return void */
function deleteWordpressPosts(GetPostsArgs $args) {
  $posts = findWordpressPostIds($args);
  $deletion = true;
  foreach ($posts as $post) $deletion = $deletion && deleteWordpressPost($post);
  if (!$deletion) throw exception("Some posts were not correctly deleted.");
}

/** @return string[] */
function findDistinctWordpressTermMetaKeys() {
  $debug = isDebugModeEnabled();
  $keys = [];

  $field = 'meta_key';
  $table = WordpressDatabase::termMetaTableName();

  $results = WordpressDatabase::getResults("SELECT DISTINCT $field FROM $table");
  if ($results !== null)
  /** @var mixed */
  foreach ($results as $result) {
    $key = mixedStringPropertyAccess($debug, $result, $field);
    if ($key !== null) $keys[] = $key;
  }
  return $keys;
}
