<?php

namespace Wino;

/**
 * This class only serves as a way to access the protected
 * `parse_json_params` method from `\WP_REST_Request`.
 *
 * @extends \WP_REST_Request<array<array-key, mixed>>
 */
class WinoRestRequest extends \WP_REST_Request {
  /** @return \WP_Error|bool */
  public static function parseJsonParams(\WP_REST_Request $request) {
    return $request->parse_json_params();
  }
}

/** @return array */
function getWordpressRestRequestJsonParams(\WP_REST_Request $request) {
  $parsing = WinoRestRequest::parseJsonParams($request);
  if ($parsing instanceof \WP_Error) throw wordpressErrorException($parsing);
  return $request->get_json_params();
}

class RestResponse implements Log {
  public const ERRORS_KEY = 'errors';
  public const INFOS_KEY = 'infos';
  public const STATUS_KEY = 'status';
  public const ERRORS_CONTEXT_KEY = 'context';

  private InMemoryLog $log;
  private array $rawData;
  private int $status;

  private function __construct(
    InMemoryLog $log
  ) {
    $this->log = $log;
    $this->rawData = [];
    $this->status = 200;
  }

  /** @return self */
  public static function from(InMemoryLog $log) {
    return new self($log);
  }

  /**
   * @param mixed $data
   * @return $this
   */
  public function setData(string $key, $data) {
    $this->rawData[$key] = $data;
    return $this;
  }

  /** @return void */
  public function setStatus(int $code) {
    $this->status = $code;
  }

  /**
   * @param mixed $value
   * @return $this
   */
  public function info($value) {
    $this->log->info($value);
    return $this;
  }

  /**
   * @param mixed $value
   * @return $this
   */
  public function error($value, int $status = 500) {
    $this->log->error($value);
    $this->setStatus($status);
    return $this;
  }

  /**
   * @param mixed $data
   * @return \WP_REST_Response
   */
  public static function wpResponse(int $status, $data) {
    return new \WP_REST_Response($data, $status);
  }

  /**
   * @template T
   * @param T[] $logs
   * @return array
   */
  private static function logsToRestFormat(array $logs) {
    $formatted = [];
    foreach ($logs as $log) {
      $value = $log;
      if ($log instanceof \Error || $log instanceof \Exception) $value = $log->getMessage();
      $formatted[] = $value;
    }
    return $formatted;
  }

  /** @return \WP_REST_Response */
  public function toWpResponse() {
    $data = $this->rawData;
    $status = $this->status;

    $errors = self::logsToRestFormat($this->log->getErrors());
    $infos = self::logsToRestFormat($this->log->getInfos());

    $data[self::STATUS_KEY] = $status;
    if (count($infos) > 0) $data[self::INFOS_KEY] = $infos;
    if (count($errors) > 0) {
      $data[self::ERRORS_KEY] = $errors;
      $data[self::ERRORS_CONTEXT_KEY] = wordpressSetupDescription();
    }

    return self::wpResponse($status, $data);
  }

  /** @return self */
  public static function empty() {
    return self::from(InMemoryLog::empty());
  }

  /** @return self */
  public static function errorResponse(\Throwable $error, int $status) {
    $response = self::empty();
    $response->error($error, $status);
    return $response;
  }
}

class RestRoute {
  public string $namespace;
  public string $path;

  protected function __construct(
    string $namespace,
    string $path
  ) {
    $this->namespace = $namespace;
    $this->path = $path;
  }

  /** @return self */
  public static function from(
    string $path,
    string $namespace = Constant::VENDOR_PREFIX
  ) {
    return new self($namespace, $path);
  }

  /** @return string */
  public function url() {
    $namespace = $this->namespace;
    $path = $this->path;
    return get_rest_url(null, "$namespace/$path");
  }
}

class RestRouter {
  /** @return void */
  public static function registerRoute(RestRoute $route, array $args) {
    add_action('rest_api_init', function () use ($route, $args) {
      register_rest_route(
        $route->namespace,
        $route->path,
        $args
      );
    });
  }

  /** @return void */
  private static function registerRouteWithVerb(string $verb, RestRoute $route, array $args) {
    self::registerRoute($route, array_merge(
      ['methods' => trim(strtoupper($verb))],
      $args
    ));
  }

  /** @return void */
  public static function registerGetRoute(RestRoute $route, array $args) {
    self::registerRouteWithVerb('get', $route, $args);
  }

  /** @return void */
  public static function registerPostRoute(RestRoute $route, array $args) {
    self::registerRouteWithVerb('post', $route, $args);
  }

  /**
   * @param Closure(\WP_REST_Request): (null|RestResponse|\WP_REST_Response) $controller
   *
   * @return array
   */
  private static function routeController(callable $controller, array $args) {
    return array_merge([
      'permission_callback' => function () {
        return true;
      },
      'callback' => function (\WP_REST_Request $request) use ($controller) {
        try {
          $response = call_user_func($controller, $request);
          if ($response === null) $response = RestResponse::empty();
          if ($response instanceof RestResponse) return $response->toWpResponse();
          return $response;
        } catch (\Throwable $thrown) {
          debugLog($thrown);
          return RestResponse::errorResponse($thrown, 500)->toWpResponse();
        }
      }
    ], $args);
  }

  /**
   * @param Closure(\WP_REST_Request): (null|RestResponse|\WP_REST_Response) $controller
   *
   * @return void
   */
  public static function registerGetController(RestRoute $route, callable $controller, array $args = []) {
    self::registerGetRoute($route, self::routeController($controller, $args));
  }

  /**
   * @param Closure(\WP_REST_Request): (null|RestResponse|\WP_REST_Response) $controller
   *
   * @return void
   */
  public static function registerPostController(RestRoute $route, callable $controller, array $args = []) {
    self::registerPostRoute($route, self::routeController($controller, $args));
  }

  /**
   * @param Closure(\WP_REST_Request): (null|RestResponse|\WP_REST_Response) $controller
   *
   * @return void
   */
  public static function registerWebhookNotificationHandler(
    RestRoute $route,
    callable $controller,
    array $args = []
  ) {
    self::registerPostController(
      $route,
      $controller,
      $args
    );
  }
}
