<?php

namespace Wino;

/**
 * Using `add_rewrite_rule` might not work on some server
 * with no write access to the file system. For instance
 * Wordpress instance deployed on Cloud Run (like the staging woocommerce)
 * will not be able to actually add rewrite rule. Leading to
 * a 404 when accessing the custom route.
 *
 * For such cases, the true url without rewriting will be used.
 *
 * @return bool
 */
function rewriteRulesAdditionEnabled() {
  return getenv("WINO_DISABLE_CUSTOM_ROUTES_REWRITE_RULES_ADDITION") !== 'true';
}

/** @return void */
function registerQueryVar(string $varName) {
  add_filter('query_vars', function (array $queryVars) use ($varName) {
    if (!in_array($varName, $queryVars)) $queryVars[] = $varName;
    return $queryVars;
  });

}

/** @return string | null */
function getQueryVar(string $varName) {
  /** @var mixed */
  $value = get_query_var($varName);
  if (!is_string($value)) return null;
  $trim = trim($value);
  if ($trim === "") return null;
  return $trim;
}

/**
 * @param string[] $queryVarNames
 * @return void
 */
function registerCustomRewriteRule(string $regex, array $queryVarNames, int $priority = 10) {
  if (!rewriteRulesAdditionEnabled()) return;

  $varRegex = "([a-z0-9-]+)";
  $regex = str_replace("$", $varRegex, $regex) . "[/]?\$";
  $query = "index.php";
  $matchIndex = 1;
  foreach ($queryVarNames as $queryVarName) {
    $token = $matchIndex === 1 ? "?" : "&";
    $query .= $token . "$queryVarName=\$matches[$matchIndex]";
    $matchIndex++;
  }

  add_action('init', function () use ($regex, $query) {
    add_rewrite_rule($regex, $query, 'top');
  }, $priority, 0);
}

/** @return string */
function defaultPageQueryVarName() {
  return 'wino-plugin-page';
}

/** @return string */
function defaultPageTemplatePath(string $page) {
  return dirname(__DIR__) . "/pages/$page.php";
}

/** @return string */
function pageTemplatePath(string $page, ?string $pageTemplatePath = null) {
  return is_string($pageTemplatePath) ? $pageTemplatePath : defaultPageTemplatePath($page);
}

/** @return string */
function pageQueryVarName(?string $pageQueryVarName = null) {
  return is_string($pageQueryVarName) ? $pageQueryVarName : defaultPageQueryVarName();
}

class RegisterCustomTemplateIncludeCache {
  /** @var string[] */
  public static array $pages = [];
}

/** @return void */
function registerCustomTemplateInclude(
  string $page,
  ?string $pageTemplatePath,
  ?string $pageQueryVarName
) {
  RegisterCustomTemplateIncludeCache::$pages[] = $page;
  $pageTemplate = pageTemplatePath($page, $pageTemplatePath);
  $queryVarName = pageQueryVarName($pageQueryVarName);
  registerQueryVar($queryVarName);
  add_filter('template_include', function (string $template) use ($queryVarName, $page, $pageTemplate) {
    $pageVar = getQueryVar($queryVarName);
    if ($pageVar === null) return $template;
    $registered = in_array($pageVar, RegisterCustomTemplateIncludeCache::$pages);
    if (!$registered) {
      status_header(404);
      nocache_headers();
      return get_query_template('404');
    }
    return $pageVar !== $page ? $template : $pageTemplate;
  });
}

class CustomRoute {
  public string $pageId;
  public string $templatePath;
  public string $pageQueryVarName;

  public string $queryVarRegex;
  /** @var string[] */
  public array $queryVarNames;
  public int $priority;
  public bool $enabled;

  /** @param string[] $queryVarNames */
  private function __construct(
    string $pageId,
    string $queryVarRegex,
    ?string $templatePath,
    ?string $pageQueryVarName,
    array $queryVarNames,
    ?int $priority
  ) {
    $this->pageId = $pageId;
    $this->queryVarRegex = $queryVarRegex;
    $this->templatePath = pageTemplatePath($pageId, $templatePath);
    $this->pageQueryVarName = pageQueryVarName($pageQueryVarName);
    $this->queryVarNames = $queryVarNames;
    $this->priority = is_int($priority) ? $priority : 10;
    $this->enabled = true;
  }

  /** @return self */
  public static function from(
    string $pageId,
    string $queryVarRegex = "",
    string ...$queryVarNames
  ) {
    return new self($pageId, $queryVarRegex, null, null, $queryVarNames, null);
  }

  /** @return $this */
  public function enable(bool $enabled) {
    $this->enabled = $enabled;
    return $this;
  }
}

/** @return string */
function customRouteRegex(CustomRoute $route) {
  return "wino.$" . $route->queryVarRegex;
}

/** @return string[] */
function customRouteQueryVarNames(CustomRoute $route) {
  $names = [$route->pageQueryVarName];
  $queryVarNames = $route->queryVarNames;
  foreach ($queryVarNames as $queryVarName) {
    $names[] = $queryVarName;
    registerQueryVar($queryVarName);
  }
  return $names;
}

/** @return void */
function registerCustomRoute(CustomRoute $route) {
  if (!$route->enabled) return;
  registerCustomRewriteRule(customRouteRegex($route), customRouteQueryVarNames($route), $route->priority);
  registerCustomTemplateInclude($route->pageId, $route->templatePath, $route->pageQueryVarName);
}

/** @return string */
function canonicalCustomRoutePath(CustomRoute $route, string ...$params) {
  $vars = array_merge([$route->pageId], $params);
  $path = customRouteRegex($route);

  foreach ($vars as $var) $path = strReplaceFirst("$", $var, $path);
  return $path;
}

/** @return string */
function fallbackCustomRoutePath(CustomRoute $route, string ...$params) {
  $queryVar = $route->pageQueryVarName;
  $name = $route->pageId;
  $path = "index.php?$queryVar=$name";
  $varNames = $route->queryVarNames;
  $i = -1;
  foreach ($params as $param) {
    $i++;
    $varName = $varNames[$i];
    $path .= "&$varName=$param";
  }
  return $path;
}

/** @return string */
function canonicalCustomRouteUrl(CustomRoute $route, string ...$params) {
  return site_url(canonicalCustomRoutePath($route, ...$params));
}

/** @return string */
function fallbackCustomRouteUrl(CustomRoute $route, string ...$params) {
  return site_url(fallbackCustomRoutePath($route, ...$params));
}

/** @return string */
function customRouteUrl(CustomRoute $route, string ...$params) {
  return rewriteRulesAdditionEnabled()
  ? canonicalCustomRouteUrl($route, ...$params)
  : fallbackCustomRouteUrl($route, ...$params);
}
